/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
* basic_overlayer.cpp
*
* Basic implementation of v4d_Overlayer
*
* Ondrej Danek (xdanek2@fi.muni.cz) 2007
*/

#include "v4d_basic.h"
#include <i3d/transform.h>
#include <i3d/basic.h>
#include "basic_overlayer.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

#undef RGB

namespace v4d 
{

    ///////////////////////////////////////////////////////////////////////////////
    //
    // v4d_OverlayBuffer
    //
    ///////////////////////////////////////////////////////////////////////////////
    
	/******************************************************************************/
    
    template <class LABEL>
    size_t v4d_OverlayBuffer<LABEL>::GetWidth() const
    {
        if (GetWindowType() != v4d_OVERLAY_WIN_YZ && GetWindowType() != v4d_OVERLAY_WIN_YZ_AF)
            return GetLayer(0).GetSizeX();
        return GetLayer(0).GetSizeY();
    }

	/******************************************************************************/

    template <class LABEL>
    size_t v4d_OverlayBuffer<LABEL>::GetHeight() const
    {
        if (m_win_type != v4d_OVERLAY_WIN_AF && m_win_type != v4d_OVERLAY_WIN_XY &&
            m_win_type != v4d_OVERLAY_WIN_XY_AF)
            return GetLayer(0).GetSizeZ();
        return GetLayer(0).GetSizeY();
    }

	/******************************************************************************/

    ///////////////////////////////////////////////////////////////////////////////
    //
    // v4d_BasicOverlayingModule
    //
    ///////////////////////////////////////////////////////////////////////////////

	/******************************************************************************/

    template <class LABEL, class VOXEL>
    v4d_BasicOverlayingModule<LABEL, VOXEL>::v4d_BasicOverlayingModule
        (wxFrame *parent, i4d::LabeledImage4d<LABEL,VOXEL> *limg, bool del_image)
        : v4d_OverlayingModule<LABEL, VOXEL> (parent), m_limg(limg), m_del_lab(del_image)
    { 
		if (limg != nullptr && limg -> GetNumFrames())
		{
			size_t num = limg -> GetNumFrames();

			for (size_t i = 0; i < 2; ++i)
			{
				m_af_xy[i].resize(num);
				m_af_xz[i].resize(num);
				m_af_yz[i].resize(num);
			}

			m_valid_af_xy.resize(num);
			m_valid_af_xz.resize(num);
			m_valid_af_yz.resize(num);
		}
	}

	/******************************************************************************/

	template <class LABEL, class VOXEL>
    v4d_BasicOverlayingModule<LABEL, VOXEL>::~v4d_BasicOverlayingModule()
    { 
        if (m_del_lab) 
            delete m_limg; 
        for (size_t i = 0; i < m_buf_vec.size(); i++)
            delete m_buf_vec[i];
    }

	/******************************************************************************/

	template <class LABEL, class VOXEL>
    size_t v4d_BasicOverlayingModule<LABEL, VOXEL>::Buffer_Create (v4d_OverlayWindowType win_type)
    {
        m_buf_vec.push_back (new v4d_OverlayBuffer<LABEL> (win_type));
        return m_buf_vec.size() - 1;
    }

	/******************************************************************************/

    template <class LABEL, class VOXEL>
    void v4d_BasicOverlayingModule<LABEL, VOXEL>::Buffer_SetDim (size_t buf_id, const i3d::Vector3d<size_t> &sz, 
        const i3d::Offset &ofs, const i3d::Vector3d<float> &res)
    {
        for (size_t i = 0; i < 2; i++)
        {
            m_buf_vec[buf_id]->GetLayer(i).MakeRoom (sz);
            m_buf_vec[buf_id]->GetLayer(i).SetOffset (ofs);
            m_buf_vec[buf_id]->GetLayer(i).SetResolution (res);
        }
    }

	/******************************************************************************/

	template <class LABEL, class VOXEL>
    void v4d_BasicOverlayingModule<LABEL, VOXEL>::Overlay (size_t buf_id, Buffer2d<i3d::RGB> &buf) const
	{   
		if (!this->GetOverlaying() || buf_id >= m_buf_vec.size())
            return;

        const v4d_OverlayBuffer<LABEL> &overlay = *m_buf_vec[buf_id];
        float alpha = this->GetTransparency(), nalpha = 1 - alpha;
        LABEL bdr_mask = GetHighestBit();

        if (overlay.GetLayersCount() == 1)
        {
            const LABEL *src = overlay.GetLayer(0).GetFirstVoxelAddr();
            i3d::RGB *first = buf.GetData();

            for (int i = 0; i < buf.GetSize(); i++, src++, first++)
            {
                LABEL label = *src & ~bdr_mask;

                if (label > LABEL(0))
				{
					if (*src > bdr_mask || alpha == 0.0f)
					{
						*first = this->GetLabelColour(label, buf_id);
					}
					else if (alpha < 1.0f)
					{
						i3d::RGB col = this->GetLabelColour(label, buf_id);
						*first = i3d::RGB(i3d::GRAY8(alpha * first->red + nalpha * col.red),
							i3d::GRAY8(alpha * first->green + nalpha * col.green),
							i3d::GRAY8(alpha * first->blue + nalpha * col.blue));
					}
				}
            }
        }
        else if (overlay.GetLayersCount() == 2)
        {
            const LABEL *src1 = overlay.GetLayer(1).GetFirstVoxelAddr();
            const LABEL *src2 = overlay.GetLayer(0).GetFirstVoxelAddr();
            i3d::RGB *first = buf.GetData();

            for (int i = 0; i < buf.GetSize(); i++, src1++, src2++, first++)
            {
                LABEL label1 = *src1 & ~bdr_mask,
                      label2 = *src2;

                if (label2 > LABEL(0))
                {
                    // First draw boundary
                    if (label1)
                        *first = this -> GetLabelColour(label1, buf_id);

                    // Next overlay component with given transparency
                    if (alpha == 0.0f)
                        *first = this -> GetLabelColour(label2, buf_id);
                    else if (alpha < 1.0f)
                    {
                        i3d::RGB col = this -> GetLabelColour(label2, buf_id);
                        *first = i3d::RGB (i3d::GRAY8(alpha * first->red + nalpha * col.red),
                                           i3d::GRAY8(alpha * first->green + nalpha * col.green),
                                           i3d::GRAY8(alpha * first->blue + nalpha * col.blue));
                    }
                }
            }
        }
	}

	/******************************************************************************/

	template <class LABEL, class VOXEL>
	void v4d_BasicOverlayingModule<LABEL, VOXEL>::Buffer_Render (size_t buf_id, size_t time)
    {
        if (buf_id >= m_buf_vec.size())
            return;

        v4d_OverlayBuffer<LABEL> &buf = *m_buf_vec[buf_id];
        v4d_OverlayWindowType win_type = buf.GetWindowType();

        if (win_type == v4d_OVERLAY_WIN_AF)
        {
			i3d::CopyToImage(*GetLabeledImage(true, false, time), buf.GetLayer(0));
			i3d::CopyToImage(*GetLabeledImage(true, true, time), buf.GetLayer(1));
            buf.SetLayersCount(2);
        }
        else if (win_type == v4d_OVERLAY_WIN_XY || win_type == v4d_OVERLAY_WIN_XZ || win_type == v4d_OVERLAY_WIN_YZ)
        {
            i3d::CopyToImage (*GetLabeledImage(false, false, time), buf.GetLayer(0));
            ComputeBoundary (buf_id);
            buf.SetLayersCount(1);
        }
        else if (win_type == v4d_OVERLAY_WIN_XY_AF)
        {
            if (!m_valid_af_xy[time])
            {
                i3d::Get_Labeled_AF_XY(*GetLabeledImage(false, false, time), m_af_xy[0][time], nullptr, false);
                i3d::Get_Labeled_AF_XY(*GetLabeledImage(false, false, time), m_af_xy[1][time], nullptr, true);
                m_valid_af_xy[time] = true;
            }
            i3d::CopyToImage (m_af_xy[0][time], buf.GetLayer(0));
            i3d::CopyToImage (m_af_xy[1][time], buf.GetLayer(1));
            buf.SetLayersCount(2);
        }
        else if (win_type == v4d_OVERLAY_WIN_XZ_AF)
        {
            if (!m_valid_af_xz[time])
            {
                i3d::Get_Labeled_AF_XZ(*GetLabeledImage(false, false, time), m_af_xz[0][time], nullptr, false);
                i3d::Get_Labeled_AF_XZ(*GetLabeledImage(false, false, time), m_af_xz[1][time], nullptr, true);
                m_valid_af_xz[time] = true;
            }
            i3d::CopyToImage (m_af_xz[0][time], buf.GetLayer(0));
            i3d::CopyToImage (m_af_xz[1][time], buf.GetLayer(1));
            buf.SetLayersCount(2);
        }
        else if (win_type == v4d_OVERLAY_WIN_YZ_AF)
        {
            if (!m_valid_af_yz[time])
            {
                i3d::Get_Labeled_AF_YZ(*GetLabeledImage(false, false, time), m_af_yz[0][time], nullptr, false);
                i3d::Get_Labeled_AF_YZ(*GetLabeledImage(false, false, time), m_af_yz[1][time], nullptr, true);
                m_valid_af_yz[time] = true;
            }
            i3d::CopyToImage (m_af_yz[0][time], buf.GetLayer(0));
            i3d::CopyToImage (m_af_yz[1][time], buf.GetLayer(1));
            buf.SetLayersCount(2);
        }
    }

	/******************************************************************************/

    #define BORDER(x,y) if (*(x)) *(x) |= y

    template <class LABEL, class VOXEL>
    void v4d_BasicOverlayingModule<LABEL, VOXEL>::ComputeBoundary (size_t buf_id)
    {
        LABEL *p = m_buf_vec[buf_id]->GetLayer(0).GetFirstVoxelAddr(), bdr_mask = GetHighestBit();
        size_t sz_x = m_buf_vec[buf_id]->GetWidth(), sz_y = m_buf_vec[buf_id]->GetHeight();

        for (size_t i = 0; i < sz_x; i++, p++)
            BORDER(p, bdr_mask);

        for (size_t y = 1; y < sz_y; y++)
        {
            BORDER(p, bdr_mask);
            p++;

            for (size_t x = 1; x < sz_x; x++, p++)
            {
                if (*p != (*(p - sz_x) & ~bdr_mask))
                {
                    BORDER(p, bdr_mask);
                    BORDER(p - sz_x, bdr_mask);
                }

                if (*p != (*(p - 1) & ~bdr_mask))
                {
                    BORDER(p, bdr_mask);
                    BORDER(p - 1, bdr_mask);
                }
            }

            BORDER(p - 1, bdr_mask);
        }

        p = m_buf_vec[buf_id]->GetLayer(0).GetFirstVoxelAddr() + sz_x * (sz_y - 1) + 1;
        for (size_t i = 1; i < sz_x - 1; i++, p++)
            BORDER(p, bdr_mask);
    }

	/******************************************************************************/

	// Explicit instantiations
    template class v4d_OverlayBuffer<i3d::GRAY16>;
    template class v4d_BasicOverlayingModule<i3d::GRAY16, bool>;
	template class v4d_BasicOverlayingModule<i3d::GRAY16, i3d::GRAY8>;
	template class v4d_BasicOverlayingModule<i3d::GRAY16, i3d::GRAY16>;

} // namespace v4d
