/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
* basic_overlayer.h
*
* Basic implementation of v4d_Overlayer
*
* Ondrej Danek (xdanek2@fi.muni.cz) 2007
*/

#ifndef V4D_BASIC_OVERLAYER_H
#define V4D_BASIC_OVERLAYER_H

#include <vector>
#include <i3d/regions.h>
#include "v4d_impexp.h"
#include "overlaying.h"
#include "v4d_labeledimage4d.h"

namespace v4d 
{
    //////////////////////////////////////////////////////////////////////////////////////////
    //
    // v4d_OverlayBuffer class
    //
    //////////////////////////////////////////////////////////////////////////////////////////
    template <class LABEL> class V4D_DLLEXPORT v4d_OverlayBuffer
    {
    protected:
        /** Overlay buffer layer */
        i3d::Image3d<LABEL> m_layer[2];
        /** Layers count */
        size_t m_layers;
        /** Buffer window type */
        v4d_OverlayWindowType m_win_type;

    public:
        /** Constructor */
        v4d_OverlayBuffer (v4d_OverlayWindowType win_type) : m_win_type(win_type) {}

        /** Get reference to given buffer layer */
        i3d::Image3d<LABEL> &GetLayer(size_t i) { return m_layer[i]; }
        /** Get const reference to given buffer layer */
        const i3d::Image3d<LABEL> &GetLayer(size_t i) const { return m_layer[i]; }
        
        /** Get number of layers */
        size_t GetLayersCount() const { return m_layers; }
        /** Set number of layers */
        void SetLayersCount(size_t num) { m_layers = num; }
        
        /** Get window type */
        v4d_OverlayWindowType GetWindowType () const { return m_win_type; }

        /** Get width */
        size_t GetWidth() const;
        /** Get width */
        size_t GetHeight() const;
    };

    //////////////////////////////////////////////////////////////////////////////////////////
    //
    // v4d_BasicOverlayingModule
    // Basic overlaying module that is used in standard viewer3d for overlaying.
    //
    //////////////////////////////////////////////////////////////////////////////////////////
	template <class LABEL, class VOXEL> 
        class V4D_DLLEXPORT v4d_BasicOverlayingModule: public v4d_OverlayingModule<LABEL, VOXEL>
	{
    protected:
        /** Pointer to labeled image */
        i4d::LabeledImage4d<LABEL, VOXEL> *m_limg;
        /** Should the image be deleted in destructor? */
        bool m_del_lab;

        /** Precomputed AF projections for xy window - 0=components, 1=boundary */
        std::vector<i3d::Image3d<LABEL> > m_af_xy[2];
        /** Precomputed AF projections for xz window - 0=components, 1=boundary */
        std::vector<i3d::Image3d<LABEL> > m_af_xz[2];
        /** Precomputed AF projections for yz window - 0=components, 1=boundary */
        std::vector<i3d::Image3d<LABEL> > m_af_yz[2];
        /** Validness flags for AF projections along the z axis. */
		std::vector<bool> m_valid_af_xy;
		/** Validness flags for AF projections along the y axis. */
		std::vector<bool> m_valid_af_xz;
		/** Validness flags for AF projections along the x axis. */
		std::vector<bool> m_valid_af_yz;
        /** Vector of overlay buffers */
        std::vector<v4d_OverlayBuffer<LABEL> *> m_buf_vec;

 	public:
		/** Constructor */
        v4d_BasicOverlayingModule(wxFrame *parent, i4d::LabeledImage4d<LABEL,VOXEL> *limg, bool del_image);
        /** Destructor */
        virtual ~v4d_BasicOverlayingModule ();

        ////////////////////////////////
        // Overlayer virtual methods
        ////////////////////////////////
        /** The pure virtual function intending for overlaying data with given buffer */
        void Overlay (size_t buf_id, Buffer2d<i3d::RGB> &buf) const;
        /** Change position for af or 3d view. */
        bool SetPosition (bool af, const wxString &pos, const wxRect *rect) 
        { return false; }
        /** Load 3d/af overlay */
        bool LoadOverlay (bool af, const i3d::Vector3d<size_t> &sz, const i3d::Offset &ofs, 
            const i3d::Resolution &res) 
        { return false; }
        /** Return true if the af/3d overlay image is empty */
		bool IsEmpty(bool af, size_t time) const 
		{ 
			return time >= m_limg->GetNumFrames() || GetLabeledImage(af, false, time)->IsEmpty(); 
		}
        
        /** Create/allocate buffer for given window type and return its ID */
        size_t Buffer_Create (v4d_OverlayWindowType win_type);
        /** Set buffer dimensions */
        virtual void Buffer_SetDim (size_t buf_id, const i3d::Vector3d<size_t> &sz, 
            const i3d::Offset &ofs, const i3d::Vector3d<float> &res);
        /** Render buffer image */
        virtual void Buffer_Render (size_t buf_id, size_t time);
		/** Return the buffer image value at given position. */
		virtual LABEL Buffer_GetValue(size_t buf_id, const i3d::Vector3d<size_t> &pos) const
		{
			return m_buf_vec[buf_id]->GetLayer(0).GetVoxel(pos) & ~GetHighestBit();
		}

    protected:        
        /** Get original 3d/af labeled image containing components/boundary. 
		Returns NULL if it is out of image count. */
        virtual const i3d::LabeledImage3d<LABEL,VOXEL> *GetLabeledImage 
            (bool af, bool boundary, size_t time) const 
		{
			if (time >= m_limg->GetNumFrames())
				return NULL;
			return m_limg->GetImage(time); 
		}

        /** Find and highlight boundary labels in given image */
        void ComputeBoundary (size_t buf_id);
        /** Get a number of type LABEL with highest bit set to one */
        static LABEL GetHighestBit () { return std::numeric_limits<LABEL>::max() / 2 + 1; }
	};

} // namespace v4d

#endif


