/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * dlg_exportsequence.cpp
 *
 * Implementation of the SUB-Dialog for setting the export of image sequence.
 *
 * Vladimr Ulman (ulman@fi.muni.cz) 2015
 */

#include "wdr/dlg_decor_wdr.h"
#include "dlg_exportsequence.h"

namespace v4d {

/*******************************************************************/

ExportSequenceDialog::ExportSequenceDialog(wxWindow *parent_, v4d_Viewer3dCanvas* _canvas)
	: wxPanel(parent_), parent(parent_), m_canvas(_canvas)
{
	//don't draw/show anything, it has already been created;
	//we just should take care of the created stuff

	//set up the pointers...
	m_saveSingle=(wxCheckBox*)parent->FindWindow(ID_SAVESINGLEIMAGE);
	m_firstFrame=(wxSpinCtrl*)parent->FindWindow(ID_SP_FIRSTFRAME_VALUE);
	m_lastFrame= (wxSpinCtrl*)parent->FindWindow(ID_SP_LASTFRAME_VALUE);
	m_stepFrame= (wxSpinCtrl*)parent->FindWindow(ID_SP_STEP_VALUE);
	m_radio_img=(wxRadioButton*)parent->FindWindow(ID_SP_FORMAT_IMAGES);
	m_radio_avi=(wxRadioButton*)parent->FindWindow(ID_SP_FORMAT_AVI);
	m_radio_gif=(wxRadioButton*)parent->FindWindow(ID_SP_FORMAT_GIF);
	m_lastline_label=(wxStaticText*)parent->FindWindow(ID_SP_OUTPUT_LABEL);
	m_lastline_value=(wxTextCtrl*)parent->FindWindow(ID_SP_OUTPUT_VALUE);

	//makes an internal copy of the image iterator
	m_ii=m_canvas->GetImageIterator();

	//set up internal data
	nameTemplate=wxString(m_ii.GetTemplateName());
	FPS=m_ii.GetFPS();

	//update the dialog with valid values
	//(assuming the images output format is on by default)
	m_saveSingle->SetValue(false);
	m_firstFrame->SetRange(m_ii.GetFirstIndex(), m_ii.GetLastIndex());
	m_firstFrame->SetValue(m_ii.GetFirstRangeIndex());
	m_lastFrame->SetRange(m_ii.GetFirstIndex(), m_ii.GetLastIndex());
	m_lastFrame->SetValue(m_ii.GetLastRangeIndex());
	m_stepFrame->SetRange(1,m_ii.GetRangeLength());
	m_stepFrame->SetValue(m_ii.GetStepRangeIndex());
	m_lastline_value->ChangeValue(nameTemplate);

	//attach handler to the three radio buttons
	parent->Bind(wxEVT_RADIOBUTTON,
			&ExportSequenceDialog::OnOutputFormatRadioCtrl,
			this,ID_SP_FORMAT_IMAGES,ID_SP_FORMAT_GIF);

	//attach handler to the "save single image" item
	parent->Bind(wxEVT_CHECKBOX,
			&ExportSequenceDialog::OnSingleImageCtrl,
			this,ID_SAVESINGLEIMAGE);

	//attach handler to the final OK button
	parent->Bind(wxEVT_BUTTON,
			&ExportSequenceDialog::OnOKbutton,
			this,wxID_OK);
}

ExportSequenceDialog::~ExportSequenceDialog()
{
	//the event handlers should live longer than the object
	//to which they are attached; if this is not the case,
	//the handlers must be Unbind(); and this is the case
	//since this class is a helper class inside a dialog
	//class to which our handlers are Bind()
	parent->Unbind(wxEVT_RADIOBUTTON,
			&ExportSequenceDialog::OnOutputFormatRadioCtrl,
			this,ID_SP_FORMAT_IMAGES,ID_SP_FORMAT_GIF);
	parent->Unbind(wxEVT_CHECKBOX,
			&ExportSequenceDialog::OnSingleImageCtrl,
			this,ID_SAVESINGLEIMAGE);
	parent->Unbind(wxEVT_BUTTON,
			&ExportSequenceDialog::OnOKbutton,
			this,wxID_OK);
}


void ExportSequenceDialog::ProcessAfterShowModal(
              const ExportData_CanvasProcessor& workhorse)
{
	if (this->m_saveSingle->GetValue()) //single image is checked
	{
		//selected to save just the single file that is currently visible
		wxFileDialog f_dlg(this, "Save single 3D image", wxEmptyString, wxEmptyString,
			"Image cytometry standard (*.ics)|*.ics|"
			"TIFF image (*.tif)|*.tif",
			wxFD_SAVE | wxFD_OVERWRITE_PROMPT | wxFD_CHANGE_DIR);

		if (f_dlg.ShowModal() == wxID_OK)
			workhorse(f_dlg.GetPath());
	}
	else
	{
		//selected to save the sequence
		try
		{
		 //should also backpropagate the ImageIterator setting to the m_canvas...
		 //(cause the OK button has been pressed = setting is confirmed)
		 this->CommitDialogToII();
		 m_canvas->UpdateImageIterator(m_ii);
		 //may also change content (viewed frame) of the canvas

		 //prepare the local ImageIterator copy for sweeping
		 m_ii.ResetForForwardSweep(true); //to keep in sync with m_canvas
		 m_ii.SetLoopMode(false); //to sweep just once
		}
		catch (const i3d::InternalException& e)
		{
		 wxMessageBox(e.what,wxString("Error processing the dialog"),
			wxOK | wxCENTRE | wxICON_EXCLAMATION);
		 return; //close the processing function without opening any file dialogs...
		}

		//suppose the setting is OKay now

		if (this->m_radio_img->GetValue()) //export as image sequence is checked
		{
			//save as a image sequence
			//selected to save sequence
			wxDirDialog f_dlg(this, "Choose folder to save 3D image sequence", wxEmptyString,
			                  wxDD_CHANGE_DIR | wxDD_DIR_MUST_EXIST);

			if (f_dlg.ShowModal() == wxID_OK)
			{
				wxString folder( f_dlg.GetPath() );
				folder+=wxFileName::GetPathSeparator();
				std::string file;
				do {
					m_ii.GetCurrentName(file);
					workhorse(folder+file);

					m_canvas->ShowNextImage();
					m_ii.AdvanceIndex();
				} while (m_ii.IndexWasUpdated());
				//should have the ShowNextImage() returned false here...
			}
		}
		else
		{
			//save as a video, single file...
			wxMessageBox(wxString("This function is not yet implemeted, sorry.\nYou may only save image sequence at the moment."),
							 wxString("Not implemented yet"));
		}
	}
}


bool ExportSequenceDialog::CheckLineForFPSvalue(void)
{
	//backup previous value (=FPS) before setting a new one (=nameTemplate)
	//(if we manage to convert it...)
	bool valueOK=false;

	long tmp;
	if ( (m_lastline_value->GetValue()).ToLong(&tmp) )
	{
		if (tmp < 1)
valueOK=(wxMessageBox(wxString("The FPS value must not be zero or negative.\
\n\nContinue with the previous value of the FPS instead?"),
wxString("Strange FPS"),wxYES_NO | wxCENTRE | wxICON_EXCLAMATION) == wxYES);
		else { FPS=tmp; valueOK=true; }
	}
	else
valueOK=(wxMessageBox(wxString("Unable to read the FPS which should be a \
positive integer number.\n\nContinue with the previous value of the FPS instead?"),
wxString("Strange FPS"),wxYES_NO | wxCENTRE | wxICON_EXCLAMATION) == wxYES);

	return (valueOK);
}

bool ExportSequenceDialog::CheckLineForTemplateString(void)
{

	//backup previous value (=nameTemplate) before setting a new one (=FPS)
	//(if we manage to convert it...)
	bool valueOK=false;

	wxString tmp=m_lastline_value->GetValue();
	if (tmp.Find('X') == wxNOT_FOUND)
valueOK=(wxMessageBox(wxString("The template name does not include \
a number-replacement character 'X'.\n\nContinue with the previous template name instead?"),
wxString("Strange teplate name"),wxYES_NO | wxCENTRE | wxICON_EXCLAMATION) == wxYES);
	else { nameTemplate=tmp; valueOK=true; }

	return (valueOK);
}


void ExportSequenceDialog::CommitDialogToII(void)
{
	m_ii.SetFirstRangeIndex(m_firstFrame->GetValue());
	m_ii.SetLastRangeIndex(m_lastFrame->GetValue());
	m_ii.SetStepRangeIndex(m_stepFrame->GetValue());

	//note that dialog "automatically" feeds the nameTemplate and FPS
	//atributes with valid values from the last line, so we should use these
	if (m_radio_img->GetValue())
	{
		//current output format was left at images,
		//hence, the template name string is considered
		//valid and FPS value is ignored
		m_ii.SetTemplateName(nameTemplate.ToStdString());
	}
	else
	{
		//vice verse...
		m_ii.SetFPS(FPS);
	}
}

void ExportSequenceDialog::OnOutputFormatRadioCtrl(wxCommandEvent& evt)
{
	//remembers last/previous state of the radio button
	static int PrevRadio=0; //0 images, 1 avi or 2 gif
	switch (evt.GetId())
	{
	case ID_SP_FORMAT_IMAGES:
		//if images were not selected previously and we are able to read
		//the value in the line at the bottom, than keep the change
		if ((PrevRadio != 0) && CheckLineForFPSvalue())
		{
			//also toggle the last line
			m_lastline_label->SetLabel(wxString("Name template:"));
			m_lastline_value->ChangeValue(nameTemplate);
			PrevRadio=0;
		}
		else
		{
			//uncheck this button and check the one checked before
			m_radio_img->SetValue(false);
			if (PrevRadio == 1) m_radio_avi->SetValue(true);
			else m_radio_gif->SetValue(true);
		}
		break;
	case ID_SP_FORMAT_AVI:
	case ID_SP_FORMAT_GIF:
		//switching from image radio button?
		if (PrevRadio == 0)
		{
			//yes, try to change the last line
			if (CheckLineForTemplateString())
			{
				//also toggle the last line
				m_lastline_label->SetLabel(wxString("FPS:"));
				m_lastline_value->ChangeValue(wxString() << FPS);
				PrevRadio=evt.GetId()-ID_SP_FORMAT_IMAGES;
			}
			else
			{
				//unsuccessful, revert radio buttons
				m_radio_img->SetValue(true);
				if (evt.GetId() == ID_SP_FORMAT_AVI) m_radio_avi->SetValue(false);
				else m_radio_gif->SetValue(false);
			}
		}
		//only remember the currently switched radio button
		else PrevRadio=evt.GetId()-ID_SP_FORMAT_IMAGES;
		break;
	}
}

void ExportSequenceDialog::OnSingleImageCtrl(wxCommandEvent& evt)
{
	//remembers last/previous state of the check item
	static bool PrevState=false; //false = not checked
	if (PrevState != evt.IsChecked())
	{
		//the item state has changed
		PrevState=evt.IsChecked();

		m_firstFrame->Enable(!PrevState);
		m_lastFrame->Enable(!PrevState);
		m_stepFrame->Enable(!PrevState);
		m_radio_img->Enable(!PrevState);
		m_radio_avi->Enable(!PrevState);
		m_radio_gif->Enable(!PrevState);
		m_lastline_label->Enable(!PrevState);
		m_lastline_value->Enable(!PrevState);
	}
}

void ExportSequenceDialog::OnOKbutton(wxCommandEvent& evt)
{
	if (this->m_saveSingle->GetValue() == false)
	{
		//single image is NOT checked, we need to check the bottom last line
		//whether it contains some valid input...
		if (m_radio_img->GetValue())
		{
			//image sequence is selected, check the template string
			//(as if we were to toggle to avi or gif buttons...)
			if (CheckLineForTemplateString() == false) return;
		}
		else
		{
			//avi or gif is selected, check the FPS value
			if (CheckLineForFPSvalue() == false) return;
		}
	}

	//if we got here, all is OKay, close the dialog then
	static_cast<wxDialog*>(parent)->EndModal(wxID_OK);
}

}
