/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * dlg_exportsequence.h
 *
 * SUB-Dialog for setting the export of image sequence.
 *
 * Vladimr Ulman (ulman@fi.muni.cz) 2015
 */

#ifndef V4D_DLG_EXPORTSEQUENCE_H
#define V4D_DLG_EXPORTSEQUENCE_H

#include <wx/checkbox.h>
#include <wx/spinctrl.h>
#include <wx/radiobut.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/panel.h>
#include <wx/valnum.h>
#include <wx/valtext.h>

#include "types/v4d_imageiterator.h"
#include "v4d_canvas3d.h"

namespace v4d 
{
	class ExportData_CanvasProcessor;

	/** The class that handles the Sequence parameters dialog. */
	class ExportSequenceDialog: public wxPanel
	{
	  public:
		/**
		 * Constructor: finds "its" controls and initiate them
		 * according to the ImageIterator found in the \e _canvas.
		 */
		ExportSequenceDialog(wxWindow *parent_, v4d_Viewer3dCanvas* _canvas);

        /** Destructor. */
		~ExportSequenceDialog();

		/**
		 * this is the common code that is always executed
		 * after: if (dlg.ShowModal() == wxID_OK)
		 *
		 * ... to avoid having the same code once for every control
		 * that can export a sequence of images
		 */
		void ProcessAfterShowModal(const ExportData_CanvasProcessor& workhorse);

		///handler for the three radio buttons
		void OnOutputFormatRadioCtrl(wxCommandEvent& evt);

		///handler for the "save single image"
		void OnSingleImageCtrl(wxCommandEvent& evt);

		///handler for the "OK button"
		void OnOKbutton(wxCommandEvent& evt);

	  private:
		/**
		 * Reads out current values from the dialog and
		 * saves them into the ImageIterator.
		 */
		void CommitDialogToII(void);

		///checks the "last line" whether input is valid template string
		bool CheckLineForTemplateString(void);
		///checks the "last line" whether input is valid FPS value
		bool CheckLineForFPSvalue(void);

	   ///used only in the destructor
		wxWindow* const parent;

		///pointers to the relevant controls
		wxCheckBox* m_saveSingle;
		wxSpinCtrl* m_firstFrame;
		wxSpinCtrl* m_lastFrame;
		wxSpinCtrl* m_stepFrame;
		wxRadioButton* m_radio_img;
		wxRadioButton* m_radio_avi;
		wxRadioButton* m_radio_gif;
		wxStaticText* m_lastline_label;
		wxTextCtrl* m_lastline_value;

		wxString nameTemplate;
		int FPS;

		///the reference to the ImageIterator
		i4d::ImageIterator m_ii;

		/** Pointer to the canvas object. */
		v4d_Viewer3dCanvas* const m_canvas;
	};


	/**
	 * The class that basically allows one to hook up different
	 * methods with various parameter lists in a uniform way
	 * (via the operator() method).
	 *
	 * background: 
	 * The ExportSequenceDialog::ShowModalAndProcess() method needs
	 * to call specific m_canvas method to actually export some image
	 * data -- in fact, the workhorse of the export itself.
	 *
	 * Since the m_canvas's workhorse methods have different parameters,
	 * we cannot (simply!) overload just one virtual function. We rather
	 * prepare derive and specialize this class: feed workhorse parameters
	 * via derived class'es parameter and call operator() to actually
	 * execute the workhorse on the parameters.
	 */
	class ExportData_CanvasProcessor
	{
	  public:
		///empty constructor
		ExportData_CanvasProcessor() {};

		///empty destructor (needs not be virtual)
		~ExportData_CanvasProcessor() {};

		virtual void operator() (const wxString& filename) const = 0;
	};

	///specialized class for the Save3d export dialog
	class ExportSave3dData_CanvasProcessor : public ExportData_CanvasProcessor
	{
	  public:
	  	///constructor: feed relevant required parameters
		ExportSave3dData_CanvasProcessor(v4d_Viewer3dCanvas* _c)
		 : m_canvas(_c) {};

		///do the export job
		virtual void operator() (const wxString& filename) const
		{ this->m_canvas->Save3dData(filename); }

	  private:
		/** Pointer to the canvas object. */
		v4d_Viewer3dCanvas* m_canvas;
	};

	///specialized class for the Save View export dialog
	class ExportSaveActualView_CanvasProcessor : public ExportData_CanvasProcessor
	{
	  public:
	  	///constructor: feed relevant required parameters
		ExportSaveActualView_CanvasProcessor(v4d_Viewer3dCanvas* _c,
			const int _d, const v4d_ProjectionType _p)
		 : m_canvas(_c), d(_d), projection(_p) {};

		///do the export job
		virtual void operator() (const wxString& filename) const
		{ this->m_canvas->SaveActualView(filename.mb_str(wxConvFile),
		                      d,projection); }

	  private:
		/** Pointer to the canvas object. */
		v4d_Viewer3dCanvas* m_canvas;

		int d;
		v4d_ProjectionType projection;
	};

	///specialized class for the Save Gallery export dialog
	class ExportSaveGalleryView_CanvasProcessor : public ExportData_CanvasProcessor
	{
	  public:
	  	///constructor: feed relevant required parameters
		ExportSaveGalleryView_CanvasProcessor(v4d_Viewer3dCanvas* _c,
			const int _start, const int _end,
			const int _step, const bool _fR,
			const int _count, const int _spacing)
		 : m_canvas(_c), start(_start), end(_end),
		   step(_step), fixedRows(_fR),
			count(_count), spacing(_spacing) {};

		///do the export job
		virtual void operator() (const wxString& filename) const
		{ this->m_canvas->SaveGalleryView(filename.mb_str(wxConvFile),
		                      start,end,step,fixedRows,count,spacing); }

	  private:
		/** Pointer to the canvas object. */
		v4d_Viewer3dCanvas* m_canvas;

		int start, end, step;
		bool fixedRows;
		int count, spacing;
	};
}

#endif
