/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * dlg_levels.cpp
 *
 * Implementation of levels dialog
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006, based on Jan Hubeny code (xhubeny@fi.muni.cz) 2004
 */

#include "v4d_basic.h"
#include <math.h>
#include <algorithm>
#include <wx/valgen.h>
#include "dlg_levels.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

#ifdef _MSC_VER
	inline double trunc(double v) { return floor(v); }
	inline double roundL(double v) { return (double) ((v >= 0.0f) ? ((int) (v + 0.5f)) : ((int) (v - 0.5f))); }
	#define ROUND roundL	
#else
	#define ROUND round
#endif

#define ID_MINMAX_MAX 20000
#define ID_MINMAX_MIN 20001
#define ID_GAMMA 20002

namespace v4d
{
    wxDEFINE_EVENT(v4d_EVT_VARIABLE_TEXTCTRL, v4d_VariableTextCtrlEvent);
    wxDEFINE_EVENT(v4d_EVT_LEVELS_PREVIEW, v4d_LevelsEvent);
    wxDEFINE_EVENT(v4d_EVT_TRIANGLE, v4d_TriangleEvent);
    wxDEFINE_EVENT(v4d_EVT_MINMAXCTRL, v4d_MinMaxCtrlEvent);

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                      GradStripe class                       //
    //                                                             //
    /////////////////////////////////////////////////////////////////

    BEGIN_EVENT_TABLE(GradStripe, wxPanel)
        EVT_PAINT(GradStripe::OnPaint)
        END_EVENT_TABLE()

        /*******************************************************************/
        GradStripe::GradStripe(wxWindow *parent, bool draw_gradient, const wxColour &min_colour, const wxColour &max_colour, 
        const wxColour &center_colour, int numberOfPos, double gamma, wxWindowID id, const wxPoint &pos, 
        const wxSize &size, long style)
        : wxPanel(parent, id, pos, size, style), _draw_gradient(draw_gradient), _min_colour(min_colour),
        _max_colour(max_colour), _center_colour(center_colour), _gamma(gamma), _numberOfPos(numberOfPos), 
        _minPos(0), _maxPos(numberOfPos - 1)
    {
    }

    /*******************************************************************/
    inline void GradStripe::SetColours(const wxColour &min_colour, const wxColour &max_colour, const wxColour &center_colour)
    {
        _min_colour = min_colour;
        _max_colour = max_colour;
        _center_colour = center_colour;
    }

    /*******************************************************************/
    inline void GradStripe::Reset()
    {
        _gamma = 1.0;
        _minPos = 0;
        _maxPos = _numberOfPos - 1;
        Refresh(FALSE);
    }

    /*******************************************************************/
    void GradStripe::DrawStripe(wxDC &dc, wxCoord x, wxCoord y, wxCoord width, wxCoord height, const wxColour &colour)
    {
        wxPen pen(colour, 1, wxPENSTYLE_SOLID);
        wxBrush brush(colour, wxBRUSHSTYLE_SOLID);

        dc.SetPen(pen);
        dc.SetBrush(brush);
        dc.DrawRectangle(x, y, width, height);
    }

    /*******************************************************************/
    void GradStripe::OnPaint(wxPaintEvent &evt)
    {
        int cw, ch; // client width and height

        wxPaintDC dc(this);
        GetClientSize(&cw, &ch);

        wxBitmap *bmp = new wxBitmap(cw, ch);
        wxMemoryDC memDC;
        memDC.SelectObject(*bmp);

        int start_x = 0, end_x = cw;
        float ratio = (float) cw / (float) _numberOfPos;
        float pos_x;

        // Draw the stripes in the space between border points and _minPos and _maxPos
        if (_minPos <= _maxPos)
        {
            // Draw the rectangle between 0 -- _minPos
            if (_minPos != 0) 
            {
                pos_x = (float) _minPos * ratio;
                DrawStripe(memDC, 0, 0, (int) ceil(pos_x), ch, _min_colour);
                start_x = (int) trunc(pos_x);
            }

            // Draw the rectangle between _maxPos -- (_numberOfPos - 1)
            if (_maxPos != (_numberOfPos - 1)) 
            {
                pos_x = (float) _maxPos * ratio;
                DrawStripe(memDC, (int) trunc(pos_x), 0, (int) ceil((float) cw - pos_x), ch, _max_colour);
                end_x = (int) ceil(pos_x);
            }
        }
        else
        {
            // Draw the rectangle between 0 -- _maxPos
            if (_maxPos != 0) 
            {
                pos_x = (float) _maxPos * ratio;
                DrawStripe(memDC, 0, 0, (int) ceil(pos_x), ch, _max_colour);
                start_x = (int) trunc(pos_x);
            }

            // Draw the rectangle between _minPos -- (_numberOfPos - 1)
            if (_minPos != (_numberOfPos - 1)) 
            {
                pos_x = (float) _minPos * ratio;
                DrawStripe(memDC, (int) trunc(pos_x), 0, (int) ceil((float) cw - pos_x), ch, _min_colour);
                end_x = (int) ceil(pos_x);
            }
        }

        if (_draw_gradient)
        {
            int i = 1, length = end_x - start_x, pos;
            float value;
            unsigned char r, g, b;
            wxPen pen(_minPos, 1, wxPENSTYLE_SOLID);
            wxBrush brush(_minPos, wxBRUSHSTYLE_SOLID);

            // Draw the gradient stripe in the space between _minPos and _maxPos 
            while (i <= length)
            {
                value = pow(((double) (i - 1) / (double) length), (double) (1.0 / _gamma));
                r = (unsigned char) ROUND(((double) _center_colour.Red()) * value);
                g = (unsigned char) ROUND(((double) _center_colour.Green()) * value);
                b = (unsigned char) ROUND(((double) _center_colour.Blue()) * value);

                pen.SetColour(r, g, b);
                brush.SetColour(r, g, b);
                memDC.SetPen(pen);
                memDC.SetBrush(brush);

                if (_minPos <= _maxPos)
                {
                    pos = start_x + i;
                }
                else
                {
                    pos = end_x - i;
                }

                memDC.DrawLine(pos, 0, pos, ch);
                i++;
            }
        }
        else
        {
            // Draw the rectangle between _minPos -- _maxPos
            DrawStripe(memDC, start_x, 0, end_x - start_x + 1, ch, _center_colour);
        }

        dc.Blit(0, 0, cw, ch, &memDC, 0, 0, wxCOPY);
        memDC.SelectObject(wxNullBitmap);
        delete bmp;
    }

    /*******************************************************************/

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                       HistGraph class                       //
    //                                                             //
    /////////////////////////////////////////////////////////////////

    /*******************************************************************/
    template <class T> 
    HistGraph<T>::HistGraph(wxWindow *parent, wxWindowID id, v4d::Histogram<T> *H, 
        const wxPoint &position, const wxSize &size, long style)
        : wxPanel(parent, id, position, size, style), _Histogram(H), _bgColour(wxColour(255, 255, 255)), 
        _fgColour(wxColour(255, 0, 0)), _gridColour(wxColour(158, 154, 145)), _histType(HIST_LOG), _gridCLines(10), 
        _gridRLines(5), _border_width(4), _border_height(4)
    {
        Bind(wxEVT_PAINT, &HistGraph<T>::OnPaint, this);
    }

    /*******************************************************************/
    template <class T> template <class Function> 
    void HistGraph<T>::DrawHistGraph(wxDC &dc, const size_t *begin, const size_t *end, Function func)
    {
        int cw, ch; // client width and height
        dc.GetSize(&cw, &ch);

        int num = end - begin;//data.size();
        size_t max = *std::max_element(begin, end);
        float width = (float) (cw - _border_width) / (float) num;
        int rh;

        const size_t *data_ptr = begin;
        int i = 0;

        //for (int i = 0 ; i < num; i++)
        while (data_ptr <= end)
        {
            //if (data[i] != 0)
            if (*data_ptr != 0 )
            {
                rh = ((int) (ROUND(((double) (ch - _border_height)) * func((double) /*data[i]*/*data_ptr) / func((double) max))));

                if (rh == 0)
                    rh++;
            }
            else
            {
                rh = 0;
            }

            dc.DrawRectangle((int) (trunc(((float) i) * width)) +  _border_width / 2, ch - rh - _border_height / 2, 
                (int) (ceil(width)), rh);

            data_ptr++;
            i++;
        }
    }

    /*******************************************************************/
    template <class T> void HistGraph<T>::OnPaint(wxPaintEvent &evt)
    {
        int cw, ch; // client width and height

        wxPaintDC dc(this);
        GetClientSize(&cw, &ch);

        wxBitmap *bmp = new wxBitmap(cw, ch);
        wxMemoryDC memDC;
        memDC.SelectObject(*bmp);

        // Draw the background rectangle
        wxPen pen(_bgColour, 1, wxPENSTYLE_SOLID);
        wxBrush brush(_bgColour, wxBRUSHSTYLE_SOLID);
        memDC.SetBrush(brush);
        memDC.SetPen(pen);
        memDC.DrawRectangle(0, 0, cw, ch);

        // Draw the bounding rectangle
        pen.SetColour(_gridColour);
        memDC.SetPen(pen);
        memDC.DrawRectangle(_border_width / 4, _border_height / 4, cw - _border_width / 2 , ch - _border_height / 2); 

        // Draw column grid lines
        int width = (int) (ROUND((float) (cw - _border_width / 2) / (float) _gridCLines)) + _border_width / 4;
        for(int i = 1; i < _gridCLines; i++)
            memDC.DrawLine(i * width, _border_height / 4, i * width, ch - _border_height / 4);

        // Draw row grid lines
        int height = (int) (ROUND((float) (ch - _border_height / 2) / (float) _gridRLines)) + _border_height / 4;
        for(int i = 1; i < _gridRLines; i++)
            memDC.DrawLine(_border_width / 4, i * height, cw - _border_width / 4, i * height);

        // Draw the histogram
        pen.SetColour(_fgColour);
        brush.SetColour(_fgColour);
        memDC.SetBrush(brush);
        memDC.SetPen(pen);

        if (_histType == HIST_LINEAR)
            DrawHistGraph(memDC, _Histogram -> begin(), _Histogram -> end(), LinearHist());
        else
            DrawHistGraph(memDC, _Histogram -> begin(), _Histogram -> end(), LogarithmicHist());

        dc.Blit(0, 0, cw, ch, &memDC, 0, 0, wxCOPY);
        memDC.SelectObject(wxNullBitmap);
        delete bmp;
    }

    /*******************************************************************/

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                  VariableTextCtrl class                     //
    //                                                             //
    /////////////////////////////////////////////////////////////////

    /*******************************************************************/
    template <class T> VariableTextCtrl<T>::VariableTextCtrl(wxWindow *parent, wxWindowID id, T *variable, 
        T min_val, T max_val, T def_val, int num_pos, const wxPoint &pos, const wxSize &size, long style)
        : wxTextCtrl(parent, id, wxEmptyString, pos, wxSize(40, -1), wxTE_PROCESS_ENTER | wxTE_RIGHT), 
        _variable(variable), _min_val(min_val), _max_val(max_val), _def_val(def_val), _num_pos(num_pos)
    {
        //HELP \todo napsat obecne - stejne tak pro velikost controlu
        SetMaxLength(10);

        UpdateIncrement();

        Bind(wxEVT_KILL_FOCUS, &VariableTextCtrl<T>::OnCorrectFE, this);
        Bind(wxEVT_COMMAND_TEXT_ENTER, &VariableTextCtrl<T>::OnCorrectCE, this);
        Bind(wxEVT_CHAR, &VariableTextCtrl<T>::OnChar, this);
    }

    /*******************************************************************/
    template <class T> void VariableTextCtrl<T>::SetVariableParams(T  *variable, T min_val, T max_val, T def_val, int num_pos)
    {
        _variable = variable;
        _min_val = min_val;
        _max_val = max_val;
        _def_val = def_val;
        _num_pos = num_pos;

        SetVariableValidator();
        UpdateIncrement();
    }

    /*******************************************************************/
    template <class T> int VariableTextCtrl<T>::GetVariablePos()
    {
        T val = _min_val;
        GetVariableValue(val);

        return static_cast<int>((val - _min_val) / _increment);
    }

    /*******************************************************************/
    template <class T> T VariableTextCtrl<T>::GetValue(int pos)
    {
        //HELP \todo test intervalu platnych hodnot
        return (_min_val + pos * _increment);
    }

    /*******************************************************************/
    template <class T> void VariableTextCtrl<T>::Reset()
    {
        SetVariableValue(_def_val);
    }

    /*******************************************************************/
    template <class T> void VariableTextCtrl<T>::SetVariableValidator()
    {
        SetValidator(IntValidator<T>(*_variable, _min_val, _max_val));
    }

    /*******************************************************************/
    template <> void VariableTextCtrl<float>::SetVariableValidator()
    {
        SetValidator(FloatValidator(*_variable, _min_val, _max_val));
    }

    /*******************************************************************/
    template <> void VariableTextCtrl<double>::SetVariableValidator()
    {
        SetValidator(DoubleValidator(*_variable, _min_val, _max_val));
    }

    /*******************************************************************/
    template <class T> wxString VariableTextCtrl<T>::VariableToString(T value)
    {
        return wxString::Format("%d", value);
    }

    /*******************************************************************/
    template <> wxString VariableTextCtrl<float>::VariableToString(float value)
    {
        return wxString::Format("%2.2f", value);
    }

    /*******************************************************************/
    template <> wxString VariableTextCtrl<double>::VariableToString(double value)
    {
        return wxString::Format("%2.2g", value);
    }

    /*******************************************************************/
    template <class T> bool VariableTextCtrl<T>::StringToVariable(const wxString &str, T &value)
    {
        long tmp;
        if (str.ToLong(&tmp))
        {
            if (tmp >= static_cast<long>(std::numeric_limits<T>::min()) && tmp <= static_cast<long>(std::numeric_limits<T>::max()))
            {
                value = tmp;
                return true;
            }

            return false;
        }

        return false;
    }

    /*******************************************************************/
    template <> bool VariableTextCtrl<float>::StringToVariable(const wxString &str, float &value)
    {
        double tmp;
        if (str.ToDouble(&tmp))
        {
            value = tmp;
            return true;
        }

        return false;
    }

    /*******************************************************************/
    template <> bool VariableTextCtrl<double>::StringToVariable(const wxString &str, double &value)
    {
        return str.ToDouble(&value);
    }

    /*******************************************************************/
    template <class T> void VariableTextCtrl<T>::UpdateIncrement()
    {
        _increment = (_max_val - _min_val + 1) / _num_pos;
    }

    /*******************************************************************/
    template <> void VariableTextCtrl<float>::UpdateIncrement()
    {
        _increment = (_max_val - _min_val) / _num_pos;
    }

    /*******************************************************************/
    template <> void VariableTextCtrl<double>::UpdateIncrement()
    {
        _increment = (_max_val - _min_val) / _num_pos;
    }

    /*******************************************************************/
    template <class T> void VariableTextCtrl<T>::Refresh(T value)
    {
        v4d_VariableTextCtrlEvent e(v4d_EVT_VARIABLE_TEXTCTRL, static_cast<double>(value), GetId());
        GetParent() -> ProcessWindowEvent(e);
    }

    /*******************************************************************/
    template <class T> void VariableTextCtrl<T>::Correct(wxEvent &evt)
    {
        T value;

        if (IsModified())
        {
            if (StringToVariable(wxTextCtrl::GetValue(), value))
            {
                if (value < _min_val)
                {
                    value = _min_val;
                }

                if (value > _max_val)
                {
                    value = _max_val;
                }
            }
            else
            {
                //HELP \todo tezko rict co sem dat
                value = *_variable;
            }

            Refresh(value);
        }
    }

    /*******************************************************************/
    template <class T> void VariableTextCtrl<T>::OnChar(wxKeyEvent &evt)
    {
        T value=0;

        if (StringToVariable(wxTextCtrl::GetValue(), value))
        {
            switch (evt.GetKeyCode())
            {
            case WXK_UP:

                if (value < _max_val)	
                {
                    value += _increment;

                    if (value > _max_val) 
                    {
                        value = _max_val;
                    }

                    Refresh(value);
                }		

                break;

            case WXK_DOWN:

                if (value > _min_val)
                {
                    value -= _increment;

                    if (value < _min_val)
                    {
                        value = _min_val;
                    }

                    Refresh(value);
                }

                break;

            default:

                evt.Skip();
            }
        }
        else
        {
            evt.Skip();
        }
    }

    /*******************************************************************/

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                       Triangles class                       //
    //                                                             //
    /////////////////////////////////////////////////////////////////

    BEGIN_EVENT_TABLE(Triangles, wxPanel)
        EVT_PAINT(Triangles::OnPaint)
        EVT_MOUSE_EVENTS(Triangles::OnMouse)
    END_EVENT_TABLE()

    /*******************************************************************/
    Triangles::Triangles(wxWindow *parent, wxWindow *evt_handler, bool input, bool can_cross, 
        const wxColour &min_colour, const wxColour &max_colour, int numOfPos, 
        double gamma, wxWindowID id, const wxPoint &pos, const wxSize &size, long style)
        : wxPanel(parent, id, pos, size, style), _numOfPos(numOfPos), _min_pos(0), _max_pos(numOfPos - 1), 
        _gamma(gamma), _input(input), _can_cross(can_cross), _evt_handler(evt_handler), 
        _min_colour(min_colour), _max_colour(max_colour)
    {
        _gamma_colour = wxColour((max_colour.Red() - min_colour.Red()) / 2,
            (max_colour.Green() - min_colour.Green()) / 2,
            (max_colour.Blue() - min_colour.Blue()) / 2);
        if (input)
        {
            RecomputeGammaPos();
        }
    }

    /*******************************************************************/
    void Triangles::RecomputeGammaPos()
    {
        _gamma_pos = (_min_pos + (int) (ROUND(((float) (_max_pos - _min_pos) / 2.0) * (1 - log10f(_gamma)))));
    }

    /*******************************************************************/
    void Triangles::Reset()
    {
        _min_pos = 0;
        _max_pos = _numOfPos - 1;

        if (_input)
        {
            _gamma = 1.0;	
            RecomputeGammaPos();
        }

        Refresh(FALSE);
    }

    /*******************************************************************/
    inline int Triangles::GetAbsoluteTrianglePosition(int pos, int cw)
    {
        float width = (float) (cw - 4) / (float) (_numOfPos - 1);

        return (int) ((float) pos * width + 2);
    };

    /*******************************************************************/
    void Triangles::DrawTriangle(wxDC &dc, int xpos, const wxColour &brush_colour, wxWindow *win)
    {
        wxPen pen;
        wxBrush brush;

        if (win == NULL)
        {
            // Set given colours
            pen.SetColour(brush_colour);
            brush.SetColour(brush_colour);
        }
        else
        {
            // Set clearing colours
            pen.SetColour(win -> GetBackgroundColour());
            brush.SetColour(win -> GetBackgroundColour());
        }

        dc.SetPen(pen);
        dc.SetBrush(brush);

        for (int y = 0; y < TRIANGLE_HEIGHT; y++)
            dc.DrawLine(xpos - y / 2, y, xpos + y / 2, y);

        if (win == NULL)		
        {	
            pen.SetColour(0,0,0);
            brush.SetColour(0,0,0);
        }

        dc.SetPen(pen);
        dc.SetBrush(brush);

        dc.DrawLine(xpos, 0, xpos - (TRIANGLE_HEIGHT - 1) / 2, TRIANGLE_HEIGHT - 1);
        dc.DrawLine(xpos, 0, xpos + (TRIANGLE_HEIGHT - 1) / 2,  TRIANGLE_HEIGHT - 1);
        dc.DrawLine(xpos - (TRIANGLE_HEIGHT - 1) / 2 , TRIANGLE_HEIGHT - 1, 
            xpos + (TRIANGLE_HEIGHT - 1) / 2 + 1, TRIANGLE_HEIGHT - 1);
    }

    /*******************************************************************/
    void Triangles::OnPaint(wxPaintEvent &ent)
    {
        int cw, ch; // client width and height

        wxPaintDC dc(this);
        GetClientSize(&cw, &ch);

        wxColour colour = GetBackgroundColour();

        wxBitmap *bmp = new wxBitmap(cw,ch);

        wxMemoryDC memDC;
        memDC.SelectObject(*bmp);
        memDC.SetPen(wxPen(colour, 1, wxPENSTYLE_SOLID));
        memDC.SetBrush(wxBrush(colour, wxBRUSHSTYLE_SOLID));
        memDC.DrawRectangle(0,0,cw,ch);

        // Draw triangle for minimal level
        DrawTriangle(memDC, GetAbsoluteTrianglePosition(_min_pos, cw), _min_colour);

        // Draw triangle for gamma variable
        if (_input)
        {
            DrawTriangle(memDC, GetAbsoluteTrianglePosition(_gamma_pos, cw), _gamma_colour);
        }

        // Draw triangle for maximal level
        DrawTriangle(memDC, GetAbsoluteTrianglePosition(_max_pos, cw), _max_colour);


        dc.Blit(0, 0, cw, ch, &memDC, 0, 0, wxCOPY);  
        memDC.SelectObject(wxNullBitmap);
        delete bmp;
    }

    /*******************************************************************/
    void Triangles::OnMouse(wxMouseEvent &evt)
    {
        int cw, ch; // client width and height
        long int x, y; // mouse position
        int pos;

        GetClientSize(&cw, &ch);
        evt.GetPosition(&x, &y);

        if (evt.m_leftDown || evt.m_rightDown || evt.m_middleDown) 
        {
            pos = (int) (((float) (x - 2) / (float) (cw - 2)) * (float) _numOfPos); 
            pos = std::min(pos, _numOfPos - 1);
            pos = std::max(pos, 0);

            if (!evt.Moving() && evt.ButtonDown()) 
            {
                // Not moving -- only button down we must find the regular control
                CaptureMouse();

                if (_input)
                {
                    if (std::abs(_min_pos - pos) <= std::abs(_gamma_pos - pos))
                    { 
                        _actualButton = MIN_POS;

                        if (((_min_pos - pos) < 0) && (_gamma_pos == _min_pos))
                        {
                            _actualButton = GAMMA_POS;
                        }
                    }
                    else
                    {
                        if (std::abs(_gamma_pos - pos) <= std::abs(_max_pos - pos))
                        {
                            _actualButton = GAMMA_POS;

                            if (((_gamma_pos - pos) < 0) && (_gamma_pos == _max_pos))
                            {
                                _actualButton = MAX_POS;
                            }
                        }
                        else
                        {
                            _actualButton = MAX_POS;
                        }
                    }
                }
                else
                {
                    if (std::abs(_max_pos - pos) < std::abs(_min_pos - pos))
                    {
                        _actualButton = MAX_POS;
                    }
                    else
                    {
                        _actualButton = MIN_POS;
                    }
                }
            }

            switch (_actualButton)
            {
            case MAX_POS:

                if (pos != _max_pos)
                {
                    if (!_can_cross)
                    {
                        pos = std::max(pos, _min_pos);
                    }

                    v4d_TriangleEvent e(v4d_EVT_TRIANGLE, pos, MAX_POS);
                    _evt_handler -> ProcessWindowEvent(e);
                }

                break;

            case MIN_POS:

                if (pos != _min_pos)
                {
                    if (!_can_cross)
                    {
                        pos = std::min(pos, _max_pos);
                    }

                    v4d_TriangleEvent e(v4d_EVT_TRIANGLE, pos, MIN_POS);
                    _evt_handler -> ProcessWindowEvent(e);
                }

                break;

            case GAMMA_POS:

                if (pos >= _min_pos && pos <= _max_pos)
                {
                    v4d_TriangleEvent e(v4d_EVT_TRIANGLE, pos, GAMMA_POS);
                    _evt_handler -> ProcessWindowEvent(e);
                }

                break;
            }
        }
        else
        {
            _actualButton = 1000;

            if (HasCapture())
            {
                ReleaseMouse();
            }
        }
    }

    /*******************************************************************/

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                   VariableMinMaxCtrl class                  //
    //                                                             //
    /////////////////////////////////////////////////////////////////

    /*******************************************************************/
    template <class T> VariableMinMaxCtrl<T>::VariableMinMaxCtrl(wxWindow *parent, wxWindowID id, bool input, 
        bool can_cross, bool draw_gradient, const wxColour &min_colour, const wxColour &max_colour,
        const wxColour &center_colour, const wxPoint &pos, const wxSize &size, long style)
        : wxPanel(parent, id, pos, size, style), _input(input), _can_cross(can_cross)
    {
        wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);

        wxPanel *zanoreny = new wxPanel(this, -1, wxDefaultPosition, wxSize(100, 60), 
            wxCLIP_CHILDREN | wxSUNKEN_BORDER | wxNO_FULL_REPAINT_ON_RESIZE);

        wxFlexGridSizer *vnitrni = new wxFlexGridSizer(1, 0, 0);
        vnitrni -> AddGrowableCol(0);

        _gradStripeUpper = new GradStripe(zanoreny, draw_gradient, min_colour, max_colour, center_colour);
        _gradStripeUpper -> Show(TRUE);

        _gradStripeLower = new GradStripe(zanoreny, true, min_colour, max_colour, center_colour);
        _gradStripeLower -> Show(TRUE);   

        vnitrni -> Add(_gradStripeUpper, 0, wxGROW | wxALIGN_CENTER_VERTICAL | wxLEFT | wxRIGHT, 2);
        vnitrni -> Add(_gradStripeLower, 0, wxGROW | wxALIGN_CENTER_VERTICAL | wxLEFT | wxRIGHT, 2);

        _triangles = new Triangles(zanoreny, this, input, can_cross, min_colour, center_colour);
        vnitrni -> Add(_triangles, 0, wxGROW | wxALIGN_CENTER_VERTICAL | wxBOTTOM, 0);

        zanoreny -> SetSizerAndFit(vnitrni);
        sizer -> Add(zanoreny, 0, wxGROW | wxALIGN_CENTER_VERTICAL, 5);

        wxGridSizer *grid = new wxGridSizer(1, 0, 0, 0);

        _min_ctrl = new VariableTextCtrl<T>(this, ID_MINMAX_MIN);
        grid -> Add(_min_ctrl, 0, wxALIGN_CENTER_VERTICAL | wxTOP | wxBOTTOM | wxRIGHT, 5);

        if (_input)
        {
            _gamma_ctrl = new VariableTextCtrl<double>(this, ID_GAMMA);
            grid -> Add(_gamma_ctrl, 0, wxALIGN_CENTER | wxTOP | wxBOTTOM, 5);
        }
        else
        {
            grid -> Add(40, -1, 0, wxALIGN_CENTER | wxTOP | wxBOTTOM, 5);
        }

        _max_ctrl = new VariableTextCtrl<T>(this, ID_MINMAX_MAX);
        grid -> Add(_max_ctrl, 0, wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL | wxLEFT | wxTOP | wxBOTTOM, 5);

        sizer -> Add(grid, 0, wxGROW | wxALIGN_CENTER_VERTICAL, 5);
        SetSizerAndFit(sizer);

        Bind(v4d_EVT_VARIABLE_TEXTCTRL, &VariableMinMaxCtrl<T>::OnVariableChange, this);
        Bind(v4d_EVT_TRIANGLE, &VariableMinMaxCtrl<T>::OnTriangleChange, this);
    }

    /*******************************************************************/
    template <class T> void VariableMinMaxCtrl<T>::SetParams(T *min_var, T *max_var, double *gamma_var, 
        T min_value, T max_value, int num_steps)
    { 
        if (_min_ctrl -> HasVariable() && _max_ctrl -> HasVariable())
        {
            TransferDataFromWindow();
        }

        _min_ctrl -> SetVariableParams(min_var, min_value, max_value, min_value, num_steps);
        _max_ctrl -> SetVariableParams(max_var, min_value, max_value, max_value, num_steps);

        if (_input)
        {
            _gamma_ctrl -> SetVariableParams(gamma_var, 0.1, 10.0, 1.0, 99);
        }

        _gradStripeUpper -> SetNumberOfPos(num_steps);
        _triangles -> SetNumberOfPos(num_steps);

        UpdateMin(*min_var);
        UpdateMax(*max_var);

        if (_input)
        {
            UpdateGamma(*gamma_var);
        }
    }

    /*******************************************************************/
    template <class T> inline void VariableMinMaxCtrl<T>::Reset()
    { 
        _min_ctrl -> Reset();
        _max_ctrl -> Reset();

        if (_input)
        {
            _gamma_ctrl -> Reset();
        }

        _gradStripeUpper -> Reset();
        _triangles -> Reset();

        TransferDataFromWindow();
    }

    /*******************************************************************/
    template <class T> inline void VariableMinMaxCtrl<T>::UpdateMin(T m) 
    {
        int pos;

        _min_ctrl -> SetVariableValue(m);
        pos = _min_ctrl -> GetVariablePos();

        _gradStripeUpper -> SetMinPos(pos);
        _gradStripeUpper -> Refresh(FALSE);

        _triangles -> SetMinPos(pos);

        if (_input)
        {
            _triangles -> RecomputeGammaPos();
        }

        _triangles -> Refresh(FALSE);
    }

    /*******************************************************************/
    template <class T> inline void VariableMinMaxCtrl<T>::UpdateMax(T m) 
    {
        int pos;

        _max_ctrl -> SetVariableValue(m);
        pos = _max_ctrl -> GetVariablePos();

        _gradStripeUpper -> SetMaxPos(pos);
        _gradStripeUpper -> Refresh(FALSE);

        _triangles -> SetMaxPos(pos);

        if (_input)
        {
            _triangles -> RecomputeGammaPos();
        }

        _triangles -> Refresh(FALSE);
    }

    /*******************************************************************/
    template <class T> inline void VariableMinMaxCtrl<T>::UpdateGamma(double g) 
    {
        if (_input)
        {
            _gamma_ctrl -> SetVariableValue(g);

            _gradStripeUpper -> SetGamma(g);
            _gradStripeUpper -> Refresh(FALSE);

            _triangles -> SetGamma(g);
            _triangles -> Refresh(FALSE);
        }
    }

    /*******************************************************************/
    template <class T> void VariableMinMaxCtrl<T>::GeneratePreviewEvent()
    {
        if (_min_ctrl -> HasVariable() && _max_ctrl -> HasVariable())
        {
            TransferDataFromWindow();

            v4d_MinMaxCtrlEvent evt(v4d_EVT_MINMAXCTRL);	
            GetParent() -> ProcessWindowEvent(evt);
        }
    }

    /*******************************************************************/
    template <class T> void VariableMinMaxCtrl<T>::OnVariableChange(v4d_VariableTextCtrlEvent &evt)
    {
        switch (evt.GetId())
        {
        case ID_MINMAX_MIN:

				{
            T min = static_cast<T>(evt.GetValue());
				T max=min;
            _max_ctrl -> GetVariableValue(max);

            if (!_can_cross)
            {
                min = std::min(min, max);
            }

            UpdateMin(min);
            GeneratePreviewEvent();
				}
            break;

        case ID_MINMAX_MAX:

				{
            T max = static_cast<T>(evt.GetValue());
				T min=max;
            _min_ctrl -> GetVariableValue(min);

            if (!_can_cross)
            {
                max = std::max(max, min);
            }

            UpdateMax(max);
            GeneratePreviewEvent();
				}
            break;

        case ID_GAMMA:

            UpdateGamma(evt.GetValue());
            GeneratePreviewEvent();
            break;
        }
    }

    /*******************************************************************/
    template <class T> void VariableMinMaxCtrl<T>::OnTriangleChange(v4d_TriangleEvent &evt)
    {
        int min_pos = _triangles -> GetMinPos();
        int max_pos = _triangles -> GetMaxPos();
        float center = (max_pos - min_pos) / 2.0;
        double gamma, gamma_min, gamma_max;

        switch (evt.GetId())
        {
        case MIN_POS:

            UpdateMin(_min_ctrl -> GetValue(evt.GetValue()));
            GeneratePreviewEvent();
            break;

        case MAX_POS:

            UpdateMax(_max_ctrl -> GetValue(evt.GetValue()));
            GeneratePreviewEvent();
            break;

        case GAMMA_POS:

            _gamma_ctrl -> GetRange(gamma_min, gamma_max);
            gamma = powf(gamma_max, ((float) (min_pos + ROUND(center) - evt.GetValue()) / center));
            if (gamma > gamma_max)
            {
                gamma = gamma_max;
            }
            else
            {
                std::max(gamma, gamma_min);
            }

            UpdateGamma(gamma);
            GeneratePreviewEvent();

            break;
        }
    }

    /*******************************************************************/

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                     LevelsWidget class                      //
    //                                                             //
    /////////////////////////////////////////////////////////////////

    /*******************************************************************/
    template <class ITYPE, class OTYPE> 
    LevelsWidget<ITYPE, OTYPE>::LevelsWidget(wxWindow *parent, wxWindowID id, 
        const std::vector<v4d::Histogram<ITYPE> *> &Histdata_vector, 
        const std::vector<TLevelsParams<ITYPE, OTYPE> > &params_vector, 
        const wxArrayString &channel_names, const wxString &title, const wxPoint &position, const wxSize& size, long style)
        : wxDialog(parent, id, title, position, size,  style), _HistData_vector(Histdata_vector), All_params(0), All_histdata(0) 
    {
        wxASSERT(_HistData_vector.size());
        wxASSERT(_HistData_vector.size() == params_vector.size());
        wxASSERT(_HistData_vector.size() == channel_names.GetCount());

        _LevelsParams_vector.clear();
        for (size_t i = 0; i < params_vector.size(); i++)
        {
            _LevelsParams_vector.push_back(params_vector[i]);
        }

        _hist = new HistGraph<ITYPE>(this, ID_LEV_HIST, _HistData_vector[0]);
        _hist -> SetGridLines(8, 4);
        _hist -> SetFgColour(wxColour(0, 0, 0));

        _minmaxi = new VariableMinMaxCtrl<ITYPE>(this, ID_MINMAXCTRLI, true, false, true, *wxBLACK, *wxWHITE, *wxWHITE);
        _minmaxo = new VariableMinMaxCtrl<OTYPE>(this, ID_MINMAXCTRLO, false, true, true, *wxBLACK, *wxWHITE, *wxWHITE); 

        LevelsFunc(this);

        _HChoice = (wxChoice *) FindWindow(ID_LEV_HCHOICE);
        _HChoice -> SetSelection(_hist -> GetHistType());

        _ChannelChoice = (wxChoice *) FindWindow(ID_LEV_CHOICE_CHANNEL);
        _ChannelChoice -> Clear();
        for (size_t i = 0; i < channel_names.GetCount(); i++)
        {
            _ChannelChoice -> Append(channel_names.Item(i));
        }
        CreateChannelAll();
        
        _preview = (wxCheckBox *) FindWindow(ID_LEV_PREVIEW);
        _preview -> SetValue(TRUE);

		if (_HistData_vector.size() > 1)
		{
			_ChannelChoice -> SetStringSelection("All");
			SelectChannel(-1);
		}
		else
		{
			_ChannelChoice -> SetSelection(0);
			SelectChannel(0);
		}

        Bind(wxEVT_COMMAND_CHOICE_SELECTED, &LevelsWidget::OnHChoice, this, ID_LEV_HCHOICE);
        Bind(wxEVT_COMMAND_CHOICE_SELECTED, &LevelsWidget::OnChannelChoice, this, ID_LEV_CHOICE_CHANNEL);
        Bind(wxEVT_COMMAND_CHECKBOX_CLICKED, &LevelsWidget::OnPreviewCheckBox, this, ID_LEV_PREVIEW);
        Bind(wxEVT_COMMAND_BUTTON_CLICKED, &LevelsWidget::OnReset, this, ID_LEV_RESET);
        Bind(wxEVT_COMMAND_BUTTON_CLICKED, &LevelsWidget::OnOK, this, wxID_OK);	
        Bind(v4d_EVT_MINMAXCTRL, &LevelsWidget::OnPreview, this);
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE> 
    LevelsWidget<ITYPE,OTYPE>::~LevelsWidget() 
    { 
        if (All_histdata) 
        {
            delete All_histdata; 
        }

        if (All_params)
        {
            delete All_params;
        }
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE> 
    void LevelsWidget<ITYPE,OTYPE>::GetParamsVector(std::vector<TLevelsParams<ITYPE, OTYPE> > &params)
    {
        if (_LevelsParams_vector.size() == 1)
        {
            params.clear();
            params.insert(params.begin(), 3, _LevelsParams_vector[0]);
        }
        else
        {
            params = _LevelsParams_vector;
        }
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE> 
    bool LevelsWidget<ITYPE, OTYPE>::CreateChannelAll()
    {
        if (_HistData_vector.size() > 1)
        {
            // check if all histograms have same parameters
            for (size_t i = 0; i < _HistData_vector.size() - 1; i++) 
            {
                if (*_HistData_vector[i] != *_HistData_vector[i + 1])
                {
                    return false;
                }
            }

            // create new histogram
            All_histdata = new v4d::Histogram<ITYPE>(*_HistData_vector[0]);

            for (size_t i = 1; i < _HistData_vector.size(); i++)
            {
                *All_histdata += *_HistData_vector[i];
            }

            // compute "AF" values for levels params 
            ITYPE imin = std::numeric_limits<ITYPE>::min(); 
            ITYPE imax = std::numeric_limits<ITYPE>::max(); 
            if (_LevelsParams_vector.size()) 
            {
                imin = _LevelsParams_vector[0].inputMin; 
                imax = _LevelsParams_vector[0].inputMax; 

                for (size_t i = 1; i < _LevelsParams_vector.size(); ++i) 
                {
                    imin = std::min(imin, _LevelsParams_vector[i].inputMin);
                    imax = std::max(imax, _LevelsParams_vector[i].inputMax);
                }
            }

            // create new structure with parameters
            All_params = new TLevelsParams<ITYPE, OTYPE>(imin, imax);

            // add new item to list of channels
            _ChannelChoice -> Append("All");

            return true;
        }

        return false;
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE> 
    void LevelsWidget<ITYPE, OTYPE>::SetChannelParams(const TLevelsParams<ITYPE, OTYPE> &params)
    {
        for (size_t i = 0; i < _LevelsParams_vector.size(); i++)
        {
            _LevelsParams_vector[i] = params;
        }
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE>
    void LevelsWidget<ITYPE, OTYPE>::DoPreview()
    {
        TLevelsParams<ITYPE, OTYPE> params;
        if (_ChannelChoice -> GetStringSelection() == "All")
        {
            wxASSERT(All_params);
            params = *All_params;
            SetChannelParams(params);
        }
        else
        {
            params = _LevelsParams_vector[_ChannelChoice -> GetSelection()];
        }

        if (_preview -> IsChecked())
        {
            v4d_LevelsEvent levt(v4d_EVT_LEVELS_PREVIEW, params.ConvertToFloat(), _ChannelChoice -> GetSelection());

            wxEvtHandler *handler = GetParent() -> GetEventHandler();

            if (handler)
            {
                handler -> ProcessEvent(levt);
            }
        }
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE> 
    void LevelsWidget<ITYPE, OTYPE>::GeneratePreviewEvents(bool allChannels)
    {
        wxEvtHandler *handler = GetParent() -> GetEventHandler();

        if (allChannels)
        {
            wxASSERT(All_params);
            v4d_LevelsEvent some_level_event(v4d_EVT_LEVELS_PREVIEW, (*All_params).ConvertToFloat(), _ChannelChoice -> GetSelection());

            if (handler)
            {
                handler -> ProcessEvent(some_level_event);
            }
        }
        else
        {
            for (size_t i = 0; i < _LevelsParams_vector.size(); i++)
            {
                v4d_LevelsEvent some_level_event(v4d_EVT_LEVELS_PREVIEW, _LevelsParams_vector[i].ConvertToFloat(), (int)i);

                if (handler)
                {
                    handler -> ProcessEvent(some_level_event);
                }
            }
        }
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE>
    void LevelsWidget<ITYPE, OTYPE>::OnHChoice(wxCommandEvent &evt)
    {
        if (_hist -> GetHistType() != evt.GetInt())
        {
            _hist -> SetHistType(evt.GetInt());	
            _hist -> Refresh(FALSE);
        }
    }

    /*******************************************************************/

    template <class ITYPE, class OTYPE>
    void LevelsWidget<ITYPE, OTYPE>::SelectChannel(int channel)
    {
        v4d::Histogram<ITYPE> *histdata;
        TLevelsParams<ITYPE, OTYPE> *params;

        if (channel == -1)
        {
            wxASSERT(All_histdata);
            histdata = All_histdata;
            wxASSERT(All_params);
            params = All_params;
        }
        else
        {
            histdata = _HistData_vector[channel];
            params = &_LevelsParams_vector[channel];
        }

		_hist -> SetHistPointer(histdata);
		_minmaxi -> SetParams(&(params -> inputMin), &(params -> inputMax), &(params -> inputGamma), _hist -> GetMin(), _hist -> GetMax(), (int)_hist -> GetSize());
		_minmaxo -> SetParams(&(params -> outputMin), &(params -> outputMax), NULL, std::numeric_limits<OTYPE>::min(), std::numeric_limits<OTYPE>::max(), std::numeric_limits<OTYPE>::max() + 1);
		_hist -> Refresh(FALSE);

        if (_preview -> IsChecked())
        {
            GeneratePreviewEvents(channel == -1);
        }
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE>
    void LevelsWidget<ITYPE, OTYPE>::OnChannelChoice(wxCommandEvent &evt)
    {
        int channel = (evt.GetString() == "All") ? -1 : evt.GetInt();
        SelectChannel(channel);
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE> 
    void LevelsWidget<ITYPE, OTYPE>::OnPreviewCheckBox(wxCommandEvent &evt)
    {
        if (_preview -> IsChecked())
        {
            GeneratePreviewEvents(_ChannelChoice->GetStringSelection() == "All");
        }
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE>
    void LevelsWidget<ITYPE, OTYPE>::OnPreview(v4d_MinMaxCtrlEvent &evt)
    {
        DoPreview();
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE>
    void LevelsWidget<ITYPE, OTYPE>::OnReset(wxCommandEvent &evt)
    {
        _minmaxi -> Reset();
        _minmaxo -> Reset();
        _hist -> SetHistType(HIST_LOG);
        _HChoice -> SetSelection(HIST_LOG);
        _hist -> Refresh(FALSE);

        DoPreview();
    }

    /*******************************************************************/
    template <class ITYPE, class OTYPE>
    void LevelsWidget<ITYPE, OTYPE>::OnOK(wxCommandEvent &evt)
    {
        if (_minmaxi -> Validate() && _minmaxi -> TransferDataFromWindow() && 
            _minmaxo -> Validate() && _minmaxo -> TransferDataFromWindow())
        {
            if (IsModal())
            {
                EndModal(wxID_OK);
            }
            else
            {
                SetReturnCode(wxID_OK);
                Show(FALSE);
            }
        }
    }

    /*******************************************************************/
    template class HistGraph<i3d::GRAY8>;
    template class HistGraph<i3d::GRAY16>;
    template class HistGraph<int>;
    template class HistGraph<float>;

    template class VariableMinMaxCtrl<i3d::GRAY8>;
    template class VariableMinMaxCtrl<i3d::GRAY16>;
    template class VariableMinMaxCtrl<int>;
    template class VariableMinMaxCtrl<float>;

    template class VariableTextCtrl<i3d::GRAY8>;
    template class VariableTextCtrl<i3d::GRAY16>;
    template class VariableTextCtrl<int>;

    template class LevelsWidget<i3d::GRAY8, i3d::GRAY8>;
    template class LevelsWidget<i3d::GRAY16, i3d::GRAY8>;
    template class LevelsWidget<int, i3d::GRAY8>;
    template class LevelsWidget<float, i3d::GRAY8>;

    /*******************************************************************/
}
