/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * dlg_levels.h
 *
 * Dialog for levels adjusting  
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006, based on Jan Hubeny code (xhubeny@fi.muni.cz) 2004
 */

#ifndef V4D_DLG_LEVELS_H
#define V4D_DLG_LEVELS_H

#include <vector>
#include "v4d_impexp.h"
#include "types/tlevelsparams.h"
#include "thistogram.h"
#include "wdr/dlg_levels_wdr.h"
#include "wxtools.h"


/////////////////////////////////////////////////////////////////
//                                                             //
//                       Basic constants                       //
//                                                             //
/////////////////////////////////////////////////////////////////

// the default width of the Histogram graph
#define HIST_WIDTH 256
// the default height of the Histogram graph
#define HIST_HEIGHT 64
// the default height of the Gradient stripe
#define GRAD_HEIGHT 12
// the default height of the triangle controls
#define TRIANGLE_HEIGHT 8
// type of the Histogram graph
#define HIST_LINEAR 1
#define HIST_LOG 0
// type of control for mouse events
#define MAX_POS 0
#define MIN_POS 1
#define GAMMA_POS 2

namespace v4d
{
    /** An event sent to the MinMaxCtrl as a request for refresh. */
    class v4d_VariableTextCtrlEvent 
        : public wxEvent
    {
    private:
        double m_value;

    public:
        /** Constructor. */
        v4d_VariableTextCtrlEvent(wxEventType eventType, double value, int id) 
            : wxEvent(id, eventType), m_value(value) 
        {}

        v4d_VariableTextCtrlEvent(const v4d_VariableTextCtrlEvent &evt)
            : wxEvent(evt), m_value(evt.m_value) 
        {}

        double GetValue() 
        { 
            return m_value; 
        }

        virtual wxEvent *Clone() const 
        { 
            return new v4d_VariableTextCtrlEvent(*this); 
        }
    };

    wxDECLARE_EXPORTED_EVENT(V4D_DLLEXPORT, v4d_EVT_VARIABLE_TEXTCTRL, v4d_VariableTextCtrlEvent);

    /** An event sent to the MinMaxCtrl as a request for refresh. */
    class v4d_TriangleEvent 
        : public wxEvent
    {
    private:
        int m_value; 	

    public:
        /** Constructor. */
        v4d_TriangleEvent(wxEventType eventType, int value, int id) 
            : wxEvent(id, eventType), m_value(value) 
        {}

        v4d_TriangleEvent(const v4d_TriangleEvent &evt)
            : wxEvent(evt), m_value(evt.m_value) 
        {}

        int GetValue() 
        { 
            return m_value; 
        }

        virtual wxEvent *Clone() const 
        { 
            return new v4d_TriangleEvent(*this); 
        }
    };

    wxDECLARE_EXPORTED_EVENT(V4D_DLLEXPORT, v4d_EVT_TRIANGLE, v4d_TriangleEvent);

    /** An event sent to the LevelsDialog as a notification about changing values. */
    class v4d_MinMaxCtrlEvent 
        : public wxNotifyEvent
    {
    public:
        /** Constructor. */
        v4d_MinMaxCtrlEvent(wxEventType eventType, int id = 0) 
            : wxNotifyEvent(eventType, id) 
        {}

        v4d_MinMaxCtrlEvent(const v4d_MinMaxCtrlEvent &evt)
            : wxNotifyEvent(evt) 
        {}

        virtual wxEvent *Clone() const 
        { 
            return new v4d_MinMaxCtrlEvent(*this); 
        }
    };

    wxDECLARE_EXPORTED_EVENT(V4D_DLLEXPORT, v4d_EVT_MINMAXCTRL, v4d_MinMaxCtrlEvent);

    /** Helper class for drawing a linear histogram graph. */
    class LinearHist
    {
    public:
        double operator()(const double &val) { return val; };
    };

    /** Helper class for drawing a logarithmic histogram graph. */
    class LogarithmicHist
    {
    public:
        double operator()(const double &val) { return log(val); };
    };

    /** Class representing the gradient stripe. */
    class GradStripe: public wxPanel
    {
    protected:       
        /** Do you want to draw a gradient stripe? If do not, a stripe with the _center_colour is drawn. */
        bool _draw_gradient;
        /** A colour for the minimal level. This is implicitly black. */
        wxColour _min_colour;
        /** A colour for the maximal level. This is implicitly white. */
        wxColour _max_colour;
        /** A colour for the levels belonging to the range from the minimal level to the maximal one. This colour is implicitly white. */
        wxColour _center_colour;
        /** Value of gamma variable. */
        double _gamma;
        /** Number of position in the gradient stripe. */
        int _numberOfPos; 
        /** End position of minimal colour. Minimal colour is implicitly black. */
        int _minPos;
        /** Start position of maximal colour. Maximal colour is implicitly white. */
        int _maxPos; 

    public:
        /** Constructor. */
        GradStripe(wxWindow *parent,
            bool draw_gradient,
            const wxColour &min_colour = *wxBLACK,
            const wxColour &max_colour = *wxWHITE,
            const wxColour &center_colour = *wxWHITE,
            int numberOfPos = 256,
            double gamma = 1.0,
            wxWindowID id = -1,
            const wxPoint &pos = wxDefaultPosition,
            const wxSize &size = wxSize(HIST_WIDTH, GRAD_HEIGHT / 2),
            long style = wxTAB_TRAVERSAL | wxCLIP_CHILDREN);
        /** Destructor. */
        ~GradStripe() {};
        /** Get end position of minimal colour. */
        int GetMinPos() { return _minPos; };
        /** Get start position of maximal colour. */
        int GetMaxPos() { return _maxPos; };
        /** Get gamma value. */
        double GetGamma() { return _gamma; };
        /** Set end position of minimal colour. */
        void SetMinPos(int p) {if ((p >= 0) && (p < _numberOfPos)) _minPos = p; };
        /** Set start position of maximal colour. */
        void SetMaxPos(int p) {if ((p >= 0) && (p < _numberOfPos)) _maxPos = p; };
        /** Set gamma value. */
        void SetGamma(double g) { _gamma = g; };
        /** Set number of position. */
        void SetNumberOfPos(int numberOfPos) { _numberOfPos = numberOfPos; };
        /** Set the minimal, maximal, and center colour. */
        void SetColours(const wxColour &min_colour, const wxColour &max_colour, const wxColour &center_colour);
        /** Reset all parameters. */
        void Reset();

    protected:
        /** Draw a stripe onto device context by given colour. The stripe is determined by top left corner. */
        void DrawStripe(wxDC &dc, wxCoord x, wxCoord y, wxCoord width, wxCoord height, const wxColour &colour);

    private:
        /** Handler for wxPaint event. */
        void OnPaint(wxPaintEvent &evt);
        /** Handler for wxEraseBackground event. */
        void OnErase(wxEraseEvent &evt) {};

        DECLARE_EVENT_TABLE()
    };


    /** Class representing the histogram graph. */
    template <class T> class HistGraph
        : public wxPanel
    { 
    private:
        /** Pointer to the v4d_Histogram structure. */
        v4d::Histogram<T> *_Histogram;
        /** Background colour. */
        wxColour _bgColour;
        /** Foreground colour. */
        wxColour _fgColour;
        /** Grid colour. */
        wxColour _gridColour;
        /** Type of the histogram graph. */
        int _histType;
        /** Number of column grid lines. */
        int _gridCLines;
        /** Number of row grid lines. */
        int _gridRLines;
        /** Width of border.*/
        int _border_width;
        /** Height of border.*/
        int _border_height;

    public:
        /** Constructor. */
        HistGraph(wxWindow *parent, 
            wxWindowID id = -1,
            v4d::Histogram<T> *H = NULL,
            const wxPoint &pos = wxDefaultPosition,
            const wxSize &size = wxSize(HIST_WIDTH + 8, HIST_HEIGHT + 8),
            long style = wxTAB_TRAVERSAL | wxSUNKEN_BORDER | wxCLIP_CHILDREN);
        /** Destructor. */
        ~HistGraph() {};
        /** Get a pointer to the histogram structure. */
        v4d::Histogram<T> *GetHistPointer() { return _Histogram; };
        /** Get backgroung colour. */
        wxColour GetBgColour() { return _bgColour; };
        /** Get foreground colour. */
        wxColour GetFgColour() { return _fgColour; };
        /** Set grid colour. */
        wxColour GetGridColour() { return _gridColour; };
        /** Get type of histogram graph. */
        int GetHistType() { return _histType; };
        /** Get minimal value of histogram. */
        T GetMin() { return _Histogram -> GetLeftCenter(); };
        /** Get maximal value of histogram. */
        T GetMax() { return _Histogram -> GetRightCenter(); };
        /** Get the size of histogram. */
        size_t GetSize() { return _Histogram -> size(); };
        /** Set a pointer to the histogram structure. */
        void SetHistPointer(v4d::Histogram<T> *hp) { _Histogram = hp; };
        /** Set backgroung colour. */
        void SetBgColour(wxColour b) { _bgColour = b; };
        /** Set foreground colour. */
        void SetFgColour(wxColour f) { _fgColour = f; };
        /** Set grid colour. */
        void SetGridColour(wxColour g) { _gridColour = g; };
        /** Set type of histogram graph. */
        void SetHistType(int t) { _histType = t; };
        /** Set type of histogram graph as linear. */
        void SetHistLinear() { _histType = HIST_LINEAR; };
        /** Set type of histogram graph as logarithmic. */
        void SetHistLogarithmic() { _histType = HIST_LOG; };
        /** Set number of row and column grid lines. */
        void SetGridLines(int c, int r) { _gridCLines = c; _gridRLines = r; };
        /** Set border width and height. */
        void SetBorder(int w, int h) { _border_width = w; _border_height = h; };

    private:
        /** Draw the histogram graph of given type onto given device context. */
        template <class Function> void DrawHistGraph(wxDC &dc, const size_t *begin, const size_t *end, Function func);
        /** Handler for wxPaint event. */
        void OnPaint(wxPaintEvent& evt);
        /** Handler for wxEraseBackground event. */
        void OnErase(wxEraseEvent &evt) {};

    };

    /** Class representing the wxTextCtrl control used for changing the value of given variable with up and down arrow controls. */
    template <class T> class VariableTextCtrl : public wxTextCtrl
    {
    private:
        /** Pointer to a variable. */
        T *_variable;
        /** Minimal value of a variable. */
        T _min_val;
        /** Maximal value of a variable. */
        T _max_val;
        /** Default value of a variable. */
        T _def_val;
        /** A number of possible positions (magnitudes) of a variable. */
        int _num_pos;
        /** An increment of a variable. */
        T _increment;

    public:
        /** Constructor. */
        VariableTextCtrl(wxWindow *parent, 
            wxWindowID id, 
            T *variable = NULL, 
            T min_val = std::numeric_limits<T>::min(),
            T max_val = std::numeric_limits<T>::max(), 
            T def_val = std::numeric_limits<T>::min(),
            int num_pos = 1,
            const wxPoint &pos = wxDefaultPosition, 
            const wxSize &size = wxDefaultSize, 
            long style = wxTE_PROCESS_ENTER);
        /** Destructor. */
        ~VariableTextCtrl() {};
        /** Get an interval of possible variable values. */
        void GetRange(T &min_val, T &max_val) { min_val = _min_val; max_val = _max_val; };
        /** Return true if the control has a valid pointer to a variable. */
        bool HasVariable() { return (_variable != NULL); };
        /** Set parameters of a variable. */
        void SetVariableParams(T *variable, T min_val, T max_val, T def_val, int num_pos); 
        /** Get value from the control. */
        void GetVariableValue(T &value) { StringToVariable(wxTextCtrl::GetValue(), value); };
        /** Set value of the control. */
        void SetVariableValue(T value) { wxTextCtrl::SetValue(VariableToString(value)); };
        /** Get a position of a variable magnitude. */
        int GetVariablePos();
        /** Get a magnitude of a variable for given position. */
        T GetValue(int pos);
        /** Reset value of the control to the default value. */
        void Reset();

    private:
        /** Set validator for a variable. */
        void SetVariableValidator();
        /** Convert a value of the variable into a string. */
        wxString VariableToString(T value);
        /** Get a value of the variable from given string. */
        bool StringToVariable(const wxString &str, T &value);
        /** Update a value of an increment according to the current variable parameters. */
        void UpdateIncrement();
        /** Refresh value in the control and refresh values in the MinMaxControl. */
        void Refresh(T value);
        /** Handler for changing the value of a variable. */
        void Correct(wxEvent &evt);
        /** Routine called when the control loses focus. It processes a wxEVT_KILL_FOCUS event. */  
        void OnCorrectFE(wxFocusEvent &evt) { Correct(evt); };
        /** Handler for a wxEVT_COMMAND_TEXT_ENTER event, generated when ENTER is pressed in the control. */
        void OnCorrectCE(wxCommandEvent &evt) { Correct(evt); };
        /** Handler for updating the value according to up and down arrow pressed. */
        void OnChar(wxKeyEvent &evt);
    };

    /** Class representing triangles in instances of MinMaxControl class. */
    class Triangles: public wxPanel
    {
    private:
        /** Number of positions. */
        int _numOfPos;
        /** Position of triangle for minimal level. */
        int _min_pos;
        /** Position of triangle for maximal level. */
        int _max_pos;
        /** Value of gamma variable. */
        double _gamma;
        /** Position of triangle for gamma variable. */
        int _gamma_pos;
        /** Type of control - input or output levels. */
        bool _input;
        /** Can the minimal and maximal levels be crossed? (i.e. Can the maximal level be lower than the minimal one?) */
        bool _can_cross;
        /** Actual pressed mouse button. */ 
        int _actualButton;
        /** A pointer to window that is able to handle v4d_TriangleEvents. */
        wxWindow *_evt_handler;
        /** A colour of the minimal triagle. */
        wxColour _min_colour;
        /** A colour of the maximal triagle. */
        wxColour _max_colour;
        /** A colour of the gamma triagle. */
        wxColour _gamma_colour;



    public:
        /** Constructor. */
        Triangles(wxWindow *parent, 
            wxWindow *evt_handler,
            bool input,
            bool can_cross,
            const wxColour &min_colour,
            const wxColour &max_colour,
            int numOfPos = 256,
            double gamma = 1.0,
            wxWindowID id = -1,
            const wxPoint &pos = wxDefaultPosition, 
            const wxSize &size = wxSize(HIST_WIDTH + 4, TRIANGLE_HEIGHT), 
            long style = wxTAB_TRAVERSAL | wxCLIP_CHILDREN);
        /** Destructor. */
        ~Triangles() {};
        /** Set position of triangle for minimal level. */
        void SetMinPos(int p) { if ((p >= 0) && (p < _numOfPos)) _min_pos = p; };
        /** Set position of triangle for maximal level. */
        void SetMaxPos(int p) { if ((p >= 0) && (p < _numOfPos)) _max_pos = p; };
        /** Set gamma value. */
        void SetGamma(double g) { _gamma = g; RecomputeGammaPos(); };
        /** Set number of positions. */
        void SetNumberOfPos(int num) { _numOfPos = num; };
        /** Get position of triangle for minimal level. */
        int GetMinPos() { return _min_pos; };
        /** Get position of triangle for maxiaml level. */
        int GetMaxPos() { return _max_pos; };
        /** Recompute position of triangle for gamma variable. */
        void RecomputeGammaPos();
        /** Reset all parameters. */
        void Reset(); 

    private:
        /** Get absolute position of triangle. */
        inline int GetAbsoluteTrianglePosition(int pos, int cw);
        /** Draw a triangle given by its upper point position onto given device context. */
        void DrawTriangle(wxDC &dc, int xpos, const wxColour &brush_colour, wxWindow *win = NULL);
        /** Handler for wxPaint event. */
        void OnPaint(wxPaintEvent &evt);
        /** Handler for wxMouseEvent. */
        void OnMouse(wxMouseEvent &evt);
        /** Handler for wxEraseBackground event. */
        void OnErase(wxEraseEvent &evt) {};

        DECLARE_EVENT_TABLE()
    };

    /** Class representing the control for adjusting minimal and maximal values of a variable. */
    template <class T> class VariableMinMaxCtrl : public wxPanel
    {
    private:
        /** Type of control - input or output levels. */
        bool _input;
        /** Can the minimal and maximal levels be crossed? (i.e. Can the maximal level be lower than the minimal one?) */
        bool _can_cross;
        /** Text control for minimal level. */
        VariableTextCtrl<T> *_min_ctrl;
        /** Text control for maximal level. */
        VariableTextCtrl<T> *_max_ctrl;
        /** Text control for value of gamma. */
        VariableTextCtrl<double> *_gamma_ctrl;
        /** Controls for graphical representation of adjusted levels. */
        GradStripe *_gradStripeUpper;
        GradStripe *_gradStripeLower;
        Triangles *_triangles;

    public:
        /** Constructor. */
        VariableMinMaxCtrl(wxWindow *parent, 
            wxWindowID id,
            bool input,
            bool can_cross,
            bool draw_gradient,
            const wxColour &min_colour,
            const wxColour &max_colour,
            const wxColour &center_colour,
            const wxPoint &pos = wxDefaultPosition,
            const wxSize &size = wxDefaultSize,
            long style = wxTAB_TRAVERSAL | wxCLIP_CHILDREN);
        /** Destructor. */
        ~VariableMinMaxCtrl() {};
        /** Set pointer to structure with levels parameters, value of minimal level and value of maximal level. */
        void SetParams(T *min_var, T *max_var, double *gamma_var, T min_val, T max_val, int num_steps);
        /** Reset all levels parameters. */
        void Reset();

    private:
        /** Update minimal level. */
        void UpdateMin(T m);
        /** Update maximal level. */
        void UpdateMax(T m);
        /** Update value of gamma. */
        void UpdateGamma(double g);
        /** Generate v4d_LevelsEvent as a request for preview drawing. */
        void GeneratePreviewEvent();
        /** Function called when a magnitude of a variable is changed. */
        void OnVariableChange(v4d_VariableTextCtrlEvent &evt);
        /** Function called when a position of triangle is changed. */
        void OnTriangleChange(v4d_TriangleEvent &evt);

    };

    /** An event sent by LevelsWidget dialog as a request for preview drawing. */
    class v4d_LevelsEvent 
        : public wxEvent
    {
    private:
        /** Structure with levels parameters. */
        TLevelsParams<float, float> params;
        /** Selected channel. */
        int channel;

    public:
        /** Constructor. */
        v4d_LevelsEvent(wxEventType commandType, const TLevelsParams<float, float> &par, int ch, int id = 0): 
          wxEvent(id, commandType), params(par), channel(ch) {}
          /** Copy constructor. */
          v4d_LevelsEvent(const v4d_LevelsEvent &evt):
          wxEvent(evt), params(evt.params), channel(evt.channel) {}
          /** Return channel number. */
          int GetChannel() const { return channel; }
          /** Return structure with parameters. */
          const TLevelsParams<float, float>& GetParams() const { return params; }
          /** Set channel number. */
          void SetChannel(int ch) { channel = ch; }
          /** Set structure with parameters. */
          void SetParams(const TLevelsParams<float, float> &par) { params = par; }
          /** Clone method. */
          virtual wxEvent *Clone() const { return new v4d_LevelsEvent(*this); }
    };

    wxDECLARE_EXPORTED_EVENT(V4D_DLLEXPORT,v4d_EVT_LEVELS_PREVIEW, v4d_LevelsEvent);

    /** The main widget for Adjust Levels dialog. If the number of available channels is greater than one, the dialog
    * creates new channel called All that provides with a possibility to adjust levels of all channels together.
    * This dialog generates v4d_LevelsEvent as a request for preview drawing. The event contains structure TLevelsParams
    * with levels parameters and index of selected channel (the value from the interval [0, n - 1] where n is the number 
    * of available channels). If an user selects channel All then channel index of v4d_LevelsEvent is n. */
    template <class ITYPE, class OTYPE> class LevelsWidget : public wxDialog
    {
    private:
        /** Pointer to the HistGraph class. */
        HistGraph<ITYPE> *_hist;
        /** List of histograms for all available channels. */
        const std::vector<v4d::Histogram<ITYPE> *> &_HistData_vector;
        /** List of parameters for all avaible channels. */
        std::vector<TLevelsParams<ITYPE, OTYPE> > _LevelsParams_vector;
        /** Control for histogram type selection. */
        wxChoice *_HChoice;
        /** Control for channel selection. */
        wxChoice *_ChannelChoice;
        /** Control for preview. */
        wxCheckBox *_preview;
        /** Control for adjusting input levels. */
        VariableMinMaxCtrl<ITYPE> *_minmaxi;
        /** Control for adjusting output levels. */
        VariableMinMaxCtrl<OTYPE> *_minmaxo;
        /** Pointer to structure with parameters for channel All. */
        TLevelsParams<ITYPE, OTYPE> *All_params;
        /** Pointer to histogram for channel All. */
        v4d::Histogram<ITYPE> *All_histdata;

    public:
        /** Constructor. */
        LevelsWidget(
            wxWindow *parent, 
            wxWindowID id, 
            const std::vector<v4d::Histogram<ITYPE> *> &Histdata_vector,
            const std::vector<TLevelsParams<ITYPE, OTYPE> > &params_vector,
            const wxArrayString &channel_names,
            const wxString &title = wxString("Adjust Levels"),
            const wxPoint &pos = wxDefaultPosition,
            const wxSize &size = wxDefaultSize,
            long style = wxRESIZE_BORDER | wxDEFAULT_DIALOG_STYLE | wxCLIP_CHILDREN);
        /** Destructor. */
        ~LevelsWidget();
        /** Return a list of parameters for all available channels. */
        void GetParamsVector(std::vector<TLevelsParams<ITYPE, OTYPE> > &params);

    private:
        /** If dialog contains at least two channels then the new channel All is created. This channel provides possibility to 
        * change options of all channels together. Returns true if the all channel has been created. */
        bool CreateChannelAll();
        /** Set level parameters of all available channels according to given level parameters. */
        void SetChannelParams(const TLevelsParams<ITYPE, OTYPE> &params);
        /** Update level parameters according to the currently selected channel and generate a preview event for him. */
        void DoPreview();
        /** Generate sequence of preview events according to given string. */
        void GeneratePreviewEvents(bool allChannels);
        /** Change current channel. */
        void SelectChannel(int channel);

        /** Handler for type selection of histogram graph. */
        void OnHChoice(wxCommandEvent &evt);
        /** Handler for channel selection. */
        void OnChannelChoice(wxCommandEvent &evt);
        /** Handler for preview selection. */
        void OnPreviewCheckBox(wxCommandEvent &evt);
        /** Handler for preview drawing. */ 
        void OnPreview(v4d_MinMaxCtrlEvent &evt);
        /** Handler for Reset button. */
        void OnReset(wxCommandEvent &evt);
        /** Handler for OK button. */
        void OnOK(wxCommandEvent &evt);
    };
}

#endif
