/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
* dlg_openfile.h
*
* Dialog for opening files (be it a sequence, overlays, single files...)
*
* Vladimir Ulman , ulman@fi.muni.cz , 2015
*/

#include "dlg_openfile.h"

#include <wx/regex.h>
#include "wdr/dlg_openfile_wdr.h"
#include "wxtools.h"
#include <i3d/basic.h>

namespace v4d
{
	BEGIN_EVENT_TABLE(OpenFileDialog, wxDialog)
		//adding files to the List
		EVT_DIRCTRL_FILEACTIVATED(ID_OD_FILECTRL, OpenFileDialog::OnAddToListClicked)
		EVT_BUTTON(ID_OD_ADDTOLIST, OpenFileDialog::OnAddToListButton)
		EVT_BUTTON(ID_OD_SETFOLDER, OpenFileDialog::OnChangeFolderButton)
		EVT_SPINCTRL(ID_OD_SEPCNT, OpenFileDialog::OnSeparatorChange)

		//selecting in the List
		EVT_BUTTON(ID_OD_SELECT, OpenFileDialog::OnSelectButton)
		EVT_TEXT_ENTER(ID_OD_SELECTEXP, OpenFileDialog::OnSelectButton)
		EVT_BUTTON(ID_OD_INVERT, OpenFileDialog::OnInvertButton)
		EVT_BUTTON(ID_OD_DESELECT, OpenFileDialog::OnDeselectButton)

		//modifying the List
		EVT_BUTTON(ID_OD_SORT, OpenFileDialog::OnSortButton)
		EVT_BUTTON(ID_OD_GROUPS_1, OpenFileDialog::OnSlot1Button)
		EVT_BUTTON(ID_OD_GROUPS_2, OpenFileDialog::OnSlot2Button)
		EVT_BUTTON(ID_OD_GROUPS_3, OpenFileDialog::OnSlot3Button)
		EVT_BUTTON(ID_OD_MOVEUP, OpenFileDialog::OnMoveUpButton)
		EVT_BUTTON(ID_OD_MOVEDOWN, OpenFileDialog::OnMoveDownButton)
		EVT_BUTTON(ID_OD_REMOVE, OpenFileDialog::OnRemoveButton)

		//open mode params
		EVT_RADIOBOX(ID_OD_GROUP_MODE, OpenFileDialog::OnGroupModeChange)
		EVT_BUTTON(ID_OD_SAMPLINGOPTIONS, OpenFileDialog::OnSamplingDialogButton)

		// update ui events
		EVT_UPDATE_UI_RANGE(ID_OD_FILECTRL, ID_OD_SAMPLINGOPTIONS, OpenFileDialog::OnUpdateUI)
		//EVT_UPDATE_UI(ID_OD_ADDTOLIST, OpenFileDialog::OnUpdateUI)
		//EVT_UPDATE_UI(ID_OD_REMOVE, OpenFileDialog::OnUpdateUI)
		//EVT_UPDATE_UI(ID_OD_GROUPS_1, OpenFileDialog::OnUpdateUI)

		//exit (and validate) button
		EVT_BUTTON(wxID_OK, OpenFileDialog::OnOK)
	END_EVENT_TABLE();

	/*******************************************************************/

	OpenFileDialog::OpenFileDialog(wxWindow *parent,
		               const wxString& folder,
		               const wxArrayString* filenames)
	 : wxDialog(parent, wxID_ANY, "Open files..."), m_grp_list(),
	   samplingDlg(this, -1, "Downsampling rates", wxDefaultPosition, wxDefaultSize)
	{
		OpenSequenceFunc(this);

#ifdef VLADO_PRIVATE_TWEAKS
		//adjust the width of the dialog a little bit, 'cause I have just too narrow LCD :-(
		wxSize mySize=this->GetMinSize();
		mySize.SetWidth(int((float)mySize.GetWidth()*1.2f));
		this->SetMinSize(mySize);
#endif

		//get handlers on the controls:
		od_filectrl=(wxGenericDirCtrl*) FindWindowById(ID_OD_FILECTRL,this);

		od_sepcnt=(wxSpinCtrl*) FindWindowById(ID_OD_SEPCNT,this);
		od_listctrl=(wxListCtrl*) FindWindowById(ID_OD_LISTCTRL,this);
		od_select_exp=(wxTextCtrl*) FindWindowById(ID_OD_SELECTEXP,this);
		od_movegroup=(wxCheckBox*) FindWindowById(ID_OD_MOVEGROUP,this);
		od_listedfiles=(wxStaticText*) FindWindowById(ID_OD_LISTEDFILES,this);

		od_grp_mode = (wxRadioBox *) FindWindowById(ID_OD_GROUP_MODE, this);
		od_swapzt=(wxCheckBox*) FindWindowById(ID_OD_SWAPZT,this);
		od_ownwindow=(wxCheckBox*) FindWindowById(ID_OD_OWNWINDOW,this);

		if ((!od_filectrl) || (!od_sepcnt) || (!od_listctrl)
		 || (!od_select_exp) || (!od_movegroup) || (!od_listedfiles)
		 || (!od_grp_mode) || (!od_swapzt) || (!od_ownwindow))
			throw i3d::InternalException("OpenDialog: Couldn't retrieve handlers to some control...");

		//initiate the controls

		//ask the left panel to start in this folder
		//(except for the very first time)
		od_filectrl -> UnselectAll();
		od_filectrl -> CollapseTree();
		od_filectrl -> ExpandPath(folder.IsEmpty() ? wxGetCwd() : folder.BeforeLast(wxFILE_SEP_PATH));

		//init the List to have two columns
		//TODO: still a third column is there on Linux
		//TODO: can't get rid of it
		//TODO: -> hack for now: make it display away from the visible region
		int width,height;
		od_listctrl->GetSize(&width,&height);
		od_listctrl->InsertColumn(0,wxString("Type"),wxLIST_FORMAT_LEFT,80);
		od_listctrl->InsertColumn(1,wxString("Image files"),wxLIST_FORMAT_RIGHT,width-85);

		/* TODO: this is an official way according to some wxWidget.wiki tutorial
		//init the List to have two columns
		wxListItem col;
		col.SetId(0);
		col.SetText(wxString("Type"));
		col.SetAlign(wxLIST_FORMAT_LEFT);
		col.SetWidth(70);
		od_listctrl->InsertColumn(0,col);

		col.SetId(1);
		col.SetText(wxString("Image files"));
		col.SetAlign(wxLIST_FORMAT_RIGHT);
		col.SetWidth(wxLIST_AUTOSIZE);
		od_listctrl->InsertColumn(1,col);
		*/


		//populate and init the sampling dialog
		SamplingFunc(&samplingDlg);
		m_xSampling=m_ySampling=m_zSampling=1.f;

		//look for the three main controls in the sampling dialog
		wxTextCtrl* xTxtCtrl=(wxTextCtrl*)samplingDlg.FindWindowById(ID_PREF_DOWNSAMPLE_X);
		wxTextCtrl* yTxtCtrl=(wxTextCtrl*)samplingDlg.FindWindowById(ID_PREF_DOWNSAMPLE_Y);
		wxTextCtrl* zTxtCtrl=(wxTextCtrl*)samplingDlg.FindWindowById(ID_PREF_DOWNSAMPLE_Z);
		if ((!xTxtCtrl) || (!yTxtCtrl) || (!zTxtCtrl))
			throw i3d::InternalException("OpenDialog: Couldn't retrieve handlers to the sampling control...");

		//associate the dialog controls with the sampling rates variables
		xTxtCtrl->SetValidator(FloatValidator(m_xSampling,1.f,1000000.f)); //,true,true,wxString("%.2f")));
		yTxtCtrl->SetValidator(FloatValidator(m_ySampling,1.f,1000000.f)); //,true,true,wxString("%.2f")));
		zTxtCtrl->SetValidator(FloatValidator(m_zSampling,1.f,1000000.f)); //,true,true,wxString("%.2f")));


		//okay, once every thing is setup, add possibly some files to begin with
		if (filenames)
		{
			//add these on the local list
			for (size_t i=0; i < filenames->GetCount(); ++i)
				if ( wxFileName::IsFileReadable((*filenames)[i]) )
					listedFiles.Add((*filenames)[i]);
				else
					wxLogWarning(wxString("File not readable: ") + (*filenames)[i]);

			//make it display
			this->UpdateFilesOnList();
		}
	}

	OpenFileDialog::~OpenFileDialog()
	{
	}

	/********************* Populating the List *********************/

	// User wants to change the current folder in the Tree Dir list
	void OpenFileDialog::OnChangeFolderButton(wxCommandEvent& evt)
	{
		wxDirDialog dlg(this, "Choose folder to look for images", wxEmptyString,
			                  wxDD_CHANGE_DIR | wxDD_DIR_MUST_EXIST);
		if (dlg.ShowModal() == wxID_OK)
		{
			od_filectrl -> UnselectAll();
			od_filectrl -> CollapseTree();
			od_filectrl -> ExpandPath(dlg.GetPath());
			//od_filectrl -> SelectPath(dlg.GetPath());
		}
	}

	// User wants to add files on the List
	void OpenFileDialog::AddFiles()
	{
		//read all selected files
		wxArrayString selectedFiles;
		od_filectrl->GetFilePaths(selectedFiles);

		//last size of listedFiles
		const long prevFiles=listedFiles.GetCount();

		//add these on the local list
		for (size_t i=0; i < selectedFiles.GetCount(); ++i)
			if ( wxFileName::IsFileReadable(selectedFiles[i]) )
				listedFiles.Add(selectedFiles[i]);
			else
				wxLogWarning(wxString("File not readable: ") + selectedFiles[i]);

		//commit the local list on the display
		long index=this->UpdateFilesOnList();

		//make the added item(s) on the List selected
		std::map<long,long>::const_iterator lM_iter=this->listMap.begin();
		while (lM_iter != listMap.end())
		{
			if (lM_iter->second >= prevFiles)
				od_listctrl->SetItemState(lM_iter->first, wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);

			++lM_iter;
		}

		//make the last added item visible
		//(hoping this will eventually scroll down the List)
		if (index > -1) od_listctrl->EnsureVisible(index);
	}

	/********************* Selecting in the List *********************/

	void OpenFileDialog::OnSelectButton(wxCommandEvent& evt)
	{
		wxRegEx rExp;
		/*
		if (! rExp.Compile(od_select_exp->GetValue()) )
		{
			wxMessageBox("Could not parse search string as a standard regular expression.\nKeeping current selection in the Image list.", "Incorrect search regular expression", wxICON_EXCLAMATION, this);
			return;
		}
		*/
		try {
			rExp.Compile(od_select_exp->GetValue());
		}
		catch (...)
		{
			wxMessageBox("Could not parse search string as a standard regular expression.\nKeeping current selection in the Image list.", "Incorrect search regular expression", wxICON_EXCLAMATION, this);
			return;
		};

		//to keep the first matching item visible:
		bool foundFirstMatching=false;

		//iterate over all items in the List
		long itemIndex = -1;
		while ((itemIndex = od_listctrl->GetNextItem(itemIndex,
		           wxLIST_NEXT_ALL, wxLIST_STATE_DONTCARE)) != wxNOT_FOUND)
		{
			long state=( rExp.Matches(od_listctrl->GetItemText(itemIndex,1)) )? wxLIST_STATE_SELECTED: 0;
			od_listctrl->SetItemState(itemIndex, state, wxLIST_STATE_SELECTED);

			//to make the first matching item visible:
			if ((!foundFirstMatching) && (state == wxLIST_STATE_SELECTED))
			{
				od_listctrl->EnsureVisible(itemIndex);
				foundFirstMatching=true;
			}
		}
	}

	void OpenFileDialog::OnInvertButton(wxCommandEvent& evt)
	{
		//iterate over all items in the List
		long itemIndex = -1;
		while ((itemIndex = od_listctrl->GetNextItem(itemIndex,
		           wxLIST_NEXT_ALL, wxLIST_STATE_DONTCARE)) != wxNOT_FOUND)
		{
			long state=od_listctrl->GetItemState(itemIndex, wxLIST_STATE_SELECTED);
			if (state) state=0; else state=wxLIST_STATE_SELECTED;
			od_listctrl->SetItemState(itemIndex, state, wxLIST_STATE_SELECTED);
		}
	}

	void OpenFileDialog::OnDeselectButton(wxCommandEvent& evt)
	{
		//iterate over all selected items in the List
		long itemIndex = -1;
		while ((itemIndex = od_listctrl->GetNextItem(itemIndex,
		           wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED)) != wxNOT_FOUND)
		{
			od_listctrl->SetItemState(itemIndex, 0, wxLIST_STATE_SELECTED);
		}
	}

	/********************* Manipulating the List *********************/

	void OpenFileDialog::GetActiveListContent(std::list<std::pair<long,wxString> >& content)
	{
		const int desiredState=(od_listctrl->GetSelectedItemCount() > 0)?
			wxLIST_STATE_SELECTED : wxLIST_STATE_DONTCARE ;

		content.clear();

		//fill only with the items of the particular state
		long itemIndex = -1;
		while ((itemIndex = od_listctrl->GetNextItem(itemIndex,
		           wxLIST_NEXT_ALL, desiredState)) != wxNOT_FOUND)
		{
			content.push_back( std::make_pair(itemIndex,od_listctrl->GetItemText(itemIndex,1)) );
		}
	}

	//helper class to aid comparing in the List
	class myListCompare {
	 public:
		bool operator() (const std::pair<long,wxString>& a, const std::pair<long,wxString>& b)
		{
			//should return true if a < b
			return (a.second.Cmp(b.second) < 0);
		}
	} myListComparator;

	void OpenFileDialog::OnSortButton(wxCommandEvent& evt)
	{
		// to speed up inserting we hide the control temporarily (from wx samples)
		// and also to make sure no one modifies it in the meantime
		od_listctrl->Hide();

		//get data to work with
		std::list<std::pair<long,wxString> > sortData;
		this->GetActiveListContent(sortData);

		//sort it
		sortData.sort(myListComparator);

		//place it back:
		//prepare sorted selected/active data
		std::list<std::pair<long,wxString> >::const_iterator sortedIt = sortData.begin();

		//where to place the sorted active data?
		const bool changeAll=(od_listctrl->GetSelectedItemCount() == 0);

		//prepare new list of full paths and the mapping
		wxArrayString listedFiles_backup(this->listedFiles);
		this->listedFiles.Empty();

		std::map<long,long> listMap_backup(this->listMap);
		this->listMap.clear();

		//iterate over the whole List and update it where required
		long itemIndex = -1;
		while ((itemIndex = od_listctrl->GetNextItem(itemIndex,
		           wxLIST_NEXT_ALL, wxLIST_STATE_DONTCARE)) != wxNOT_FOUND)
		{
			if ((od_listctrl->GetItemState(itemIndex,wxLIST_STATE_SELECTED))
			  || (changeAll))
			{
				//came across an "active" item:

				//update the List
				od_listctrl->SetItem(itemIndex,1,sortedIt->second);
				//the content now at itemIndex is the one that used to be at sortedIt->first
				listedFiles.Add( listedFiles_backup[sortedIt->first] );
				listMap[itemIndex]=listedFiles.GetCount()-1;

				sortedIt++;
			}
			else
			{
				//came across some other/remaining item - no change here
				listedFiles.Add( listedFiles_backup[listMap_backup[itemIndex]] );
				listMap[itemIndex]=listedFiles.GetCount()-1;
				//listMap[itemIndex]=listMap_backup[itemIndex]; -- should work the same, however
			}
		}

		//"un-hide"
		od_listctrl->Show();
	}

	void OpenFileDialog::OnSlotHelper(const int beforeInGroup,const int afterInGroup)
	{
		//plan:
		//iterate over the selected ones and move them in the listedFiles arrays
		// (but moving them breaks the association between Image list control and listedFiles array)
		//note down, which indices are to be selected
		//than let the List of images regenerate according to the one
		//select the indices that were selected before


		//so, first copy out items to move (=selected items) and the rest...
		wxArrayString toMoveItems, toKeepItems;

		//iterate over all items in the List, and copy them out seperately
		long itemIndex = -1;
		while ((itemIndex = od_listctrl->GetNextItem(itemIndex,
		           wxLIST_NEXT_ALL, wxLIST_STATE_DONTCARE)) != wxNOT_FOUND)
		{
			long state=od_listctrl->GetItemState(itemIndex, wxLIST_STATE_SELECTED);
			if (state)
				toMoveItems.Add( this->listedFiles[this->listMap[itemIndex]] );
			else
				toKeepItems.Add( this->listedFiles[this->listMap[itemIndex]] );
		}

		//clear the original list and restore it "our way"
		//note: this only invalidates content (to prevent of re-allocating memory)
		this->listedFiles.Empty();

		//remember selected items (vector stores the order within the new listedFiles)
		std::vector<size_t> indicesToSelect;

		//we first try to interleave,
		//then we add what has remained (from the remaining single list)
		while ((toKeepItems.GetCount() > 0) && (toMoveItems.GetCount() > 0))
		{
			int i=beforeInGroup;
			while ((i > 0) && (toKeepItems.GetCount() > 0))
			{
				listedFiles.Add(toKeepItems[0]);
				toKeepItems.RemoveAt(0);
				--i;
			}

			listedFiles.Add(toMoveItems[0]);
			toMoveItems.RemoveAt(0);

			//note down, which indices to select
			indicesToSelect.push_back(listedFiles.GetCount()-1);

			i=afterInGroup;
			while ((i > 0) && (toKeepItems.GetCount() > 0))
			{
				listedFiles.Add(toKeepItems[0]);
				toKeepItems.RemoveAt(0);
				--i;
			}
		}

		//now, add what has remained
		while (toKeepItems.GetCount() > 0)
		{
			listedFiles.Add(toKeepItems[0]);
			toKeepItems.RemoveAt(0);
		}
		while (toMoveItems.GetCount() > 0)
		{
			listedFiles.Add(toMoveItems[0]);
			toMoveItems.RemoveAt(0);

			//note down, which indices to select
			indicesToSelect.push_back(listedFiles.GetCount()-1);
		 }

		//the listedFiles is ready, redraw it in the control
		this->UpdateFilesOnList();

		//finally, highlight the originally selected items:
		//iterate over all items in the List, and select appropriate
		std::vector<size_t>::const_iterator toSelect=indicesToSelect.begin();
		size_t processedItems=0;

		itemIndex = -1;
		while ((itemIndex = od_listctrl->GetNextItem(itemIndex,
		           wxLIST_NEXT_ALL, wxLIST_STATE_DONTCARE)) != wxNOT_FOUND)
		{
			//if we are on the appropriate item, select/highlight it
			if (processedItems == *toSelect)
			{
				//highlight:
				od_listctrl->SetItemState(itemIndex,
				                          wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);
				++toSelect;
				//TODO, make more robust: indicesToSelect.end()
			}

			++processedItems;
		}
	}

	void OpenFileDialog::OnSlot1Button(wxCommandEvent& evt)
	{
		//if nothing selected, we have nothing to do...
		if (od_listctrl->GetSelectedItemCount() == 0) return;

		//now, determine the number of slots
		const long groupSize=(long)m_grp_list.Get(od_grp_mode -> GetSelection()).Size();

		//sanity check, but this should't happen (see this->OnUpdateUI())
		if (groupSize < 2) return;

		this->OnSlotHelper(0,groupSize-1);
	}

	void OpenFileDialog::OnSlot2Button(wxCommandEvent& evt)
	{
		//if nothing selected, we have nothing to do...
		if (od_listctrl->GetSelectedItemCount() == 0) return;

		//now, determine the number of slots
		const long groupSize=(long)m_grp_list.Get(od_grp_mode -> GetSelection()).Size();

		//sanity check, but this should't happen (see this->OnUpdateUI())
		if (groupSize < 2) return;

		this->OnSlotHelper(1,groupSize-2);
	}

	void OpenFileDialog::OnSlot3Button(wxCommandEvent& evt)
	{
		//if nothing selected, we have nothing to do...
		if (od_listctrl->GetSelectedItemCount() == 0) return;

		//now, determine the number of slots
		const long groupSize=(long)m_grp_list.Get(od_grp_mode -> GetSelection()).Size();

		//sanity check, but this should't happen (see this->OnUpdateUI())
		if (groupSize < 3) return;

		this->OnSlotHelper(2,groupSize-3);
	}

	void OpenFileDialog::OnMoveHelper(const int direction)
	{
		//if nothing selected, we have nothing to do...
		if (od_listctrl->GetSelectedItemCount() == 0) return;

		// to speed up inserting we hide the control temporarily (from wx samples)
		od_listctrl->Hide();

		//how many lines together form one group?
		int groupSize=1;

		//if we should work at the level of groups, we fake it
		//by selecting the whole groups and submitting the job
		//to the routine below
		if (od_movegroup->GetValue())
		{
			groupSize = m_grp_list.Get(od_grp_mode -> GetSelection()).Size();

			//to sweep the List "group-wise"
			long groupIndices[3];
			bool groupIndicesSelected[3]={false,false,false};
			int currGroup=0;

			//current index and position
			long itemIndex = -1;
			int position=0;

			while ((itemIndex = od_listctrl->GetNextItem(itemIndex,
					  wxLIST_NEXT_ALL, wxLIST_STATE_DONTCARE)) != wxNOT_FOUND)
			{
				//have we moved into the region of another group?
				if ((position/groupSize) != currGroup)
				{
					//yes, new group opened, check the old one
					if (groupIndicesSelected[0] || groupIndicesSelected[1]
					 || groupIndicesSelected[2])
						for (int i=0; i < groupSize; ++i)
							od_listctrl->SetItemState(groupIndices[i],
								wxLIST_STATE_SELECTED,wxLIST_STATE_SELECTED);
					++currGroup;
				}

				//record situation in the "current" group
				groupIndices[position % groupSize]=itemIndex;
				groupIndicesSelected[position % groupSize]=
					(od_listctrl->GetItemState(itemIndex,wxLIST_STATE_SELECTED) > 0);

				++position;
			}
		}

		//since in the direction end-to-front, the listing is not really working
		//we need to clone the item list into a better data structure...
		std::list<long> itemIndicies;
		long itemIndex=-1; //indicated beginning of the list
		while ((itemIndex = od_listctrl->GetNextItem(itemIndex,
					  wxLIST_NEXT_ALL, wxLIST_STATE_DONTCARE)) != wxNOT_FOUND)
			itemIndicies.push_back(itemIndex);

		//if moving down with groups, shorten the list virtually
		//to end at boundary of the last complete group
		if ((od_movegroup->GetValue()) && (direction == wxLIST_NEXT_ABOVE))
		{
			while ((itemIndicies.size() % groupSize) != 0)
				itemIndicies.pop_back();
		}

		//find index of the border item
		std::list<long>::const_iterator itemIndiciesIt
			=(direction == wxLIST_NEXT_BELOW)?
				itemIndicies.begin() : --itemIndicies.end();
		itemIndex=*itemIndiciesIt;

		//if the "destination border" item is selected, how can we move then?
		if (od_listctrl->GetItemState(itemIndex,wxLIST_STATE_SELECTED))
		{
			od_listctrl->Show();
			return;
		}
		//else: there is a (complete) room to swap (to move), which is good...

		//scan the List in the given direction and swap by one
		long prevItemIndex[3]={-1,-1,-1};
		prevItemIndex[groupSize-1]=itemIndex;

		//advance the iterator in the given direction
		if (direction == wxLIST_NEXT_BELOW)
		{ itemIndiciesIt++; }
		else
		{
			if (itemIndiciesIt == itemIndicies.begin())
				itemIndiciesIt=itemIndicies.end();
			else
				itemIndiciesIt--;
		}

		//to keep moved selection visible:
		bool foundFirstSelected=false;
		long firstSelectedItemIndex=-1;

		//while there is an index to advance into, do...
		while (itemIndiciesIt != itemIndicies.end())
		{
			itemIndex=*itemIndiciesIt;

			if (od_listctrl->GetItemState(itemIndex,wxLIST_STATE_SELECTED))
			{
				//swap prevItemIndex with itemIndex
				wxString tmp;

				//the List itself
				tmp=od_listctrl->GetItemText(itemIndex,1);
				od_listctrl->SetItem(itemIndex,1, od_listctrl->GetItemText(prevItemIndex[0],1) );
				od_listctrl->SetItem(prevItemIndex[0],1, tmp);

				//also update the selection...
				od_listctrl->SetItemState(prevItemIndex[0],wxLIST_STATE_SELECTED,wxLIST_STATE_SELECTED);
				od_listctrl->SetItemState(itemIndex,0,wxLIST_STATE_SELECTED);

				//adjust the mapping as well
				tmp=listedFiles[ listMap[itemIndex] ];
				listedFiles[ listMap[itemIndex] ]=listedFiles[ listMap[prevItemIndex[0]] ];
				listedFiles[ listMap[prevItemIndex[0]] ]=tmp;

				//to keep moved selection visible:
				//remember the index where the first selected item has been moved to
				if (!foundFirstSelected)
				{
					firstSelectedItemIndex=prevItemIndex[0];
					foundFirstSelected=true;
				}
			}

			//rotate the buffer...
			for (int i=1; i < groupSize; ++i)
				prevItemIndex[i-1]=prevItemIndex[i];
			prevItemIndex[groupSize-1]=itemIndex;

			//advance the current index in the given direction
			if (direction == wxLIST_NEXT_BELOW)
			{ itemIndiciesIt++; }
			else
			{
				if (itemIndiciesIt == itemIndicies.begin())
					itemIndiciesIt=itemIndicies.end();
				else
					itemIndiciesIt--;
			}
		}

		//to keep moved selection visible:
		if (foundFirstSelected) od_listctrl->EnsureVisible(firstSelectedItemIndex);

		//"un-hide"
		od_listctrl->Show();
	}

	void OpenFileDialog::OnRemoveButton(wxCommandEvent& evt)
	{
		//if nothing selected, we have nothing to do...
		if (od_listctrl->GetSelectedItemCount() == 0) return;

		//get items to be deleted
		std::list<std::pair<long,wxString> > toRemoveData;
		this->GetActiveListContent(toRemoveData);

		std::list<std::pair<long,wxString> >::const_iterator removeIt = toRemoveData.begin();
		long deleted=0;
		while (removeIt != toRemoveData.end())
		{
			//the indices of the subsequent items decrease by one with every single removal
			this->listedFiles.RemoveAt(listMap[removeIt->first]-deleted);
			++removeIt;
			++deleted;
		}

		this->UpdateFilesOnList();
	}

	/********************* Re-drawing the List *********************/

	long OpenFileDialog::UpdateFilesOnList(bool keepSelection)
	{
		//indicator whether the current view should be kept as well
		//(which makes sense only in the cases when the redrawing of the control
		//is not due to a significant change (such as adding or removing) of the
		//content of the listedFiles, which is indicated with the parameter)
		const bool keepView=keepSelection;

		//check whether the number of items currently on the list
		//matches the desired number of items
		if (od_listctrl->GetItemCount() != (int)listedFiles.GetCount()) keepSelection=false;

		//if nothing is selected, we don't have to preserve it either
		if (od_listctrl->GetSelectedItemCount() == 0) keepSelection=false;

		// the selected group mode
		const GroupMode &grp_mode = m_grp_list.Get(od_grp_mode -> GetSelection());

		// to speed up inserting we hide the control temporarily (from wx samples)
		od_listctrl->Hide();

		//if selection should be preserved, remember what it looked like
		std::vector<size_t> indicesToSelect;
		const long topViewItem=od_listctrl->GetTopItem();
		if (keepSelection)
		{
			//fill the vector by scanning the whole control
			size_t scanPos=0;
			long itemIndex = -1;
			while ((itemIndex = od_listctrl->GetNextItem(itemIndex,
						  wxLIST_NEXT_ALL, wxLIST_STATE_DONTCARE)) != wxNOT_FOUND)
			{
				if (od_listctrl->GetItemState(itemIndex,wxLIST_STATE_SELECTED) > 0)
					indicesToSelect.push_back(scanPos);

				++scanPos;
			}
		}

		//init the List
		od_listctrl->DeleteAllItems();
		this->listMap.clear();

		//helper variables required only when re-selecting some items
		std::vector<size_t>::const_iterator toSelect=indicesToSelect.begin();

		//set content of the rows:
		long index=-1;
		for (size_t i=0; i < listedFiles.GetCount(); ++i)
		{
			//fill type column
			index=od_listctrl->InsertItem(i, wxString::Format("%d. %s", (i / grp_mode.Size()) + 1, grp_mode.Annotation(i % grp_mode.Size())));

			//fill file name column (including separators limit)
			wxString inFilename = listedFiles[i];
			wxString outFilename = wxEmptyString;
			int sepRemaining = od_sepcnt -> GetValue() + 1;
			int num_sep = inFilename.Freq(wxFILE_SEP_PATH);

			if (sepRemaining > num_sep)
			{
				outFilename = inFilename;
			}
			else
			{
				int index;

				do
				{
					index = inFilename.Find(wxFILE_SEP_PATH, true);
					outFilename.Prepend(outFilename.IsEmpty() ? inFilename.Mid(index + 1) : inFilename.Mid(index + 1) + wxFILE_SEP_PATH);
					inFilename = inFilename.Mid(0, index);
					--sepRemaining;
				} while (sepRemaining > 0);
			}
			
			od_listctrl->SetItem(index, 1, outFilename);
			//set background colour do enhance individual groups
			od_listctrl->SetItemBackgroundColour(index, wxColour((i / grp_mode.Size()) % 2 * 55 + 200, 255, 255));

			//re-select this one?
			if ((keepSelection) && (i == *toSelect))
			{
				od_listctrl->SetItemState(index,
				                          wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);
				++toSelect;
				//TODO, make more robust: indicesToSelect.end()
			}

			//store the binding to the complete paths
			this->listMap[index]=i;
		}

		if (keepView)
		{
			//ensure both the first line and bottom line items are visible
			od_listctrl->EnsureVisible(topViewItem);
			od_listctrl->EnsureVisible(topViewItem + od_listctrl->GetCountPerPage()-1);
		}

		//"un-hide"
		od_listctrl->Show();

		return (index);
	}

	void OpenFileDialog::OnUpdateUI(wxUpdateUIEvent &evt)
	{
		wxArrayString fnames;

		switch (evt.GetId())
		{
			case ID_OD_ADDTOLIST:
				od_filectrl -> GetFilePaths(fnames);
				evt.Enable(fnames.GetCount() > 0);
				//break;

			case ID_OD_REMOVE:
			case ID_OD_LISTEDFILES:
				od_listedfiles->SetLabel(wxString::Format("%d", od_listctrl -> GetItemCount()));
				break;

			case ID_OD_GROUPS_1:
				evt.Enable((int)m_grp_list.Get(od_grp_mode -> GetSelection()).Size() > 1);
				break;

			case ID_OD_GROUPS_2:
			case ID_OD_GROUPS_3:
				evt.Enable((int)m_grp_list.Get(od_grp_mode -> GetSelection()).Size() > (evt.GetId()-ID_OD_GROUPS_1));
				break;

			default: break;
		}
	}

	/********************* Closing the dialog *********************/

	void OpenFileDialog::OnOK(wxCommandEvent &evt)
	{
		bool mayClose=true;

		//some common tests:
		//empty list?
		if (od_listctrl->GetItemCount() == 0)
		{
			wxMessageBox("You must add at least one file on the Image list.",
			  "Inconsistent request", wxICON_EXCLAMATION, this);
			mayClose=false;
		}
		else
		//correctly long list?
		if ((od_listctrl->GetItemCount() % m_grp_list.Get(od_grp_mode->GetSelection()).Size()) != 0)
		{
			wxMessageBox("The number of files on the Image list\nmust align with group boundary.\n"
			  "It must be divisible with the number of images\nthat make up one time point in the given group.",
			  "Inconsistent request", wxICON_EXCLAMATION, this);
			mayClose=false;
		}

		//notification that something is not implemented yet...
		//unready opening modes
		if (od_grp_mode -> GetSelection() > 1)
		{
			wxMessageBox("Sorry, this variant is not implemented yet.",
			             "Not implemented", wxICON_EXCLAMATION, this);
			mayClose=false;
		}

		//unready opening options
		if (mayClose)
		{
			if ((m_xSampling > 1.0f) || (m_ySampling > 1.0f) || (m_zSampling > 1.0f)
			  || (od_swapzt->GetValue()))
			{
				wxMessageBox("Sorry, downsampling and swapping axes is not implemented yet.",
								 "Not implemented", wxICON_EXCLAMATION, this);
				mayClose=false;
			}
		}

		if (mayClose) EndDialog(wxID_OK);
	}
}
