/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
* dlg_openfile.h
*
* Dialog for opening files (be it a sequence, overlays, single files...)
*
* Vladimir Ulman , ulman@fi.muni.cz , 2015
*/

#ifndef V4D_DLG_OPENFILE_H
#define V4D_DLG_OPENFILE_H

#include <wx/dialog.h>
#include <wx/dirctrl.h>
#include <wx/listctrl.h>
#include <wx/spinctrl.h>
#include <wx/button.h>
#include <wx/textctrl.h>
#include <wx/radiobut.h>
#include <wx/checkbox.h>
#include <wx/stattext.h>
#include <wx/radiobox.h>
#include <list>
#include <map>
#include <utility>
#include <vector>

namespace v4d
{
	/** Group mode representation. */
	class GroupMode
	{
	private:
		/** Annotation list. */
		wxArrayString m_annotation;

	public:
		/** Constructor. */
		GroupMode(const wxString &ann) : m_annotation() { m_annotation.Add(ann); };
		/** Constructor. */
		GroupMode(const wxString &ann1, const wxString &ann2) : m_annotation() { m_annotation.Add(ann1); m_annotation.Add(ann2); };
		/** Constructor. */
		GroupMode(const wxString &ann1, const wxString &ann2, const wxString &ann3) : m_annotation() { m_annotation.Add(ann1); m_annotation.Add(ann2); m_annotation.Add(ann3); };
		/** Get the group mode size. */
		size_t Size() const { return m_annotation.GetCount(); };
		/** Get the i-th annotation. */
		const wxString &Annotation(size_t i) const { return m_annotation.Item(i); };
	};

	/*******************************************************************/

	/** A list of supported group modes. */
	class GroupModeList
	{
	private:
		/** A list of supported group modes. */
		std::vector<GroupMode> m_list;

	public:
		/** Constructor. */
		GroupModeList() : m_list()
		{
			m_list.push_back(GroupMode("IMG"));
			m_list.push_back(GroupMode("IMG", "MASK"));
			m_list.push_back(GroupMode("REAL", "IMAG"));
			m_list.push_back(GroupMode("REAL", "IMAG", "MASK"));
			m_list.push_back(GroupMode("RED", "GREEN"));
			m_list.push_back(GroupMode("RED", "GREEN", "BLUE"));
		}
		/** Get the i-th supported group mode. */
		const GroupMode &Get(size_t i) const { return m_list[i]; };
	};

	/*******************************************************************/

	/** The main widget for Adjust Levels dialog. */
	class OpenFileDialog: public wxDialog
	{
	public:
		/** Constructor
		* \param[in] parent		pointer to the parent ;-)
		* \param[in] folder		folder where the left panel should start
		* \param[in] filenames	list of files to populate the dialog initially with
		*/
		OpenFileDialog(wxWindow *parent,
		               const wxString& folder,
		               const wxArrayString* filenames);
		/** Destructor */
		~OpenFileDialog(); 

		///Returns the list of files on the List
		void GetFileNames(wxArrayString &names) const
		{ names=listedFiles; }

		///Returns the current folder used in the left panel
		void GetCurrentFolder(wxString& folder) const
		{ if (od_filectrl) folder = od_filectrl->GetPath(); }

		///Returns the current state of SeparateWindows checkbox
		bool GetCurrentSepWinMode(void)
		{ return( (od_ownwindow)? od_ownwindow->GetValue() : false ); }

		///Returns the currently selected ``group mode'' of the image sequence
		int GetCurrentGroupMode(void)
		{ return( (od_grp_mode)? od_grp_mode->GetSelection() : 0 ); }

		///Returns the current state of the SwapZT checkbox
		bool GetCurrentSwapZTMode(void)
		{ return( (od_swapzt)? od_swapzt->GetValue() : false ); }

		///Returns the current downsampling values
		void GetCurrentDownsampleRates(float& x,float& y,float& z) const
		{ x=m_xSampling; y=m_ySampling; z=m_zSampling; }

		///Returns the current number of separators displayed
		int GetCurrentPathSeparatorsUsed(void)
		{ return( (od_sepcnt)? od_sepcnt->GetValue() : 0 ); }

	private:
	/********************* the List *********************/

		/// User wants to add files on the List by clicking
		void OnAddToListClicked(wxTreeEvent& evt) { AddFiles(); };

		/// User wants to add files on the List by pressing a button
		void OnAddToListButton(wxCommandEvent& evt) { AddFiles(); };

		/// User wants to change the current folder in the Tree Dir list
		void OnChangeFolderButton(wxCommandEvent& evt);

		/// Update after the separator has changed
		void OnSeparatorChange(wxSpinEvent& evt)
		{ this->UpdateFilesOnList(true); }

		/** Add the files selected in the left panel to the list. */
		void AddFiles();

		/// User wants to select files on the List that match reg exp
		void OnSelectButton(wxCommandEvent& evt);

		/// User wants to invert current selection of files on the List
		void OnInvertButton(wxCommandEvent& evt);

		/// User wants to select no files on the List
		void OnDeselectButton(wxCommandEvent& evt);

		/**
		 * helper internal function that fills the \e content
		 * with file names from the List that user wants to
		 * manipulate somehow: this is usually only the selected
		 * items or the whole List if nothing is selected
		 *
		 * the content is the filename (wxString) together
		 * with its List's index (int)
		 */
		void GetActiveListContent(std::list<std::pair<long,wxString> >& content);

		/// User wants to add files on the List by pressing a button
		void OnSortButton(wxCommandEvent& evt);

		/**
		 * A helper routine to put selected items (in the Image list)
		 * at appropriate slot within the group. This is achieved by interleaving
		 * selected items with the rest. In fact, the relative position of the
		 * selected item is given with two params:
		 *
		 * \param[in] beforeInGroup   how many unselected items before a selected one
		 * \param[in] afterInGroup    how many unselected items after a selected one
		 *
		 * The group size should, therefore, be \e beforeInGroup + 1 + \e afterInGroup.
		 */
		void OnSlotHelper(const int beforeInGroup,const int afterInGroup);

		/// User wants selected items to occupy 1st slots within a group
		void OnSlot1Button(wxCommandEvent& evt);
		/// User wants selected items to occupy 2nd slots within a group
		void OnSlot2Button(wxCommandEvent& evt);
		/// User wants selected items to occupy 3rd slots within a group
		void OnSlot3Button(wxCommandEvent& evt);

		/**
		 * User wants to move up the selected items
		 * (or whole groups they are members of)
		 */
		void OnMoveHelper(const int direction);

		/// Wrapper for the moving up the selected items
		void OnMoveUpButton(wxCommandEvent& evt)
		{ this->OnMoveHelper(wxLIST_NEXT_BELOW); }

		/// Wrapper for the moving down the selected items
		void OnMoveDownButton(wxCommandEvent& evt)
		{ this->OnMoveHelper(wxLIST_NEXT_ABOVE); }

		/// User wants to remove the selected items
		void OnRemoveButton(wxCommandEvent& evt);

		/** The handler for wxUpdateUI events. */
		void OnUpdateUI(wxUpdateUIEvent &evt);

		/**
		 * Update the content of the List.
		 *
		 * \param[in] keepSelection	preserve original selection
		 *
		 * The parameter \e keepSelection indicates whether the selected
		 * items should be selected even after redrawing the control.
		 * This is, however, possible only when the number of items
		 * on the list has not changed, e.g., when redrawing only due to
		 * different grouping mode or when more path separators should
		 * be visible. Default value is false.
		 *
		 * Returns the index of the last inserted item.
		 */
		long UpdateFilesOnList(bool keepSelection=false);

		/**
		 * stores the association between indices that the List
		 * gave us (when adding new rows) and indices of this->listedFiles
		 */ 
		std::map<long,long> listMap;

		///the full path list of files currently on the List
		wxArrayString listedFiles;

	/********************* managing opening modes *********************/
	/*   incl. group modes, sampling dialog, various checkboxes...    */

		/// User wants to change current group mode
		void OnGroupModeChange(wxCommandEvent& evt)
		{ this->UpdateFilesOnList(true); };

		/// user wants to adjust sub-sampling options
		void OnSamplingDialogButton(wxCommandEvent& evt)
		{ samplingDlg.ShowModal(); }

	/********************* closing the dialog *********************/

		/** User pressed OK */
		void OnOK(wxCommandEvent& evt);

	/********************* handlers to controls *********************/
		///handlers on the main controls
		wxGenericDirCtrl* od_filectrl;

		wxSpinCtrl* od_sepcnt;
		wxListCtrl* od_listctrl;
		wxStaticText* od_listedfiles;
		wxTextCtrl* od_select_exp;
		wxCheckBox* od_movegroup;

		wxRadioBox *od_grp_mode;
		wxCheckBox *od_ownwindow, *od_swapzt;
		GroupModeList m_grp_list;

	/************* sampling dialog *************/
		///the sampling dialog itself
		wxDialog samplingDlg;

		///current sampling rates
		float m_xSampling, m_ySampling, m_zSampling;

		DECLARE_EVENT_TABLE();
	};
}

#endif
