/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * dlg_pref.h
 *
 * Dialog for changing screen options
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006
 * Adriana Smijakova 2015
 */

#ifndef __DLG_PREF_H__
#define __DLG_PREF_H__

#ifndef WX_PRECOMP
    #include <wx/wx.h>
#endif

#include <wx/colourdata.h>
#include <wx/spinctrl.h>
#include "v4d_canvas3d.h"

namespace v4d {
	
	/** The main widget for Adjust Levels dialog. */
	class PreferencesDialog: public wxDialog
	{
	  private:
		/** The pointer to structure with canvas settings. */
		v4d_CanvasSettings *m_settings;
		/** The pointer to radio box item with cut pointer types. */
		wxRadioBox *m_cut_ctrl;
		/** The pointer to static bitmap representing the cut pointer colour. */
		wxStaticBitmap *m_cut_bitmap;
		/** The colour palette for the cut pointer colour. */
		wxColourData *m_palette;
		/** The pointer to text control for resolution in x axis. */
		wxTextCtrl *m_res_x_ctrl;
		/** The pointer to text control for resolution in y axis. */
		wxTextCtrl *m_res_y_ctrl;
		/** The pointer to text control for resolution in z axis. */
		wxTextCtrl *m_res_z_ctrl;
		/** The pointer to text control for aspect ratio in x axis. */
		wxTextCtrl *m_rat_x_ctrl;
		/** The pointer to text control for aspect ratio in y axis. */
		wxTextCtrl *m_rat_y_ctrl;
		/** The pointer to text control for aspect ratio in z axis. */
		wxTextCtrl *m_rat_z_ctrl;

		////////////////////// Projection annotation /////////////////////

		/** The pointer to text control for xy annotation text. */
		wxTextCtrl *m_anot_xy_text;
		/** The pointer to text control for xz annotation text. */
		wxTextCtrl *m_anot_xz_text;
		/** The pointer to text control for yz annotation text. */
		wxTextCtrl *m_anot_yz_text;
		/** The pointer to radio box item with positions of xy annotation text. */
		wxRadioBox *m_anot_xypos_radiobox;
		/** The pointer to radio box item with positions of xz annotation text. */
		wxRadioBox *m_anot_xzpos_radiobox;
		/** The pointer to radio box item with positions of yz annotation text. */
		wxRadioBox *m_anot_yzpos_radiobox;
		/** The pointer to spin control for xy's x offset of annotation text. */
		wxSpinCtrl *m_pref_xyann_offset_x;
		/** The pointer to spin control for xy's y offset of annotation text. */
		wxSpinCtrl *m_pref_xyann_offset_y;
		/** The pointer to spin control for xz's x offset of annotation text. */
		wxSpinCtrl *m_pref_xzann_offset_x;
		/** The pointer to spin control for xz's y offset of annotation text. */
		wxSpinCtrl *m_pref_xzann_offset_y;
		/** The pointer to spin control for yz's x offset of annotation text. */
		wxSpinCtrl *m_pref_yzann_offset_x;
		/** The pointer to spin control for yz's y offset of annotation text. */
		wxSpinCtrl *m_pref_yzann_offset_y;

		////////////////////// Scale bar /////////////////////

		/** The pointer to text control for scale bar's length. */
		wxTextCtrl *m_pref_sblength_value;
		/** The pointer to spin control for scale bar's width. */
		wxSpinCtrl *m_pref_sbwidth_value;
		/** The pointer to radio box item with positions of scale bar. */
		wxRadioBox *m_pref_sb_corner;
		/** The pointer to spin control for scale bar's x offset. */
		wxSpinCtrl *m_pref_sb_offset_x;
		/** The pointer to spin control for scale bar's y offset. */
		wxSpinCtrl *m_pref_sb_offset_y;

		////////////////////// Time bar /////////////////////

		/** The pointer to text control for time bar's length. */
		wxSpinCtrl *m_pref_tblength_value;
		/** The pointer to spin control for time bar's width. */
		wxSpinCtrl *m_pref_tbwidth_value;
		/** The pointer to radio box item with positions of time bar. */
		wxRadioBox *m_pref_tb_corner;
		/** The pointer to spin control for time bar's x offset. */
		wxSpinCtrl *m_pref_tb_offset_x;
		/** The pointer to spin control for time bar's y offset. */
		wxSpinCtrl *m_pref_tb_offset_y;

	  public:
		/** Constructor. */
		PreferencesDialog(wxWindow *parent, v4d_CanvasSettings *settings, wxColourData *palette);
		/** Destructor. */
		~PreferencesDialog();
		/** Function called when a wxTextCtrl losing focus. */
		void OnKillFocus(wxFocusEvent &event);
		/** Function called when a wxSpinCtrl is changing value. */
		void OnSpinEvent(wxSpinEvent &event);
        
	  private:
		/** Function called when an user wants change the cut pointer style. */
		void OnChangePointerStyle(wxCommandEvent &event);
		/** Function called when an user wants change the cut pointer colour. */
		void OnChangePointerColour(wxCommandEvent &event);
		/** Set background colour of static bitmap to cut pointer colour. */
		void SetBitmapColour();
		/** Function called when an user wants change the font of annotation. */
		void OnChangeAnnotationFont(wxCommandEvent &event);
		/** Function called when an user wants change the colour of annotation. */
		void OnChangeAnnotationColour(wxCommandEvent &event);
		/** Update value in text control for aspect ratio according to value in text control for resolution. */
		void UpdateAspectRatioControl(const wxTextCtrl *res_ctrl, wxTextCtrl *ratio_ctrl);
		/** Update value in text control for resolution according to value in text control for aspect ratio. */
		void UpdateResolutionControl(const wxTextCtrl *ratio_ctrl, wxTextCtrl *res);
		/** Function called when an user wants change the xy annotation position. */
		void OnChangeXyAnnotPosition(wxCommandEvent &event);
		/** Function called when an user wants change the xz annotation position. */
		void OnChangeXzAnnotPosition(wxCommandEvent &event);
		/** Function called when an user wants change the yz annotation position. */
		void OnChangeYzAnnotPosition(wxCommandEvent &event);
		/** Function called when an user wants change the scale bar colour. */
		void OnChangeScaleBarColour(wxCommandEvent &event);
		/** Function called when an user wants change the time bar colour. */
		void OnChangeTimeBarColour(wxCommandEvent &event);
		/** Function called when an user wants change the scale bar position. */
		void OnChangeScaleBarCorner(wxCommandEvent &event);
		/** Function called when an user wants change the time bar position. */
		void OnChangeTimeBarCorner(wxCommandEvent &event);

   	    DECLARE_EVENT_TABLE()
	};
}

#endif
