/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * dlg_thres.cpp
 *
 * Implementation of thresholding dialog
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006
 */

#include "v4d_basic.h"
#include "dlg_thres.h"
#include "wdr/dlg_thres_wdr.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

#define ID_THRES_MINMAX_MAX 30000
#define ID_THRES_MINMAX_MIN 30001

namespace v4d
{
    wxDEFINE_EVENT(v4d_EVT_THRESHOLDING_PREVIEW, v4d_ThresholdingEvent);

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                   ThresholdingDialog class                  //
    //                                                             //
    /////////////////////////////////////////////////////////////////

    /*******************************************************************/
    template <class T> 
    ThresholdingDialog<T>::ThresholdingDialog(wxWindow *parent, wxWindowID id, 
        const std::vector<v4d::Histogram<T> *> &Histdata_vector,
        const std::vector<ThresParams<T> > &params_vector, 
        const wxArrayString &channel_names, const wxString &title, const wxPoint &pos, const wxSize& size, long style)
        : wxDialog(parent, id, title, pos, size, style), _HistData_vector(Histdata_vector), All_params(0), All_histdata(0) 
    {
        wxASSERT(_HistData_vector.size());
        wxASSERT(_HistData_vector.size() == params_vector.size());
        wxASSERT(_HistData_vector.size() == channel_names.GetCount());

        _ThresParams_vector.clear();
        for (size_t i = 0; i < params_vector.size(); i++)
        {
            _ThresParams_vector.push_back(params_vector[i]);
        }

        _hist = new HistGraph<T>(this, ID_THRES_HIST, _HistData_vector[0]);
        _hist -> SetGridLines(8, 4);
        _hist -> SetFgColour(wxColour(0, 0, 0));

        _minmax = new VariableMinMaxCtrl<T>(this, ID_THRES_MINMAX, false, false, false, *wxBLACK, *wxBLACK, *wxWHITE);
        _minmax -> SetParams(&(_ThresParams_vector[0].min_threshold), &(_ThresParams_vector[0].max_threshold), NULL,
            _hist -> GetMin(), _hist -> GetMax(), (int)_hist -> GetSize()); 

        ThresholdFunc(this);

        _HChoice = (wxChoice *) FindWindow(ID_THRES_HCHOICE);
        _HChoice -> SetSelection(_hist -> GetHistType());

        _ChannelChoice = (wxChoice *) FindWindow(ID_THRES_CHOICE_CHANNEL);
        _ChannelChoice -> Clear();
        for (size_t i = 0; i < channel_names.GetCount(); i++)
        {
            _ChannelChoice -> Append(channel_names.Item(i));
        }
        CreateChannelAll();

        _preview = (wxCheckBox *) FindWindow(ID_THRES_PREVIEW);
        _preview -> SetValue(TRUE);

		if (_HistData_vector.size() > 1)
		{
			_ChannelChoice -> SetStringSelection("All");
			SelectChannel(-1);
		}
		else
		{
			_ChannelChoice -> SetSelection(0);
			SelectChannel(0);
		}

        Bind(wxEVT_COMMAND_CHOICE_SELECTED, &ThresholdingDialog<T>::OnHChoice, this, ID_THRES_HCHOICE);
        Bind(wxEVT_COMMAND_CHOICE_SELECTED, &ThresholdingDialog<T>::OnChannelChoice, this, ID_THRES_CHOICE_CHANNEL);
        Bind(wxEVT_COMMAND_CHECKBOX_CLICKED, &ThresholdingDialog<T>::OnPreviewCheckBox, this, ID_THRES_PREVIEW);
        Bind(wxEVT_COMMAND_BUTTON_CLICKED, &ThresholdingDialog<T>::OnReset, this, ID_THRES_RESET);
        Bind(wxEVT_COMMAND_BUTTON_CLICKED, &ThresholdingDialog<T>::OnOK, this, wxID_OK);
        Bind(v4d_EVT_THRESHOLDING_PREVIEW, &ThresholdingDialog<T>::OnPreview, this);
        Bind(v4d_EVT_MINMAXCTRL, &ThresholdingDialog<T>::OnPreviewMinMax, this);
    }

    /*******************************************************************/
    template <class T> 
    ThresholdingDialog<T>::~ThresholdingDialog() 
    { 
        if (All_histdata) 
        {
            delete All_histdata; 
        }

        if (All_params)
        {
            delete All_params;
        }
    }

    /*******************************************************************/
    template <class T>
    void ThresholdingDialog<T>::GetParamsVector(std::vector<ThresParams<T> > &params)
    {
        if (_ThresParams_vector.size() == 1)
        {
            params.clear();
            params.insert(params.begin(), 3, _ThresParams_vector[0]);
        }
        else
        {
            params = _ThresParams_vector;
        }
    }

    /*******************************************************************/
    template <class T>
    int ThresholdingDialog<T>::CreateChannelAll()
    {
        if (_HistData_vector.size() > 1)
        {
            // check if all histograms have same parameters
            for (size_t i = 0; i < _HistData_vector.size() - 1; i++) 
            {
                if (*_HistData_vector[i] != *_HistData_vector[i + 1])
                {
                    return 0;
                }
            }

            // add new item to list of channels
            _ChannelChoice -> Append("All");

            // create new histogram
            All_histdata = new v4d::Histogram<T>(*_HistData_vector[0]);

            for (size_t i = 1; i < _HistData_vector.size(); i++)
            {
                *All_histdata += *_HistData_vector[i];
            }

            // check if all channels have the same minimal and maximal thresholds
            bool same_min = true, same_max = true;
            for (size_t i = 0; i < _ThresParams_vector.size() - 1; i++)
            {
                if (_ThresParams_vector[i].min_threshold != _ThresParams_vector[i + 1].min_threshold) 
                {
                    same_min = false;
                    break;
                }

                if (_ThresParams_vector[i].max_threshold != _ThresParams_vector[i + 1].max_threshold) 
                {
                    same_max = false;
                    break;
                }
            }

            // create new structure with parameters
            if (same_min && same_max)
            {
                All_params = new ThresParams<T>(_ThresParams_vector[0].min_threshold, 
                    _ThresParams_vector[0].max_threshold);
                return (_ChannelChoice -> GetCount() - 1);
            }
            else
            {
                All_params = new ThresParams<T>(_hist -> GetMin(), _hist -> GetMax());
                return 0;
            }
        }

        return 0;
    }

    /*******************************************************************/
    template <class T>
    void ThresholdingDialog<T>::SetChannelParams(const ThresParams<T> &params)
    {
        for (size_t i = 0; i < _ThresParams_vector.size(); i++)
        {
            _ThresParams_vector[i] = params;
        }
    }

    /*******************************************************************/
    template <class T>
    void ThresholdingDialog<T>::DoPreview()
    {
        ThresParams<T> params;
        if (_ChannelChoice -> GetStringSelection() == "All")
        {
            wxASSERT(All_params);
            params = *All_params;
            SetChannelParams(params);
        }
        else
        {
            params = _ThresParams_vector[_ChannelChoice -> GetSelection()];
        }

        if (_preview -> IsChecked())
        {
            v4d_ThresholdingEvent levt(v4d_EVT_THRESHOLDING_PREVIEW, params.ConvertToFloat(), _ChannelChoice -> GetSelection());

            wxEvtHandler *handler = GetParent() -> GetEventHandler();

            if (handler)
            {
                handler -> ProcessEvent(levt);
            }
        }
    }

    /*******************************************************************/
    template <class T>
    void ThresholdingDialog<T>::GeneratePreviewEvents(bool allChannels)
    {
        wxEvtHandler *handler = GetParent() -> GetEventHandler();

        if (allChannels)
        {
            wxASSERT(All_params);
            v4d_ThresholdingEvent some_event(v4d_EVT_THRESHOLDING_PREVIEW, (*All_params).ConvertToFloat(), 
                _ChannelChoice -> GetSelection());

            if (handler)
                handler -> ProcessEvent(some_event);
        }
        else
        {
            for (size_t i = 0; i < _ThresParams_vector.size(); i++)
            {
                v4d_ThresholdingEvent some_event(v4d_EVT_THRESHOLDING_PREVIEW, 
                    _ThresParams_vector[i].ConvertToFloat(), (int)i);

                if (handler)
                    handler -> ProcessEvent(some_event);
            }
        }
    }

    /*******************************************************************/
    template <class T>
    void ThresholdingDialog<T>::OnHChoice(wxCommandEvent &evt)
    {
        if (_hist -> GetHistType() != evt.GetInt())
        {
            _hist -> SetHistType(evt.GetInt());	
            _hist -> Refresh(FALSE);
        }
    }

    /*******************************************************************/

    template <class T>
    void ThresholdingDialog<T>::SelectChannel(int channel)
    {
        v4d::Histogram<T> *histdata;
        ThresParams<T> *params;

        if (channel == -1)
        {
            wxASSERT(All_histdata);
            histdata = All_histdata;
            wxASSERT(All_params);
            params = All_params;
        }
        else
        {
            histdata = _HistData_vector[channel];
            params = &_ThresParams_vector[channel];
        }

        _hist -> SetHistPointer(histdata);
		_minmax -> SetParams(&(params -> min_threshold), &(params -> max_threshold), NULL, _hist -> GetMin(), _hist -> GetMax(), (int)_hist -> GetSize());
		_hist -> Refresh(FALSE);

        if (_preview -> IsChecked())
        {
            GeneratePreviewEvents(channel == -1);
        }
    }

    /*******************************************************************/

    template <class T>
    void ThresholdingDialog<T>::OnChannelChoice(wxCommandEvent &evt)
    {
        int channel = (evt.GetString() == "All") ? -1 : evt.GetInt();
        SelectChannel(channel);
    }

    /*******************************************************************/
    template <class T>
    void ThresholdingDialog<T>::OnPreviewCheckBox(wxCommandEvent &evt)
    {
        if (_preview -> IsChecked())
        {
            GeneratePreviewEvents(_ChannelChoice->GetStringSelection() == "All");
        }
    }

    /*******************************************************************/
    template <class T> 
    void ThresholdingDialog<T>::OnPreview(v4d_ThresholdingEvent &evt)
    {
        DoPreview();
    }

    /*******************************************************************/
    template <class T> 
    void ThresholdingDialog<T>::OnPreviewMinMax(v4d_MinMaxCtrlEvent &evt)
    {
        DoPreview();
    }    

    /*******************************************************************/
    template <class T>
    void ThresholdingDialog<T>::OnReset(wxCommandEvent &evt)
    {
        _minmax -> Reset();
        _hist -> SetHistType(HIST_LOG);
        _HChoice -> SetSelection(HIST_LOG);
        _hist -> Refresh(FALSE);

        DoPreview();
    }

    /*******************************************************************/
    template <class T>
    void ThresholdingDialog<T>::OnOK(wxCommandEvent &evt)
    {
        if (_minmax -> Validate() && _minmax -> TransferDataFromWindow())
        {
            if (IsModal())
            {
                EndModal(wxID_OK);
            }
            else
            {
                SetReturnCode(wxID_OK);
                Show(FALSE);
            }
        }
    }

    /*******************************************************************/

    template class ThresholdingDialog<i3d::GRAY8>;
    template class ThresholdingDialog<i3d::GRAY16>;
    template class ThresholdingDialog<int>;
    template class ThresholdingDialog<float>;
}
