/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * dlg_thres.h
 *
 * Dialog for image thresholding
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006
 */

#ifndef V4D_DLG_THRES_H
#define V4D_DLG_THRES_H

#include "v4d_impexp.h"
#include "dlg_levels.h"
#include "types/thresparams.h"

namespace v4d
{
    /** An event sent by Thresholding dialog as a request for preview drawing. */
    class v4d_ThresholdingEvent : public wxEvent
    {
    private:
        /** Structure with thresholding parameters. */
        ThresParams<float> params;
        /** Selected channel. */
        int channel;

    public:
        /** Constructor. */
        v4d_ThresholdingEvent(wxEventType evtType, const ThresParams<float> &par, int ch, int id = 0): 
          wxEvent(id, evtType), params(par), channel(ch) {};
          /** Copy constructor. */
          v4d_ThresholdingEvent(const v4d_ThresholdingEvent &evt):
          wxEvent(evt), params(evt.params), channel(evt.channel) {}; 
          /** Return channel number. */
          int GetChannel() const { return channel; };
          /** Return structure with parameters. */
          const ThresParams<float> &GetParams() const { return params; };
          /** Set channel number. */
          void SetChannel(int ch) { channel = ch; };
          /** Set structure with parameters. */
          void SetParams(const ThresParams<float> &par) { params = par; };
          /** Clone method. */
          virtual wxEvent *Clone() const { return new v4d_ThresholdingEvent(*this); };
    };

    wxDECLARE_EXPORTED_EVENT(V4D_DLLEXPORT, v4d_EVT_THRESHOLDING_PREVIEW, v4d_ThresholdingEvent);

    /** The class representing the dialog for image thresholding. If the number of available channels is greater 
    * than one, the dialog creates new channel called All that provides possibility to threshold all channels 
    * together. This dialog generates v4d_ThresholdingEvent as a request for preview drawing. The event contains
    * structure ThresParams with thresholding parameters and index of selected channel (the value from the interval
    * [0, n - 1] where n is the number of available channels). If an user selects channel All then channel index of
    * v4d_ThresholdingEvent is n. */

    template <class T> class ThresholdingDialog : public wxDialog
    {
    private:
        /** Pointer to the HistGraph class. */
        HistGraph<T> *_hist;
        /** List of histograms for all available channels. */
        const std::vector<v4d::Histogram<T> *> &_HistData_vector;
        /** List of parameters for all avaible channels. */
        std::vector<ThresParams<T> > _ThresParams_vector;
        /** Control for histogram type selection. */
        wxChoice *_HChoice;
        /** Control for channel selection. */
        wxChoice *_ChannelChoice;
        /** Control for preview. */
        wxCheckBox *_preview;
        /** Control for thresholds adjusting. */
        VariableMinMaxCtrl<T> *_minmax;
        /** Pointer to structure with parameters for channel All. */
        ThresParams<T> *All_params;
        /** Pointer to histogram for channel All. */
        v4d::Histogram<T> *All_histdata;

    public:
        /** Constructor. */
        ThresholdingDialog(
            wxWindow *parent, 
            wxWindowID id, 
            const std::vector<v4d::Histogram<T> *> &Histdata_vector,
            const std::vector<ThresParams<T> > &params_vector,
            const wxArrayString &channel_names,
            const wxString &title = wxString("Adjust Thresholds"),
            const wxPoint& pos = wxDefaultPosition,
            const wxSize& size = wxDefaultSize,
            long style = wxRESIZE_BORDER | wxDEFAULT_DIALOG_STYLE | wxCLIP_CHILDREN);
        /** Destructor. */
        ~ThresholdingDialog(); 
        /** Return a list of parameters for all available channels. */
        void GetParamsVector(std::vector<ThresParams<T> > &params);

    private:
        /** If dialog contains at least two channels with the same ranges then the new channel All is created. 
        * This channel provides possibility to threshold all channels together. The return value determines 
        * selected channel during the first dialog displaying. */
        int CreateChannelAll();
        /** Set thresholding parameters of all available channels according to given thresholding parameters. */
        void SetChannelParams(const ThresParams<T> &params);
        /** Update level parameters according to the currently selected channel and generate a preview event for him. */
        void DoPreview();
        /** Generate sequence of preview events according to given string. */
        void GeneratePreviewEvents(bool allChannels);
        /** Change current channel. */
        void SelectChannel(int channel);

        /** Handler for type selection of histogram graph. */
        void OnHChoice(wxCommandEvent &evt);
        /** Handler for channel selection. */
        void OnChannelChoice(wxCommandEvent &evt);
        /** Handler for preview selection. */
        void OnPreviewCheckBox(wxCommandEvent &evt);
        /** Handler for preview drawing. */ 
        void OnPreview(v4d_ThresholdingEvent &evt);
        /** Handler for preview drawing. */ 
        void OnPreviewMinMax(v4d_MinMaxCtrlEvent &evt);
        /** Handler for Reset button. */
        void OnReset(wxCommandEvent &evt);
        /** Handler for OK button. */
        void OnOK(wxCommandEvent &evt);
    };
}

#endif
