/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * bcrop_image_cropper.h
 *
 * Implementation of ImageCropper class
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006
 */

// For compilers that support precompilation
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#include <new>
#include <fstream>
#include <typeinfo>
#include <wx/filename.h>
#include <wx/dir.h>
#include <wx/fileconf.h>
#include <i3d/image3d.h>
#include <i3d/imgI3D.h>
#include "image_cropper_by_slices.h"
#include "wxtools.h"
#include "batchcrop.h"

namespace bcrop {
	/*****************************************************************************\
	*
	*                            Local constants
	*
	\******************************************************************************/
	//--------------------------------------------------------------------------
	// output_file
	//--------------------------------------------------------------------------
	const wxString output_file = "crop.txt";

	/*****************************************************************************\
	*
	*                            Local functions
	*
	\******************************************************************************/
	//--------------------------------------------------------------------------
	// FilenameToFSpec
	//--------------------------------------------------------------------------
	bool FilenameToFSpec(const wxString &fname, wxString &filespec, const wxString &wildcard)
	{
		wxString af;   
		wxString fn = wxFileName(fname).GetFullName();

		if (fn.Contains("af."))
			af = "_af.";
		else if (filespec.Contains("_AF."))
			af = "_AF.";
		else if (filespec.Contains("_01."))
			af = "_01.";
		else
			return false;

		filespec = fn;
		return filespec.Replace(af, wildcard) == 1;
	} // FilenameToFSpec

	//--------------------------------------------------------------------------
	// GetFilesToCrop
	//--------------------------------------------------------------------------
	size_t GetFilesToCrop(const wxString &fname, wxArrayString &files, bool &use_mask)
	{
		wxString filespec;
		size_t numslc = 0;
		files.Clear();

		if (use_mask && FilenameToFSpec(fname, filespec, "_*."))
		{
			// get all files to crop at current position
			wxDir::GetAllFiles(fname.BeforeLast('\\'), &files, filespec, wxDIR_FILES);
			
			// count number of slices (no. of files that do not contain af substring)
			for (size_t i = 0; i < files.GetCount(); ++i)
			{
				if (!files[i].MakeLower().Contains("af")) 
					++numslc;
			}

			if (numslc == 0)
			{
				use_mask = false;
			}
		}
		else
		{
			files.Add(fname);
			use_mask = false;
		}

		return numslc;

	} // GetFilesToCrop

	//--------------------------------------------------------------------------
	// CreateI3D
	//--------------------------------------------------------------------------
	bool CreateI3D(
		const wxString &fname, 
		const wxString &mask, 
		size_t sx, 
		size_t sy, 
		size_t sz, 
		const char *i3dtype, 
		const i3d::Resolution &res)
	{
		std::ofstream ofs(fname.mb_str());
		ofs << "size: (" << sx << "," << sy << "," << sz << ")" << std::endl
			<< "offset: (0,0,0)" << std::endl
			<< "type: " << i3dtype << std::endl
			<< "resolution: " << res.GetRes() << std::endl
			<< "mask: ^" << mask.c_str() << std::endl;

		return true;
	} // CreateI3D

	/*****************************************************************************\
	*
	*                            ImageCropperBySlices
	*
	\******************************************************************************/
	//--------------------------------------------------------------------------
	// LoadAllBoxes
	//--------------------------------------------------------------------------
	bool ImageCropperBySlices::LoadAllBoxes(const wxString &odir, v4d::MapOfPositionBoxes &mp)
	{
		wxString last_pos, pos;
		float x, y, dx, dy;
		size_t id;

		mp.clear();
		
		std::ifstream ifs(wxString::Format("%s\\%s", odir.c_str(), output_file.c_str()).mb_str());
		while (ifs.good())
		{   
			v4d::LoadwxString(pos, ifs);
			ifs >> x >> y >> dx >> dy >> id;

			if (ifs.good())
			{
				if (pos == "-")
					pos = last_pos;
				else
					last_pos = pos;

				mp[pos.BeforeLast('\\')][pos].push_back(v4d::Box<float>(id, x, y, dx, dy));
			}
		}

		return true;
	} // LoadCropBoxes

	//--------------------------------------------------------------------------
	// SaveAllBoxes
	//--------------------------------------------------------------------------
	bool ImageCropperBySlices::SaveAllBoxes(const wxString &odir, const v4d::MapOfPositionBoxes &mp)
	{
		v4d::PositionBoxes boxes;
		v4d::PositionBoxes::iterator pit;

		v4d::ListOfBoxes list;
		v4d::ListOfBoxes::iterator lit;

		v4d::MapOfPositionBoxes::const_iterator mit = mp.begin();

		std::ofstream ofs(wxString::Format("%s\\%s", odir.c_str(), output_file.c_str()).mb_str());

		while (mit != mp.end())
		{
			boxes = mit -> second;
			pit = boxes.begin();

			while (pit != boxes.end())
			{
				list = pit -> second;
				lit = list.begin();

				while (lit != list.end())
				{
					EnsurePositiveSize(*lit);
					wxString str = wxString::Format("%s", (pit -> first).c_str());
					v4d::SavewxString(str, ofs);
					ofs << "\t" << lit -> offset.x 
						<< "\t" << lit -> offset.y 
						<< "\t" << lit -> size.x
						<< "\t" << lit -> size.y 
						<< "\t" << wxString::Format("%06u", lit -> id_box)
						<< std::endl;

					lit++;
				}

				pit++;
			}

			mit++;
		}

		return true;

	} // SaveCropBoxes

	//--------------------------------------------------------------------------
	// HasParams
	//--------------------------------------------------------------------------
	bool ImageCropperBySlices::HasParams()
	{
		return true;
	} // HasParams

	//--------------------------------------------------------------------------
	// SetParams
	//--------------------------------------------------------------------------
	void ImageCropperBySlices::SetParams(wxWindow *parent)
	{
		if (typeid(*parent) == typeid(BatchcropFrame))
		{
			BatchcropFrame *frame = dynamic_cast<BatchcropFrame *>(parent);

			wxString old_odir = m_odir;
			m_odir = frame -> GetLastOutputDir();
			m_res = frame -> GetRes();

			if (old_odir != m_odir)
			{
				LoadChannels();
			}

			ShowChannelDialog(parent);
		}
		else
		{
			wxLogError("ImageCropperBySlices::SetParams: Parent isn't BatchcropFrame!");
		}
		
	} // SetParams

	//--------------------------------------------------------------------------
	// CroppingAvail
	//--------------------------------------------------------------------------
	bool ImageCropperBySlices::CroppingAvail()
	{
		for (size_t i = 0; i < m_channels.size(); i++)
		{
			if (m_channels[i].comp != cc_NONE)
			{
				return true;
			}
		}

		return false;
	} // CroppingAvail

	//--------------------------------------------------------------------------
	// CropPosition
	//--------------------------------------------------------------------------
	void ImageCropperBySlices::CropPosition(const wxString &position, const v4d::ListOfBoxes &boxes)
	{
		if (boxes.empty())
		{
			// there is no box to crop
			return;
		}

		// get names of files to crop
		wxArrayString files; 
        bool use_mask = true;			
		size_t numslc = GetFilesToCrop(position, files, use_mask);

		// create output subdirectory 'regions'
		wxString dir = wxString::Format("%s\\regions", m_odir.c_str());
		if (!wxDirExists(dir))
		{
			if (!wxMkdir(dir))
			{
				wxLogError("bcrop_ImageCropper::CropPosition: Directory %s cannot be created! Cropping of position %s is aborted.", 
						   dir.c_str(), position.c_str());
				return;
			}
		}

		// create output subdirectories for boxes
		wxString box_dir;
		v4d::Box<float> box;
		v4d::ListOfBoxes boxes_to_crop;
		v4d::ListOfBoxes::const_iterator it = boxes.begin();
		while (it != boxes.end())
		{
			box_dir = wxString::Format("%s\\%06u", dir.c_str(), it -> id_box);

			if (!::wxDirExists(box_dir))
			{
				if (!wxMkdir(box_dir))
				{
					wxLogWarning("bcrop_ImageCropper::CropPosition: Directory %s cannot be created! Box #%u is ignored.", 
								 box_dir.c_str(), it -> id_box);
					it++;
					continue;
				}
			}

			// delete old data in the subdirectory for current box
			if (m_delete_old)
			{
				wxArrayString old_files;
				wxDir::GetAllFiles(box_dir, &old_files);
				for (size_t i = 0; i < old_files.GetCount(); i++)
				{
					if (!wxRemoveFile(old_files[i]))
					{
						wxLogWarning("bcrop_ImageCropper::CropPosition: File %s cannot be removed!", old_files[i].c_str());
					}
				}
			}

			// create i3d description files for current box
			if (use_mask)
			{
				wxString fname;
				wxString mask;
				size_t size_z = numslc == 0 ? 1 : numslc;

				// create i3d file for each defined channel
				for (size_t i = 0; i < m_channels.size(); i++)
				{
					if (m_channels[i].comp != cc_NONE)
					{
						fname = wxString::Format("%s\\c%02lu.i3d", box_dir.c_str(), static_cast<unsigned long>(i));

						if (m_delete_old || !wxFileExists(fname))
						{
							mask = wxString::Format("c%02lu_[0-9][0-9].tif", static_cast<unsigned long>(i));
							if (!CreateI3D(fname, mask, static_cast<size_t>(it -> size.x * m_res.GetX()), 
										   static_cast<size_t>(it -> size.y * m_res.GetY()), size_z, "RGB", m_res))
							{
								wxLogWarning("bcrop_ImageCropper::CropPosition: File %s cannot be created!", fname.c_str());
							}
						}
					}
				}
			}
			
			box = *it;
			PrepareBox(box);

			if (box.size.x >= 0 && box.size.y >= 0)
			{
				boxes_to_crop.push_back(box);
			}
			else
			{
				wxLogWarning("bcrop_ImageCropper::CropPosition: Box #%u is ignored because it doesn't belong to the input image!", 
					         it -> id_box);
			}

			it++;
		}

		bool is_decomposition_needed = IsDecompositionNeeded();

		// crop slice by slice
		for (size_t i = 0; i < files.GetCount(); i++)
		{
			CropSingleSlice(files[i], boxes_to_crop, is_decomposition_needed);
		}

		// convert image files to ICS format
		if (use_mask)
		{
			wxString i3dfn, icsfn, tmp;
			i3d::Image3d<i3d::GRAY8> img;
			v4d::ListOfBoxes::iterator bit = boxes_to_crop.begin();
			while (bit != boxes_to_crop.end())
			{
				for(size_t i = 0; i < m_channels.size(); i++)
				{
					if(m_channels[i].comp == cc_NONE || m_channels[i].comp == cc_RGB)	
						continue;

					tmp = wxString::Format("%s\\regions\\%06u\\c%02ld", m_odir.c_str(), bit -> id_box, static_cast<unsigned long>(i));
					i3dfn = tmp + ".i3d";
					icsfn = tmp + ".ics";

					if (m_delete_old || !wxFileExists(icsfn))
					{
						try
						{
							i3d::ImgVoxelType vt = i3d::ReadImageType(i3dfn.mb_str());
							
							switch (vt)
							{
							case i3d::Gray8Voxel:
								img.ReadImage(i3dfn.mb_str());
								img.SaveImage(icsfn.mb_str(), i3d::IMG_ICSv2);
								break;
							default:
								wxLogWarning("bcrop_ImageCropper::CropPosition: Only GRAY8 images can currently be converted to ICS format!");
							}
                            
						}
						catch(i3d::IOException &e)
						{
							wxLogError("bcrop_ImageCropper::CropPosition: %s", e.what.c_str());
						}
						catch(std::bad_alloc &)
						{
							wxLogError("bcrop_ImageCropper::CropPosition: Not enough memory to load image %s!", i3dfn.c_str());
						}
						catch(...)
						{
							wxLogError("bcrop_ImageCropper::CropSingleSlice: Unknown exception raised!");
						}
					}

					//delete the i3d file and its slices
					i3d::FileList slices; 
					{
						i3d::I3DReader ir(i3dfn.mb_str());
						ir.LoadImageInfo();
						ir.GetFileNames(slices);
					}

					for (size_t j = 0; j < slices.size(); j++)
					{
						if (!wxRemoveFile(wxString(slices[j].c_str(), wxConvUTF8)))
						{
							wxLogError("bcrop_ImageCropper::CropSingleSlice: File %s cannot be removed!", slices[j].c_str());
						}
					}
				}

				bit++;
			}
		}
	} // CropPosition

	//--------------------------------------------------------------------------
	// IsDecompositionNeeded
	//--------------------------------------------------------------------------
	bool ImageCropperBySlices::IsDecompositionNeeded()
	{
		for (size_t i = 0; i < m_channels.size(); i++)
		{
			if (m_channels[i].comp != cc_NONE && m_channels[i].comp != cc_RGB)
			{
				return true;
			}
		}

		return false;
	} // IsDecompositionNeeded

	//--------------------------------------------------------------------------
	// CropSingleSlice
	//--------------------------------------------------------------------------
	void ImageCropperBySlices::CropSingleSlice(
		const wxString &fname, 
		const v4d::ListOfBoxes &boxes,
		bool is_decomposition_needed)
	{
		if (boxes.empty())
		{
			// there is no box to crop
			return;
		}

		i3d::Image3d<i3d::RGB> img, cim;
		i3d::Image3d<i3d::GRAY8> r, g, b;

		wxString dir, name, ext;
		wxFileName::SplitPath(fname, &dir, &name, &ext);

		wxString cut = name.AfterLast('_');
		if (cut == name) 
		{
			// there is no '_'
			cut = wxEmptyString;
		}
		else
		{
			cut = "_" + cut;
		}

		// load input image
		try
		{
			img.ReadImage(fname.mb_str());
			img.SetResolution(m_res);
		}
		catch(i3d::IOException &e)
		{
			wxLogError("bcrop_ImageCropper::CropSingleSlice: %s!", e.what.c_str());
			return;
		}
		catch(std::bad_alloc &)
		{
			wxLogError("bcrop_ImageCropper::CropSingleSlice: Not enough memory to load image %s!", fname.c_str());
			return;
		}
		catch(...)
		{
			wxLogError("bcrop_ImageCropper::CropSingleSlice: Unknown exception raised!");
			return;
		}

		// crop all boxes
		v4d::ListOfBoxes::const_iterator it = boxes.begin();
		while (it != boxes.end())
		{
			// create a cropped image 
			i3d::VOI<i3d::MICRONS> voi(it -> offset.x, it -> offset.y, img.GetOffset().z, 
									   it -> size.x, it -> size.y, 1.0f / img.GetResolution().GetZ());

			try
			{
				cim = i3d::Image3d<i3d::RGB>(img, voi);
			}
			catch (i3d::InternalException &e)
			{
				wxLogError("bcrop_ImageCropper::CropSingleSlice: %s. Box #%u wasn't cropped successfully!",
					       e.what.c_str(), it -> id_box);
				it++;
				continue;
			}

			// decompose cropped image if it's required
			if (is_decomposition_needed)
			{
				i3d::RGBtoGray(cim, r, g, b);
			}
			
			// go through all channels and save requested images
			for (size_t i = 0; i < m_channels.size(); i++)
			{
				wxString fn = wxString::Format("%s\\regions\\%06u\\c%02lu%s.tif", 
											   m_odir.c_str(), it -> id_box, static_cast<unsigned long>(i), cut.c_str());

				if (m_delete_old || !wxFileExists(fn))
				{
					try 
					{
						switch (m_channels[i].comp)
						{
						case cc_RED:
							r.SaveImage(fn.mb_str());
							break;
						case cc_GREEN:
							g.SaveImage(fn.mb_str());
							break;
						case cc_BLUE:
							b.SaveImage(fn.mb_str());
							break;
						case cc_RGB:
							cim.SaveImage(fn.mb_str());
							break;
						default:
							break;
						}
					}
					catch(i3d::IOException &e)
					{
						wxLogError("bcrop_ImageCropper::CropSingleSlice: %s. Cropped image of box #%u for channel %s wasn't saved successfully!", 
								   e.what.c_str(), it -> id_box, m_channels[i].name.c_str());
					}
				}
			}

			it++;
		}
	} // CropSingleSlice

	//--------------------------------------------------------------------------
	// PrepareBox
	//--------------------------------------------------------------------------
	void ImageCropperBySlices::PrepareBox(v4d::Box<float> &box)
	{
		v4d::EnsurePositiveSize(box);

		if (box.offset.x < 0)
		{
			box.size.x += box.offset.x;
			box.offset.x = 0;
		}

		if (box.offset.y < 0)
		{
			box.size.y += box.offset.y;
			box.offset.y = 0;
		}
	} // PrepareBox

	//--------------------------------------------------------------------------
	// LoadChannels
	//--------------------------------------------------------------------------
	void ImageCropperBySlices::LoadChannels()
	{
		wxString fname = wxString::Format("%s\\channels.cfg", m_odir.c_str());
		wxFileConfig file(wxEmptyString, wxEmptyString, fname, wxEmptyString, wxCONFIG_USE_RELATIVE_PATH);

		if (file.HasGroup("names"))
		{
			wxString name;
			int i = 0;

			m_channels.clear();
			file.SetPath("/names");

			while (file.Read(wxString::Format("%d", i), &name))
			{
				m_channels.push_back(bcrop_Channel(name));
				i++;
			}
		}
	} // LoadChannels

	//--------------------------------------------------------------------------
	// SaveChannels
	//--------------------------------------------------------------------------
	void ImageCropperBySlices::SaveChannels()
	{
		wxString fname = wxString::Format("%s\\channels.cfg", m_odir.c_str());
		wxFileConfig file(wxEmptyString, wxEmptyString, fname, wxEmptyString, wxCONFIG_USE_RELATIVE_PATH);

		if (file.HasGroup("names"))
		{
			file.DeleteGroup("names");
		}

		file.SetPath("/names");

		for (size_t i = 0; i < m_channels.size(); i++)
		{
			file.Write(wxString::Format("%lu", static_cast<unsigned long>(i)), m_channels[i].name);
		}
	} // SaveChannels

	//--------------------------------------------------------------------------
	// ShowChannelDialog
	//--------------------------------------------------------------------------
	void ImageCropperBySlices::ShowChannelDialog(wxWindow *parent)
	{
		if (ReadChannelDlg(parent, m_channels))
		{
			SaveChannels();
		}
	} // ShowChannelDialog

} // namespace bcrop

