/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "v4d_basic.h"
#include <wx/frame.h>
#include <wx/numdlg.h>
#include <wx/busyinfo.h>
#include <wx/filedlg.h>
#include <algorithm>
#include "mapping.h"
#include "types/evt_mapping.h"
#include "dlg_levels.h"
#include "dlg_thres.h"
#include "dlg_set_range.h"
#include "palette.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

#define CUTOFF_LEFT(x,leftbound) if ((x) < (leftbound)) x = (leftbound)
#define CUTOFF_RIGHT(x,rightbound) if ((x) > (rightbound)) x = (rightbound)

namespace v4d 
{
    wxDEFINE_EVENT(v4d_EVT_MAPPING_CHANGED, v4d_MappingEvent);
    wxDEFINE_EVENT(v4d_EVT_MAPPING_COMMAND, v4d_MappingCommandEvent);

	enum {
		// bit depth
		MM_SET_BITDEPTH = 0,
		MM_SET_MIN_MAX,
		// stretch
		MM_NO_STRETCH,
		MM_LINEAR_STRETCH,
		MM_PERCENTIL_STRETCH,
		MM_HISTEQUAL_STRETCH,
		MM_LEVELS,
		MM_THRESHOLD,
		// colour mapping
		MM_GRAY,
		MM_SATURATION,
		MM_LABELS,
		MM_PERIODIC,
		MM_PALETTE,
		// channels
		MM_SHOW_RGB,
		MM_SHOW_RED,
		MM_SHOW_GREEN,
		MM_SHOW_BLUE,
		// channels for complex
		MM_SHOW_REAL,
		MM_SHOW_IMAGINARY,
		MM_SHOW_PHASE,
		MM_SHOW_AMPLITUDE,
		MM_LAST_ID
	};

    //#define IsBitDepthId(id) ((id) >= MM_8BIT && (id) <= MM_16BIT)
    #define IsStretchModeId(id) ((id) >= MM_NO_STRETCH && (id) <= MM_THRESHOLD)
    #define IsColourModeId(id) ((id) >= MM_GRAY && (id) <= MM_PALETTE)
    #define IsChannelId(id) ((id) >= MM_SHOW_RGB && (id) <= MM_SHOW_BLUE)
	#define IsComplexShowId(id) ((id) >= MM_SHOW_REAL && (id) <= MM_SHOW_AMPLITUDE)

	/////////////////////////////////////////////////////
	//
	// Auxiliary methods
	//
	/////////////////////////////////////////////////////

    /** Returns the number of bits needed to store the given number */
    static int BitCount(int num)
    {    
        int bits = 0;
        do
        {
            num >>= 1;
            bits++;
        } while (num != 0);
        return bits;
    }

	/////////////////////////////////////////////////////
	//
	// Mapping operators
	//
	/////////////////////////////////////////////////////
	template <class T> 
	class GrayLUTMapper 
    {
    private:
		const i3d::LUT<T, i3d::GRAY8> &m_lut;

	public:
		GrayLUTMapper(const i3d::LUT<T, i3d::GRAY8> &lut)
            : m_lut(lut) 
        {}

		void operator()(i3d::RGB &dst, const T &src) const
		{
			dst.red   = m_lut.Get(src);
			dst.green = m_lut.Get(src);
			dst.blue  = m_lut.Get(src);
		}
	};

	template <class T> 
	class PaletteMapper 
    {
    private:
		const i3d::LUT<i3d::GRAY8, i3d::RGB> &m_pal;
		const i3d::LUT<T, i3d::GRAY8> &m_lut;

	public:
		PaletteMapper(const i3d::LUT<T, i3d::GRAY8> &lut, const i3d::LUT<i3d::GRAY8, i3d::RGB> &pal)
            : m_pal(pal), m_lut(lut) 
        {}

		void operator()(i3d::RGB &dst, const T &src) const
		{
			dst = m_pal.Get(m_lut.Get(src));
		}
	};

	template <class T>
	class ComplexToRealLUTMapper
	{
	private:
		const i3d::LUT<T, i3d::GRAY8> &m_lut;
		const i3d::LUT<i3d::GRAY8, i3d::RGB> *m_pal;
	public:
		ComplexToRealLUTMapper(const i3d::LUT<T, i3d::GRAY8> &lut, const i3d::LUT<i3d::GRAY8, i3d::RGB> *pal = NULL)
			: m_lut(lut), m_pal(pal)
		{}

		void operator()(i3d::RGB &dst, const std::complex<T> &src) const
		{
			if (m_pal == NULL)
			{
				dst.red = m_lut.Get(src.real());
				dst.green = dst.red;
				dst.blue = dst.red;
			}
			else
			{
				dst = m_pal->Get(m_lut.Get(src.real()));
			}
		}
	};

	template <class T>
	class ComplexToImagLUTMapper
	{
	private:
		const i3d::LUT<T, i3d::GRAY8> &m_lut;
		const i3d::LUT<i3d::GRAY8, i3d::RGB> *m_pal;
	public:
		ComplexToImagLUTMapper(const i3d::LUT<T, i3d::GRAY8> &lut, const i3d::LUT<i3d::GRAY8, i3d::RGB> *pal = NULL)
			: m_lut(lut), m_pal(pal)
		{}

		void operator()(i3d::RGB &dst, const std::complex<T> &src) const
		{
			if (m_pal == NULL)
			{
				dst.red = m_lut.Get(src.imag());
				dst.green = dst.red;
				dst.blue = dst.red;
			}
			else
			{
				dst = m_pal->Get(m_lut.Get(src.imag()));
			}					
		}
	};

	template <class T>
	class ComplexToPhaseLUTMapper
	{
	private:
		const i3d::LUT<T, i3d::GRAY8> &m_lut;
		const i3d::LUT<i3d::GRAY8, i3d::RGB> *m_pal;
	public:
		ComplexToPhaseLUTMapper(const i3d::LUT<T, i3d::GRAY8> &lut, const i3d::LUT<i3d::GRAY8, i3d::RGB> *pal = NULL)
			: m_lut(lut), m_pal(pal)
		{}

		void operator()(i3d::RGB &dst, const std::complex<T> &src) const
		{
			//atan2 treats signs properly (and also when src.real() == 0)
			//and returns values from [-PI; PI] (phase in radians)
			if (m_pal == NULL)
			{
				dst.red = m_lut.Get(std::atan2(src.imag(), src.real()));
				dst.green = dst.red;
				dst.blue = dst.red;
			}
			else
			{
				dst = m_pal->Get(m_lut.Get(std::atan2(src.imag(), src.real())));
			}
		}
	};

	template <class T>
	class ComplexToAmpLUTMapper
	{
	private:
		const i3d::LUT<T, i3d::GRAY8> &m_lut;
		const i3d::LUT<i3d::GRAY8, i3d::RGB> *m_pal;
	public:
		ComplexToAmpLUTMapper(const i3d::LUT<T, i3d::GRAY8> &lut, const i3d::LUT<i3d::GRAY8, i3d::RGB> *pal = NULL)
			: m_lut(lut), m_pal(pal)
		{}

		void operator()(i3d::RGB &dst, const std::complex<T> &src) const
		{
			const T real = src.real();
			const T imag = src.imag();

			if (m_pal == NULL)
			{
				dst.red = m_lut.Get(sqrt(real*real + imag*imag));
				dst.green = dst.red;
				dst.blue = dst.red;
			}
			else
			{
				dst = m_pal->Get(m_lut.Get(sqrt(real*real + imag*imag)));
			}
		}
	};

	template <class T> class RGBLUTMapper
    {
    private:
		const i3d::LUT<T, i3d::GRAY8> &m_rlut;
		const i3d::LUT<T, i3d::GRAY8> &m_glut;
		const i3d::LUT<T, i3d::GRAY8> &m_blut;

	public:
		RGBLUTMapper(const i3d::LUT<T, i3d::GRAY8> &r, 
			      const i3d::LUT<T, i3d::GRAY8> &g, 
				  const i3d::LUT<T, i3d::GRAY8> &b)
			:  m_rlut(r), m_glut(g), m_blut(b)
        {}

		void operator()(i3d::RGB &dst, const i3d::RGB_generic<T> &src) const
		{
			dst.red   = m_rlut.Get(src.red);
			dst.green = m_glut.Get(src.green);
			dst.blue  = m_blut.Get(src.blue);
		}
	};

    /** Map image data to rgb buffer. */
    template <class VOXEL, class Mapper>
    void MapToRGBBuf(const VOXEL *src, Buffer2d<i3d::RGB> &buf, const Mapper &map)
    {
        i3d::RGB *last = buf.GetData() + buf.GetSize();		
        for (i3d::RGB *dst = buf.GetData(); dst != last; ++dst)
        {
            map(*dst,*src);
            ++src;
        }
    }

    /******************************************************************************/
    //                          Generic Mapping Module
    /******************************************************************************/
    template <class T>
    v4d_MappingModule<T>::v4d_MappingModule(wxFrame *parent, const HistogramSet<T> &hist_set, double percentil_val)
        : v4d_ExtensionModule(parent), v4d_Mapper<T>(), m_hist_set(hist_set), m_stretch_mode(MM_NO_STRETCH), m_percentil_value(percentil_val) 
    {}

    template <class T>
    int v4d_MappingModule<T>::GetStretchMode()
    {
        return m_stretch_mode - MM_NO_STRETCH;
    }

    template <class T>
    wxMenu *v4d_MappingModule<T>::CreateMenu(int &last_id)
    {
        // the menu can be created only once
        wxASSERT(this->GetMenuId(0) == 0);
        SetFirstMenuId(++last_id);
        wxMenu *menu = new wxMenu();

        menu->Append(this->GetMenuId(MM_SET_MIN_MAX), "Set image ran&ge...", "Set the range of displayed image intensities");
        menu->AppendSeparator();
        menu->Append(this->GetMenuId(MM_NO_STRETCH), "&No stretch\tAlt-1", "Apply no stretching", wxITEM_CHECK);
        menu->Append(this->GetMenuId(MM_LINEAR_STRETCH), "&Linear stretch\tAlt-2", "Apply linear stretch", wxITEM_CHECK);
		menu->Append(this->GetMenuId(MM_PERCENTIL_STRETCH), "&Percentil stretch\tAlt-3", "Apply percentil stretch", wxITEM_CHECK);
		
		//PATCH BECAUSE OF LINUX WXGTK PORT...
		//(which allows to change state of disabled menu item with its accelerator key, regardless of UpdateUI()...)
        if (typeid(T) != typeid(i3d::RGB) && typeid(T) != typeid(i3d::RGB16))
		{
			//no RGB image
			menu->Append(this->GetMenuId(MM_HISTEQUAL_STRETCH), "&Histogram equalization\tAlt-4", "Apply histogram equalization", wxITEM_CHECK);
		}
		else
		{
			//yes RGB image
			menu->Append(this->GetMenuId(MM_HISTEQUAL_STRETCH), "&Histogram equalization", "Apply histogram equalization", wxITEM_CHECK);

			wxLogMessage("Disabling Histogram equalization as it is not implemented for color images.");
			menu->Enable(GetMenuId(MM_HISTEQUAL_STRETCH),false);
		}

        menu->Append(this->GetMenuId(MM_LEVELS), "Levels\tAlt-5", "Show levels dialog", wxITEM_CHECK);
        menu->AppendSeparator();
        menu->Append(this->GetMenuId(MM_THRESHOLD), "Threshold\tAlt-6", "Apply thresholding", wxITEM_CHECK);
        last_id = this->GetMenuId(MM_LAST_ID);

		//connect update events with enabled menu items only
        Bind(wxEVT_UPDATE_UI, &v4d_MappingModule<T>::OnUpdateUI, this, GetMenuId(MM_SET_BITDEPTH), GetMenuId(MM_THRESHOLD));

        m_menu_ptr = menu;

        return menu;
    }

    template <class T>
    void v4d_MappingModule<T>::OnMapping(wxCommandEvent &evt)
    {
        int cid = GetCommandId(evt.GetId());

        if (cid == MM_SET_MIN_MAX)
        {
            SetRange();            
        }

        if (cid == MM_SET_BITDEPTH)
        {
            SetBitdepth();
        }

        if (IsStretchModeId(cid))
        {
            bool apply = true;

            switch (cid)
            {
            case MM_NO_STRETCH:
                SetIdentity();
                break;

            case MM_LINEAR_STRETCH:
                SetLinearStretch();
                break;

            case MM_PERCENTIL_STRETCH:
                SetPercentilStretch();
                break;

            case MM_HISTEQUAL_STRETCH:
                SetHistEqualStretch();
                break;

            case MM_LEVELS:
                apply = SetLevelsStretch();
                break;

            case MM_THRESHOLD:
                apply = SetThreshold();
                break;
            }

            if (apply)
                m_stretch_mode = cid;
            else
                m_menu_ptr->Check(this->GetMenuId(m_stretch_mode), true);
        }
    }

    template <class T>
    void v4d_MappingModule<T>::OnUpdateUI(wxUpdateUIEvent &evt)
    {
        int cid = GetCommandId(evt.GetId());

        //evt.Enable(HasData());

        if (IsStretchModeId(cid))
        {
            evt.Check(cid == m_stretch_mode);
        }
    }

    template class V4D_DLLEXPORT v4d_MappingModule<i3d::GRAY8>;
    template class V4D_DLLEXPORT v4d_MappingModule<i3d::GRAY16>;
    template class V4D_DLLEXPORT v4d_MappingModule<i3d::RGB>;
    template class V4D_DLLEXPORT v4d_MappingModule<i3d::RGB16>;

    /******************************************************************************/
    //                            RGB Mapping Module
    /******************************************************************************/
	
    template <class T>
    v4d_RGBMappingModule<T>::v4d_RGBMappingModule(wxFrame *parent, const HistogramSet<i3d::RGB_generic<T> > &hist_set)
        : v4d_MappingModule< i3d::RGB_generic<T> > (parent, hist_set), m_show_red(true), m_show_green(true), m_show_blue(true)
    {
        m_min_value.red = hist_set[0].GetLeftCenter();
        m_min_value.green = hist_set[1].GetLeftCenter();
        m_min_value.blue = hist_set[2].GetLeftCenter();

        m_max_value.red = hist_set[0].GetRightCenter();
        m_max_value.green = hist_set[1].GetRightCenter();
        m_max_value.blue = hist_set[2].GetRightCenter();

        m_levels_arr.resize(3);
        m_thres_arr.resize(3);

        for (int i = 0; i < 3; ++i)
        {
            m_levels_arr[i] = TLevelsParams<T, i3d::GRAY8>(m_min_value.GetComponent(i), m_max_value.GetComponent(i));
            m_thres_arr[i] = ThresParams<T>(m_min_value.GetComponent(i), m_max_value.GetComponent(i));
        }

        ResetMapping();
    }

    template <class T> bool v4d_RGBMappingModule<T>::HasData()
    {
        return (!m_hist_set[0].empty() || !m_hist_set[1].empty() || !m_hist_set[2].empty());
    }

    template <class T> 
    wxMenu* v4d_RGBMappingModule<T>::CreateMenu(int &last_id)
    {
        wxMenu *menu = v4d_MappingModule< i3d::RGB_generic<T> >::CreateMenu(last_id);

	    if (typeid(T) == typeid(i3d::GRAY8) || typeid(T) == typeid(i3d::GRAY16)) 
		{
            menu->Insert(1, GetMenuId(MM_SET_BITDEPTH), "Set bitdepth...", "Set \"real\" image bitdepth");
        }

        menu->AppendSeparator();
        menu->Append(GetMenuId(MM_SHOW_RGB), "&Show RGB\tF5", "Show all three rgb channels");
        menu->AppendSeparator();
        menu->Append(GetMenuId(MM_SHOW_RED), "&Red channel\tF6", "Show/hide red channel", wxITEM_CHECK);
        menu->Append(GetMenuId(MM_SHOW_GREEN), "&Green channel\tF7", "Show/hide green channel", wxITEM_CHECK);
        menu->Append(GetMenuId(MM_SHOW_BLUE), "&Blue channel\tF8", "Show/hide blue channel", wxITEM_CHECK);

        Bind(wxEVT_COMMAND_MENU_SELECTED, &v4d_RGBMappingModule<T>::OnMapping, this, GetMenuId(0), last_id);
        Bind(wxEVT_UPDATE_UI, &v4d_RGBMappingModule<T>::OnUpdateUI, this, GetMenuId(MM_SHOW_RGB), GetMenuId(MM_SHOW_BLUE));
        Bind(v4d_EVT_LEVELS_PREVIEW, &v4d_RGBMappingModule<T>::OnPreviewLevels, this);
        Bind(v4d_EVT_THRESHOLDING_PREVIEW, &v4d_RGBMappingModule<T>::OnPreviewThresholding, this);

        return menu;
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::Map(const i3d::RGB_generic<T> *src, Buffer2d<i3d::RGB> &buf) const
    {
        MapToRGBBuf(src, buf, RGBLUTMapper<T>(
            m_show_red   ? m_rlut : m_null,
            m_show_green ? m_glut : m_null,
            m_show_blue  ? m_blut : m_null));
    }

    template <class T>
    void v4d_RGBMappingModule<T>::SetRange()
    {
        i3d::RGB_generic<T> min_limit(m_hist_set[0].GetLeftCenter(), m_hist_set[1].GetLeftCenter(), m_hist_set[2].GetLeftCenter());
        i3d::RGB_generic<T> max_limit(m_hist_set[0].GetRightCenter(), m_hist_set[1].GetRightCenter(), m_hist_set[2].GetRightCenter());
        i3d::RGB_generic<T> lminv = m_min_value, lmaxv = m_max_value;

        if (GetRGBRangeFromUser<T>(this->GetParent(), min_limit, max_limit, lminv, lmaxv))
        {
            m_min_value = lminv;
            m_max_value = lmaxv;
            ResetMapping();
        }        
    }

    template <class T>
    void v4d_RGBMappingModule<T>::SetBitdepth()
    {
        int max_val = std::min(m_max_value.red, std::min(m_max_value.green, m_max_value.blue));
        int max_limit = std::min(m_hist_set[0].GetRightCenter(), std::min(m_hist_set[1].GetRightCenter(), m_hist_set[2].GetRightCenter()));

        int bit_depth = BitCount(max_val);
        int max_bit_depth = BitCount(max_limit);

        long ret = ::wxGetNumberFromUser("Set image bit depth", "Bit depth = ", "Image Bit Depth", 
            bit_depth, 1, max_bit_depth, this->GetParent());

        if (ret != -1 && ret != bit_depth)
        {
            m_max_value.red = T((1 << ret) - 1);
            m_max_value.green = T((1 << ret) - 1);
            m_max_value.blue = T((1 << ret) - 1);

            ResetMapping();
        }
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::SetIdentity()
    {
        for (int i = 0; i < 3; i++)
        {
            m_levels_arr[i] = TLevelsParams<T, i3d::GRAY8>(m_min_value.GetComponent(i), m_max_value.GetComponent(i));
        }

        UpdateLUTs();
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::UpdateLUTs()
    {
        UpdateChannelLUT(m_rlut, m_levels_arr[0]);
        UpdateChannelLUT(m_glut, m_levels_arr[1]);
        UpdateChannelLUT(m_blut, m_levels_arr[2]);
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::UpdateChannelLUT(i3d::LUT<T, i3d::GRAY8> &lut, const TLevelsParams<T, i3d::GRAY8> &params)
    {
        lut.CreateGamma(
            params.inputMin, 
            params.inputMax, 
            params.inputGamma, 
            params.outputMin, 
            params.outputMax); 
    }

    template <class T>
    void v4d_RGBMappingModule<T>::DoThresholding()
    {
        DoChannelThresholding(m_rlut, m_thres_arr[0], 0);
        DoChannelThresholding(m_glut, m_thres_arr[1], 1);
        DoChannelThresholding(m_blut, m_thres_arr[2], 2);
    }

    template <class T>
    void v4d_RGBMappingModule<T>::DoChannelThresholding(i3d::LUT<T, i3d::GRAY8> &lut, 
        const ThresParams<T> &params, int channel)
    {
        // Do thresholding of the channel only if at least one threshold differs from
        // its default
        if (params.min_threshold != this->m_min_value.GetComponent(channel) ||
            params.max_threshold != this->m_max_value.GetComponent(channel))
        {
            lut.CreateThresholding(params.min_threshold, params.max_threshold);
        }
        else
        {
            UpdateChannelLUT(lut, m_levels_arr[channel]);
        }
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::SetLinearStretch()
    {
        for (int i = 0; i < 3; ++i)
        {
            T l1, l2;

            Histogram<T> h;
            h.CreateSubHist(m_hist_set[i], this->m_min_value.GetComponent(i), this->m_max_value.GetComponent(i));

            l1 = h.GetBinCenter(h.FindFirstNonZero());
            l2 = h.GetBinCenter(h.FindLastNonZero());
            if (l1 >= l2) { // we have zero channel
                l1 = h.GetLeftCenter();
                l2 = h.GetRightCenter();
            }
            m_levels_arr[i] = TLevelsParams<T, i3d::GRAY8>(l1, l2);
            wxLogDebug("linear stretch %d - %u %u", i, l1, l2);
        }

        UpdateLUTs();
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::SetPercentilStretch()
    {
        for (int i = 0; i < 3; ++i)
        {
            T l1, l2;

            Histogram<T> h;
            h.CreateSubHist(m_hist_set[i], this->m_min_value.GetComponent(i), this->m_max_value.GetComponent(i));

            l1 = h.GetBinCenter(h.FindLeftPercentil(m_percentil_value));
            l2 = h.GetBinCenter(h.FindRightPercentil(m_percentil_value));
            m_levels_arr[i] = TLevelsParams<T, i3d::GRAY8>(l1, l2);
            wxLogDebug("percentil stretch %d - %u %u", i, l1, l2);
        }

        UpdateLUTs();
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::SetHistEqualStretch()
    {
	   wxLogError("The function does nothing. It is implemented only for GRAY or float images.");
    }

    template <class T>
    void v4d_RGBMappingModule<T>::OnHistogramChange(const HistogramSet< i3d::RGB_generic<T> > &hist_set)
    {
        m_hist_set = hist_set;

        m_min_value.red = std::max(m_min_value.red, m_hist_set[0].GetLeftCenter());
        m_min_value.green = std::max(m_min_value.green, m_hist_set[1].GetLeftCenter());
        m_min_value.blue = std::max(m_min_value.blue, m_hist_set[2].GetLeftCenter());

        m_max_value.red = std::min(m_max_value.red, m_hist_set[0].GetRightCenter());
        m_max_value.green = std::min(m_max_value.green, m_hist_set[1].GetRightCenter());
        m_max_value.blue = std::min(m_max_value.blue, m_hist_set[2].GetRightCenter());

        ResetMapping();
    }

    template <class T> void v4d_RGBMappingModule<T>::ResetMapping()
    {
        m_rlut.Init();
        m_glut.Init();
        m_blut.Init();
        m_null.Init();
        m_null.CreateAffine(0.0, 0.0);

        // update thresholding and levels params
        for (int i = 0; i < 3; i++) 
        {
            int pos = (int)m_hist_set[i].GetBinPos(this->m_min_value.GetComponent(i));
            CUTOFF_LEFT(m_levels_arr[i].inputMin, pos);
            CUTOFF_LEFT(m_levels_arr[i].inputMax, pos);
            CUTOFF_LEFT(m_thres_arr[i].min_threshold, pos);
            CUTOFF_LEFT(m_thres_arr[i].max_threshold, pos);

            pos = (int)m_hist_set[i].GetBinPos(this->m_max_value.GetComponent(i));
            CUTOFF_RIGHT(m_levels_arr[i].inputMin, pos);
            CUTOFF_RIGHT(m_levels_arr[i].inputMax, pos);
            CUTOFF_RIGHT(m_thres_arr[i].min_threshold, pos);
            CUTOFF_RIGHT(m_thres_arr[i].max_threshold, pos);
        }

        switch (this->m_stretch_mode) 
        {
			case MM_NO_STRETCH: SetIdentity(); break;
			case MM_LINEAR_STRETCH: SetLinearStretch(); break;
			case MM_PERCENTIL_STRETCH: SetPercentilStretch(); break;
			case MM_HISTEQUAL_STRETCH: SetHistEqualStretch(); break;
			case MM_LEVELS: UpdateLUTs(); break;
			case MM_THRESHOLD: DoThresholding(); break;
			default: break;
        }
    }

    template <class T> 
    bool v4d_RGBMappingModule<T>::SetLevelsStretch()
    {
        wxArrayString channel_names;
        channel_names.Add("Red");
        channel_names.Add("Green");
        channel_names.Add("Blue");

        std::vector<Histogram<T> *> histdata_vector;
        Histogram<T> *hist = NULL;

        for (int i = 0; i < 3; i++)
        {
            hist = new Histogram<T>();
            hist -> CreateSubHist(m_hist_set[i], this->m_min_value.GetComponent(i), this->m_max_value.GetComponent(i));
            histdata_vector.push_back(hist);
        }

        LevelsWidget<T, i3d::GRAY8> *lw = new LevelsWidget<T, i3d::GRAY8>(GetParent(), -1, histdata_vector, 
            m_levels_arr, channel_names);

        bool dlg_ok = (lw -> ShowModal() == wxID_OK);
        if (dlg_ok)
        {
            lw -> GetParamsVector(m_levels_arr);
            UpdateLUTs();
        }
        else
        {
            ResetMapping();
        }

        for (size_t i = 0; i < 3; i++)
        {
            delete histdata_vector[i];
        }

        return dlg_ok;
    }

    template <class T> 
    bool v4d_RGBMappingModule<T>::SetThreshold()
    {
        wxArrayString channel_names;
        channel_names.Add("Red");
        channel_names.Add("Green");
        channel_names.Add("Blue");

        std::vector<Histogram<T> *> histdata_vector;
        Histogram<T> *hist = NULL;

        for (int i = 0; i < 3; i++)
        {
            hist = new Histogram<T>();
            hist -> CreateSubHist(m_hist_set[i], this->m_min_value.GetComponent(i), this->m_max_value.GetComponent(i));
            histdata_vector.push_back(hist);
        }

        ThresholdingDialog<T> *dlg = new ThresholdingDialog<T>(GetParent(), -1, histdata_vector, m_thres_arr, channel_names);

        bool dlg_ok = (dlg -> ShowModal() == wxID_OK);
        if (dlg_ok)
        {
            dlg -> GetParamsVector(m_thres_arr);
            DoThresholding();
        }
        else
        {
            ResetMapping();
        }

        for (size_t i = 0; i < 3; ++i)
        {
            delete histdata_vector[i];
        }

        return dlg_ok;
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::OnMapping(wxCommandEvent &evt)
    {
        // call common mapping handler first
        v4d_MappingModule< i3d::RGB_generic<T> >::OnMapping(evt);

        int id = this->GetCommandId(evt.GetId());

		if (id == MM_SHOW_RGB)
        {
            m_show_red = true;
			m_show_green = true;
			m_show_blue = true;
        }
        switch (id)
        {
        case MM_SHOW_RED:
            m_show_red = evt.IsChecked();
            break;
        case MM_SHOW_GREEN:
            m_show_green = evt.IsChecked();
            break;
        case MM_SHOW_BLUE:
            m_show_blue = evt.IsChecked();
            break;
        }

        v4d_MappingEvent mevt(v4d_EVT_MAPPING_CHANGED);
        GetParent() -> ProcessWindowEvent(mevt);

        if (evt.ShouldPropagate())
        {
            v4d_MappingCommandEvent mcevt(v4d_EVT_MAPPING_COMMAND);
            mcevt.SetAttachedEvent(evt);
            mcevt.SetClientData(this);
            GetParent() -> ProcessWindowEvent(mcevt);
        }
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::OnUpdateUI(wxUpdateUIEvent &evt)
    {
        int cid = this->GetCommandId(evt.GetId());

        if (IsChannelId(cid))
        {
            evt.Enable(this->HasData());

            switch (cid)
            {
            case MM_SHOW_RED:
				evt.Check(m_show_red);
                break;
            case MM_SHOW_GREEN:
                evt.Check(m_show_green);
                break;
            case MM_SHOW_BLUE:
                evt.Check(m_show_blue);
                break;
            }
        }
        else
		{
            evt.Skip();
		}

    }

    template <class T> 
    void v4d_RGBMappingModule<T>::OnPreviewLevels(v4d_LevelsEvent &evt)
    {
        TLevelsParams<T, i3d::GRAY8> params(evt.GetParams());
        switch (evt.GetChannel())
        {
        case 0: 
            UpdateChannelLUT(m_rlut, params); 
            break;

        case 1: 
            UpdateChannelLUT(m_glut, params); 
            break;

        case 2: 
            UpdateChannelLUT(m_blut, params); 
            break;

        case 3: 
            UpdateChannelLUT(m_rlut, params);
            UpdateChannelLUT(m_glut, params); 
            UpdateChannelLUT(m_blut, params);
            break;
        }

        v4d_MappingEvent mevt(v4d_EVT_MAPPING_CHANGED, 0, false);
        GetParent() -> ProcessWindowEvent(mevt);
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::OnPreviewThresholding(v4d_ThresholdingEvent &evt)
    {
        ThresParams<T> params(evt.GetParams());
        switch (evt.GetChannel())
        {
        case 0: 
            DoChannelThresholding(m_rlut, params, 0); 
            break;

        case 1: 
            DoChannelThresholding(m_glut, params, 1); 
            break;

        case 2: 
            DoChannelThresholding(m_blut, params, 2); 
            break;

        case 3: 
            DoChannelThresholding(m_rlut, params, 0);
            DoChannelThresholding(m_glut, params, 1); 
            DoChannelThresholding(m_blut, params, 2);
            break;
        }

        v4d_MappingEvent mevt(v4d_EVT_MAPPING_CHANGED, 0, false);
        GetParent() -> ProcessWindowEvent(mevt);
    }

    template <class T>
    static inline void ReadIntegerFromConfig(wxFileConfig &conf, const wxString &sec, T *val, T defVal)
    {
        // wxWidgets do not support loading of all types so we have to perform
        // conversion to a "universal" long
        long tmpVal;
        conf.Read(sec, &tmpVal, (long)defVal);
        *val = T(tmpVal);
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::ReadFromConfig(wxFileConfig &fc)
    {
        fc.Read("RGBMAP_STRETCH_MODE", &m_stretch_mode, MM_NO_STRETCH);
        fc.Read("RGBMAP_PERCENTIL_VALUE", &m_percentil_value, 0.001);

        ReadIntegerFromConfig(fc, "RGBMAP_MIN_RED", &m_min_value.red, T(0));
        ReadIntegerFromConfig(fc, "RGBMAP_MIN_GREEN", &m_min_value.green, T(0));
        ReadIntegerFromConfig(fc, "RGBMAP_MIN_BLUE", &m_min_value.blue, T(0));
        
        ReadIntegerFromConfig(fc, "RGBMAP_MAX_RED", &m_max_value.red, std::numeric_limits<T>::max());
        ReadIntegerFromConfig(fc, "RGBMAP_MAX_GREEN", &m_max_value.green, std::numeric_limits<T>::max());
        ReadIntegerFromConfig(fc, "RGBMAP_MAX_BLUE", &m_max_value.blue, std::numeric_limits<T>::max());

        fc.Read("RGBMAP_SHOW_RED", &m_show_red, true);
        fc.Read("RGBMAP_SHOW_GREEN", &m_show_green, true);
        fc.Read("RGBMAP_SHOW_BLUE", &m_show_blue, true);

        ReadIntegerFromConfig(fc, "RGBMAP_THRES_RMIN", &m_thres_arr[0].min_threshold, T(0));
        ReadIntegerFromConfig(fc, "RGBMAP_THRES_RMAX", &m_thres_arr[0].max_threshold, std::numeric_limits<T>::max());
        ReadIntegerFromConfig(fc, "RGBMAP_THRES_GMIN", &m_thres_arr[1].min_threshold, T(0));
        ReadIntegerFromConfig(fc, "RGBMAP_THRES_GMAX", &m_thres_arr[1].max_threshold, std::numeric_limits<T>::max());
        ReadIntegerFromConfig(fc, "RGBMAP_THRES_BMIN", &m_thres_arr[2].min_threshold, T(0));
        ReadIntegerFromConfig(fc, "RGBMAP_THRES_BMAX", &m_thres_arr[2].max_threshold, std::numeric_limits<T>::max());

        GetParent()->GetMenuBar()->Check(this->GetMenuId(m_stretch_mode), true);
        GetParent()->GetMenuBar()->Check(this->GetMenuId(MM_SHOW_RED), m_show_red);
        GetParent()->GetMenuBar()->Check(this->GetMenuId(MM_SHOW_GREEN), m_show_green);
        GetParent()->GetMenuBar()->Check(this->GetMenuId(MM_SHOW_BLUE), m_show_blue);

        ReadLevelsFromConfig(fc);
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::ReadLevelsFromConfig(wxFileConfig &fc)
    {
        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_IMIN_RED", &m_levels_arr[0].inputMin, T(0));
        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_IMAX_RED", &m_levels_arr[0].inputMax, std::numeric_limits<T>::max());
        fc.Read("RGBMAP_LEVELS_IGAMMA_RED", &m_levels_arr[0].inputGamma, 1.0);
        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_OMIN_RED", &m_levels_arr[0].outputMin, i3d::GRAY8(0));
        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_OMAX_RED", &m_levels_arr[0].outputMax, i3d::GRAY8(255));

        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_IMIN_GREEN", &m_levels_arr[1].inputMin, T(0));
        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_IMAX_GREEN", &m_levels_arr[1].inputMax, std::numeric_limits<T>::max());
        fc.Read("RGBMAP_LEVELS_IGAMMA_GREEN", &m_levels_arr[1].inputGamma, 1.0);
        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_OMIN_GREEN", &m_levels_arr[1].outputMin, i3d::GRAY8(0));
        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_OMAX_GREEN", &m_levels_arr[1].outputMax, i3d::GRAY8(255));

        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_IMIN_BLUE", &m_levels_arr[2].inputMin, T(0));
        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_IMAX_BLUE", &m_levels_arr[2].inputMax, std::numeric_limits<T>::max());
        fc.Read("RGBMAP_LEVELS_IGAMMA_BLUE", &m_levels_arr[2].inputGamma, 1.0);
        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_OMIN_BLUE", &m_levels_arr[2].outputMin, i3d::GRAY8(0));
        ReadIntegerFromConfig(fc, "RGBMAP_LEVELS_OMAX_BLUE", &m_levels_arr[2].outputMax, i3d::GRAY8(255));
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::WriteToConfig(wxFileConfig &fc)
    {
        fc.Write("RGBMAP_STRETCH_MODE", m_stretch_mode);
        fc.Write("RGBMAP_PERCENTIL_VALUE", m_percentil_value);
        
        fc.Write("RGBMAP_MIN_RED", m_min_value.red);
        fc.Write("RGBMAP_MIN_GREEN", m_min_value.green);
        fc.Write("RGBMAP_MIN_BLUE", m_min_value.blue);
        
        fc.Write("RGBMAP_MAX_RED", m_max_value.red);
        fc.Write("RGBMAP_MAX_GREEN", m_max_value.green);
        fc.Write("RGBMAP_MAX_BLUE", m_max_value.blue);

        fc.Write("RGBMAP_SHOW_RED", m_show_red);
        fc.Write("RGBMAP_SHOW_GREEN", m_show_green);
        fc.Write("RGBMAP_SHOW_BLUE", m_show_blue);

        fc.Write("RGBMAP_THRES_RMIN", m_thres_arr[0].min_threshold);
        fc.Write("RGBMAP_THRES_RMAX", m_thres_arr[0].max_threshold);
        fc.Write("RGBMAP_THRES_GMIN", m_thres_arr[1].min_threshold);
        fc.Write("RGBMAP_THRES_GMAX", m_thres_arr[1].max_threshold);
        fc.Write("RGBMAP_THRES_BMIN", m_thres_arr[2].min_threshold);
        fc.Write("RGBMAP_THRES_BMAX", m_thres_arr[2].max_threshold);

        WriteLevelsToConfig(fc);
    }

    template <class T> 
    void v4d_RGBMappingModule<T>::WriteLevelsToConfig(wxFileConfig &fc)
    {
        fc.Write("RGBMAP_LEVELS_IMIN_RED", m_levels_arr[0].inputMin);
        fc.Write("RGBMAP_LEVELS_IMAX_RED", m_levels_arr[0].inputMax);
        fc.Write("RGBMAP_LEVELS_IGAMMA_RED", m_levels_arr[0].inputGamma);
        fc.Write("RGBMAP_LEVELS_OMIN_RED", m_levels_arr[0].outputMin);
        fc.Write("RGBMAP_LEVELS_OMAX_RED", m_levels_arr[0].outputMax);

        fc.Write("RGBMAP_LEVELS_IMIN_GREEN", m_levels_arr[1].inputMin);
        fc.Write("RGBMAP_LEVELS_IMAX_GREEN", m_levels_arr[1].inputMax);
        fc.Write("RGBMAP_LEVELS_IGAMMA_GREEN", m_levels_arr[1].inputGamma);
        fc.Write("RGBMAP_LEVELS_OMIN_GREEN", m_levels_arr[1].outputMin);
        fc.Write("RGBMAP_LEVELS_OMAX_GREEN", m_levels_arr[1].outputMax);

        fc.Write("RGBMAP_LEVELS_IMIN_BLUE", m_levels_arr[2].inputMin);
        fc.Write("RGBMAP_LEVELS_IMAX_BLUE", m_levels_arr[2].inputMax);
        fc.Write("RGBMAP_LEVELS_IGAMMA_BLUE", m_levels_arr[2].inputGamma);
        fc.Write("RGBMAP_LEVELS_OMIN_BLUE", m_levels_arr[2].outputMin);
        fc.Write("RGBMAP_LEVELS_OMAX_BLUE", m_levels_arr[2].outputMax);
    }

    /******************************************************************************/
    //                            Gray Mapping Module
    /******************************************************************************/

    template <class T> 
    v4d_GrayMappingModule<T>::v4d_GrayMappingModule(wxFrame *parent, const HistogramSet<T> &hist_set)
        : v4d_MappingModule<T>(parent, hist_set), 
        m_colour_mode(MM_GRAY)
    {
        m_min_value = hist_set.Get().GetLeftCenter();
        m_max_value = hist_set.Get().GetRightCenter();

        ResetMapping();

        m_pal.Init();

        m_levels = TLevelsParams<T, i3d::GRAY8>(m_min_value, m_max_value);
        m_thres = ThresParams<T>(m_min_value, m_max_value);
        UpdateLUT();
    }

	template <class T> 
    wxMenu* v4d_GrayMappingModule<T>::CreateMenu(int &last_id)
    {
        wxMenu *menu = v4d_MappingModule<T>::CreateMenu(last_id);
        if (typeid(T) == typeid(i3d::GRAY8) || typeid(T) == typeid(i3d::GRAY16)) 
		{
            menu->Insert(1, GetMenuId(MM_SET_BITDEPTH), "Set bitdepth...", "Set \"real\" image bitdepth");
        }

        menu->AppendSeparator();
        menu->Append(GetMenuId(MM_GRAY), "Gray", "View the data as labels", wxITEM_CHECK);
        menu->Append(GetMenuId(MM_SATURATION), "Saturation", "Show saturated values in colours", wxITEM_CHECK);
        menu->Append(GetMenuId(MM_LABELS), "Labels", "View the data as labels", wxITEM_CHECK);
        menu->Append(GetMenuId(MM_PERIODIC), "Periodic", "View the data with a periodic palette", wxITEM_CHECK);
        menu->Append(GetMenuId(MM_PALETTE), "Palette...", "Select custom palette", wxITEM_CHECK);

        Bind(wxEVT_COMMAND_MENU_SELECTED, &v4d_GrayMappingModule<T>::OnMapping, this, GetMenuId(0), last_id);
        Bind(wxEVT_UPDATE_UI, &v4d_GrayMappingModule<T>::OnUpdateUI, this, GetMenuId(MM_GRAY), GetMenuId(MM_PALETTE));

        Bind(v4d_EVT_LEVELS_PREVIEW, &v4d_GrayMappingModule<T>::OnPreviewLevels, this);
        Bind(v4d_EVT_THRESHOLDING_PREVIEW, &v4d_GrayMappingModule<T>::OnPreviewThresholding, this);

        return menu;
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::Map(const T *src, Buffer2d<i3d::RGB> &buf) const
    {
        if (m_colour_mode == MM_GRAY)
        {
            MapToRGBBuf(src, buf, GrayLUTMapper<T>(m_lut));
        }
        else
        {
            MapToRGBBuf(src, buf, PaletteMapper<T>(m_lut, m_pal));
        }
    }

    template <class T> 
    bool v4d_GrayMappingModule<T>::HasData()
    {
        return (!m_hist_set.Get().empty());
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::UpdateLUT()
    {
        try 
        {
            m_lut.CreateGamma(m_levels.inputMin, m_levels.inputMax, m_levels.inputGamma, m_levels.outputMin, m_levels.outputMax);
        } 
        catch (std::exception &e) 
        {
            wxMessageBox(wxString(e.what(),wxConvUTF8));
        }
    }

    template <class T>
    void v4d_GrayMappingModule<T>::OnHistogramChange(const HistogramSet<T> &hist_set)
    {
        m_hist_set = hist_set;

        m_min_value = std::max(m_min_value, hist_set.Get().GetLeftCenter());
        m_max_value = std::min(m_max_value, hist_set.Get().GetRightCenter());

        ResetMapping();
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::ResetMapping()
    {
        if (m_hist_set.Get().size() <= 1) { // we do not get valid histogram yet. LUT can't  be initialised
            m_lut.Init();
        } else {
            m_lut.Init((long)m_hist_set.Get().size(), m_hist_set.Get().GetLeftCenter(), m_hist_set.Get().GetRightCenter());
        }
        switch (this->m_stretch_mode) 
		{
			case MM_NO_STRETCH: SetIdentity(); break;
			case MM_LINEAR_STRETCH: SetLinearStretch(); break;
			case MM_PERCENTIL_STRETCH: SetPercentilStretch(); break;
			case MM_HISTEQUAL_STRETCH: SetHistEqualStretch(); break;
			case MM_LEVELS:
            {
                int pos = (int)m_hist_set.Get().GetBinPos(m_min_value);
                CUTOFF_LEFT(m_levels.inputMin, pos);

                pos = (int)m_hist_set.Get().GetBinPos(m_max_value);
                CUTOFF_RIGHT(m_levels.inputMax, pos);

                UpdateLUT();
            }
            break;
			case MM_THRESHOLD: 
	        {
                int pos = (int)m_hist_set.Get().GetBinPos(m_min_value);
                CUTOFF_LEFT(m_thres.min_threshold, pos);
                CUTOFF_LEFT(m_thres.max_threshold, pos);

                pos = (int)m_hist_set.Get().GetBinPos(m_max_value);
                CUTOFF_RIGHT(m_thres.min_threshold, pos);
                CUTOFF_RIGHT(m_thres.max_threshold, pos);

                DoThresholding();
            }
            break;
			default: break;
        }
    }

	template <class T>
    void v4d_GrayMappingModule<T>::DoThresholding()
	{
		m_lut.CreateThresholding(m_thres.min_threshold, m_thres.max_threshold);
	}

    template <>
    void v4d_GrayMappingModule<float>::SetRange()
    {
        float min_limit = m_hist_set.Get().GetLeftCenter();
        float max_limit = m_hist_set.Get().GetRightCenter();
        float lminv = m_min_value;
        float lmaxv = m_max_value;

        if (GetFloatRangeFromUser(this->GetParent(), min_limit, max_limit, lminv, lmaxv)) 
        {
            m_min_value = lminv;
            m_max_value = lmaxv;
            ResetMapping();
        }        
    }

    template <class T>
    void v4d_GrayMappingModule<T>::SetRange()
    {
        long min_limit = m_hist_set.Get().GetLeftCenter();
        long max_limit = m_hist_set.Get().GetRightCenter();
        long lminv = m_min_value;
        long lmaxv = m_max_value;

        if (GetLongRangeFromUser(this->GetParent(), min_limit, max_limit, lminv, lmaxv)) 
        {
            m_min_value = lminv;
            m_max_value = lmaxv;
            ResetMapping();
        }        
    }

    template <class T>
    void v4d_GrayMappingModule<T>::SetBitdepth()
    {
        int bit_depth = BitCount((int)m_max_value);
        int max_bit_depth = BitCount((int)m_hist_set.Get().GetRightCenter());

        long ret = ::wxGetNumberFromUser("Set image bit depth", "Bit depth = ", "Image Bit Depth", 
            bit_depth, 1, max_bit_depth, this->GetParent());

        if (ret != -1 && ret != bit_depth)
        {
            m_max_value = T((1 << ret) - 1);
            ResetMapping();
        }
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::SetIdentity()
    {
        m_levels = TLevelsParams<T, i3d::GRAY8>(m_min_value, m_max_value);
        UpdateLUT();
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::SetLinearStretch()
    {
        T l1, l2;

        Histogram<T> h;
        h.CreateSubHist(m_hist_set.Get(), this->m_min_value, this->m_max_value);

        l1 = h.GetBinCenter (h.FindFirstNonZero());
        l2 = h.GetBinCenter (h.FindLastNonZero());
        /*if (l1 >= l2) // we have zero or constant channel
        { 
            l1 = h.GetLeftCenter();
            l2 = h.GetRightCenter();
        }*/
        m_levels = TLevelsParams<T, i3d::GRAY8>(l1, l2);
        wxLogDebug("linear stretch %lu %lu", static_cast<unsigned long>(l1), static_cast<unsigned long>(l2));

        UpdateLUT();
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::SetPercentilStretch()
    {
        T l1, l2;

        Histogram<T> h;
        h.CreateSubHist(m_hist_set.Get(), this->m_min_value, this->m_max_value);

        l1 = h.GetBinCenter (h.FindLeftPercentil(m_percentil_value));
        l2 = h.GetBinCenter (h.FindRightPercentil(m_percentil_value));
        m_levels = TLevelsParams<T, i3d::GRAY8>(l1, l2);
        wxLogDebug("percentil stretch %lu %lu", static_cast<unsigned long>(l1), static_cast<unsigned long>(l2));

        UpdateLUT();
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::SetHistEqualStretch()
    {
        Histogram<T> h;
        h.CreateSubHist(m_hist_set.Get(), this->m_min_value, this->m_max_value);

		  //get number of voxels stored in this sub-histogram
		  size_t tmp=0;
		  const size_t* histBin=h.begin() + h.GetBinPos(this->m_min_value);
		  const size_t* const histEnd=h.begin() + h.GetBinPos(this->m_max_value);

		  //number of bins in the sub-histogram
		  size_t histSize=0;

		  while ( (histBin != histEnd) && (histBin != h.end()) ) {
		  		tmp += *histBin;
				++histBin;

				++histSize;
		  }
		  tmp += *histBin; //histEnd and h.end() point on last valid item
		  ++histSize;

		  //sum of the bin values in the sub-histogram
		  const size_t histSum=tmp;
		  
		  if ((unsigned)m_lut.size() >= histSize) {
			  //hopefully, the lut is at least of the same length
			  //as is the current sub-histogram

			  //first, clear the whole lut
			  i3d::GRAY8* lutPtr=m_lut.begin();
			  while (lutPtr != m_lut.end()) {
			    *lutPtr=0; ++lutPtr;
			  }
			  
			  //make sure we make use of the whole printable intensities
			  histSize = std::numeric_limits<i3d::GRAY8>::max();

			  /*
			  attempt to ignore the first 5%-percentil of pixels... should be a GUI parameter
			  //now, skip the 5% of small values in the sub-histogram related part of the lut
			  size_t i=h.GetBinPos(this->m_min_value);
			  tmp=h[i]; i++;
			  for (; i <= h.GetBinPos(this->m_max_value) && (20*tmp < histSum); ++i) tmp+=h[i];

			  histSum-=tmp;
			  tmp=0;

			  //now, finish/set the sub-histogram related part of the lut
			  for (; i <= h.GetBinPos(this->m_max_value); ++i)
			  {
					tmp += h[i];
					m_lut.Set( h.GetBinCenter(i), histSize * tmp / histSum );
			  }
			  */

			  //now, set the sub-histogram related part of the lut
			  tmp = 0;
			  for (size_t i = h.GetBinPos(this->m_min_value);
			    i <= h.GetBinPos(this->m_max_value); ++i) {
			    tmp += h[i];
			    m_lut.Set(h.GetBinCenter(i), histSize * tmp / histSum);
			  }

			  m_levels = TLevelsParams<T, i3d::GRAY8>(m_min_value, m_max_value);
		  } else {
			  wxLogError("The function does nothing because sub-histogram is longer than the current lut.");
		  }
#if 0
		//this appears to work perfectly until the SetRange() or the SetBitdepth() is used...

		  //get number of voxels stored in the histogram
		  size_t tmp=0;
		  const size_t* histBin=h.begin();
		  while (histBin != h.end()) {
		  		tmp += *histBin;
				++histBin;
		  }
		  tmp += *histBin; //h.end() points on last valid item

		  const size_t histSum=tmp;
		  const size_t histSize=h.size();
		  tmp=0;
		  
		  if ((unsigned)m_lut.size() == histSize) {
			  //hopefully, the lut is of the same length as is the current histogram
			  i3d::GRAY8* lutPtr=m_lut.begin();

			  for (size_t i=0; i < histSize; ++i) {
					tmp += h[i];
					lutPtr[i]= (histSize-1) * tmp / histSum;
			  }

			  m_levels = TLevelsParams<T, i3d::GRAY8>(m_min_value, m_max_value);
		  } else {
			  wxLogError("The function does nothing because histogram and lut sizes mismatch.");
		  }
#endif
    }

    template <class T> 
    bool v4d_GrayMappingModule<T>::SetLevelsStretch()
    {
        wxArrayString channel_names;
        channel_names.Add("Gray");

        std::vector<Histogram<T> *> histdata_vector;
        std::vector<TLevelsParams<T, i3d::GRAY8> > params_vector;

        Histogram<T> hist;
        hist.CreateSubHist(m_hist_set.Get(), this->m_min_value, this->m_max_value);
        histdata_vector.push_back(&hist);

        params_vector.push_back(m_levels);

        LevelsWidget<T, i3d::GRAY8> *lw = new LevelsWidget<T, i3d::GRAY8>(GetParent(), -1, histdata_vector, 
            params_vector, channel_names);

        bool dlg_ok = (lw -> ShowModal() == wxID_OK);
        if (dlg_ok)
        {
            lw -> GetParamsVector(params_vector);
            m_levels = params_vector[0];
			   UpdateLUT();
        }
		  else
		  {
		  		ResetMapping();
		  }

        return dlg_ok;
    }

    template <class T> 
    bool v4d_GrayMappingModule<T>::SetThreshold()
    {
        wxArrayString channel_names;
        channel_names.Add("Gray");

        std::vector<Histogram<T> *> histdata_vector;
        std::vector<ThresParams<T> > params_vector;

        Histogram<T> hist;
        hist.CreateSubHist(m_hist_set.Get(), this->m_min_value, this->m_max_value);
        histdata_vector.push_back(&hist);

        params_vector.push_back(m_thres);

        ThresholdingDialog<T> *dlg = new ThresholdingDialog<T>(GetParent(), -1, histdata_vector, params_vector, channel_names);

        bool dlg_ok = (dlg -> ShowModal() == wxID_OK);
        if (dlg_ok)
        {
            dlg -> GetParamsVector(params_vector);
            m_thres = params_vector[0];
            DoThresholding();
        }
        else
        {
            UpdateLUT();
        }

        return dlg_ok;
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::SetPal(const v4d_Palette &pal)
    {
        try {
				m_pal.Set(0, i3d::RGB(0,0,0));
            for (long i = 1; i <= m_pal.max_elem(); ++i)
            {
                m_pal.Set(i, pal[i % pal.size()]);
            }
        } catch (i3d::LibException &e) {
            wxMessageBox(wxString(e.what.c_str(),wxConvUTF8));
        }
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::OnMapping(wxCommandEvent &evt)
    {
        // call common mapping handler first
        v4d_MappingModule<T>::OnMapping(evt);

        int cid = this->GetCommandId(evt.GetId());

        if (IsColourModeId(cid))
        {
            m_colour_mode = cid;
        }

        v4d_Palette pal;
        switch (cid)
        {
        case MM_PALETTE:
            {
                // TODO let the user to read a palette from disk
                wxString fname = wxFileSelector("Load palette", wxEmptyString, wxEmptyString, "pal",
                    "Palettes (*.pal)|*.pal| Text files (*.txt)|*.txt|All files (*.*)|*.*", 
                    wxFD_OPEN | wxFD_FILE_MUST_EXIST, GetParent());

                if (!fname.IsEmpty())
                {
                    if (LoadPalette(pal, fname))
                    {
                        SetPal(pal);
                    }
                }
            }
            break;

        case MM_PERIODIC:
            v4d_GetPalette(v4dPAL_PERIODIC, pal);
            SetPal(pal);
            break;

        case MM_LABELS:
            v4d_GetPalette(v4dPAL_LABELS, pal);
            SetPal(pal);
            break;

        case MM_SATURATION:
            v4d_GetPalette(v4dPAL_SATURATION, pal);				
            SetPal(pal);
            break;
        }

        v4d_MappingEvent mevt(v4d_EVT_MAPPING_CHANGED);
        GetParent() -> ProcessWindowEvent(mevt);

        if (evt.ShouldPropagate())
        {
            v4d_MappingCommandEvent mcevt(v4d_EVT_MAPPING_COMMAND);
            mcevt.SetAttachedEvent(evt);
            mcevt.SetClientData(this);
            GetParent() -> ProcessWindowEvent(mcevt);
        }
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::OnUpdateUI(wxUpdateUIEvent &evt)
    {
        int cid = this->GetCommandId(evt.GetId());

        if (IsColourModeId(cid))
        {
            evt.Check(m_colour_mode == cid);
        }
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::OnPreviewLevels(v4d_LevelsEvent &evt)
    {
        TLevelsParams<T, i3d::GRAY8> params(evt.GetParams());
        m_lut.CreateGamma(params.inputMin, params.inputMax, params.inputGamma, params.outputMin, params.outputMax);

        v4d_MappingEvent mevt(v4d_EVT_MAPPING_CHANGED, 0, false);
        GetParent() -> ProcessWindowEvent(mevt);
    }

    template <class T>
    void v4d_GrayMappingModule<T>::OnPreviewThresholding(v4d_ThresholdingEvent &evt)
    {
        ThresParams<T> params(evt.GetParams());
        m_lut.CreateThresholding(params.min_threshold, params.max_threshold);

        v4d_MappingEvent mevt(v4d_EVT_MAPPING_CHANGED, 0, false);
        GetParent() -> ProcessWindowEvent(mevt);
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::ReadFromConfig(wxFileConfig &fc)
    {
        wxLogError("%s: Not implemented", __FUNCTION__);
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::ReadLevelsFromConfig(wxFileConfig &fc)
    {
        wxLogError("%s: Not implemented", __FUNCTION__);
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::WriteToConfig(wxFileConfig &fc)
    {
        wxLogError("%s: Not implemented", __FUNCTION__);
    }

    template <class T> 
    void v4d_GrayMappingModule<T>::WriteLevelsToConfig(wxFileConfig &fc)
    {
        wxLogError("%s: Not implemented", __FUNCTION__);
    }

	/******************************************************************************/
	//                            Complex Mapping Module
	/******************************************************************************/

	template <class T>
	v4d_ComplexMappingModule<T>::v4d_ComplexMappingModule(wxFrame *parent, 
		const HistogramSet<std::complex<T> > &hist_set,
		const bool nonemptyR, const bool nonemptyI,
		const bool nonemptyP, const bool nonemptyA)
		: v4d_MappingModule<std::complex<T> >(parent, hist_set),
		  m_colour_mode(MM_GRAY), showPart(REAL),
		  m_enabled_real(nonemptyR),  m_enabled_imag(nonemptyI),
		  m_enabled_phase(nonemptyP), m_enabled_amp(nonemptyA)
	{
		//find first available mode for displaying
		if (!m_enabled_real)
		{
			showPart=IMAGINARY;
			if (!m_enabled_imag)
			{
				showPart=PHASE;
				if (!m_enabled_phase)
				{
					showPart=AMPLITUDE;
					if (!m_enabled_amp) throw("Complex: Shouldn't get here");
				}
			}
		}

		for (size_t i = 0; i < 4; i++)
		{
			m_min_value[i] = hist_set[i].GetLeftCenter();
			m_max_value[i] = hist_set[i].GetRightCenter();
		}		

		ResetMapping();

		m_pal.Init();
		for (size_t i = 0; i < 4; i++)
		{
			m_levels[i] = TLevelsParams<T, i3d::GRAY8>(m_min_value[i], m_max_value[i]);
			m_thres[i] = ThresParams<T>(m_min_value[i], m_max_value[i]);
		}		
		UpdateLUT();
	};

	template <class T>
	wxMenu* v4d_ComplexMappingModule<T>::CreateMenu(int &last_id)
	{
		wxMenu *menu = v4d_MappingModule<std::complex<T> >::CreateMenu(last_id);

		menu->AppendSeparator();
		menu->Append(GetMenuId(MM_GRAY), "Gray", "View the data as labels", wxITEM_CHECK);
		menu->Append(GetMenuId(MM_SATURATION), "Saturation", "Show saturated values in colours", wxITEM_CHECK);
		menu->Append(GetMenuId(MM_LABELS), "Labels", "View the data as labels", wxITEM_CHECK);
		menu->Append(GetMenuId(MM_PERIODIC), "Periodic", "View the data with a periodic palette", wxITEM_CHECK);
		menu->Append(GetMenuId(MM_PALETTE), "Palette...", "Select custom palette", wxITEM_CHECK);

		Bind(wxEVT_COMMAND_MENU_SELECTED, &v4d_ComplexMappingModule<T>::OnMapping, this, GetMenuId(0), last_id);
		Bind(wxEVT_UPDATE_UI, &v4d_ComplexMappingModule<T>::OnUpdateUI, this, GetMenuId(MM_GRAY), GetMenuId(MM_PALETTE));

		Bind(v4d_EVT_LEVELS_PREVIEW, &v4d_ComplexMappingModule<T>::OnPreviewLevels, this);
		Bind(v4d_EVT_THRESHOLDING_PREVIEW, &v4d_ComplexMappingModule<T>::OnPreviewThresholding, this);

		menu->AppendSeparator();
		//PATCH BECAUSE OF LINUX WXGTK PORT...
		//(which allows to change state of disabled menu item with its accelerator key, regardless of UpdateUI()...)
		if (m_enabled_real) menu->Append(GetMenuId(MM_SHOW_REAL), "&Real\tF5", "Show real part", wxITEM_RADIO);
		               else menu->Append(GetMenuId(MM_SHOW_REAL), "&Real", "Show real part", wxITEM_RADIO);
		if (m_enabled_imag) menu->Append(GetMenuId(MM_SHOW_IMAGINARY), "&Imaginary\tF6", "Show imaginary part", wxITEM_RADIO);
		               else menu->Append(GetMenuId(MM_SHOW_IMAGINARY), "&Imaginary", "Show imaginary part", wxITEM_RADIO);
		if (m_enabled_phase) menu->Append(GetMenuId(MM_SHOW_PHASE), "&Phase\tF7", "Show phase", wxITEM_RADIO);
		                else menu->Append(GetMenuId(MM_SHOW_PHASE), "&Phase", "Show phase", wxITEM_RADIO);
		if (m_enabled_amp) menu->Append(GetMenuId(MM_SHOW_AMPLITUDE), "&Amplitude\tF8", "Show amplitude", wxITEM_RADIO);
		              else menu->Append(GetMenuId(MM_SHOW_AMPLITUDE), "&Amplitude", "Show amplitude", wxITEM_RADIO);

		//disable some mode if required...
		//(note: constructor took care of chosing appropriate/avaliable initial mode)
		if (!m_enabled_real)  menu->Enable(GetMenuId(MM_SHOW_REAL),     false);
		if (!m_enabled_imag)  menu->Enable(GetMenuId(MM_SHOW_IMAGINARY),false);
		if (!m_enabled_phase) menu->Enable(GetMenuId(MM_SHOW_PHASE),    false);
		if (!m_enabled_amp)   menu->Enable(GetMenuId(MM_SHOW_AMPLITUDE),false);

		Bind(wxEVT_UPDATE_UI, &v4d_ComplexMappingModule<T>::OnShowPartChange, this, GetMenuId(MM_SHOW_REAL), GetMenuId(MM_SHOW_AMPLITUDE));
		
		return menu;
	}
	
	template <class T>
	void v4d_ComplexMappingModule<T>::Map(const std::complex<T> *src, Buffer2d<i3d::RGB> &buf) const
	{
		const i3d::LUT<i3d::GRAY8, i3d::RGB> *pallete = NULL;
		if (m_colour_mode != MM_GRAY)
		{
			pallete = &m_pal;
		}
		switch (showPart)
		{
		case REAL:
			MapToRGBBuf(src, buf, ComplexToRealLUTMapper<T>(m_lut, pallete));
			break;
		case IMAGINARY:
			MapToRGBBuf(src, buf, ComplexToImagLUTMapper<T>(m_lut, pallete));
			break;
		case PHASE:
			MapToRGBBuf(src, buf, ComplexToPhaseLUTMapper<T>(m_lut, pallete));
			break;
		case AMPLITUDE:
			MapToRGBBuf(src, buf, ComplexToAmpLUTMapper<T>(m_lut, pallete));
			break;
		default:
			wxLogError("Don't know how to display this mode of complex numbers.");
		}
	}

	template <class T>
	bool v4d_ComplexMappingModule<T>::HasData()
	{
		return (!m_hist_set[0].empty() || !m_hist_set[1].empty() 
			|| !m_hist_set[2].empty() || !m_hist_set[3].empty());
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::UpdateLUT()
	{

		try
		{
			m_lut.CreateGamma(m_levels[showPart].inputMin, m_levels[showPart].inputMax, 
				m_levels[showPart].inputGamma, m_levels[showPart].outputMin, m_levels[showPart].outputMax);
		}
		catch (std::exception &e)
		{
			wxMessageBox(wxString(e.what(), wxConvUTF8));
		}
	}

	template <>
	void v4d_ComplexMappingModule<float>::SetLinearStretch()
	{
		float l1, l2;

		Histogram<float> h;
		h.CreateSubHist(m_hist_set[showPart], this->m_min_value[showPart], this->m_max_value[showPart]);

		l1 = h.GetBinCenter(h.FindFirstNonZero());
		l2 = h.GetBinCenter(h.FindLastNonZero());
		if (l1 >= l2) // we have zero or constant channel
		{
			l1 = h.GetLeftCenter();
			l2 = h.GetRightCenter();
		}
		m_levels[showPart] = TLevelsParams<float, i3d::GRAY8>(l1, l2);
		wxLogDebug("linear stretch %lu %lu", static_cast<unsigned long>(l1), static_cast<unsigned long>(l2));

		UpdateLUT();
	}

	template <>
	void v4d_ComplexMappingModule<float>::SetPercentilStretch()
	{
		float l1, l2;

		Histogram<float> h;
		h.CreateSubHist(m_hist_set[showPart], this->m_min_value[showPart], this->m_max_value[showPart]);

		l1 = h.GetBinCenter(h.FindLeftPercentil(m_percentil_value));
		l2 = h.GetBinCenter(h.FindRightPercentil(m_percentil_value));
		m_levels[showPart] = TLevelsParams<float, i3d::GRAY8>(l1, l2);
		wxLogDebug("percentil stretch %lu %lu", static_cast<unsigned long>(l1), static_cast<unsigned long>(l2));

		UpdateLUT();
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::OnHistogramChange(const HistogramSet<std::complex<T> > &hist_set)
	{
		m_hist_set = hist_set;

		for (size_t i = 0; i < 4; i++)
		{
			m_min_value[i] = std::max(m_min_value[i], m_hist_set[i].GetLeftCenter());
			m_max_value[i] = std::min(m_max_value[i], m_hist_set[i].GetRightCenter());
		}

		ResetMapping();
	}

	template <>
	void v4d_ComplexMappingModule<float>::ResetMapping()
	{
		if (m_hist_set[showPart].size() <= 1) { // we do not get valid histogram yet. LUT can't  be initialised
			m_lut.Init();
		}
		else {
			m_lut.Init((long)m_hist_set[showPart].size(), m_hist_set[showPart].GetLeftCenter(), m_hist_set[showPart].GetRightCenter());
		}
		switch (this->m_stretch_mode) {
		case MM_NO_STRETCH: SetIdentity(); break;
		case MM_LINEAR_STRETCH: SetLinearStretch(); break;
		case MM_PERCENTIL_STRETCH: SetPercentilStretch(); break;
		case MM_HISTEQUAL_STRETCH: SetHistEqualStretch(); break;
		case MM_LEVELS:
		{
			int pos = (int)m_hist_set[showPart].GetBinPos(m_min_value[showPart]);
			CUTOFF_LEFT(m_levels[showPart].inputMin, pos);

			pos = (int)m_hist_set[showPart].GetBinPos(m_max_value[showPart]);
			CUTOFF_RIGHT(m_levels[showPart].inputMax, pos);

			UpdateLUT();
		}
			break;
		case MM_THRESHOLD:
		{
			int pos = (int)m_hist_set[showPart].GetBinPos(m_min_value[showPart]);
			CUTOFF_LEFT(m_thres[showPart].min_threshold, pos);
			CUTOFF_LEFT(m_thres[showPart].max_threshold, pos);

			pos = (int)m_hist_set[showPart].GetBinPos(m_max_value[showPart]);
			CUTOFF_RIGHT(m_thres[showPart].min_threshold, pos);
			CUTOFF_RIGHT(m_thres[showPart].max_threshold, pos);

			DoThresholding();
		}
			break;
		default: break;
		}
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::DoThresholding()
	{
		m_lut.CreateThresholding(m_thres[showPart].min_threshold, m_thres[showPart].max_threshold);
	}

	template <>
	void v4d_ComplexMappingModule<float>::SetRange()
	{
		float min_limit = m_hist_set[showPart].GetLeftCenter();
		float max_limit = m_hist_set[showPart].GetRightCenter();
		float lminv = m_min_value[showPart];
		float lmaxv = m_max_value[showPart];

		if (GetFloatRangeFromUser(this->GetParent(), min_limit, max_limit, lminv, lmaxv))
		{
			m_min_value[showPart] = lminv;
			m_max_value[showPart] = lmaxv;
			ResetMapping();
		}
	}

	template <>
	void v4d_ComplexMappingModule<float>::SetBitdepth()
	{
		int bit_depth = BitCount((int)(m_max_value[showPart]));
		int max_bit_depth = BitCount((int)m_hist_set[showPart].GetRightCenter());

		long ret = ::wxGetNumberFromUser("Set image bit depth", "Bit depth = ", "Image Bit Depth",
			bit_depth, 1, max_bit_depth, this->GetParent());

		if (ret != -1 && ret != bit_depth)
		{
			m_max_value[showPart] = float((1 << ret) - 1);
			ResetMapping();
		}
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::SetIdentity()
	{
		m_levels[showPart] = TLevelsParams<T, i3d::GRAY8>(m_min_value[showPart], m_max_value[showPart]);
		UpdateLUT();
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::SetHistEqualStretch()
	{
		Histogram<T> h;
		h.CreateSubHist(m_hist_set[showPart], this->m_min_value[showPart], this->m_max_value[showPart]);

		//get number of voxels stored in this sub-histogram
		size_t tmp = 0;
		const size_t* histBin = h.begin() + h.GetBinPos(this->m_min_value[showPart]);
		const size_t* const histEnd = h.begin() + h.GetBinPos(this->m_max_value[showPart]);

		//number of bins in the sub-histogram
		size_t histSize = 0;

		while ((histBin != histEnd) && (histBin != h.end())) {
			tmp += *histBin;
			++histBin;

			++histSize;
		}
		tmp += *histBin; //histEnd and h.end() point on last valid item
		++histSize;

		//sum of the bin values in the sub-histogram
		const size_t histSum = tmp;

		if ((unsigned)m_lut.size() >= histSize) {
			//hopefully, the lut is at least of the same length
			//as is the current sub-histogram

			//first, clear the whole lut
			i3d::GRAY8* lutPtr = m_lut.begin();
			while (lutPtr != m_lut.end()) {
				*lutPtr = 0; ++lutPtr;
			}

			//make sure we make use of the whole printable intensities
			histSize = std::numeric_limits<i3d::GRAY8>::max();

			//now, set the sub-histogram related part of the lut
			tmp = 0;
			for (size_t i = h.GetBinPos(this->m_min_value[showPart]);
				i <= h.GetBinPos(this->m_max_value[showPart]); ++i) {
				tmp += h[i];
				m_lut.Set(h.GetBinCenter(i), histSize * tmp / histSum);
			}

			m_levels[showPart] = TLevelsParams<T, i3d::GRAY8>(m_min_value[showPart], m_max_value[showPart]);
		}
		else {
			wxLogError("The function does nothing because sub-histogram is longer than the current lut.");
		}
#if 0
		//this appears to work perfectly until the SetRange() or the SetBitdepth() is used...

		//get number of voxels stored in the histogram
		size_t tmp = 0;
		const size_t* histBin = h.begin();
		while (histBin != h.end()) {
			tmp += *histBin;
			++histBin;
		}
		tmp += *histBin; //h.end() points on last valid item

		const size_t histSum = tmp;
		const size_t histSize = h.size();
		tmp = 0;

		if ((unsigned)m_lut.size() == histSize) {
			//hopefully, the lut is of the same length as is the current histogram
			i3d::GRAY8* lutPtr = m_lut.begin();

			for (size_t i = 0; i < histSize; ++i) {
				tmp += h[i];
				lutPtr[i] = (histSize - 1) * tmp / histSum;
			}

			m_levels[showPart] = TLevelsParams<T, i3d::GRAY8>(m_min_value[showPart], m_max_value[showPart]);
		}
		else {
			wxLogError("The function does nothing because histogram and lut sizes mismatch.");
		}
#endif
	}

	template <class T>
	bool v4d_ComplexMappingModule<T>::SetLevelsStretch()
	{
		wxArrayString channel_names;
		channel_names.Add("Gray");

		std::vector<Histogram<T> *> histdata_vector;
		std::vector<TLevelsParams<T, i3d::GRAY8> > params_vector;

		Histogram<T> hist;
		hist.CreateSubHist(m_hist_set[showPart], this->m_min_value[showPart], this->m_max_value[showPart]);
		histdata_vector.push_back(&hist);

		params_vector.push_back(m_levels[showPart]);

		LevelsWidget<T, i3d::GRAY8> *lw = new LevelsWidget<T, i3d::GRAY8>(GetParent(), -1, histdata_vector,
			params_vector, channel_names);

		bool dlg_ok = (lw->ShowModal() == wxID_OK);
		if (dlg_ok)
		{
			lw->GetParamsVector(params_vector);
			m_levels[showPart] = params_vector[0];
			UpdateLUT();
		}
		else
		{
			ResetMapping();
		}

		return dlg_ok;
	}

	template <class T>
	bool v4d_ComplexMappingModule<T>::SetThreshold()
	{
		wxArrayString channel_names;
		channel_names.Add("Gray");

		std::vector<Histogram<T> *> histdata_vector;
		std::vector<ThresParams<T> > params_vector;

		Histogram<T> hist;
		hist.CreateSubHist(m_hist_set[showPart], this->m_min_value[showPart], this->m_max_value[showPart]);
		histdata_vector.push_back(&hist);

		params_vector.push_back(m_thres[showPart]);

		ThresholdingDialog<T> *dlg = new ThresholdingDialog<T>(GetParent(), -1, histdata_vector, params_vector, channel_names);

		bool dlg_ok = (dlg->ShowModal() == wxID_OK);
		if (dlg_ok)
		{
			dlg->GetParamsVector(params_vector);
			m_thres[showPart] = params_vector[0];
			DoThresholding();
		}
		else
		{
			UpdateLUT();
		}

		return dlg_ok;
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::SetPal(const v4d_Palette &pal)
	{
		try {
			m_pal.Set(0, i3d::RGB(0,0,0));
			for (long i = 1; i <= m_pal.max_elem(); ++i)
			{
				m_pal.Set(i, pal[i % pal.size()]);
			}
		}
		catch (i3d::LibException &e) {
			wxMessageBox(wxString(e.what.c_str(), wxConvUTF8));
		}
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::OnMapping(wxCommandEvent &evt)
	{
		// call common mapping handler first
		v4d_MappingModule<std::complex<T> >::OnMapping(evt);

		int cid = this->GetCommandId(evt.GetId());

		if (IsColourModeId(cid))
		{
			m_colour_mode = cid;
		}

		//wxLogMessage(wxString::Format("%d == %d", cid, MM_PALETTE));
		v4d_Palette pal;
		switch (cid)
		{
			case MM_PALETTE:
			{
				// TODO let the user to read a palette from disk
				wxString fname = wxFileSelector("Load palette", wxEmptyString, wxEmptyString, "pal",
					"Palettes (*.pal)|*.pal| Text files (*.txt)|*.txt|All files (*.*)|*.*",
					wxFD_OPEN | wxFD_FILE_MUST_EXIST, GetParent());

				if (!fname.IsEmpty())
				{
					if (LoadPalette(pal, fname))
					{
						SetPal(pal);
					}
				}
			}
			break;

			case MM_PERIODIC:
				v4d_GetPalette(v4dPAL_PERIODIC, pal);
				SetPal(pal);
				break;

			case MM_LABELS:
				v4d_GetPalette(v4dPAL_LABELS, pal);
				SetPal(pal);
				break;

			case MM_SATURATION:
				v4d_GetPalette(v4dPAL_SATURATION, pal);
				SetPal(pal);
				break;
			case MM_SHOW_REAL:
				if ((evt.IsChecked()) && (m_enabled_real))
					showPart = REAL;
				break;
			case MM_SHOW_IMAGINARY:
				if ((evt.IsChecked()) && (m_enabled_imag))
					showPart = IMAGINARY;		
				break;
			case MM_SHOW_PHASE:
				if ((evt.IsChecked()) && (m_enabled_phase))
					showPart = PHASE;
				break;
			case MM_SHOW_AMPLITUDE:
				if ((evt.IsChecked()) && (m_enabled_amp))
					showPart = AMPLITUDE;
				break;

		}

		if (IsComplexShowId(cid))
		{
			ResetMapping();
			UpdateLUT();
		}		

		v4d_MappingEvent mevt(v4d_EVT_MAPPING_CHANGED);
		GetParent()->ProcessWindowEvent(mevt);
		
		if (evt.ShouldPropagate())
		{
			v4d_MappingCommandEvent mcevt(v4d_EVT_MAPPING_COMMAND);
			mcevt.SetAttachedEvent(evt);
			mcevt.SetClientData(this);
			GetParent()->ProcessWindowEvent(mcevt); 
		}		
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::OnUpdateUI(wxUpdateUIEvent &evt)
	{
		int cid = this->GetCommandId(evt.GetId());

		if (IsColourModeId(cid))
		{			
			evt.Check(m_colour_mode == cid);
		}
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::OnPreviewLevels	(v4d_LevelsEvent &evt)
	{
		TLevelsParams<T, i3d::GRAY8> params(evt.GetParams());
		m_lut.CreateGamma(params.inputMin, params.inputMax, params.inputGamma, params.outputMin, params.outputMax);

		v4d_MappingEvent mevt(v4d_EVT_MAPPING_CHANGED, 0, false);
		GetParent()->ProcessWindowEvent(mevt);
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::OnPreviewThresholding(v4d_ThresholdingEvent &evt)
	{
		ThresParams<T> params(evt.GetParams());
		m_lut.CreateThresholding(params.min_threshold, params.max_threshold);

		v4d_MappingEvent mevt(v4d_EVT_MAPPING_CHANGED, 0, false);
		GetParent()->ProcessWindowEvent(mevt);
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::OnShowPartChange(wxUpdateUIEvent &evt)
	{
		int cid = this->GetCommandId(evt.GetId());
		if (IsComplexShowId(cid))
		{	
			//evt.Enable(this->HasData());
			switch (cid)
			{	
				case MM_SHOW_REAL:
					evt.Check(REAL == showPart);					
					break;
				case MM_SHOW_IMAGINARY:
					evt.Check(IMAGINARY == showPart);
					break;
				case MM_SHOW_PHASE:
					evt.Check(PHASE == showPart);
					break;
				case MM_SHOW_AMPLITUDE:
					evt.Check(AMPLITUDE == showPart);
					break;
				default:
					break;
			}
		}
		else
			evt.Skip();
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::ReadFromConfig(wxFileConfig &fc)
	{
		wxLogError("%s: Not implemented", __FUNCTION__);
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::ReadLevelsFromConfig(wxFileConfig &fc)
	{
		wxLogError("%s: Not implemented", __FUNCTION__);
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::WriteToConfig(wxFileConfig &fc)
	{
		wxLogError("%s: Not implemented", __FUNCTION__);
	}

	template <class T>
	void v4d_ComplexMappingModule<T>::WriteLevelsToConfig(wxFileConfig &fc)
	{
		wxLogError("%s: Not implemented", __FUNCTION__);
	}

    /******************************************************************************/
    //                            Create Mapping Module
    /******************************************************************************/

	template <class T>
	V4D_DLLEXPORT v4d_MappingModule<T> *v4d_CreateMapperForType(wxFrame *parent, const i4d::Image4d<T> &data)
	{
		wxBusyInfo wait("Creating histogram...");

		// create the histogram for the whole sequence
		int last_id = data.GetNumFrames() - 1;
		Histogram<T> complete_hist;
		Histogram<T> current_hist;

		ComputeHist(complete_hist, data.GetImage(last_id)->begin(), data.GetImage(last_id)->end(), (size_t)std::numeric_limits<T>::max() + 1);
		for (int i = last_id - 1; i >= 0; i--)
		{
			ComputeHist(current_hist, data.GetImage(i)->begin(), data.GetImage(i)->end(), (size_t)std::numeric_limits<T>::max() + 1);
			complete_hist += current_hist;
			current_hist.Clear();
		}

		HistogramSet<T> hist_set;
		hist_set.Set(complete_hist);

		return v4d_CreateGrayMapperForType<T>(parent, hist_set);
	}

	template <>
	v4d_MappingModule<float> *v4d_CreateMapperForType(wxFrame *parent, const i4d::Image4d<float> &data)
	{
		wxBusyInfo wait("Creating histogram...");

		const float min_value = data.GetCachedMinValue();
		const float max_value = data.GetCachedMaxValue();

		//check for single-valued data (an image or even a whole image sequence)
		if (min_value == max_value)
		  	throw i3d::InternalException( wxString::Format(
			   "Consistency check failed. Constant images are not supported! (Value=%f)",
				min_value
				).ToStdString() );

		//now, create the histogram for the whole sequence
		const int last_id = data.GetNumFrames() - 1;
		Histogram<float> complete_hist;
		Histogram<float> current_hist;

		ComputeHist(complete_hist, data.GetImage(last_id)->begin(), data.GetImage(last_id)->end(), std::numeric_limits<i3d::GRAY8>::max() + 1, min_value, max_value);
		for (int i = last_id - 1; i >= 0; i--)
		{
			ComputeHist(current_hist, data.GetImage(i)->begin(), data.GetImage(i)->end(), std::numeric_limits<i3d::GRAY8>::max() + 1, min_value, max_value);
			complete_hist += current_hist;
			current_hist.Clear();
		}

		HistogramSet<float> hist_set;
		hist_set.Set(complete_hist);

		return v4d_CreateGrayMapperForType<float>(parent, hist_set);
	}

	template <>
	v4d_MappingModule<int> *v4d_CreateMapperForType(wxFrame *parent, const i4d::Image4d<int> &data)
	{
		wxBusyInfo wait("Creating histogram...");

		const int min_value = data.GetCachedMinValue();
		const int max_value = data.GetCachedMaxValue();

		//check for single-valued data (an image or even a whole image sequence)
		if (min_value == max_value)
		  	throw i3d::InternalException( wxString::Format(
			   "Consistency check failed. Constant images are not supported! (Value=%d)",
				min_value
				).ToStdString() );

		//now, create the histogram for the whole sequence
		const int last_id = data.GetNumFrames() - 1;
		Histogram<int> complete_hist;
		Histogram<int> current_hist;

		ComputeHist(complete_hist, data.GetImage(last_id)->begin(), data.GetImage(last_id)->end(), max_value-min_value+1, min_value, max_value);
		for (int i = last_id - 1; i >= 0; i--)
		{
			ComputeHist(current_hist, data.GetImage(i)->begin(), data.GetImage(i)->end(), max_value-min_value+1, min_value, max_value);
			complete_hist += current_hist;
			current_hist.Clear();
		}

		HistogramSet<int> hist_set;
		hist_set.Set(complete_hist);

		return v4d_CreateGrayMapperForType<int>(parent, hist_set);
	}

	template <class T>
	v4d_MappingModule<i3d::RGB_generic<T> > *v4d_CreateMapperForType
		(wxFrame *parent, const i4d::Image4d< i3d::RGB_generic<T> > &data)
	{
		wxBusyInfo wait("Creating histogram...");

		// create the channel histograms for the whole sequence
		int last_id = data.GetNumFrames() - 1;
		Histogram<T> complete_hist[3];
		Histogram<T> current_hist[3];

		for (size_t i = 0; i < 3; i++)
		{
			IntensityHist(*data.GetImage(last_id), complete_hist[i], i);
		}
		
		for (int j = last_id - 1; j >= 0; j--)
		{
			for (size_t i = 0; i < 3; i++)
			{
				IntensityHist(*data.GetImage(j), current_hist[i], i);
				complete_hist[i] += current_hist[i];
				current_hist[i].Clear();
			}
		}

		HistogramSet<i3d::RGB_generic<T> > hist_set;
		for (size_t i = 0; i < 3; i++)
		{
			hist_set.Set(i, complete_hist[i]);
		}		

		return v4d_CreateRGBMapperForType<T>(parent, hist_set);
	}

	template <>
	V4D_DLLEXPORT v4d_MappingModule<i3d::RGB_generic<i3d::GRAY8> > *v4d_CreateMapperForType<i3d::RGB_generic<i3d::GRAY8> >
		(wxFrame *parent, const i4d::Image4d< i3d::RGB_generic<i3d::GRAY8> > &data)
	{
		return v4d_CreateMapperForType<i3d::GRAY8>(parent, data);
	}

	template <>
	V4D_DLLEXPORT v4d_MappingModule<i3d::RGB_generic<i3d::GRAY16> > *v4d_CreateMapperForType<i3d::RGB_generic<i3d::GRAY16> >
		(wxFrame *parent, const i4d::Image4d< i3d::RGB_generic<i3d::GRAY16> > &data)
	{
		return v4d_CreateMapperForType<i3d::GRAY16>(parent, data);
	}
		
	template <class T>
	V4D_DLLEXPORT v4d_MappingModule<std::complex<T> > *v4d_CreateMapperForType
		(wxFrame *parent, const i4d::Image4d<std::complex<T> > &data)
	{
		wxBusyInfo wait("Creating histogram...");

		i4d::ExtremaValues<std::complex<T> > extremes = data.GetExtremes();
		const T min_values[4] = { extremes.minValue.real(), extremes.minValue.imag(),
			extremes.minP, extremes.minM };
		const T max_values[4] = { extremes.maxValue.real(), extremes.maxValue.imag(),
			extremes.maxP, extremes.maxM };

		//check for single-valued data (an image or even a whole image sequence)
		int problematicModes=0;
		if (min_values[0] == max_values[0])
		{
			++problematicModes;
			wxLogMessage("Disabling constant real part. (value=%f)",min_values[0]);
			//however, the real 'disabling' happens later when calling the constructor...
		}
		if (min_values[1] == max_values[1])
		{
			++problematicModes;
			wxLogMessage("Disabling constant imaginary part. (value=%f)",min_values[1]);
		}
		if (min_values[2] == max_values[2])
		{
			++problematicModes;
			wxLogMessage("Disabling constant phase. (value=%f)",min_values[2]);
		}
		if (min_values[3] == max_values[3])
		{
			++problematicModes;
			wxLogMessage("Disabling constant amplitude. (value=%f)",min_values[3]);
		}

		if (problematicModes == 4)
		  	throw i3d::InternalException( wxString::Format(
			   "Consistency check failed. Complex images with both parts, phase and amplitude constant are not supported! (real value=%f, imaginary value=%f, phase value=%f, amplitude value=%f)",
				min_values[0], min_values[1], min_values[2], min_values[3]
				).ToStdString() );

		//now, create the histogram for the whole sequence
		int last_id = data.GetNumFrames() - 1;
		Histogram<T> complete_hist[4];
		Histogram<T> current_hist[4];

		HistogramSet<std::complex<T> > continuous_histset;
		continuous_histset.CreateFromImage(*data[last_id], min_values, max_values);
		for (size_t i = 0; i < 4; i++)
		{
			complete_hist[i] = continuous_histset[i];
		}

		for (int j = last_id - 1; j >= 0; j--)
		{
			continuous_histset.CreateFromImage(*data[j], min_values, max_values);
			for (size_t i = 0; i < 4; i++)
			{
				current_hist[i] = continuous_histset[i];
				complete_hist[i] += current_hist[i];
				current_hist[i].Clear();
			}
		}
		
		HistogramSet<std::complex<T> > hist_set;
		for (size_t i = 0; i < 4; i++)
		{
			hist_set.Set(i, complete_hist[i]);
		}
		v4d_ComplexMappingModule<T> *cmplex
			= v4d_CreateComplexMapperForType<T>(parent, hist_set,
			         min_values[0] < max_values[0],
			         min_values[1] < max_values[1],
			         min_values[2] < max_values[2],
			         min_values[3] < max_values[3]);
		return cmplex;
	}

	template <>
	V4D_DLLEXPORT v4d_MappingModule<std::complex<float> > *v4d_CreateMapperForType<std::complex<float> >
		(wxFrame *parent, const i4d::Image4d<std::complex<float> > &data)
	{
		return v4d_CreateMapperForType<float>(parent, data);
	}

	// --- CreateXYZMappersForType here --- (yet more specific versions of the v4d_CreateMapperForType())

	template <class T> 
    v4d_GrayMappingModule<T> *v4d_CreateGrayMapperForType
        (wxFrame *parent, const HistogramSet<T> &hist_set)
	{		
		return new v4d_GrayMappingModule<T> (parent, hist_set);
	}

	template <class T> 
    v4d_RGBMappingModule<T> *v4d_CreateRGBMapperForType
        (wxFrame *parent, const HistogramSet<i3d::RGB_generic<T> > &hist_set)
	{
		return new v4d_RGBMappingModule<T>(parent, hist_set);
	}

	template <class T>
	v4d_ComplexMappingModule<T> *v4d_CreateComplexMapperForType
		(wxFrame *parent, const HistogramSet<std::complex<T> > &hist_set,
		 const bool nonemptyR, const bool nonemptyI,
		 const bool nonemptyP, const bool nonemptyA)
	{
		return new v4d_ComplexMappingModule<T>(parent, hist_set, nonemptyR,nonemptyI,nonemptyP,nonemptyA);
	}

	// Explicit instantiations
    template class V4D_DLLEXPORT v4d_RGBMappingModule<i3d::GRAY8>;
    template class V4D_DLLEXPORT v4d_RGBMappingModule<i3d::GRAY16>;

	template class V4D_DLLEXPORT v4d_GrayMappingModule<i3d::GRAY8>;
	template class V4D_DLLEXPORT v4d_GrayMappingModule<i3d::GRAY16>;
    template class V4D_DLLEXPORT v4d_GrayMappingModule<int>;
    template class V4D_DLLEXPORT v4d_GrayMappingModule<float>;

	template class V4D_DLLEXPORT v4d_ComplexMappingModule<float>;

	template V4D_DLLEXPORT v4d_MappingModule<i3d::GRAY8> *v4d_CreateMapperForType(wxFrame *parent, const i4d::Image4d<i3d::GRAY8> &data);
	template V4D_DLLEXPORT v4d_MappingModule<i3d::GRAY16> *v4d_CreateMapperForType(wxFrame *parent, const i4d::Image4d<i3d::GRAY16> &data);
}
