/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
* mapping.h
*
* Mapping functions
*
* Petr Matula (pem@fi.muni.cz) 2006
* Pavel Matula (pam@fi.muni.cz) 2007
*/

#ifndef V4D_MAPPING_H
#define V4D_MAPPING_H

#include <wx/fileconf.h>
#include <i3d/image3d.h>
#include <i3d/lut.h>
#include "v4d_impexp.h"
#include "types/mapper.h"
#include "types/extension.h"
#include "types/tlevelsparams.h"
#include "types/thresparams.h"
#include "thistogram.h"
#include "palette.h"
#include "dlg_levels.h"
#include "dlg_thres.h"
#include "v4d_image4d.h"

namespace v4d 
{
	/** Base class of all mapping classes. This is an abstract class. */
	template <class T> 
    class V4D_DLLEXPORT v4d_MappingModule
        : public v4d_ExtensionModule, public v4d_Mapper<T>
	{
	protected:
        /** Current set of histograms. */
        HistogramSet<T> m_hist_set;
        /** Current stretch mode. */
		int m_stretch_mode;
        /** Menu pointer. */
        wxMenu *m_menu_ptr;
        /** Percentil stretch. */
        double m_percentil_value;

	public:
		v4d_MappingModule(wxFrame *parent, const HistogramSet<T> &hist_set, double percentil_val = 0.001);
		
		virtual wxMenu* CreateMenu(int &last_id);

        virtual void SetRange() = 0;
        virtual void SetBitdepth() = 0;
		virtual void SetIdentity() = 0;
		virtual void SetLinearStretch() = 0;
        virtual void SetPercentilStretch() = 0;
		virtual void SetHistEqualStretch() = 0;
		virtual bool SetLevelsStretch() = 0;
		virtual bool SetThreshold() = 0;

        // Get current mapping mode
        int GetStretchMode();

        // Read all mapper settings from a config file
        virtual void ReadFromConfig(wxFileConfig &fc) = 0;
        // Write all mapper settings to a config file
        virtual void WriteToConfig(wxFileConfig &fc) = 0;
        // Read levels dialog settings from a config file
        virtual void ReadLevelsFromConfig(wxFileConfig &fc) = 0;
        // Write levels dialog settings to a config file
        virtual void WriteLevelsToConfig(wxFileConfig &fc) = 0;
        
        /** Update set of histograms and bit depths - called when new position is loaded in af_frame */
        virtual void OnHistogramChange(const HistogramSet<T> &hist_set) = 0;

		/** This function is called when something (histogram, min_value, max_value) has changed and reseting of 
			mapping module is necessary.
		*/
		virtual void ResetMapping() = 0;
		virtual bool HasData() = 0;

		void OnMapping(wxCommandEvent &evt);
		void OnUpdateUI(wxUpdateUIEvent &evt);

        void SetPercentilValue(double pv)
        {
            m_percentil_value = pv;
        }

		using v4d_ExtensionModule::GetParent;
		using v4d_ExtensionModule::GetMenuId;
	};

	/** RGB mapping module. This class implements mappers from RGB images to RGB buffers. 
	    It creates a menu of commands and implements all handlers of these commands. When a 
	    command from the menu is selected, the class changes its atributes and v4d_MappingChanged 
	    event is sent to the parent. 
	*/
	template <class T> 
    class V4D_DLLEXPORT v4d_RGBMappingModule 
        : public v4d_MappingModule<i3d::RGB_generic<T> >
	{
    private:
        /** Minimum mapped value. */
		i3d::RGB_generic<T> m_min_value;
        /** Maximum mapped value. */
		i3d::RGB_generic<T> m_max_value;

        /** Show red channel. */
		bool m_show_red;
        /** Show green channel. */
		bool m_show_green;
        /** Show blue channel. */
		bool m_show_blue;

		/** Levels params for each channel. */
		std::vector<TLevelsParams<T, i3d::GRAY8> > m_levels_arr;
		/** Thesholding params for each channel. */
		std::vector<ThresParams<T> > m_thres_arr;

		/** LUT for red channel*/
		i3d::LUT<T, i3d::GRAY8> m_rlut;
		/** LUT for green channel*/
		i3d::LUT<T, i3d::GRAY8> m_glut;
		/** LUT for blue channel*/
		i3d::LUT<T, i3d::GRAY8> m_blut;
		/** LUT that maps all intensities to zero */
		i3d::LUT<T, i3d::GRAY8> m_null;

    private:
		/** recomputes LUTSs for all three channels according to m_levels_params */
		void UpdateLUTs();
		/** recomputes LUT for according to given TLevelParams */
		void UpdateChannelLUT(i3d::LUT<T, i3d::GRAY8> &lut, const TLevelsParams<T, i3d::GRAY8> &params);
		/** all three LUTs are set according to m_thres_arr. If m_thres_arr[i] is not set for a i-th channel then mapping 
			given by m_levels_params is applied */
		void DoThresholding();
		/** set LUT according to params. If it has its default values then levels params of the given channel are applied. */
		void DoChannelThresholding(i3d::LUT<T, i3d::GRAY8> &lut, const ThresParams<T> &params, int channel);

	public:
		v4d_RGBMappingModule(wxFrame *parent, const HistogramSet<i3d::RGB_generic<T> > &hist_set);

		virtual wxMenu* CreateMenu(int &last_id);
		virtual void Map(const i3d::RGB_generic<T> *src, Buffer2d<i3d::RGB> &buf) const;

        virtual void SetRange();
        virtual void SetBitdepth();
		virtual void SetIdentity();
		virtual void SetLinearStretch();
        virtual void SetPercentilStretch();
		virtual void SetHistEqualStretch();
		virtual bool SetLevelsStretch();
		virtual bool SetThreshold();

        virtual void ReadFromConfig(wxFileConfig &fc);
        virtual void WriteToConfig(wxFileConfig &fc);
        virtual void ReadLevelsFromConfig(wxFileConfig &fc);
        virtual void WriteLevelsToConfig(wxFileConfig &fc);

        virtual void OnHistogramChange(const HistogramSet< i3d::RGB_generic<T> > &hist_set);
		virtual void ResetMapping();
		virtual bool HasData();

		void OnMapping(wxCommandEvent &evt);
		void OnUpdateUI(wxUpdateUIEvent &evt);
		void OnPreviewLevels(v4d_LevelsEvent &evt);
		void OnPreviewThresholding(v4d_ThresholdingEvent &evt);

		using v4d_MappingModule< i3d::RGB_generic<T> >::GetParent;
		using v4d_MappingModule< i3d::RGB_generic<T> >::GetMenuId;
		using v4d_MappingModule< i3d::RGB_generic<T> >::Bind;
        using v4d_MappingModule< i3d::RGB_generic<T> >::m_hist_set;
        using v4d_MappingModule< i3d::RGB_generic<T> >::m_stretch_mode;
        using v4d_MappingModule< i3d::RGB_generic<T> >::m_percentil_value;
		using v4d_ExtensionModule::GetCommandId;
	};

	/** Mapping module for one channel (usually class T = GRAY8, GRAY16, float, or int). */
	template <class T> 
    class V4D_DLLEXPORT v4d_GrayMappingModule 
        : public v4d_MappingModule<T>
	{
    private:
        /** Minimum mapped value. */
		T m_min_value;
        /** Maximum mapped value. */
		T m_max_value;

        /** Colour mode. */
		int m_colour_mode;
	protected:
        /** Look-up table. */
		i3d::LUT<T, i3d::GRAY8> m_lut;
	private:
        /** Current palette. */
		i3d::LUT<i3d::GRAY8, i3d::RGB> m_pal;

		/** Levels params. */
		TLevelsParams<T, i3d::GRAY8> m_levels;
		/** Thresholding params. */
		ThresParams<T> m_thres;

    private:
		void SetPal(const v4d_Palette &smallpal);
		void UpdateLUT();
		void DoThresholding();

    public:
		v4d_GrayMappingModule(wxFrame *parent, const HistogramSet<T> &hist_set);
		
        virtual wxMenu* CreateMenu(int &last_id);
		virtual void Map(const T *src, Buffer2d<i3d::RGB> &buf) const;

        virtual void SetRange();
        virtual void SetBitdepth();
		virtual void SetIdentity();
		virtual void SetLinearStretch();
        virtual void SetPercentilStretch();
		virtual void SetHistEqualStretch();
		virtual bool SetLevelsStretch();
		virtual bool SetThreshold();

        virtual void ReadFromConfig(wxFileConfig &fc);
        virtual void WriteToConfig(wxFileConfig &fc);
        virtual void ReadLevelsFromConfig(wxFileConfig &fc);
        virtual void WriteLevelsToConfig(wxFileConfig &fc);

		/** \brief reset LUTs */
        virtual void OnHistogramChange(const HistogramSet<T> &hist_set);
		virtual void ResetMapping();
		virtual bool HasData();

		void OnMapping(wxCommandEvent &evt);
		void OnUpdateUI(wxUpdateUIEvent &evt);
		void OnPreviewLevels(v4d_LevelsEvent &evt);
		void OnPreviewThresholding(v4d_ThresholdingEvent &evt);

		using v4d_MappingModule<T>::m_hist_set;
		using v4d_MappingModule<T>::GetParent;
		using v4d_MappingModule<T>::GetMenuId;
        using v4d_MappingModule<T>::m_percentil_value;
		using v4d_ExtensionModule::GetCommandId;
		using v4d_MappingModule<T>::Bind;
	};

	/** Mapping module for Complex images. */
	template <class T>
	class V4D_DLLEXPORT v4d_ComplexMappingModule
		: public v4d_MappingModule<std::complex<T> >
	{
	public:
		/** Represent which part of complex imagine is shown */
		enum ComplexMode
		{
			REAL=0,
			IMAGINARY,
			PHASE,
			AMPLITUDE
		};

	private:
		/** Minimum mapped value. */
		T m_min_value[4];
		/** Maximum mapped value. */
		T m_max_value[4];

		/** Colour mode. */
		int m_colour_mode;
		/** Information which part of complex image is shown */
		ComplexMode showPart;

      /** May we even show real part? If false, disable the corresponding menu item as well. */
		const bool m_enabled_real;
      /** May we even show imaginary part? If false, disable the corresponding menu item as well. */
		const bool m_enabled_imag;
      /** May we even show computed phases? If false, disable the corresponding menu item as well. */
		const bool m_enabled_phase;
      /** May we even show computed amplitudes? If false, disable the corresponding menu item as well. */
		const bool m_enabled_amp;

	protected:
		/** Look-up table. */
		i3d::LUT<T, i3d::GRAY8> m_lut;
	private:
		/** Current palette. */
		i3d::LUT<i3d::GRAY8, i3d::RGB> m_pal;

		/** Levels params. */
		TLevelsParams<T, i3d::GRAY8> m_levels[4];
		/** Thresholding params. */
		ThresParams<T> m_thres[4];

	private:
		void SetPal(const v4d_Palette &smallpal);
		void UpdateLUT();
		void DoThresholding();

	public:
		v4d_ComplexMappingModule(wxFrame *parent, const HistogramSet<std::complex<T> > &hist_set,
		  const bool nonemptyR, const bool nonemptyI, const bool nonemptyP, const bool nonemptyA);

		virtual wxMenu* CreateMenu(int &last_id);
		virtual void Map(const std::complex<T> *src, Buffer2d<i3d::RGB> &buf) const;

		virtual void SetRange();
		virtual void SetBitdepth();
		virtual void SetIdentity();
		virtual void SetLinearStretch();
		virtual void SetPercentilStretch();
		virtual void SetHistEqualStretch();
		virtual bool SetLevelsStretch();
		virtual bool SetThreshold();

		virtual void ReadFromConfig(wxFileConfig &fc);
		virtual void WriteToConfig(wxFileConfig &fc);
		virtual void ReadLevelsFromConfig(wxFileConfig &fc);
		virtual void WriteLevelsToConfig(wxFileConfig &fc);

		/** \brief reset LUTs */
		virtual void OnHistogramChange(const HistogramSet<std::complex<T> > &hist_set);
		virtual void ResetMapping();
		virtual bool HasData();

		void OnMapping(wxCommandEvent &evt);
		void OnUpdateUI(wxUpdateUIEvent &evt);
		void OnPreviewLevels(v4d_LevelsEvent &evt);
		void OnPreviewThresholding(v4d_ThresholdingEvent &evt);
		void OnShowPartChange(wxUpdateUIEvent &evt);

		using v4d_MappingModule<std::complex<T> >::m_hist_set;
		using v4d_MappingModule<std::complex<T> >::GetParent;
		using v4d_MappingModule<std::complex<T> >::GetMenuId;
		using v4d_MappingModule<std::complex<T> >::m_percentil_value;
		using v4d_ExtensionModule::GetCommandId;
		using v4d_MappingModule<std::complex<T> >::Bind;

		/** Tell the current mode, see the v4d_ComplexMappingModule::showPart attribute. */
		ComplexMode GetShowMode(void) const
			{ return(this->showPart); };
	};

    /** Create the most appropriate mapper for given type */
	template <class T>
	V4D_DLLEXPORT v4d_MappingModule<T> *v4d_CreateMapperForType
		(wxFrame *parent, const i4d::Image4d<T> &data);
    template <class T> 
    V4D_DLLEXPORT v4d_GrayMappingModule<T> *v4d_CreateGrayMapperForType
        (wxFrame *parent, const HistogramSet<T> &hist_set); 
    template <class T> 
    V4D_DLLEXPORT v4d_RGBMappingModule<T> *v4d_CreateRGBMapperForType
        (wxFrame *parent,  const HistogramSet<i3d::RGB_generic<T> > &hist_set,
		  const bool nonemptyR, const bool nonemptyG, const bool nonemptyB);
	template <class T>
	V4D_DLLEXPORT v4d_ComplexMappingModule<T> *v4d_CreateComplexMapperForType
		(wxFrame *parent, const HistogramSet<std::complex<T> > &hist_set,
		 const bool nonemptyR, const bool nonemptyI,
		 const bool nonemptyP, const bool nonemptyA);

}

#endif
