/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
* overlaying.h
*
* Overlaying functions
*
* Martin Maka (xmaska@fi.muni.cz) 2007
*/

#include "v4d_basic.h"
#include <wx/cmndata.h>
#include <wx/colordlg.h>
#include <wx/frame.h>
#include "overlaying.h"
#include "palette.h"
#include "types/evt_overlaying.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

namespace v4d 
{
	// define event for signaling changes in overlaying
	wxDEFINE_EVENT(v4d_EVT_OVERLAYING_CHANGED, v4d_OverlayingEvent);
	wxDEFINE_EVENT(v4d_EVT_OVERLAYING_COMMAND, v4d_OverlayingCommandEvent);

	enum {
        // Overlaying
        OM_OVERLAY_ENABLE = 0,
		OM_OVERLAY_RELOAD,
        OM_OVERLAY_COMPONENTS,
        OM_OVERLAY_BOUNDARY,

		// colour mode
		OM_COL_DEFAULT,		// default palette
		OM_COL_DEFINED,     // user-defined colour
		OM_COL_CHANGE,
		OM_COL_TRANS_MORE,
		OM_COL_TRANS_LESS,
        
        // the last identifier
		OM_LAST_ID,
	};

	#define IsColourModeId(id) ((id) >= OM_COL_DEFAULT && (id) <= OM_COL_DEFINED)

	/******************************************************************************/
	//                          Generic Overlaying Module
	/******************************************************************************/

	template <class LABEL, class VOXEL> v4d_OverlayingModule<LABEL, VOXEL>::v4d_OverlayingModule(wxFrame *parent) : v4d_ExtensionModule(parent), v4d_Overlayer<LABEL, VOXEL>(), 
		    m_default_colour_mode(true), m_user_colour("RED")
	{
		SetPal();
	}

	/******************************************************************************/

	template <class LABEL, class VOXEL>
		wxMenu *v4d_OverlayingModule<LABEL, VOXEL>::CreateMenu(int &last_id)
	{
		// the menu can be created only once
		wxASSERT(GetMenuId(0) == 0);
		SetFirstMenuId(++last_id);
		wxMenu *menu = new wxMenu();
        
        // Overlay
        menu->AppendCheckItem(this->GetMenuId(OM_OVERLAY_ENABLE), "&Enable\tCtrl-1", "Enable/disable overlaying");

        // Appearance
        //wxMenu *ap_menu = new wxMenu();
        menu->AppendSeparator();
        menu->AppendRadioItem(this->GetMenuId(OM_OVERLAY_COMPONENTS), "&Components\tCtrl-2", "Overlay components (zero transparency)");
        menu->AppendRadioItem(this->GetMenuId(OM_OVERLAY_BOUNDARY), "&Boundary\tCtrl-3", "Overlay boundary (full transparency)");
		menu->AppendSeparator();
        menu->Append(this->GetMenuId(OM_COL_TRANS_MORE), "&Increase transparency\tCtrl-4", "Increase overlay transparency");
        menu->Append(this->GetMenuId(OM_COL_TRANS_LESS), "D&ecrease transparency\tCtrl-5", "Decrease overlay transparency");
        menu->AppendSeparator();
		menu->AppendCheckItem(this->GetMenuId(OM_COL_DEFAULT), "&Default palette\tCtrl-6", "Use default palette");
		menu->AppendCheckItem(this->GetMenuId(OM_COL_DEFINED), "&User-defined colour\tCtrl-7", "Use selected colour");
		menu->AppendSeparator();
		menu->Append(this->GetMenuId(OM_COL_CHANGE), "&Choose colour...\tCtrl-8", "Select user-defined colour");
        //menu->AppendSubMenu(ap_menu, "&View");

        // Reload
        menu->AppendSeparator();		
        menu->Append(this->GetMenuId(OM_OVERLAY_RELOAD), "&Reload\tCtrl-R", "Reload");        
        
        last_id = this->GetMenuId(OM_LAST_ID);

        Bind(wxEVT_COMMAND_MENU_SELECTED, &v4d_OverlayingModule::OnOverlaying, this, this->GetMenuId(0), last_id);
        Bind(wxEVT_UPDATE_UI, &v4d_OverlayingModule::OnUpdateUI, this, this->GetMenuId(0), last_id);

        m_menu_ptr = menu;

		return menu;
	}

	/******************************************************************************/
	
		template <class LABEL, class VOXEL>
		void v4d_OverlayingModule<LABEL, VOXEL>::OnOverlaying(wxCommandEvent &evt)
		{
			int cid = GetCommandId(evt.GetId());

			if (cid == OM_OVERLAY_ENABLE)
			{
				this->SetOverlaying (!this->GetOverlaying());
			}
			else if(cid == OM_OVERLAY_RELOAD)
			{
                v4d_OverlayingEvent oevt(v4d_EVT_OVERLAYING_CHANGED);
				oevt.SetReload(true, true);
				this -> GetParent() -> ProcessWindowEvent(oevt);
				return;
			}
			else if (cid == OM_OVERLAY_COMPONENTS)
			{
				this->SetOverlaying (true);
				this->SetTransparency (0);
			}
			else if (cid == OM_OVERLAY_BOUNDARY)
			{
				this->SetOverlaying (true);
				this->SetTransparency (1);
			}
			else if (IsColourModeId(cid))
			{
				m_default_colour_mode = (cid == OM_COL_DEFAULT);
				SetPal();
			}
			else if (cid == OM_COL_CHANGE)
			{
				ChangeColour();
			}

			float alpha = this->GetTransparency();
			if (cid == OM_COL_TRANS_MORE)
				this->SetTransparency (std::min (1.0f, alpha + 0.05f));
			if (cid == OM_COL_TRANS_LESS)
				this->SetTransparency (std::max (0.0f, alpha - 0.05f));

            v4d_OverlayingEvent oevt(v4d_EVT_OVERLAYING_CHANGED);
			this -> GetParent() -> ProcessWindowEvent(oevt);

			if (evt.ShouldPropagate())
			{
                v4d_OverlayingCommandEvent ocevt(v4d_EVT_OVERLAYING_COMMAND);
				ocevt.SetAttachedEvent(evt);
				ocevt.SetClientData(this);
				GetParent() -> ProcessWindowEvent(ocevt);
			}
		}

	/******************************************************************************/

	template <class LABEL, class VOXEL>
		void v4d_OverlayingModule<LABEL, VOXEL>::OnUpdateUI(wxUpdateUIEvent &evt)
	{
		int cid = GetCommandId(evt.GetId());

		if (cid == OM_OVERLAY_ENABLE)
		{
			evt.Check(this->GetOverlaying());
		}

		if (IsColourModeId(cid))
		{
			switch (cid)
			{
			  case OM_COL_DEFAULT:
				  evt.Check(m_default_colour_mode); 
				  break;

			  case OM_COL_DEFINED:
				  evt.Check(!m_default_colour_mode);
				  break;
			}
		}
	}

	/******************************************************************************/

	template <class LABEL, class VOXEL>
		void v4d_OverlayingModule<LABEL, VOXEL>::ChangeColour()
	{
		wxColourData data;
		data.SetChooseFull(false);
		data.SetColour(m_user_colour);

		wxColourDialog dlg(GetParent(), &data);

		if (dlg.ShowModal() == wxID_OK)
		{
			SetColour(dlg.GetColourData().GetColour());

			m_default_colour_mode = false;
            m_menu_ptr->Check (GetMenuId (OM_COL_DEFINED), true);
		    SetPal();
 	    }
	}

	/******************************************************************************/

	template <class LABEL, class VOXEL>
		void v4d_OverlayingModule<LABEL, VOXEL>::SetPal()
	{
        if (this->m_pal != NULL)
        {
            delete [] this->m_pal;
        }

		if (m_default_colour_mode)
		{
			v4d_Palette pal;
			v4d_GetPalette(v4dPAL_OVERLAY, pal);

            this->m_pal_size = pal.size();
            this->m_pal = new i3d::RGB[this->m_pal_size];
			
            for (size_t i = 0; i < this->m_pal_size; i++)
                this->m_pal[i] = pal[i];
		}
		else
		{
            this->m_pal_size = 1;
            this->m_pal = new i3d::RGB[this->m_pal_size];
            this->m_pal[0] = i3d::RGB(m_user_colour.Red(), m_user_colour.Green(), m_user_colour.Blue());
		}
	}

	/******************************************************************************/

	// Explicit instantiations
	template class v4d_OverlayingModule<i3d::GRAY16, bool>;
	template class v4d_OverlayingModule<i3d::GRAY16, i3d::GRAY8>;
	template class v4d_OverlayingModule<i3d::GRAY16, i3d::GRAY16>;

} // namespace v4d
