/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
* palette.cpp
*
* Palettes
*
* Petr Matula (pem@fi.muni.cz) 2006
*/

#include "v4d_basic.h"
#include <wx/textfile.h>
#include <wx/tokenzr.h>
#include "palette.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

#define COL_BLUE    i3d::RGB(0, 0, 255)
#define COL_RED     i3d::RGB(255, 0, 0)
#define COL_GREEN   i3d::RGB(0, 255, 0)
#define COL_MAGENTA i3d::RGB(255, 0, 255)
#define COL_YELLOW  i3d::RGB(255, 255, 0)
#define COL_CYAN    i3d::RGB(0, 255, 255)

namespace v4d 
{
	bool str2rgb(const wxString &str, i3d::RGB &rgb)
	{
		wxStringTokenizer tkz(str.Strip(wxString::both), " :,");
		wxString n1 = tkz.GetNextToken();
		wxString n2 = tkz.GetNextToken();
		wxString n3 = tkz.GetNextToken();
		long val1, val2, val3;
		if (n1.ToLong(&val1) && n2.ToLong(&val2) && n3.ToLong(&val3))
		{
			rgb = i3d::RGB(val1, val2, val3);
			return true;
		}
		else 
			return false;
	}

	bool LoadPalette(v4d_Palette &pal, const wxString &fname)
	{
		wxTextFile file;
		
		if (!file.Open(fname))
			return false;

		pal.resize(file.GetLineCount() > 256 ? 256 : file.GetLineCount());

		for (size_t i = 0; i < pal.size(); ++i)
		{
			if (!str2rgb(file.GetLine(i), pal[i]))
				return false;
		}

		return true;
	}

	inline i3d::RGB MultColour(const i3d::RGB &c, double factor)
	{
		return i3d::RGB(static_cast<i3d::GRAY8>(double(c.red) * factor), 
			static_cast<i3d::GRAY8>(double(c.green) * factor), 
			static_cast<i3d::GRAY8>(double(c.blue) * factor));
	}

	void AddTransition(v4d_Palette::iterator &p, const i3d::RGB &c1, 
		const i3d::RGB &c2, long num)
	{
        for (long i = 0; i < num; i++)
		{
            i3d::RGB c;
            float factor = float(i) / float(num-1);

            c.red = i3d::GRAY8(c1.red + (c2.red - c1.red) * factor);
            c.green = i3d::GRAY8(c1.green + (c2.green - c1.green) * factor);
            c.blue = i3d::GRAY8(c1.blue + (c2.blue - c1.blue) * factor);

            *p = c;
            ++p;
		}
	}

	void AddMultiple(v4d_Palette::iterator &p, v4d_Palette::iterator end, 
		const i3d::RGB col, long num)
	{
		long i = 1;
		while (p != end && i <= abs(num))
		{
			*p = col;
			++p;
			++i;
		}
	}


	void v4d_GetPalette(size_t idx, v4d_Palette &pal)
	{
		v4d_Palette::iterator p;

		switch (idx)
		{
			case v4dPAL_LABELS:
			
				pal.clear();
				//pal.push_back(i3d::RGB(0,   0,   0));
				pal.push_back(i3d::RGB(255,   0,   0));
				pal.push_back(i3d::RGB(  0, 255,   0));
				pal.push_back(i3d::RGB(  0,   0, 255));
				pal.push_back(i3d::RGB(255, 255,   0));
				pal.push_back(i3d::RGB(  0, 255, 255));
				pal.push_back(i3d::RGB(255,   0, 255));
				pal.push_back(i3d::RGB(128,   0,   0));
				pal.push_back(i3d::RGB(  0, 128,   0));
				pal.push_back(i3d::RGB(  0,   0, 128));
				pal.push_back(i3d::RGB(128, 128,   0));
				pal.push_back(i3d::RGB(  0, 128, 128));
				pal.push_back(i3d::RGB(128,   0, 128));
				break;

			case v4dPAL_SATURATION:

				pal.resize(256);
                for (i3d::GRAY8 i = 1; i < 255; ++i)
					pal[i] = i3d::RGB(i, i, i);

				pal[0] = i3d::RGB(0, 0, 255);
				pal[255] = i3d::RGB(255, 0, 0);
				break;

			case v4dPAL_PERIODIC:

				pal.resize(256);
				p = pal.begin();
				AddTransition(p, COL_BLUE, COL_CYAN, 43);
				AddTransition(p, COL_CYAN, COL_GREEN, 43);
				AddTransition(p, COL_GREEN, COL_YELLOW, 43);
				AddTransition(p, COL_YELLOW, COL_RED, 42);
				AddTransition(p, COL_RED, COL_MAGENTA, 43);
				AddTransition(p, COL_MAGENTA, COL_BLUE, 42);
				break;

			case v4dPAL_OVERLAY:

				pal.clear();
                pal.push_back(i3d::RGB(255, 255, 255)); // white
                pal.push_back(i3d::RGB(255,   0,   0)); // red
				pal.push_back(i3d::RGB(  0, 255,   0)); // green
				pal.push_back(i3d::RGB(  0,   0, 255)); // blue
                pal.push_back(i3d::RGB(255,   0, 255)); // magenta
				pal.push_back(i3d::RGB(  0, 255, 255)); // cyan
				pal.push_back(i3d::RGB(128,   0,   0)); // dark red
				pal.push_back(i3d::RGB(  0, 128,   0)); // dark green
				pal.push_back(i3d::RGB(  0,   0, 128)); // dark blue
				pal.push_back(i3d::RGB(128, 128, 255)); // light blue
				pal.push_back(i3d::RGB(128, 128,   0)); // beuge
                pal.push_back(i3d::RGB(128, 0,   255));	// dark magenta
                pal.push_back(i3d::RGB(255, 128, 255)); // light pink
                pal.push_back(i3d::RGB(128, 128, 128)); // grey
				break;
				
			default: // gray palette

				pal.resize(256);
				pal[255] = i3d::RGB(255, 255, 255);
		                for (i3d::GRAY8 i = 0; i < 255 ; ++i)
					pal[i] = i3d::RGB(i, i, i);
				break;
		}
	}

} // namespace v4d


