/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * FILE: thistogram.cc
 *
 * template version of histogram
 *
 * Petr Matula (pem@fi.muni.cz) 2006
 * Ondrej Danek (xdanek2@fi.muni.cz) 2007
 */

#include "v4d_basic.h"
#include <fstream>
#include <wx/busyinfo.h>
#include "thistogram.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

namespace v4d 
{
    /////////////////////////////////////////////////////////////////////
    //
    //  Histogram class
    //
    ////////////////////////////////////////////////////////////////////

	/**
		Histogram shoud not be used with complex type 
	*/
	template<>
	size_t Histogram<std::complex<float> >::GetBinPos(std::complex<float> val) const
	{
		throw "Not Implemented";
	}
	
	template<class T>
	size_t Histogram<T>::GetBinPos(T val) const
	{
		if (val <= GetLeftCenter())
			return 0;
		else if (val >= GetRightCenter())
			return size() - 1;
		else
			return size_t(
			(double(val) - double(GetLeftCenter())) /
			(double(GetRightCenter()) - double(GetLeftCenter())) *
			double(m_num_bins - 1));
	}

    template <class T>
        Histogram<T>& Histogram<T>::operator+=(const Histogram<T> &h) 
    {
        size_t max = std::min (m_num_bins, h.size());

        for (size_t i = 0; i < max; i++)
            m_bins[i] += h.m_bins[i];

        return *this;
    }

	template <class T>
		bool Histogram<T>::operator==(const Histogram<T> &h)
	{
		return (size() == h.size() && GetLeftCenter() == h.GetLeftCenter() && GetRightCenter() == h.GetRightCenter());
	}

    template <class T>
		bool Histogram<T>::operator!=(const Histogram<T> &h)
	{
		return (size() != h.size() || GetLeftCenter() != h.GetLeftCenter() || GetRightCenter() != h.GetRightCenter());
	}

    template <class T>
        void Histogram<T>::Allocate(size_t num_bins)
    {
        Dispose ();
        m_num_bins = num_bins;
        if (num_bins > 0)
        {
            m_bins = new size_t[num_bins];
            m_bin_centers = new T[num_bins];
        }
    }

    template <class T>
        void Histogram<T>::Dispose()
    { 
        if (m_bins != NULL)
        {
            delete [] m_bins;
        }
        if (m_bin_centers != NULL)
        {
            delete [] m_bin_centers;
        }
        m_bins = NULL;
        m_bin_centers = NULL;
        m_num_bins = 0;
    }

	template <class T>
        void Histogram<T>::GetCopy (const Histogram<T> &h)
	{
        Allocate (h.size());

        for (size_t i = 0; i < m_num_bins; i++)
        {
            m_bins[i] = h.m_bins[i];
            m_bin_centers[i] = h.m_bin_centers[i];
        }
    };

	template <class T>
		T ComputeBinCenter(size_t i, size_t num_bins, const T &lc, const T &rc)
	{
		if (num_bins < 2) 
			return T((double(lc) + double(rc)) / 2.0); 
		else
			return T((double(rc) - double(lc)) / double(num_bins - 1) * double(i) + double(lc));
	};

	template <class T>
		void Histogram<T>::Init(const T &minbnd, const T &maxbnd, size_t num_bins)
	{
        Allocate (num_bins);

		for (size_t i = 0; i < num_bins; ++i)
		{
			m_bins[i] = 0;
			m_bin_centers[i] = ComputeBinCenter(i, num_bins, minbnd, maxbnd);
		}
	}

    template <class T>
        size_t Histogram<T>::GetSum() const 
    { 
        size_t sum = 0;
        for (size_t i = 0; i < m_num_bins; i++)
            sum += m_bins[i];
        return sum;
    }

    template <class T>
        void Histogram<T>::Clear()
    { 
        for (size_t i = 0; i < m_num_bins; i++)
            m_bins[i] = (size_t)0;
    }

	template <class T>
		bool Histogram<T>::Save(const wxString &filename) const
	{
        // Find first and last used bin
        size_t first_bin = 0, last_bin = 0;

        if (m_num_bins > 0)
        {
            last_bin = m_num_bins - 1;
            while (first_bin < last_bin && m_bins[first_bin] == 0)
                first_bin++;
            while (last_bin > first_bin && m_bins[last_bin] == 0)
                last_bin--;
        }

        std::ofstream ofs((const char *)filename.mb_str(wxConvFile));

        if (!ofs.good())
            return false;

        // Save data
        ofs << typeid(T).name() << '\n' << m_num_bins << ' ' << first_bin << ' ' << last_bin << ' ' << '\n';
        if (typeid(T) == typeid (unsigned char))
            ofs << (int)GetLeftCenter() << ' ' << (int)GetRightCenter() << '\n';
        else
            ofs << GetLeftCenter() << ' ' << GetRightCenter() << '\n';

        for (size_t i = first_bin; i <= last_bin; i++)
            ofs << m_bins[i] << ' ';

        return true;
	}

	template <class T>
		bool Histogram<T>::Load(const wxString &filename)
	{
        size_t num_bins, first_bin, last_bin;
        T left, right;

        std::ifstream ifs((const char *)filename.mb_str(wxConvFile));

        if (!ifs.good())
            return false;

        char type_name[100];
        ifs.getline (type_name, 99);
        if (strcmp (type_name, typeid(T).name()))
            return false;

        if (typeid(T) != typeid(unsigned char))
        {
            ifs >> num_bins >> first_bin >> last_bin >> left >> right;
        }
        else
        {
            int left_int, right_int;
            ifs >> num_bins >> first_bin >> last_bin >> left_int >> right_int;
            left = (unsigned char)left_int;
            right = (unsigned char)right_int;
        }
        Init (left, right, num_bins);

        for (size_t i = 0; i < first_bin; i++)
        {
            m_bins[i] = 0;
        }
        for (size_t i = last_bin + 1; i < num_bins; i++)
        {
            m_bins[i] = 0;
        }                
        
        for (size_t i = first_bin; i <= last_bin; i++)
        {
            // Following code replaces: ifs >> m_bins[i]; which was too slow
            char c = 0, buf[20];
            size_t len = 0;

            while (len < 20)
            {
                ifs.read(&c, 1);

                if (c == ' ' || ifs.eof())
                {
                    break;
                }

                if (c != '\n' && c != '\r')
                {
                    if (c < '0' || c > '9')
                    {
                        return false;
                    }

                    buf[len++] = c;
                }
            }

            if (!len || len > 19)
            {
                return false;
            }

            // Convert string to number
            m_bins[i] = 0;
            size_t nom = 1;
            
            while (len-- > 0)
            {
                m_bins[i] += size_t(buf[len] - '0') * nom;
                nom *= 10;
            }
        }

        return true;
	}

    template <class T> 
        size_t Histogram<T>::FindFirstNonZero() const
    {
        for (size_t i = 0; i < m_num_bins; i++)
            if (m_bins[i] != 0)
                return i;

        return m_num_bins - 1;
    }

    template <class T> 
        size_t Histogram<T>::FindLastNonZero() const
    {
        for (size_t i = m_num_bins; i > 0; i--)
            if (m_bins[i - 1] != 0)
                return i - 1;

        return 0;
    }

    template <class T> 
        size_t Histogram<T>::FindLeftPercentil (double p) const
    {
        size_t pos = 0;
        double sum = GetSum();
        double accum = 0.0;

        do 
        {
            pos++;
            accum += double(m_bins[pos - 1]) / sum;
        } while (pos < m_num_bins && accum < p);

        return pos - 1;
    }

    template <class T> 
        size_t Histogram<T>::FindRightPercentil (double p) const
    {
        size_t pos = m_num_bins;
        double sum = GetSum();
        double accum = 0.0;

        do 
        {
            pos--;
            accum += double(m_bins[pos]) / sum;
        } while (pos > 0 && accum < p);

        return pos;
    }

	template <class T>
		void Histogram<T>::CreateSubHist(const Histogram<T> &hist, T min_val, T max_val)
	{
		size_t min_pos = hist.GetBinPos(min_val);
		size_t max_pos = hist.GetBinPos(max_val);
		size_t num_bins = max_pos - min_pos + 1;

		Allocate(num_bins);

		for (size_t i = 0; i < num_bins; ++i)
		{
			m_bins[i] = hist[i + min_pos];
			m_bin_centers[i] = hist.GetBinCenter(i + min_pos);
		}
	}

    /////////////////////////////////////////////////////////////////////
    //
    //  HistogramSet class
    //
    ////////////////////////////////////////////////////////////////////

    template <class T>
        void HistogramSet<T>::CreateFromImage (const i3d::Image3d<T> &img)
    {
        wxBusyInfo wait("Creating histogram...");
        ComputeHist (m_hist, img.begin(), img.end(), (size_t) std::numeric_limits<T>::max() + 1);
    }

    template <class T>
        void HistogramSet<i3d::RGB_generic<T> >::CreateFromImage(const i3d::Image3d<i3d::RGB_generic<T> > &img)
    {
	    wxBusyInfo wait1("Creating histogram (red)...");
	    IntensityHist (img, m_hist[0], 0);
	    wxBusyInfo wait2("Creating histogram (green)...");
	    IntensityHist (img, m_hist[1], 1);
	    wxBusyInfo wait3("Creating histogram (blue)...");
	    IntensityHist (img, m_hist[2], 2);
    }
	
	template <class T>
	T GetPhase(const std::complex<T> *complexNum)
	{
		return std::atan2(complexNum->imag(), complexNum->real());
	}

	template <class T>
	T GetAmplitude(const std::complex<T> *complexNum)
	{
		return sqrt(complexNum->real() * complexNum->real() + complexNum->imag() * complexNum->imag());
	}

	template <class T>
	void HistogramSet<std::complex<T> >::CreateFromImage(const i3d::Image3d<std::complex<T> > &img)
	{
		wxBusyInfo wait("Creating histogram...");
		const std::complex<T> *p = img.GetFirstVoxelAddr();

		//0-real 1-imaginary 2-phase 3-amplitude
		T min_value[4]={p->real(), p->imag(), GetPhase<T>(p), GetAmplitude<T>(p)};
		T max_value[4]={p->real(), p->imag(), GetPhase<T>(p), GetAmplitude<T>(p)};
		i3d::Image3d<T> complexPartsImgs[4];

		for (size_t i = 0; i < 4; i++)
		{
			complexPartsImgs[i].CopyMetaData(img);
		}

		T *voxelPs[4]={complexPartsImgs[0].GetFirstVoxelAddr(), complexPartsImgs[1].GetFirstVoxelAddr(),
			complexPartsImgs[2].GetFirstVoxelAddr(), complexPartsImgs[3].GetFirstVoxelAddr()};

		T phase;
		T amplitude;

		for (size_t i = 0; i < img.GetImageSize(); ++i) {
			phase = GetPhase(p);
			amplitude = GetAmplitude(p);

			if (min_value[0] > p->real()) min_value[0] = p->real();
			if (max_value[0] < p->real()) max_value[0] = p->real();

			if (min_value[1] > p->imag()) min_value[1] = p->imag();
			if (max_value[1] < p->imag()) max_value[1] = p->imag();

			if (min_value[2] > phase) min_value[2] = phase;
			if (max_value[2] < phase) max_value[2] = phase;

			if (min_value[3] > amplitude) min_value[3] = amplitude;
			if (max_value[3] < amplitude) max_value[3] = amplitude;

			*voxelPs[0] = p->real();
			*voxelPs[1] = p->imag();
			*voxelPs[2] = phase;
			*voxelPs[3] = amplitude;

			++p;
			for (size_t j = 0; j < 4; j++)
			{
				++voxelPs[j];
			}
		}

		for (size_t i = 0; i < 4; i++)
		{
			ComputeHist(m_hist[i], complexPartsImgs[i].begin(), complexPartsImgs[i].end(),
				std::numeric_limits<i3d::GRAY8>::max() + 1, min_value[i], max_value[i]);
		}
	}

	template <class T>
	void HistogramSet<std::complex<T> >::CreateFromImage(const i3d::Image3d<std::complex<T> > &img,
		const T min_values[4], const T max_values[4])
	{
		wxBusyInfo wait("Creating histogram...");

		//a helper scalar image to compute histogram upon it
		i3d::Image3d<T> tmp;
		tmp.CopyMetaData(img);

		//0-real 1-imaginary 2-phase 3-amplitude

		//real part:
		const std::complex<T>* p=img.GetFirstVoxelAddr(); //read-only pointer that can be changed
		const std::complex<T>* const pS=p + img.GetImageSize(); //read-only pointer that can't be changed
		T* t=tmp.GetFirstVoxelAddr();
		while (p != pS)
		{
			*t=p->real();
			++p; ++t;
		}
		ComputeHist(m_hist[0], tmp.begin(), tmp.end(),
			std::numeric_limits<i3d::GRAY8>::max() + 1, min_values[0], max_values[0]);

		//imaginary part:
		p=img.GetFirstVoxelAddr();
		t=tmp.GetFirstVoxelAddr();
		while (p != pS)
		{
			*t=p->imag();
			++p; ++t;
		}
		ComputeHist(m_hist[1], tmp.begin(), tmp.end(),
			std::numeric_limits<i3d::GRAY8>::max() + 1, min_values[1], max_values[1]);

		//phase part:
		p=img.GetFirstVoxelAddr();
		t=tmp.GetFirstVoxelAddr();
		while (p != pS)
		{
			*t=GetPhase(p);
			++p; ++t;
		}
		ComputeHist(m_hist[2], tmp.begin(), tmp.end(),
			std::numeric_limits<i3d::GRAY8>::max() + 1, min_values[2], max_values[2]);

		//amplitude part:
		p=img.GetFirstVoxelAddr();
		t=tmp.GetFirstVoxelAddr();
		while (p != pS)
		{
			*t=GetAmplitude(p);
			++p; ++t;
		}
		ComputeHist(m_hist[3], tmp.begin(), tmp.end(),
			std::numeric_limits<i3d::GRAY8>::max() + 1, min_values[3], max_values[3]);
	}

    // Explicit instantiations
	template class V4D_DLLEXPORT Histogram<i3d::GRAY8>;
	template class V4D_DLLEXPORT Histogram<i3d::GRAY16>;
	template class V4D_DLLEXPORT Histogram<float>;
	template class V4D_DLLEXPORT Histogram<int>;
	
    template class V4D_DLLEXPORT HistogramSet<i3d::GRAY8>;
    template class V4D_DLLEXPORT HistogramSet<i3d::GRAY16>;
    template class V4D_DLLEXPORT HistogramSet<i3d::RGB>;
    template class V4D_DLLEXPORT HistogramSet<i3d::RGB16>;
    template class V4D_DLLEXPORT HistogramSet<float>;
    template class V4D_DLLEXPORT HistogramSet<int>;
	template class V4D_DLLEXPORT HistogramSet<std::complex<float> >;
}
