/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * FILE: thistogram.h
 *
 * template version of histogram
 *
 * Petr Matula (pem@fi.muni.cz) 2006
 * Ondrej Danek (xdanek2@fi.muni.cz) 2007
 */

#ifndef V4D_THISTOGRAM_H
#define V4D_THISTOGRAM_H

/**
 * \defgroup hist Histogram manipulation
 */

#include <limits>
#include <i3d/image3d.h>
#include "v4d_impexp.h"

namespace v4d 
{
    /////////////////////////////////////////////////////////////////////
    //  Histogram class
    ////////////////////////////////////////////////////////////////////
    template <class T>
    class V4D_DLLEXPORT Histogram
    {		
    private:
        /** Number of bins */
        size_t m_num_bins;
        /** Bins array */
        size_t *m_bins;
        /** Bin centers array */
        T *m_bin_centers;

    public:
        /** Default constructor. It creates an empty histogram with zero number of bins. */
        Histogram() : m_num_bins(0), m_bins(NULL), m_bin_centers(NULL)  {};
        /** Copy constructor */
        Histogram(const Histogram<T> &h) : m_num_bins(0), m_bins(NULL), m_bin_centers(NULL) { GetCopy(h); }
        /** Destructor */
        virtual ~Histogram() { Dispose(); }

        /** Assignment operator */
        Histogram& operator= (const Histogram<T> &h) { GetCopy (h); return *this; }
        /** Add two histograms together */
        Histogram& operator+=(const Histogram<T> &h);
		/** Comparison operators. */
		bool operator==(const Histogram<T> &h);
		bool operator!=(const Histogram<T> &h);

        /** Is histogram empty ? */
        bool empty() const { return m_num_bins == 0; };
        /** Get the size of histogram, i.e. number of bins */
        size_t size() const { return m_num_bins; };

        /** Init histogram - allocate memory for data and calculate centers */
        void Init(const T &left_center, const T &right_center, size_t num_bins);
        /** Save histogram to given file \return true on success */
        bool Save (const wxString &filename) const;
        /** Load histogram from given file \return true on success */
        bool Load (const wxString &filename);

        /** Get reference to bin corresponding to this value */
        size_t &GetBin(const T& val) { return m_bins[GetBinPos(val)]; };
        /** Get constant reference to bin corresponding to this value */
        const size_t &GetBin(const T& val) const { return m_bins[GetBinPos(val)]; };
        /** Get reference to given bin */
        size_t &operator[](size_t bin) { return m_bins[bin]; };
        /** Get const reference to given bin */
        const size_t &operator[](size_t bin) const { return m_bins[bin]; };

        /** Return index of bin for given value */
		size_t GetBinPos(T val) const;

        /** Find first nonzero bin */
        size_t FindFirstNonZero() const;
        /** Find last non zero bin */
        size_t FindLastNonZero() const;
        /** Find given left percentil */
        size_t FindLeftPercentil (double p = 0.05) const;
        /** Find given right percentil */
        size_t FindRightPercentil (double p = 0.05) const;

        /** Return left bin center value */
        T GetLeftCenter() const {return m_bin_centers[0]; };
        /** Return right bin center value */
        T GetRightCenter() const {return m_bin_centers[m_num_bins - 1]; };
        /** Return bin center for given position */
        T GetBinCenter(size_t i) const {return m_bin_centers[i]; };

        /** Return begin iterator */
        size_t *begin() { return &m_bins[0]; };
        const size_t *begin() const { return &m_bins[0]; };
        /** Return end iterator */
        size_t *end()   { return &m_bins[m_num_bins - 1]; };
        const size_t *end() const { return &m_bins[m_num_bins - 1]; };

        /** Calculate histogram sum */
        size_t GetSum() const;
        /** Clear histogram - fill bins with zeroes */
        void Clear();

		/** Create subhistogram. */
		void CreateSubHist(const Histogram<T> &hist, T min_val, T max_val);

    private:
        /** Allocate memory for histogram */
        void Allocate(size_t num_bins);
        /** Free allocated memory */
        void Dispose() ;
        /** Get copy of given histogram */
        void GetCopy (const Histogram<T> &h);
    };


    /////////////////////////////////////////////////////////////////////
    //  HistogramSet 
    //
    //  Class containing histogram for each image channel.
    ////////////////////////////////////////////////////////////////////

    // One channel variant (GRAY8, GRAY16, int, float, ...)
    template <class T>
    class V4D_DLLEXPORT HistogramSet
    {
    private:
        Histogram<T>    m_hist;

    public:
        /** Default constructor - creates empty histogram */
        HistogramSet() {};
        /** Constructor - create histogram from source image */
        HistogramSet(const i3d::Image3d<T> &img) { CreateFromImage (img); }

        /** Get histogram */
        Histogram<T> &Get() { return m_hist; }
        const Histogram<T> &Get() const { return m_hist; } 
        /** Set new histogram */
        void Set (const Histogram<T> &hist) { m_hist = hist; }
        /** Create histogram from given image */
        void CreateFromImage (const i3d::Image3d<T> &img);
        /** Is histogram empty */
        bool IsEmpty() { return m_hist.empty(); }
    };

    // RGB variant
    template <class T>
    class V4D_DLLEXPORT HistogramSet<i3d::RGB_generic<T> >
    {
    private:
        Histogram<T>    m_hist[3];

    public:
        /** Default constructor - creates empty histogram */
        HistogramSet() {};
        /** Constructor - create histogram from source image */
        HistogramSet(const i3d::Image3d<i3d::RGB_generic<T> > &img) { CreateFromImage (img); }

        /** Get histogram for given channel (0=red, 1=green, 2=blue) */
        Histogram<T> &Get (int channel) { return m_hist[channel]; }
        const Histogram<T> &Get (int channel) const { return m_hist[channel]; }
        /** Acces via [] operator (0=red, 1=green, 2=blue) */
        Histogram<T> &operator[] (int channel) { return m_hist[channel]; }
        const Histogram<T> &operator[] (int channel) const { return m_hist[channel]; }
        /** Set new histogram */
        void Set (int channel, const Histogram<T> &hist) { m_hist[channel] = hist; }
        /** Create histogram from given image */
        void CreateFromImage (const i3d::Image3d<i3d::RGB_generic<T> > &img);
        /** Is histogram empty */
        bool IsEmpty() { return m_hist[0].empty() && m_hist[1].empty() && m_hist[2].empty(); }
    };

	// Complex variant
	template <class T>
	class V4D_DLLEXPORT HistogramSet<std::complex<T> >
	{
	private:
		Histogram<T>    m_hist[4];

	public:
		/** Default constructor - creates empty histogram */
		HistogramSet() {};
		/** Constructor - create histogram from source image */
		HistogramSet(const i3d::Image3d<std::complex<T> > &img) { CreateFromImage(img); }

		/** Get histogram for given channel (0=real, 1=imaginary, 2=phase, 3=amplitude) */
		Histogram<T> &Get(int channel) { return m_hist[channel]; }
		const Histogram<T> &Get(int channel) const { return m_hist[channel]; }
		/** Acces via [] operator (0=real, 1=imaginary, 2=phase, 3=amplitude) */
		Histogram<T> &operator[] (int channel) { return m_hist[channel]; }
		const Histogram<T> &operator[] (int channel) const { return m_hist[channel]; }
		/** Set new histogram */
		void Set(int channel, const Histogram<T> &hist) { m_hist[channel] = hist; }
		/** Create histogram from given image */
		void CreateFromImage(const i3d::Image3d<std::complex<T> > &img);
		/** Create histogram from given image. Min and max values for particular 
			parts of imaginary image are given in the arrays.*/
		void CreateFromImage(const i3d::Image3d<std::complex<T> > &img, const T min_values[4], const T max_values[4]);
		/** Is histogram empty */
		bool IsEmpty() { return m_hist[0].empty() && m_hist[1].empty() && m_hist[2].empty() && m_hist[3].empty(); }
	};

    /////////////////////////////////////////////////////////////////////
    //  
    // Auxiliary functions
    //  
    ////////////////////////////////////////////////////////////////////

    /** 
    * \ingroup hist
    *
    * Compute histogram for given array of values 
    */
    template <class InIt, class T>
    void ComputeHist(Histogram<T> &hist, InIt first, InIt last, size_t num_bins = 256, 
        T left_center = std::numeric_limits<T>::min(), 
        T right_center = std::numeric_limits<T>::max(),
        size_t stride = 1)
    {
        hist.Init(left_center, right_center, num_bins);

        if (stride == 1)
        {
            while (first < last)
            {
                ++hist.GetBin(*first);
                ++first;
            }
        }
        else
        {
            while (first < last)
            {
                ++hist.GetBin(*first);
                first += stride;
            }
        }
    }

    /** Computes intensity histogram of image img. Output is stored to the vector hist*/
    template <class T> 
    void IntensityHist(const i3d::Image3d<T> &img, Histogram<T> &hist) 
    {
        ComputeHist(hist, img.begin(), img.end());
    };

    /** 
    * \ingroup hist
    *
    * Computes histogram from RGB image. 
    * \p channel specifies for which channels the histogram is computed
    * (0 = red, 1 = green, 2 = blue) 
    **/
    template <class T> 
    void IntensityHist(const i3d::Image3d< i3d::RGB_generic<T> > &img, Histogram<T> &hist, int channel)
    {
        ComputeHist(hist,
            (T *)(img.begin()) + channel, 
            (T *)(img.end()) + channel, 
            std::numeric_limits<T>::max() + 1,
            std::numeric_limits<T>::min(),
            std::numeric_limits<T>::max(),
            3);
    }

	/**
	* \ingroup hist
	*
	* Computes histogram from Complex image.
	* \p channel specifies for which channels the histogram is computed
	* (0 = real, 1 = imaginary, 2 = phase, 3 = amplitude)
	**/
	template <class T>
	void IntensityHist(const i3d::Image3d< std::complex<T> > &img, Histogram<T> &hist, int channel)
	{
		ComputeHist(hist,
			(T *)(img.begin()) + channel,
			(T *)(img.end()) + channel,
			std::numeric_limits<T>::max() + 1,
			std::numeric_limits<T>::min(),
			std::numeric_limits<T>::max(),
			4);
	}

    inline void IntensityHist(const i3d::Image3d<float> &img, Histogram<float> &hist)
    {
        ComputeHist(hist,
            img.begin(),
            img.end(),
            256,
            float(std::numeric_limits<i3d::GRAY8>::min()),
            float(std::numeric_limits<i3d::GRAY8>::max()));
    }

	inline void IntensityHist(const i3d::Image3d<double> &img, Histogram<double> &hist)
	{
		ComputeHist(hist,
			img.begin(),
			img.end(),
			256,
			double(std::numeric_limits<i3d::GRAY8>::min()),
			double(std::numeric_limits<i3d::GRAY8>::max()));
	}
}

#endif
