/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * buffer2d.h
 *
 * 2d buffer
 *
 * Petr Matula (pem@fi.muni.cz) 2003
 */


#ifndef V4D_TYPES_BUFFER2D_H
#define V4D_TYPES_BUFFER2D_H

namespace v4d
{
    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                           Buffer2d                          //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    /** Two-dimensional buffer. This class represents a 2D buffer. 
        The buffer has its size dx*dy and a pointer to a continuous block
        of memory with length dx*dy. 
    */
    template <class ITEM> 
    class Buffer2d 
    {
        /** Width of the buffer */
        int dx;
        /** Height of the buffer */
        int dy;
        /** Pointer to the buffer data */
        ITEM *data;

    public:
        /** Constructor. It creates a new buffer of size \a dx_ x \a dy_. */
        explicit Buffer2d(int dx_ = 0, int dy_ = 0)
            : dx(dx_), dy(dy_) 
        {
            data = new ITEM[dx_*dy_];
        }

        /** Copy constructor. It creates a new buffer of size \a dx_ x \a dy_ using the nearest neighbour strategy. The parameters reverse_x and reverse_y
            determine the axis orientations of a new buffer with respect to the axis orientations of the source buffer. */
        Buffer2d(const Buffer2d &buff, int dx_, int dy_, bool reverse_x = false, bool reverse_y = false, bool retain_aspect = true)
            : dx(dx_), dy(dy_)
        {
            data = new ITEM[dx * dy];
            clear();

            int size_x = dx, size_y = dy;
            if (retain_aspect)
            {
                if (buff.GetWidth() > buff.GetHeight())
                    size_y = static_cast<int>((float) dy / ((float) buff.GetWidth() / (float) buff.GetHeight()));
                else
                    size_x = static_cast<int>((float) dx / ((float) buff.GetHeight() / (float) buff.GetWidth()));
            }

            if (size_x == 0 || size_y == 0)
                return;

            int rev_x = reverse_x ? 1 : 0;
            int rev_y = reverse_y ? 1 : 0;
            int mult_x = reverse_x ? -1 : 1;
            int mult_y = reverse_y ? -1 : 1;
            int step_x = buff.GetWidth() / size_x;
            int step_y = buff.GetHeight() / size_y;
            int start_x = (buff.GetWidth() - step_x * size_x) / 2;
            int start_y = (buff.GetHeight() - step_y * size_y) / 2;

            step_x *= mult_x;
            step_y *= mult_y;

            //int src_start_pos = start_y * buff.GetWidth() + (mult_x * start_x);
            int dst_start_pos = ((dy - size_y + 1) / 2) * dx + ((dx - size_x + 1) / 2);

            int src_start_pos = rev_y * buff.GetWidth() * buff.GetHeight() + mult_y * (start_y + rev_y) * buff.GetWidth() + mult_x * start_x + rev_x * (buff.GetWidth() - 1);

            ITEM *src = buff.GetData();

            for (int y = 0; y < size_y; y++)
            {
                for (int x = 0; x < size_x; x++)
                {
                    data[dst_start_pos + y * dx + x] = src[src_start_pos + y  * step_y * buff.GetWidth() + x * step_x];
                }
            }
        }

        /** Destructor. Frees the memory used by the buffer. */
        ~Buffer2d() 
        {
            if (data)
            {
                delete[] data; 
            }
        }

        /** Buffer resizing. The new size is set to \a dx_ x \a dy_. */
        void resize(int dx_, int dy_)
        {
            if (data)
            {
                delete [] data;
            }

            data = new ITEM[dx_*dy_];
            dx = dx_;
            dy = dy_;
        }

        /** Buffer clearing. All elements in the buffer are set to the value
            ITEM(). */
        void clear() 
        {
            ITEM *p = data;
            for (int i = 0; i < dx * dy; i++, ++p)
            {
                *p = ITEM();
            }
        }

        /** \returns Size of the buffer, i.e. GetWidth()*GetHeight(). */
        int GetSize(void) const
        {
            return dx * dy; 
        }

        /** \returns Width of the buffer. */
        int GetWidth(void) const
        {
            return dx; 
        }

        /** \returns Height of the buffer. */
        int GetHeight(void) const
        {
            return dy; 
        }

        /** \returns Pointer to the first element in the buffer. */
        ITEM* GetData() const 
        { 
            return &data[0]; 
        }

        ITEM GetDataAt(size_t x, size_t y) const
        {
            return data[y * dx + x];
        }
    };
}

#endif
