/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * evt_overlaying.h
 *
 * Event signaling changes in overlaying
 *
 * Martin Maka (xmaska@fi.muni.cz) 2007
 */


#ifndef V4D_TYPES_EVT_OVERLAYING_H
#define V4D_TYPES_EVT_OVERLAYING_H

#include <wx/event.h>
#include "../v4d_impexp.h"

namespace v4d 
{
    /**************************************************/
    /*                                                */
    /*               Overlaying events                */
    /*                                                */
    /**************************************************/	

    /** An event sent by Overlaying module as a request for redrawing. */
    class v4d_OverlayingEvent 
        : public wxEvent
    {
    protected:
        /** Flag signalizing that the overlay image for af frame should be reloaded from disk */
        bool m_reload_af;
        /** Flag signalizing that the overlay image for v4d frame should be reloaded from disk */
        bool m_reload_3d;

    public:
        /** Constructor. */
        v4d_OverlayingEvent(wxEventType eventType, int id = 0)
            : wxEvent(id, eventType), m_reload_af(false), m_reload_3d(false) 
        {}

        /** Clone method. */
        virtual wxEvent *Clone() const { return new v4d_OverlayingEvent(*this); }

        /** Set reloading flags */
        void SetReload (bool af, bool v4d) { m_reload_af = af; m_reload_3d = v4d; }
        /** Returns true if af overlay should be reloaded */
        bool ShouldReloadAf () { return m_reload_af; }
        /** Returns true if af overlay should be reloaded */
        bool ShouldReload3d () { return m_reload_3d; }
    };
    
    /** An event sent by Overlaying module to inform parent about taken changes. */
    class v4d_OverlayingCommandEvent 
        : public wxCommandEvent
    {
    private:
        wxEvent *_attEvent;

    public:
        /** Constructor. */
        v4d_OverlayingCommandEvent(wxEventType eventType, int id = 0)
            : wxCommandEvent(eventType, id), _attEvent(NULL) 
        {}

        /** Copy constructor */
        v4d_OverlayingCommandEvent(const v4d_OverlayingCommandEvent &copy)
            : wxCommandEvent(copy)
        { 
            _attEvent = copy._attEvent->Clone(); 
        }

        /** Destructor */
        ~v4d_OverlayingCommandEvent() 
        { 
            if (_attEvent != NULL)
            {
                delete _attEvent; 
            }
        }

        /** Clone method. */
        virtual wxEvent *Clone() const 
        { 
            return new v4d_OverlayingCommandEvent(*this); 
        }

        /** Get the attached event */
        wxEvent* GetAttachedEvent() 
        { 
            return _attEvent; 
        }

        /** Get the attached event */
        void SetAttachedEvent(const wxEvent &evt) 
        { 
            if (_attEvent != NULL)
            {
                delete _attEvent; 
            }

			_attEvent = evt.Clone(); 
        }
    };

    // Declare event tags
    wxDECLARE_EXPORTED_EVENT(V4D_DLLEXPORT, v4d_EVT_OVERLAYING_CHANGED, v4d_OverlayingEvent);
    wxDECLARE_EXPORTED_EVENT(V4D_DLLEXPORT, v4d_EVT_OVERLAYING_COMMAND, v4d_OverlayingCommandEvent);
}

#endif

