/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
* overlayer.h
*
* Abstract overlayer
*
* Martin Maka (xmaska@fi.muni.cz) 2007
* Ondrej Danek (xdanek2@fi.muni.cz) 2007
*/

#ifndef V4D_TYPES_OVERLAYER_H
#define V4D_TYPES_OVERLAYER_H

#include <i3d/image3d.h>
#include <i3d/regions.h>
#include "buffer2d.h"

namespace v4d 
{
	/** All posible window types that can be overlayed */
	enum v4d_OverlayWindowType
	{
		v4d_OVERLAY_WIN_AF = 0,
		v4d_OVERLAY_WIN_XY,
		v4d_OVERLAY_WIN_XZ,
		v4d_OVERLAY_WIN_YZ,
		v4d_OVERLAY_WIN_XY_AF,
		v4d_OVERLAY_WIN_XZ_AF,
		v4d_OVERLAY_WIN_YZ_AF,
	};

	//////////////////////////////////////////////////////////////////////////////////////////
	//
	// v4d_Overlayer abstract class
	//
	//////////////////////////////////////////////////////////////////////////////////////////
	template <class LABEL, class VOXEL> class v4d_Overlayer
	{
	private:
		/** Overlaying on/off */
		bool m_overlaying;
		/** Transparency ratio */
		float m_alpha;

	protected:
		/** The palette for mapping labels into the given colour. */
		i3d::RGB *m_pal;
		/** The size of palette - if label number is higher than the palette size then modulo is used */
		size_t m_pal_size;

	public:
		/** Constructor. */
		v4d_Overlayer() 
            : m_overlaying(false), m_alpha(0), m_pal(NULL), m_pal_size(0) 
        {}

		/** Destructor. */
		virtual ~v4d_Overlayer()
		{
			if (m_pal != NULL)
			{
				delete [] m_pal; 
			}
		}

		/** Returns true if overlaying is enabled */
		bool GetOverlaying() const { return m_overlaying; }
		/** Enable/disable overlaying */
		void SetOverlaying(bool enable) { m_overlaying = enable; }
		/** Get transparency */
		float GetTransparency () const { return m_alpha; }
		/** Set transparency */
		void SetTransparency (float alpha) { m_alpha = alpha; }
		/** Get the palette */
		i3d::RGB* GetPalette() { return m_pal; }
		/** Get palette size */
		size_t GetPaletteSize() { return m_pal_size; }

		//////////////////////// FOLLOWING SHOULD BE DEFINED IN DERIVED CLASSES ////////////////

		/** The pure virtual function intending for overlaying data with given buffer */
		virtual void Overlay (size_t buf_id, Buffer2d<i3d::RGB> &buf) const = 0;

		/** Change position for af or 3d view. These values should be remembered and used during LoadOverlay. */
		virtual bool SetPosition (bool af, const wxString &pos, const wxRect *rect) = 0;
		/** Load 3d/af overlay. The overlayed and overlaying images must have the same size, offset and resolution.
		This can be checked using the 3 last parameters that contain the dimensions of the underlying image. */
		virtual bool LoadOverlay (bool af, const i3d::Vector3d<size_t> &sz, const i3d::Offset &ofs, 
			const i3d::Resolution &res) = 0;

		/** Return true if the af/3d overlay image is empty */
		virtual bool IsEmpty (bool af, size_t time) const = 0;

		/** Create/allocate buffer for given window type and return its ID */
		virtual size_t Buffer_Create (v4d_OverlayWindowType win_type) = 0;
		/** Set buffer dimensions */
		virtual void Buffer_SetDim (size_t buf_id, const i3d::Vector3d<size_t> &sz, 
			const i3d::Offset &ofs, const i3d::Vector3d<float> &res) = 0;
		/** Render buffer image */
		virtual void Buffer_Render (size_t buf_id, size_t time) = 0;
		/** Return pointer to the loaded af or 3d overlay image */
		virtual const i3d::LabeledImage3d<LABEL, VOXEL> *GetLabeledImage (bool af, bool boundary, size_t time) const = 0;
		/** Return the buffer image value at given position. */
		virtual LABEL Buffer_GetValue(size_t buf_id, const i3d::Vector3d<size_t> &pos) const = 0;
	};
}

#endif
