/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * roi.h
 *
 * templates for rectangular regions
 *
 */

#ifndef V4D_TYPES_ROI_H
#define V4D_TYPES_ROI_H

#include "vector2d.h"

namespace v4d
{
    /** Volume of interest -- area in any image */
    template <class T> 
    struct ROI 
    {
        /** origin of volume of interest (ROI) */
        Vector2d<T> offset; 
        /** size of ROI */
        Vector2d<T> size;  

        /** constructor */
        ROI(T off_x = 0, T off_y = 0,
            T sz_x = 0, T sz_y = 0)
            : offset(off_x, off_y), size(sz_x, sz_y) 
        {}

        /** constructor */
        ROI(const Vector2d<T> &o, const Vector2d<T> &s)
            : offset(o), size(s) {}

        /** test whether the ROI is empty. i.e. size.x == size.y == size.z == 0 */
        bool is_empty(void) {return size.x == 0 && size.y == 0; };

        Vector2d<T> GetCenter() const {
            return offset + size/T(2);
        }

        Vector2d<T> TopLeft() const { 
            return Vector2d<T>(std::min<T>(offset.x, offset.x + size.x),
                std::min<T>(offset.y, offset.y + size.y));
        }

        Vector2d<T> TopRight() const { 
            return Vector2d<T>(std::max<T>(offset.x, offset.x + size.x),
                std::min<T>(offset.y, offset.y + size.y));
        }

        Vector2d<T> BottomLeft() const { 
            return Vector2d<T>(std::min<T>(offset.x, offset.x + size.x),
                std::max<T>(offset.y, offset.y + size.y));
        }

        Vector2d<T> BottomRight() const { 
            return Vector2d<T>(std::max<T>(offset.x, offset.x + size.x),
                std::max<T>(offset.y, offset.y + size.y));
        }

        /// Union of two ROIs
        ROI &operator+= (const ROI &v) {
            size   = max(offset + size, v.offset + v.size);
            offset = min(offset, v.offset);
            size  -= offset;
            return *this;
        }

        /// Intersection of two ROIs
        ROI &operator*= (const ROI &v) {
            size   = min(offset + size, v.offset + v.size);
            offset = max(offset, v.offset);
            if (size.x < offset.x || size.y < offset.y )
                size = 0;
            else
                size -= offset;
            return *this;
        }
    };

    template <class T>
    inline bool operator== (const ROI<T> &v1, const ROI<T> &v2) {
        return (v1.offset.x == v2.offset.x &&
            v1.offset.y == v2.offset.y &&
            v1.size.x == v2.size.x &&
            v1.size.y == v2.size.y );
    }

    template <class T>
    inline bool operator!= (const ROI<T> &v1, const ROI<T> &v2) {
        return (v1.offset.x != v2.offset.x ||
            v1.offset.y != v2.offset.y ||
            v1.size.x != v2.size.x ||
            v1.size.y != v2.size.y );
    }

    template <class T>
    inline ROI<T> operator* (const ROI<T> &v1, const ROI<T> &v2) {
        ROI<T> v(v1);
        v *= v2;
        return v;
    }

    template <class T>
    inline ROI<T> operator+ (const ROI<T> &v1, const ROI<T> &v2) {
        ROI<T> v(v1);
        v += v2;
        return v;
    }

    template <class T> void EnsurePositiveSize(ROI<T> &roi)
    {
        roi.offset = roi.TopLeft();
        roi.size = abs(roi.size);
    }

    template <class T>
    inline bool CloseToBorder(ROI<T> roi, const Vector2d<T> &p, const Vector2d<T> &epsilon) {
        EnsurePositiveSize(roi);
        return ((std::abs(static_cast<int>(p.x - roi.offset.x)) <= epsilon.x && roi.offset.y <= p.y && roi.offset.y + roi.size.y >= p.y) ||
            (std::abs(static_cast<int>(p.x - roi.offset.x - roi.size.x)) <= epsilon.x && roi.offset.y <= p.y && roi.offset.y + roi.size.y >= p.y) ||
            (std::abs(static_cast<int>(p.y - roi.offset.y)) <= epsilon.y && roi.offset.x <= p.x && roi.offset.x + roi.size.x >= p.x) ||
            (std::abs(static_cast<int>(p.y - roi.offset.y - roi.size.y)) <= epsilon.y && roi.offset.x <= p.x && roi.offset.x + roi.size.x >= p.x));
    }

    template <class T>
    inline bool CloseToTopLeftCorner(ROI<T> roi, const Vector2d<T> &p, const Vector2d<T> &epsilon) 
    {
        Vector2d<T> c = roi.TopLeft();
        return std::abs(static_cast<int>(p.x - c.x)) <= epsilon.x && std::abs(static_cast<int>(p.y - c.y)) <= epsilon.y;
    }

    template <class T>
    inline bool CloseToTopRightCorner(ROI<T> roi, const Vector2d<T> &p, const Vector2d<T> &epsilon) 
    {
        Vector2d<T> c = roi.TopRight();
        return std::abs(static_cast<int>(p.x - c.x)) <= epsilon.x && std::abs(static_cast<int>(p.y - c.y)) <= epsilon.y;
    }

    template <class T>
    inline bool CloseToBottomLeftCorner(ROI<T> roi, const Vector2d<T> &p, const Vector2d<T> &epsilon) 
    {
        Vector2d<T> c = roi.BottomLeft();
        return std::abs(static_cast<int>(p.x - c.x)) <= epsilon.x && std::abs(static_cast<int>(p.y - c.y)) <= epsilon.y;
    }

    template <class T>
    inline bool CloseToBottomRightCorner(ROI<T> roi, const Vector2d<T> &p, const Vector2d<T> &epsilon) 
    {
        Vector2d<T> c = roi.BottomRight();
        return std::abs(static_cast<int>(p.x - c.x)) <= epsilon.x && std::abs(static_cast<int>(p.y - c.y)) <= epsilon.y;
    }

    template <class T>
    inline bool CloseToCorner(ROI<T> roi, const Vector2d<T> &p, const Vector2d<T> &epsilon) {
        return ((std::abs(static_cast<int>(p.x - roi.offset.x)) <= epsilon.x && std::abs(static_cast<int>(p.y - roi.offset.y)) <= epsilon.y) ||
            (std::abs(static_cast<int>(p.x - roi.offset.x - roi.size.x)) <= epsilon.x && std::abs(static_cast<int>(p.y - roi.offset.y - roi.size.y)) <= epsilon.y) ||
            (std::abs(static_cast<int>(p.x - roi.offset.x)) <= epsilon.x && std::abs(static_cast<int>(p.y - roi.offset.y - roi.size.y)) <= epsilon.y) ||
            (std::abs(static_cast<int>(p.x - roi.offset.x - roi.size.x)) <= epsilon.x && std::abs(static_cast<int>(p.y - roi.offset.y)) <= epsilon.y));
    }
}

#endif
