/*
 * i4dcore library - 3D+t image representation library 
 *
 * Copyright (C) 2005-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * v4d_imageiterator.h
 *
 * A class for browsing through (any general; however, images in this context) sequence.
 *
 * Vladimir Ulman (ulman@fi.muni.cz) 2015
 */

#ifndef __IMAGEITERATOR
#define __IMAGEITERATOR


#include <string>
#include <i3d/basic.h>


namespace i4d
{
	/** \addtogroup tlimain
	 *
	 * \author Vladimir Ulman
	 * \date 2015
	 */
	// @{
	/**
	 * Class for advanced browsing through any continous sequence of images.
	 * In fact, it works with any such sequence of numbers.
	 *
	 * Main usage:
	 * There are some Get/Set functions to provide data for GUI stuff.
	 * One may/need change iteration range/span, including stepping, direction,
	 * and sweeping mode. Note that some operations (changing the range/span
	 * or reseting sweeping mode) may result in an update of the current index
	 * (so that, e.g., it is again within the iteration range). Most operations,
	 * for which it makes sense, set a flag whether the change has occurred.
	 * The status can be read with IndexWasUpdated().
	 *
	 * Current index is retrieved with GetCurrentIndex(). One then
	 * iteratively calls the AdvanceIndex() until it returns with
	 * exactly the same index that was before it was called (which indicates
	 * no advancement can be done); or until IndexWasUpdated() was false.
	 * Iterations may be initiated with ResetForForwardSweep() and ResetForBackwardSweep(),
	 * which also returns the new reseted legal index position.
	 *
	 * If single step in forward or backward direction is required, regardless
	 * of the current sweeping direction and mode, one may need to use
	 * the AdvanceIndexForward() and AdvanceIndexBackward(), respectively.
	 *
	 * At any index position, a specific filename can be derived/obtained
	 * with GetCurrentName() according to the template, which can be altered
	 * anytime with SetTemplateName("name"). Note that this "name" has to
	 * obtain a sequence of 'X's where the sequence number will be placed.
	 */
	class ImageIterator {
	public:
		/**
		 * Supply the span of the interval.
		 *
		 * \param[in] fIndex		the index of the first image
		 * \param[in] lIndex		the index of the last image
		 */
		ImageIterator(const int fIndex=0,const int lIndex=0) :
			firstIndex(fIndex), lastIndex(lIndex),
			firstRangeIndex(fIndex), lastRangeIndex(lIndex), stepRangeIndex(1),
			templateName("imageXXX.tif"), atIndex(fIndex), indexUpdated(false),
			loops(false), oscillations(false), direction(1), FPS(1)
		{
			if (lIndex < fIndex)
				throw i3d::InternalException("Last index can't be smaller than the first index.");
		}

		///copy ctor
		ImageIterator(const ImageIterator& ii)
		{
			this->firstIndex=ii.GetFirstIndex();
			this->lastIndex=ii.GetLastIndex();
			this->firstRangeIndex=ii.GetFirstRangeIndex();
			this->lastRangeIndex=ii.GetLastRangeIndex();
			this->stepRangeIndex=ii.GetStepRangeIndex();
			ii.GetTemplateName(this->templateName);
			this->atIndex=ii.GetCurrentIndexConst();
			this->indexUpdated=ii.IndexWasUpdated();
			this->loops=ii.IsInLoopMode();
			this->oscillations=ii.IsInOscillationMode();
			this->direction=ii.GetCurrentDirection();
			this->FPS=ii.GetFPS();
		}

		///get functions of the whole sequence 
		int GetFirstIndex(void) const
		{ return(firstIndex); }
		int GetLastIndex(void) const
		{ return(lastIndex); }

		///(re)set function of the whole sequence span; range view is intersected with the new setting
		void SetFirstLastIndex(const int fIndex=0,const int lIndex=0)
		{
			firstIndex=fIndex;
			lastIndex=lIndex;
			if (lIndex < fIndex)
				throw i3d::InternalException("Last index can't be smaller than the first index.");

			if (firstRangeIndex < firstIndex) firstRangeIndex=firstIndex;
			if (lastRangeIndex > lastIndex) lastRangeIndex=lastIndex;

			//sets the 'modified' flag for me...
			GetCurrentIndex();
		}


		///get/set functions for the range view 
		int GetFirstRangeIndex(void) const
		{
			return(firstRangeIndex);
		}
		void SetFirstRangeIndex(const int fIndex)
		{
			if ((fIndex < firstIndex) || (fIndex > lastRangeIndex))
				throw i3d::InternalException("Incorrect index given.");
			firstRangeIndex=fIndex;

			//sets the 'modified' flag for me...
			GetCurrentIndex();
		}

		int GetLastRangeIndex(void) const
		{
			return(lastRangeIndex);
		}
		void SetLastRangeIndex(const int lIndex)
		{
			if ((lIndex < firstRangeIndex) || (lIndex > lastIndex))
				throw i3d::InternalException("Incorrect index given.");
			lastRangeIndex=lIndex;

			//sets the 'modified' flag for me...
			GetCurrentIndex();
		}

		void SetFirstLastRangeIndex(const int fIndex,const int lIndex)
		{
			if ((fIndex < firstIndex) || (lIndex > lastIndex))
				throw i3d::InternalException("Incorrect index given.");
			firstRangeIndex=fIndex;
			lastRangeIndex=lIndex;

			//sets the 'modified' flag for me...
			GetCurrentIndex();
		}

		///get/set functions for the range stepping
		int GetStepRangeIndex(void) const
		{
			return(stepRangeIndex);
		}
		void SetStepRangeIndex(const int step)
		{
			if ((step < 1) || (step > this->GetRangeLength()))
				throw i3d::InternalException("Incorrect step size given.");
			stepRangeIndex=step;
		}


		///gets the length of the current range
		int GetRangeLength(void) const
		{ return (lastRangeIndex - firstRangeIndex +1); }

		/**
		 * gets the number of tics in the current range
		 * given the current position within
		 */
		int GetRangeTics(void) const
		{
			//number of steps to get to the beginning
			int steps=(atIndex - firstRangeIndex) / stepRangeIndex;
			//the current position
			steps++;
			//number of steps to get to the end
			steps+=(lastRangeIndex - atIndex) / stepRangeIndex;

			return (steps);
		}


		///returns current index, which may get changed if range interval has changed
		int GetCurrentIndex(void)
		{
			indexUpdated=false;

			if (atIndex < firstRangeIndex)
				{ atIndex=firstRangeIndex; indexUpdated=true; }
			if (atIndex > lastRangeIndex)
				{ atIndex=lastRangeIndex; indexUpdated=true; }

			return (atIndex);
		}

		/**
		 * returns the current index without changing anything;
		 * do not use unless you know what you are doing...
		 */
		int GetCurrentIndexConst(void) const
		{ return (atIndex); }


		/**
		 * advances \e this->atIndex by \e this->stepRangeIndex and returns the new index;
		 * advances in the \e this->direction
		 *
		 * if looping is enabled, it loops the current index when required
		 *
		 * \note: Stop calling this function if AdvanceIndex() is not changing
		 * the current index, i.e., no progress can be done.
		 */
		int AdvanceIndex(void)
		{
			//GetCurrentIndex() may possibly update the newIndex (and the indexUpdated flag)
			int newIndex=GetCurrentIndex() + direction*stepRangeIndex;
			
			//are we (positively) too far and should we manage it?
			if ((newIndex > lastRangeIndex) && (loops))
			{
				//how to manage it?
				if (oscillations)
				{
					direction*=-1;
					newIndex=atIndex + direction*stepRangeIndex;
				}
				else newIndex=ResetForForwardSweep(true);
			}
			else
			//are we (negatively) too far and should we manage it?
			if ((newIndex < firstRangeIndex) && (loops))
			{
				if (oscillations)
				{
					direction*=-1;
					newIndex=atIndex + direction*stepRangeIndex;
				}
				else newIndex=ResetForBackwardSweep(true);
			}

			//now, check again that the possibly corrected newIndex is okay
			if ((newIndex >= firstRangeIndex) && (newIndex <= lastRangeIndex))
			{
				if (newIndex != atIndex) indexUpdated=true; //the flag...
				atIndex=newIndex;
			}

			return(atIndex);
		}

		/**
		 * advances this->atIndex by this->stepRangeIndex in the positive/forward
		 * direction; it won't change the current index if the advancement
		 * is not possible (unless GetCurrentIndex() won't change it)
		 *
		 * if looping is enabled, it loops the current index when required
		 *
		 * \note: Stop calling this function if it is not changing
		 * the current index, i.e., no progress can be done.
		 */
		int AdvanceIndexForward(void)
		{
			//GetCurrentIndex() may possibly update the newIndex (and the indexUpdated flag)
			int newIndex=GetCurrentIndex() + stepRangeIndex;
			if ((newIndex > lastRangeIndex) && (loops))
				newIndex=ResetForForwardSweep(true);

			//now, check again that the possibly corrected newIndex is okay
			if ((newIndex >= firstRangeIndex) && (newIndex <= lastRangeIndex))
			{
				//this 'if' should be always true provided stepRangeIndex is never zero...
				if (newIndex != atIndex) indexUpdated=true; //the flag...
				atIndex=newIndex;
			}

			return(atIndex);
		}

		/**
		 * advances this->atIndex by this->stepRangeIndex in the negative/backward
		 * direction; it won't change the current index if the advancement
		 * is not possible (unless GetCurrentIndex() won't change it)
		 *
		 * if looping is enabled, it loops the current index when required
		 *
		 * \note: Stop calling this function if it is not changing
		 * the current index, i.e., no progress can be done.
		 */
		int AdvanceIndexBackward(void)
		{
			//GetCurrentIndex() may possibly update the newIndex (and the indexUpdated flag)
			int newIndex=GetCurrentIndex() - stepRangeIndex;
			if ((newIndex < firstRangeIndex) && (loops))
				newIndex=ResetForBackwardSweep(true);

			//now, check again that the possibly corrected newIndex is okay
			if ((newIndex >= firstRangeIndex) && (newIndex <= lastRangeIndex))
			{
				//this 'if' should be always true provided stepRangeIndex is never zero...
				if (newIndex != atIndex) indexUpdated=true; //the flag...
				atIndex=newIndex;
			}

			return(atIndex);
		}


		/**
		 * reset \e this->direction for the forward direction;
		 * if \e fromEdge is true, the \e this->atIndex is moved somewhere
		 * near the \e this->firstRangeIndex such that \e this->AdvanceIndex()
		 * will eventually step at the current index
		 */
		int ResetForForwardSweep(const bool fromEdge=false)
		{
			direction=+1;
			
			const int steps=(GetCurrentIndex() - firstRangeIndex) / stepRangeIndex;
			if ((fromEdge) && (steps*stepRangeIndex))
			{
				atIndex-=steps*stepRangeIndex;
				indexUpdated=true;
			}

			return (atIndex);
		}

		/**
		 * reset \e this->direction for the backward direction;
		 * if \e fromEdge is true, the \e this->atIndex is moved somewhere
		 * near the \e this->lastRangeIndex such that \e this->AdvanceIndex()
		 * will eventually step at the current index
		 */
		int ResetForBackwardSweep(const bool fromEdge=false)
		{
			direction=-1;
			
			const int steps=(lastRangeIndex - GetCurrentIndex()) / stepRangeIndex;
			if ((fromEdge) && (steps*stepRangeIndex))
			{
				atIndex+=steps*stepRangeIndex;
				indexUpdated=true;
			}

			return (atIndex);
		}

		/**
		 * \return Returns the current direction:
		 * +1 = positive/forward, -1 = negative/backward
		 */
		int GetCurrentDirection(void) const
		{ return (direction); }


		///non-standard regimes
		bool IsInLoopMode(void) const
		{ return (loops); }
		void SetLoopMode(const bool mode)
		{ loops=mode; }

		bool IsInOscillationMode(void) const
		{ return (oscillations); }
		void SetOscillationMode(const bool mode)
		{ oscillations=mode; }


		/**
		 * grabs the this->templateName and fills it according to the this->GetCurrentIndex() 
		 *
		 * If the this->templateName does not contain enough 'X's to hold the
		 * number, the number of truncated adequately (overwriting the previously
		 * generated images in this way). If the number of 'X's is far enough,
		 * the sequence number is padded with zeros from the left.
		 */
		void GetCurrentName(std::string& name) const
		{
			//the first occurence
			int fX=templateName.find('X');
			int lX=fX+1;
			while ((lX < (signed)templateName.length()) && (templateName[lX] == 'X')) ++lX;
			if (lX == (signed)templateName.length())
				throw i3d::InternalException("Missing suffix in the template name.");

			//lX now points just after the last consecutive 'X'...

			//construct the file name
			name=templateName.substr(0,fX);
			//name.append(lX-fX,'0'); //atIndex at lX-fX characters
			int wdth=lX-fX;
			int divisor=1;
			for (int i=1; i < wdth; ++i, divisor*=10);
			while (wdth > 0)
			{
				name.append(1, ((atIndex/divisor)%10) +48 );
				divisor/=10;
				--wdth;
			}
			name.append(templateName.substr(lX));
		}

		///gets/sets the new templateName
		void GetTemplateName(std::string& name) const
		{ name=templateName; }
		const std::string& GetTemplateName(void) const
		{ return (templateName); }
		void SetTemplateName(const std::string& name)
		{
			if (name.find('X') == std::string::npos)
				throw i3d::InternalException("Missing the number-replacement character 'X'.");

			templateName=name;
		}

		///gets/sets the FPS
		int GetFPS(void) const
		{ return (FPS); }
		void SetFPS(const int fps)
		{
			if (fps < 1)
				throw i3d::InternalException("Incorrect FPS requested.");
			FPS=fps;
		}

		/**
		 * tells the state of the \e this->indexUpdated flag,
		 * i.e., whether the \e this->atIndex was changed
		 * after the last operation
		 */
		bool IndexWasUpdated(void) const
		{ return (indexUpdated); }

	private:
		///the sequence params
		int firstIndex;
		int lastIndex;

		///current range params
		int firstRangeIndex;
		int lastRangeIndex;
		int stepRangeIndex;

		///current filename template
		std::string templateName;

		///current index
		int atIndex;

		///flags whether \e this->atIndex was changed with the last operation
		bool indexUpdated;

		///shall we be looping?
		bool loops;
		/**
		 * shall we be swinging back and forth (oscillation mode) = true,
		 * or shall we start over again = false
		 */
		bool oscillations;
		///in what directions are we advancing: forward=+1, backward=-1
		int direction;

		///FPS (frames per second) of the browsing
		int FPS;
	};

	// @}
}

#endif
