/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * vector2d.h
 *
 * 2D vector representation
 *
 * Petr Matula (pem@fi.muni.cz) 2006
 */

#ifndef V4D_TYPES_VECTOR2D_H
#define V4D_TYPES_VECTOR2D_H

#include <i3d/vector3d.h>

namespace v4d
{
    /** General structure for 2D coordinations manipulation */
    template <class T> struct Vector2d 
    {
    public:
        /// the particular coordinates
        T x, y;

        /// constructor
        Vector2d(T val = 0): x(val),y(val) {};

        /// constructor
        Vector2d(T x_ , T y_)
            : x(x_), y(y_) {}

        /// conversion constructor
        template <class T2> Vector2d(const Vector2d<T2>& v)
            : x(static_cast<T>(v.x)),y(static_cast<T>(v.y)) {}

        /// conversion constructor from 3d vectors
        template <class T2> Vector2d(const i3d::Vector3d<T2>& v)
            : x(static_cast<T>(v.x)), y(static_cast<T>(v.y)){}

        /// left addition  
        Vector2d<T> &operator+= (const Vector2d<T>& v)
        { x += v.x; y += v.y; return *this; };

        /// left subtraction
        Vector2d<T> &operator-= (const Vector2d<T>& v)
        { x -= v.x; y -= v.y; return *this; };

        /// left multiplication by another vector
        Vector2d<T> &operator*= (const Vector2d<T>& v)
        { x *= v.x; y *= v.y; return *this; };

        /// left division by another vector
        Vector2d<T> &operator/= (const Vector2d<T>& v)
        { x /= v.x; y /= v.y; return *this; };

        /// left multiplication by scalar
        Vector2d<T> &operator*= (const T& s) { x *= s; y *= s; return *this;};

        /// left division by scalar
        Vector2d<T> &operator/= (const T& s) { x /= s; y /= s; return *this;};

        /// unary minus = negate
        Vector2d<T> operator-() const { return Vector2d<T>(-x,-y); }

        /// is the vector in the interval [a,b)?
        bool InCO(const Vector2d<T>& a, const Vector2d<T>& b) const {
            return (a.x <= x && a.y <= y && b.x >  x && b.y >  y);
        }

        /// is the vector in the interval (a,b)?
        bool InOO(const Vector2d<T>& a, const Vector2d<T>& b) const {
            return (a.x < x && a.y < y && b.x >  x && b.y >  y);
        }

        /// is the vector in the interval (a,b]?
        bool InOC(const Vector2d<T>& a, const Vector2d<T>& b) const {
            return (a.x < x && a.y < y && b.x >=  x && b.y >=  y);
        }

        /// is the vector in the interval [a,b]?
        bool InCC(const Vector2d<T>& a, const Vector2d<T>& b) const {
            return (a.x <= x && a.y <= y && b.x >=  x && b.y >=  y);
        }

        T operator [] (size_t i) { 
            wxASSERT(i<2);
            return  *(((T*) (this))+i);
        }
    };


    /// Euclidean norm:
    template <class T> T Norm(const Vector2d<T>& v) {return sqrt(v.x*v.x + v.y*v.y);}

    /// minimum
    template <class T> Vector2d<T> max(const Vector2d<T> &v1,
        const Vector2d<T> &v2) {
            return Vector2d<T>( std::max(v1.x,v2.x),
                std::max(v1.y,v2.y)); }

    /// maximum
    template <class T> Vector2d<T> min(const Vector2d<T> &v1,
        const Vector2d<T> &v2) {
            return Vector2d<T>(	std::min(v1.x,v2.x),
                std::min(v1.y,v2.y)); }

    // arithmetic operations with vectors

    template <class T> inline Vector2d<T> operator+ (const Vector2d<T>& v1,
        const Vector2d<T>& v2) {
            Vector2d<T> v = v1; v += v2; return v; }
    template <class T> inline Vector2d<T> operator- (const Vector2d<T>& v1,
        const Vector2d<T>& v2) {
            Vector2d<T> v = v1; v -= v2; return v; }
    template <class T> inline Vector2d<T> operator* (const Vector2d<T>& v1,
        const Vector2d<T>& v2) {
            Vector2d<T> v = v1; v *= v2; return v; }
    template <class T> inline Vector2d<T> operator/ (const Vector2d<T>& v1,
        const Vector2d<T>& v2) {
            Vector2d<T> v = v1; v /= v2; return v; }

    // arithmetic operations with scalars
    template <class T> inline Vector2d<T> operator+ (const Vector2d<T>& v, T s) {
        return Vector2d<T>(v.x + s, v.y + s); }
    template <class T> inline Vector2d<T> operator- (const Vector2d<T>& v, T s) {
        return Vector2d<T>(v.x - s, v.y - s); }
    template <class T> inline Vector2d<T> operator* (const Vector2d<T>& v, T s) {
        return Vector2d<T>(v.x * s, v.y * s); }
    template <class T> inline Vector2d<T> operator/ (const Vector2d<T>& v, T s) {
        return Vector2d<T>(v.x / s, v.y / s); }
    template <class T> inline Vector2d<T> operator+ (T s, const Vector2d<T>& v) {
        return Vector2d<T>(s + v.x, s + v.y); }
    template <class T> inline Vector2d<T> operator- (T s, const Vector2d<T>& v) {
        return Vector2d<T>(s - v.x, s - v.y); }
    template <class T> inline Vector2d<T> operator* (T s, const Vector2d<T>& v) {
        return Vector2d<T>(s * v.x, s * v.y); }
    template <class T> inline Vector2d<T> operator/ (T s, const Vector2d<T>& v) {
        return Vector2d<T>(s / v.x, s / v.y); }


    template <class T> inline Vector2d<T> abs(const Vector2d<T> &v) {
        return Vector2d<T>(std::abs(v.x), std::abs(v.y));
    }

    // comparations 
    template <class T> inline bool operator== (const Vector2d<T>& v1,
        const Vector2d<T>& v2) {
            return (v1.x == v2.x) && (v1.y == v2.y);
    }
    template <class T> inline bool operator!= (const Vector2d<T>& v1,
        const Vector2d<T>& v2) {
            return (v1.x != v2.x) || (v1.y != v2.y);
    }

    // Lexicografical ordering
    template <class T> inline bool operator> (const Vector2d<T>& v1,
        const Vector2d<T>& v2)
    {
        return ((v1.x > v2.x) ||
            (v1.x == v2.x && v1.y > v2.y));
    }

    template <class T> inline bool operator< (const Vector2d<T>& v1,
        const Vector2d<T>& v2)
    {
        return ((v1.x < v2.x) ||
            (v1.x == v2.x && v1.y < v2.y));
    }

    /// Element-by-element partial ordering:
    template <class T> inline bool le(const Vector2d<T>& v1, const Vector2d<T>& v2)
    {
        return (v1.x <= v2.x && v1.y <= v2.y);
    }
}

#endif
