/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * v4d_canvas3d.cpp
 *
 * Canvas with 3D drawing windows for presentation of 3D data
 *
 * Petr Matula (pem@fi.muni.cz) 2006
 */

#include "v4d_basic.h"
#include <wx/msgdlg.h>
#include <wx/image.h>
#include <wx/minifram.h>
#include "v4d_canvas3d.h"
#include "v4d_frame.h"
#include "dlg_set_range.h"
#include "wdr/info_wdr.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

namespace v4d 
{
    ///////////////////////////////////////////////////
    // local functions
    ///////////////////////////////////////////////////

    /** Helper function for optimal setting of the size of drawing
        windows. It sets the patameters \a v1 and \a v2 so that their sum
        is less or equal to the parameter \a w and simultaneously the
        parameters \a v1 and \a v2 are as clase as possible to \a r1,
        and. \a r2, respectively. \sa Viewer3dCanvas::RecomputeVisible() */
    static void AdjustSize(int *v1, int *v2, int r1, int r2, int w)
    {
        int w2 = w/2;

        if (r1 >= w2 && r2 >= w2)
            *v1 = *v2 = w2;
        else if (r1 + r2 <= w)
        {
            *v1 = r1;
            *v2 = r2;
        }
        else if (r1 > r2)
        {
            *v1 = w - r2;
            *v2 = r2;
        }
        else
        {
            *v2 = w - r1;
            *v1 = r1;
        }
    }

    /** Helper function for setting of optimal offset of visible part of
        the 3d data. It sets offset \a off so that the new visible part
        with width \a new_size was centered as best as possible with the
        old visible part starting at the offset \a old_off and having
        length \a old_size and simultaneosly the following condition is
        true: The new interval [\a off, \a off + \a new_size] lies inside
        the interval [\a real_off, \a real_off + \a real_size]. \sa
        Viewer3dCanvas::RecomputeVisible() */
    static void AdjustOffset(float *off, float new_size,
        float old_off, float old_size,
        float real_off, float real_size)
    {
        *off = (old_size - new_size) / 2.0f + old_off;
        if (*off + new_size > real_off + real_size)
            *off = real_off + real_size - new_size;
        if (*off < real_off)
            *off = real_off;
    }

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                         Viewer3dCanvas                      //
    //                                                             //
    /////////////////////////////////////////////////////////////////

    v4d_Viewer3dCanvas::v4d_Viewer3dCanvas(wxWindow *parent, wxWindowID id, 
        const wxPoint& pos, const wxSize& size)
        : wxPanel(parent, id, pos, size, 0),
        m_canvas_info_flag(false), m_show_xy(true), m_show_xz(true), m_show_yz(true), 
        m_what_xy(pc_Ortho), m_what_xz(pc_Ortho), m_what_yz(pc_Ortho), m_pointer(),
        m_volVis(NULL), m_info_frame(NULL)
    {
        m_xy = new v4d_DrWinXY(this, m_vv,m_ii);
        m_xz = new v4d_DrWinXZ(this, m_vv,m_ii);
        m_yz = new v4d_DrWinYZ(this, m_vv,m_ii);
		  // Instructs the timer to sent timing-events to this object
		  m_ii_timer.SetOwner(this);

        // Compute the border_size value
        int w, h;
        int w1, h1;
        m_xy->GetClientSize (&w1, &h1);
        m_xy->GetSize (&w, &h);
        m_border_size = w - w1 ;

        // Event handlers
        Bind(wxEVT_SIZE, &v4d_Viewer3dCanvas::OnSize, this);
        Bind(wxEVT_PAINT, &v4d_Viewer3dCanvas::OnPaint, this);
        Bind(v4d_EVT_MOUSE3D, &v4d_Viewer3dCanvas::OnMouse3d, this);
		  Bind(wxEVT_TIMER, &v4d_Viewer3dCanvas::OnSequenceTimerHandler, this);

        RecomputeVisible();
        Show(TRUE);
    }

    v4d_Viewer3dCanvas::~v4d_Viewer3dCanvas()
    {
        if (m_volVis)
        {
            delete m_volVis;
        }
    }

    wxSize v4d_Viewer3dCanvas::GetOptimalSize()
    {
        wxSize sz(0,0);
        wxSize s_xy = m_xy->GetOptimalSize();
        wxSize s_xz = m_xz->GetOptimalSize();
        wxSize s_yz = m_yz->GetOptimalSize();

        const int extra_space = 6;

        if (m_xy->IsShown() || m_yz->IsShown())  {  
            sz.y += s_xy.GetHeight() + extra_space; 
        }
        if (m_xz->IsShown()) {
            sz.y += s_xz.GetHeight() + extra_space;
        }
        if (m_xy->IsShown() || m_xz->IsShown())  {  
            sz.x += s_xy.GetWidth() + extra_space; 
        }
        if (m_yz->IsShown()) {
            sz.x += s_yz.GetWidth() + extra_space;
        }

        return sz;
    }

    wxString StrPT(v4d_ProjectionContent pc)
    {
        if (pc == pc_Af)
            return wxString("A");
        else
            return wxString("O");
    }

    void v4d_Viewer3dCanvas::OnPaint(wxPaintEvent& evt)
    {
        wxPaintDC dc(this);

        if (m_xy && m_xz && m_yz)
        {
            m_xy->Refresh();
            m_xz->Refresh();
            m_yz->Refresh();
        }
    }

    void v4d_Viewer3dCanvas::OnMouse3d(v4d_Mouse3dEvent& evt)
    {
        InfoFrame_UpdatePixelInfo();
    }

    void v4d_Viewer3dCanvas::OnSize(wxSizeEvent& evt)
    {
        RecomputeVisible();
        OnVis3dChange(pt_All);
        InfoFrame_Move();
    }

    void v4d_Viewer3dCanvas::ShowVolVis(const wxIconBundle &bundle)
    {
        bool init;

        //create the window if necessary
        if (!m_volVis)
        {
            m_volVis = new vol3D_VolVisFrame(NULL, wxID_ANY, "Volume visualization", wxDefaultPosition, wxSize(600, 600));
            init = m_volVis -> GetVolVisCanvas() -> Init();

            //cleanup if initialization failed
            if (!init)
            {
                if (m_volVis)
                {
                    delete m_volVis;
                    m_volVis = NULL;
                }

                return;
            }
        }

        m_volVis -> GetVolVisCanvas() -> Reset();
        m_volVis -> SetIcons(bundle);

        m_volVis -> Show();
        UpdateVolVis();
        UpdateIsosurface(); 
    }

    void v4d_Viewer3dCanvas::RefreshVolVisIcon(const wxIconBundle &bundle)
    {
        if (m_volVis)
        {
            m_volVis -> SetIcons(bundle);
        }
    }

    void v4d_Viewer3dCanvas::GetSettings(v4d_CanvasSettings &params) const
    {
        params.show_info  = m_canvas_info_flag;  
        params.show_xy    = m_show_xy;
        params.show_xz    = m_show_xz;
        params.show_yz    = m_show_yz;

        params.pc_xy      = m_what_xy;
        params.pc_xz      = m_what_xz;
        params.pc_yz      = m_what_yz;

        params.zoom_level = m_vv.zoom_level; 
        params.cut        = m_vv.cut; 

        params.orig_res = m_vv.orig_res;
        params.cut_pointer_style = m_pointer;

        for (size_t i = 0; i < 3; i++)
        {
            params.rev_axis[i] = m_vv.rev_axis[i];
        }

		params.projection_annotation_style = m_projection_annotation;
		params.scale_bar_style = m_scale_bar_style;
		params.time_bar_style = m_time_bar_style;
    }

    void v4d_Viewer3dCanvas::SetSettings(const v4d_CanvasSettings &params)
    {
        m_canvas_info_flag = params.show_info;
        m_show_xy   = params.show_xy;
        m_show_xz   = params.show_xz;
        m_show_yz   =	params.show_yz;

        m_what_xy   =	params.pc_xy;
        m_what_xz   =	params.pc_xz;
        m_what_yz   =	params.pc_yz;

        m_vv.zoom_level = params.zoom_level;
        m_vv.cut  =	params.cut;

        m_vv.orig_res = params.orig_res;
        m_pointer = params.cut_pointer_style;

        // set cut to the center of bbox if it is out of bbox;
        if (!m_vv.cut.InOO(m_vv.bbox.offset, m_vv.bbox.offset + m_vv.bbox.size))
        {
            m_vv.cut = m_vv.bbox.offset + m_vv.bbox.size / 2.0f;
        }

        for (size_t i = 0; i < 3; i++)
        {
            m_vv.rev_axis[i] = params.rev_axis[i];
        }

		m_projection_annotation = params.projection_annotation_style;
		m_scale_bar_style = params.scale_bar_style;
		m_time_bar_style = params.time_bar_style;

        RecomputeVisible();
        OnVis3dChange(pt_All);
    }

	 void v4d_Viewer3dCanvas::OnSequenceMenuHandler(wxCommandEvent& evt,wxString& msg)
	 {
		try {
			switch (evt.GetId())
			{
			case SI_SM_PLAY_FORWARD:
				m_ii.ResetForForwardSweep();
				if (m_ii.IndexWasUpdated()) OnTimeChange();
				if (!m_ii_timer.IsRunning())
					m_ii_timer.Start(1000/m_ii.GetFPS(),wxTIMER_CONTINUOUS);
				msg << "Forward auto browsing.";
				break;
			case SI_SM_PLAY_BACKWARD:
				m_ii.ResetForBackwardSweep();
				if (m_ii.IndexWasUpdated()) OnTimeChange();
				if (!m_ii_timer.IsRunning())
					m_ii_timer.Start(1000/m_ii.GetFPS(),wxTIMER_CONTINUOUS);
				msg << "Backward auto browsing.";
				break;
			case SI_SM_PLAY_STOPNOW:
				m_ii_timer.Stop();
				msg << "Stopped auto browsing.";
				break;
			case SI_SM_SETRANGE:
				{
				//local copy of the current range
				long minVal=m_ii.GetFirstRangeIndex();
				long maxVal=m_ii.GetLastRangeIndex();
				//pop-up a range dialog for integers
				bool OKbutton=GetLongRangeFromUser(this,
					m_ii.GetFirstIndex(),m_ii.GetLastIndex(),
					minVal,maxVal);
				if (OKbutton)
				{
					//new range is possibly set
					m_ii.SetFirstLastRangeIndex(minVal,maxVal);
					if (m_ii.IndexWasUpdated()) OnTimeChange();
				}
				}
				break;
			case SI_SM_FPS_INCREASE:
				m_ii.SetFPS(m_ii.GetFPS()+1);
				if (m_ii_timer.IsRunning()) //restart the timer with the new delays
					m_ii_timer.Start(1000/m_ii.GetFPS(),wxTIMER_CONTINUOUS);
				msg << "FPS set to " << m_ii.GetFPS() << ".";
				break;
			case SI_SM_FPS_DECREASE:
				m_ii.SetFPS(m_ii.GetFPS()-1);
				if (m_ii_timer.IsRunning()) //restart the timer with the new delays
					m_ii_timer.Start(1000/m_ii.GetFPS(),wxTIMER_CONTINUOUS);
				msg << "FPS set to " << m_ii.GetFPS() << ".";
				break;
			case SI_SM_STEP_INCREASE:
				m_ii.SetStepRangeIndex(m_ii.GetStepRangeIndex()+1);
				msg << "Browsing by " << m_ii.GetStepRangeIndex() << " frames.";
				break;
			case SI_SM_STEP_DECREASE:
				m_ii.SetStepRangeIndex(m_ii.GetStepRangeIndex()-1);
				msg << "Browsing by " << m_ii.GetStepRangeIndex() << " frames.";
				break;
			case SI_SM_PLAY_LOOP:
				m_ii.SetLoopMode( !m_ii.IsInLoopMode() );
				break;
			case SI_SM_PLAY_OSCILLATION:
				m_ii.SetOscillationMode( !m_ii.IsInOscillationMode() );
				break;
			}
		}
		catch (...)
		{  /* just do nothing (and don't break the program) */
			msg << "No change.";
	 	}
	 }

	 void v4d_Viewer3dCanvas::OnSequenceTimerHandler(wxTimerEvent& evt)
	 {
		//can we move further?
		m_ii.AdvanceIndex();
		if (m_ii.IndexWasUpdated())
			//yes, let's re-draw
			OnTimeChange();
		else
			//no, makes no sense to continue with the timer
			m_ii_timer.Stop();
	 }

    void v4d_Viewer3dCanvas::CenterView()
    {
        this->m_xy->CenterView(m_vv.cut.x, m_vv.cut.y);
        this->m_xz->CenterView(m_vv.cut.x, m_vv.cut.z);
    }

    void v4d_Viewer3dCanvas::CenterViewIfOut()
    {
        if (!m_vv.cut.InCC(m_vv.vis3d.offset, m_vv.vis3d.offset + m_vv.vis3d.size))
        {
            CenterView();
        }
    }

    void v4d_Viewer3dCanvas::ChangeCut(const i3d::Vector3d<float> &new_cut)
    {
        if (new_cut.InOO(m_vv.bbox.offset, m_vv.bbox.offset + m_vv.bbox.size))
        {
            m_vv.cut = new_cut;

            OnCutChange(pt_All);
        }

    }

    //////////////////////////// Drawing, zoooming, ... ////////////////////////

    void v4d_Viewer3dCanvas::RecomputeVisible()
    {
        if (m_vv.zoom_level == 0)
            m_vv.res = m_vv.orig_res;
        else if (m_vv.zoom_level > 0)
            m_vv.res = m_vv.orig_res * float(m_vv.zoom_level + 1);
        else
            m_vv.res = m_vv.orig_res / float(-m_vv.zoom_level + 1);

        if (m_show_xy || m_show_yz || m_show_xz)
        {
            int w, h;

            GetClientSize(&w, &h);

            if (w == 0 && h == 0)
                return;
            // Leave 1 blank pixel on each side
            w -= 2;
            h -= 2;

            // Include the size of the border of the windows
            if (m_show_xy || m_show_xz || m_show_yz)
            {
                w -= m_border_size;
                h -= m_border_size;
            }
            if ((m_show_xy && m_show_yz) || (m_show_yz && m_show_xz))
                w -= m_border_size;
            if ((m_show_xy && m_show_xz) || (m_show_yz && m_show_xz))
                h -= m_border_size;

            // Don't allow width or height to be less than zero
            w = (w < 0) ? 0 : w;
            h = (h < 0) ? 0 : h;

            i3d::VOI<i3d::MICRONS> old = m_vv.vis3d;
            i3d::Vector3d<int> bbsize = m_vv.bbox.size * m_vv.res;
            //bbsize += i3d::Vector3d<float> (1.0f);

            int w1,w2,h1,h2,z;
            w1 = std::min (w, bbsize.x);
            h1 = std::min (h, bbsize.y);
            w2 = std::min (w, bbsize.z);
            h2 = std::min (h, bbsize.z);

            if ((m_show_yz && m_show_xy) || (m_show_yz && m_show_xz))
                AdjustSize(&w1, &w2, bbsize.x, bbsize.z, w);

            if ((m_show_xy && m_show_xz) || (m_show_yz && m_show_xz))
                AdjustSize(&h1, &h2, bbsize.y, bbsize.z, h);

            // Don't allow width or height to be less than one
            w1 = (w1 < 1) ? 1 : w1;
            h1 = (h1 < 1) ? 1 : h1;
            w2 = (w2 < 1) ? 1 : w2;
            h2 = (h2 < 1) ? 1 : h2;

            if (m_show_xz && m_show_yz)
                z = w2 = h2 = std::min(w2,h2);
            else if (m_show_yz)
                z = w2;
            else
                z = h2;

            // Not so sure about the -1, but this way it works
            m_vv.vis3d.size.x = float(w1) / m_vv.res.x;
            m_vv.vis3d.size.y = float(h1) / m_vv.res.y;
            m_vv.vis3d.size.z = float(z)  / m_vv.res.z;

            //	std::cout << "**************BEGIN***************\n";
            //	std::cout << "m_vv.zoom_level: " << m_vv.zoom_level << "\n";
            //	std::cout << "*******XXXX*******\n";
            AdjustOffset(&m_vv.vis3d.offset.x, m_vv.vis3d.size.x,
                old.offset.x, old.size.x,
                m_vv.bbox.offset.x, m_vv.bbox.size.x);
            //	std::cout << "*******YYYY*******\n";
            AdjustOffset(&m_vv.vis3d.offset.y, m_vv.vis3d.size.y,
                old.offset.y,old.size.y,
                m_vv.bbox.offset.y, m_vv.bbox.size.y);
            //	std::cout << "*******ZZZZ*******\n";
            AdjustOffset(&m_vv.vis3d.offset.z, m_vv.vis3d.size.z,
                old.offset.z, old.size.z,
                m_vv.bbox.offset.z, m_vv.bbox.size.z);

            w1 += m_border_size;
            h1 += m_border_size;
            w2 += m_border_size;
            h2 += m_border_size;

            m_xy->SetSize(1,1,w1,h1);
            m_xz->SetSize(1,(m_show_xy || m_show_yz) ? h1 + 1: 1, w1, h2);
            m_yz->SetSize((m_show_xy || m_show_xz) ? w1 + 1 : 1, 1, w2, h1);

            // this constant can be moved elsewhere
            const float MSTEP_PERCENTAGE = 0.2f;
            m_vv.mstep = m_vv.vis3d.size * MSTEP_PERCENTAGE;
        }

        m_xy->Show(m_show_xy);
        m_yz->Show(m_show_yz);
        m_xz->Show(m_show_xz);
    }

    v4d_ProjectionContent v4d_Viewer3dCanvas::GetContent(v4d_ProjectionType pt) const
    {
        switch (pt)
        {
        case pt_XY:
            return m_what_xy;
        case pt_XZ:
            return m_what_xz;
        case pt_YZ:
            return m_what_yz;
        default:
            break;
        }
        return m_what_xy == pc_Af && m_what_xz == pc_Af && m_what_yz == pc_Af ? pc_Af : pc_Ortho;
    }

    void v4d_Viewer3dCanvas::ChangeContent(v4d_ProjectionContent pc, v4d_ProjectionType pt)
    {
        if (pt & pt_XY)
            m_what_xy = pc;

        if (pt & pt_XZ)
            m_what_xz = pc;

        if (pt & pt_YZ)
            m_what_yz = pc;

        OnContentChange(pt);
    }

    void v4d_Viewer3dCanvas::ShowDrWin(v4d_ProjectionType pt, bool state)
    {
        if (pt & pt_XY)
            m_show_xy = state;

        if (pt & pt_XZ)
            m_show_xz = state;

        if (pt & pt_YZ)
            m_show_yz = state;

        RecomputeVisible();
        OnVis3dChange(pt_All);
        InfoFrame_Move();
    }

    void v4d_Viewer3dCanvas::ZoomIn()
    {
        ++m_vv.zoom_level;
        RecomputeVisible();
        OnVis3dChange(pt_All);
        InfoFrame_Move();
    }

    void v4d_Viewer3dCanvas::ZoomOut()
    {
        --m_vv.zoom_level;
        RecomputeVisible();
        OnVis3dChange(pt_All);
        InfoFrame_Move();
    }

    void v4d_Viewer3dCanvas::NormalZoom()
    {
        m_vv.zoom_level = 0;
        RecomputeVisible();
        OnVis3dChange(pt_All);
        InfoFrame_Move();
    }

    bool v4d_Viewer3dCanvas::IsDrWinShown(v4d_ProjectionType pt) const
    {
        bool visible = true;

        if (pt & pt_XY)
            visible = visible && m_show_xy;
        if (pt & pt_XZ)
            visible = visible && m_show_xz;
        if (pt & pt_YZ)
            visible = visible && m_show_yz;

        return visible;
    }

    bool v4d_Viewer3dCanvas::ToggleAxisReverse(int axis)
    {
        if (axis >= 0 && axis < 3)
        {
            m_vv.rev_axis[axis] = !m_vv.rev_axis[axis];
            ReverseAxis(axis);
            Refresh();

            InfoFrame_UpdatePixelInfo();

            wxUpdateUIEvent e(SI_REFRESH_ICON);
            AddPendingEvent(e);

            return m_vv.rev_axis[axis];
        }

        return false;
    }

    /////////////////////////////// Info frame ///////////////////////////////

    void v4d_Viewer3dCanvas::InfoFrame_Show(bool show)
    {
        if (show)
        {        
            if (m_info_frame == NULL)
            {
				InfoFrame_Create();
            }
            
            m_info_frame->Show(true);

            InfoFrame_Move();
            InfoFrame_UpdatePixelInfo();
            InfoFrame_UpdateImageInfo();
            InfoFrame_UpdateCutAndVisBoxInfo();
        }
        else
        {
            if (m_info_frame != NULL)
            {                
                m_info_frame->Show(false);                
            }
        }
    }

    void v4d_Viewer3dCanvas::InfoFrame_Move()
    {
        if (m_info_frame && m_info_frame->IsShown())
        {
            wxPoint dead_pos;

            if (!m_xy->IsShown() && !m_xz->IsShown() && m_yz->IsShown())
            {
                dead_pos = ClientToScreen(m_yz->GetPosition() + m_yz->GetSize());
            }
            else if (!m_xy->IsShown() && m_xz->IsShown() && !m_yz->IsShown())
            {
                dead_pos = ClientToScreen(m_xz->GetPosition() + m_xz->GetSize());
            }
            else
            {
                dead_pos = ClientToScreen(m_xy->GetPosition() + m_xy->GetSize());
            }

            m_info_frame->Move(dead_pos);
        }
    }

	void v4d_Viewer3dCanvas::InfoFrame_Create()
	{ 
		m_info_frame = new wxMiniFrame(this, wxID_ANY, "Info frame", wxDefaultPosition, wxDefaultSize, 0); 
		m_info_frame->SetMaxSize(wxSize(380,200));
		InfoWindowFunc(m_info_frame); 
	}
}
