/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * v4d_canvas3d.h
 *
 * Canvas with 3D drawing windows for presentation of 3D data
 *
 * Petr Matula (pem@fi.muni.cz) 2006
 */
#ifndef V4D_V4D_CANVAS3D_H
#define V4D_V4D_CANVAS3D_H

#include <wx/panel.h>
#include <wx/dcclient.h>
#include <wx/dc.h>
#include <wx/settings.h>
#include <wx/log.h>
#include <wx/frame.h>
#include <wx/timer.h>
#include <i3d/image3d.h>
#include "v4d_impexp.h"
#include "types/evt_mapping.h"
#include "types/evt_overlaying.h"
#include "vol3D_frame.h"
#include "types/cutpointer.h"
#include "types/projections.h"
#include "v4d_drwin.h"
#include "types/v4d_imageiterator.h"
#include "types/projectionannotationstyle.h"
#include "types/scalebarstyle.h"
#include "types/timebarstyle.h"

namespace v4d 
{
	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                    v4d Canvas Settings                      //
	//                                                             //
	/////////////////////////////////////////////////////////////////
	/** Settings of a canvas. They can be retreived from the canvas and
	    applied to the same or even a different instance. */
	struct v4d_CanvasSettings {
		bool show_info;  ///< show info about class?
		bool show_xy;    ///< display xy projection?
		bool show_xz;    ///< display xz projection?
		bool show_yz;    ///< display yz projection?
		v4d_ProjectionContent pc_xy; ///< content of xy projection (af/Ortho)
		v4d_ProjectionContent pc_xz; ///< content of xy projection (af/Ortho)
		v4d_ProjectionContent pc_yz; ///< content of xy projection (af/Ortho)
		int zoom_level;  ///< current zoom level
		i3d::Vector3d<float> cut; /// Position of a cut pointer in microns
		i3d::Vector3d<float> orig_res; /// Original resolution of the screen
		v4d_CutPointerStyle cut_pointer_style; /// Cut pointer style
		bool rev_axis[3]; ///< Which axis a user wants to be reversed? (0 - X, 1 - Y, 2 - Z)
		v4d_ProjectionAnnotationStyle projection_annotation_style;
		v4d_ScaleBarStyle scale_bar_style;
		v4d_TimeBarStyle time_bar_style;
	};

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                        Viewer3dCanvas                       //
	//                                                             //
	/////////////////////////////////////////////////////////////////
	/** Canvas of the main viewer window. It contains three drawing
	    windows (for each orthogonal projection one), the loaded image and
	    some attributes saying what to display on the screen. */
	class V4D_DLLEXPORT v4d_Viewer3dCanvas 
        : public wxPanel 
    {
	protected:
		/** Variables shared between drawing windows. */
		v4d_ViewerVars m_vv;

		/** A shared object that handles browsing through
		 * the image sequence. */
		i4d::ImageIterator m_ii;
		/** (Helper) Timer object to aid automatic sequence browsing. */
		wxTimer m_ii_timer;

		/** Pointer to the xy drawing window */
		v4d_DrWinXY *m_xy;
		/** Pointer to the xz drawing window */
		v4d_DrWinXZ *m_xz;
		/** Pointer to the yz drawing window */
		v4d_DrWinYZ *m_yz;
		
	public:
		/** Constructor. It creates an instance of the 3d canvas and
		initialize all its attributes on its defaults.  */
		v4d_Viewer3dCanvas(wxWindow *parent, wxWindowID id = -1, 
			const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxDefaultSize);

		/** Destructor. It frees memory occupied by 3d data and drawing
		windows. */
		virtual ~v4d_Viewer3dCanvas();

		/////////////////////////// Drawing, zooming //////////////////////////////
	private:
		/** Should be inforamation about canvas drawn? */
		bool m_canvas_info_flag;
		/** Should xy window be displayed? */
		bool m_show_xy;
		/** Should xz window be displayed? */
		bool m_show_xz;
		/** Should yz window be displayed? */
		bool m_show_yz;

		/** What to show in xy drawing window */
		v4d_ProjectionContent m_what_xy;
		/** What to show in xz drawing window */
		v4d_ProjectionContent m_what_xz;
		/** What to show in yz drawing window */
		v4d_ProjectionContent m_what_yz;

		/** Size of the border of the DrWin windows. Used during RecomputeVisible. */
		int m_border_size;

	protected:
		/** This function changes the sizes of drawing windows and the
		    attribute vv.vis3d according to the size of the canvas and the
		    visible part of the image. This function do not redraw the
		    content of the drawing windows. Function Redraw() must be
		    called. */
		void RecomputeVisible();
		/** Set cut pointer to the center ov vis3d, if it is out of view. */
		void CenterViewIfOut();
		/** Reverse axis orientation (0=X 1=Y 2=Z) in all projections. */
		virtual void ReverseAxis(int axis) = 0;

    private:
		/** This function is called when the size of canvas is changed. It
		    calles Viewer3dCanvas::RecomputeVisible() */
		void OnSize(wxSizeEvent& evt);
		/** This function is called on repaint events */
		void OnPaint(wxPaintEvent& evt);

	public:
		/** Set cut pointer to the center ov vis3d. */
		void CenterView();
		/** Multiply the screen resolution by 2 and redraw the drawing
		    windows. */
		void ZoomIn();
		/** Devide the screen resolution by 2 and redraw the drawing
		    windows. \sa RecomputeVisible() */
		void ZoomOut();
		/** Set the screen resolution to the default resolution and redraw
		    the drawing windows. \sa RecomputeVisible() */
		void NormalZoom();
		/** Changes the content of drawing windows. The buffers in drawing
		    windows are filled with new data. \param pc What to show in
		    drawing windows \param pt Which drawing windows to update */
		void ChangeContent(v4d_ProjectionContent pc, v4d_ProjectionType pt = pt_All);
		/** Return zoom level */
		int GetZoom() { return m_vv.zoom_level; }
		/** Return what is shown in a given window(s). Returns reasonable values also for
		    inputs such as pt_All. */
		v4d_ProjectionContent GetContent(v4d_ProjectionType pt) const;
		/** Show or hide drawing windows on the canvas. \param pt Which
		    projections are handled (visibility of others is untouched)
		    \param state \a true = show, \a false = hide */
		void ShowDrWin(v4d_ProjectionType pt = pt_All, bool state = true);
		/** Is the specified projection visible? */
		bool IsDrWinShown(v4d_ProjectionType pt) const;
		/** Get settings of the canvas. It can be used to store settings to store
		    settings somewhere and applying it later (or to a different canvas). */
		void GetSettings(v4d_CanvasSettings &params) const;
		/** Set several parameters of a canvas. */
		void SetSettings(const v4d_CanvasSettings &params);

		/** Get a copy of this->m_ii for the GUI purposes... */
		i4d::ImageIterator GetImageIterator(void) const
			{ return (m_ii); }
		/** Updates this->m_ii with *Range params, FPS and templateName.
			Also displays the first image in the range view. */
		void UpdateImageIterator(const i4d::ImageIterator& ii)
			{
				//update ranges and detected whether we need to change the current frame
				m_ii.SetFirstRangeIndex(ii.GetFirstRangeIndex());
				bool shouldChangeFrame=m_ii.IndexWasUpdated();
				m_ii.SetLastRangeIndex(ii.GetLastRangeIndex());
				shouldChangeFrame|=m_ii.IndexWasUpdated();
				m_ii.SetStepRangeIndex(ii.GetStepRangeIndex());
				shouldChangeFrame|=m_ii.IndexWasUpdated();

				m_ii.SetTemplateName(ii.GetTemplateName());
				m_ii.SetFPS(ii.GetFPS());

				m_ii.ResetForForwardSweep(true);
				if (m_ii.IndexWasUpdated() || shouldChangeFrame)
				{
					this->OnTimeChange();
					this->Update();
				}
			}
		/** Displays the next image in the sequence,
			designed for the exporting purposes...
			\return True if the was a new/next image available. */
		bool ShowNextImage(void)
			{
				m_ii.AdvanceIndexForward();
				bool newNextOne=m_ii.IndexWasUpdated();
				if (newNextOne)
				{
					this->OnTimeChange();
					this->Update();
				}
				return (newNextOne);
			}

		/** Reverse axis orientation. 0=X 1=Y 2=Z. */
		bool ToggleAxisReverse (int axis);
        /** Is given axis reversed? 0=X 1=Y 2=Z. */
        bool IsAxisReversed(size_t pid)
        {
            return m_vv.rev_axis[pid];
        }
        /** Check wheather detailed information about canvas is required. */
		bool GetInfoFlag() const { return m_canvas_info_flag; }
        /** Toggle info flag. */
		void ToggleInfoFlag() { m_canvas_info_flag = !m_canvas_info_flag; }

		/** This virtual function is called from OnPaint handlers of DrWins. 
		    You can redefine this function to draw projections. where is the 
		    window ID, from which drawing is called and is always one of the:
		    V4D_DRWIN_XY, V4D_DRWIN_YZ, or V4D_DRWIN_XZ. */
		virtual void Draw(wxDC &dc, v4d_ProjectionType where) = 0;

		/** \retuns list of wxIcons with different sizes. Each icon contains an image shown in xy window. 
		    If this window is hidden, then wxIcons remains unchanging. */
		virtual void GetIcons(wxIconBundle &bundle) {};
        virtual i3d::Vector3d<size_t> GetImageSize() = 0;
		/** Returns number of image in sequence which is currently displayed */
		virtual int GetTime() = 0;
		/** Returns total count of images in sequence */
		virtual int GetImageCount() = 0;
		/** Get resolution of images */
		virtual i3d::Resolution GetResolution() = 0;
		/** Save original 3d image data to given file. */
		virtual void Save3dData(const wxString &fname) = 0;
		/** Save original 3d image data to given file. */
		virtual void SaveActualView(const char *fname, int d, v4d_ProjectionType projection) = 0;
		/** Save slices of 3d data in chosen increments */
		virtual void SaveGalleryView(const char *fname, int start, int end, int step, bool fixedRows,
			int count, int spacing) = 0;

		/** Handles all the 'Sequence' menu commands. */
		void OnSequenceMenuHandler(wxCommandEvent& evt,wxString& msg);
		void OnSequenceTimerHandler(wxTimerEvent& evt);

		/** \return Returns indication whether automatic browsing is in effect. */
		bool OnSequenceMenuUpdateUI_PlayingForward(void) const
		{ return (m_ii_timer.IsRunning() && m_ii.GetCurrentDirection()==+1); }
		/** \return Returns indication whether automatic browsing is in effect. */
		bool OnSequenceMenuUpdateUI_PlayingBackward(void) const
		{ return (m_ii_timer.IsRunning() && m_ii.GetCurrentDirection()==-1); }
		/** \return Returns state of the m_ii loop mode. */
		bool OnSequenceMenuUpdateUI_Loop(void) const
		{ return (m_ii.IsInLoopMode()); }
		/** \return Returns state of the m_ii oscillation  mode. */
		bool OnSequenceMenuUpdateUI_Oscillation(void) const
		{ return (m_ii.IsInOscillationMode()); }

		/////////////////////////////// Cut pointer, projection annotation, scale and time bar ///////////////////////////////
	private:
		/** How to display the cut position. */
		v4d_CutPointerStyle m_pointer;
		/** How to display projection annotation. */
		v4d_ProjectionAnnotationStyle m_projection_annotation;
		/** How to display scale bar. */
		v4d_ScaleBarStyle m_scale_bar_style;
		/** How to display time bar */
		v4d_TimeBarStyle m_time_bar_style;
	public:
		/** Get optimal size of the canvas to see the whole bbox in all drawing windows */
		virtual wxSize GetOptimalSize();

		/** \returns The current cut pointer style */
		const v4d_CutPointerStyle& GetPointerStyle() const
		{ return m_pointer; }
		/** Set current cut pointer style */
		void SetPointerStyle(const v4d_CutPointerStyle &t) { m_pointer = t; }

		/** \returns The current projection annotation style */
		const v4d_ProjectionAnnotationStyle& GetProjectionAnnotationStyle() const
		{ return m_projection_annotation; }

		/** \returns The current scale bar style */
		const v4d_ScaleBarStyle& GetScaleBarStyle() const
		{ return m_scale_bar_style; }

		/** \returns The current time bar style */
		const v4d_TimeBarStyle& GetTimeBarStyle() const
		{ return m_time_bar_style; }

		/** Change cut position. Set orthogonal cutting planes to intersect in 
		    the given point (in microns). If the point is out of the bounding box nothing 
		    happens. */
		void ChangeCut(const i3d::Vector3d<float> &new_cut);

        /////////////////////////// Virtual handlers ///////////////////////////////

	public:
		/** This function is called on changes of m_vv.cut_* variables. 
		    The input parameter specifies which projections were changed. 
		    This function can be redefined in derived classes to do some 
		    extra actions on cut changes (e.g. making some caches dirty). */
		virtual void OnCutChange(v4d_ProjectionType pt) 
        { 
            Refresh(false); 
            InfoFrame_UpdateCutAndVisBoxInfo();
        }
		/** This function is called on changes of m_vv.vis3d_* variables.
		    The input parameter specifies which projections were changes.
		    This function can be redefined in derived classes to do some 
		    extra actions on changes in visible part of the scene (e.g. making 
		    some caches dirty). */
		virtual void OnVis3dChange(v4d_ProjectionType pt)
        {
            Refresh(false); 
            InfoFrame_UpdateCutAndVisBoxInfo();
        }
		/** This function is called on projection content changes.
		    The input parameter specifies which projections were changes. */
		virtual void OnContentChange(v4d_ProjectionType pt) 
        { 
            Refresh(false); 
        }
        /** Mapping change. */
		virtual void MappingChanged(v4d_MappingEvent& evt) 
        { 
            Refresh(false); 
        }
        /** Overlaying change. */
		virtual void OverlayingChanged(v4d_OverlayingEvent &evt) 
        { 
            Refresh(false); 
        }
        /** Data change. */
		virtual void OnDataChange()
        {
            InfoFrame_Move();
            InfoFrame_UpdateImageInfo();
            InfoFrame_UpdateCutAndVisBoxInfo();
        }
		/** Time point change. */
		virtual void OnTimeChange()
		{
			m_vv.time=m_ii.GetCurrentIndex();
			Refresh(false);
		}
		/** Mouse 3D event. This handler updates also the info frame so don't
            forget to call it if overloaded in a derived class. */
        virtual void OnMouse3d(v4d_Mouse3dEvent& evt);

        /////////////////////////// Volume visualization /////////////////////////

    protected:
        /** Pointer to the 3D visualization window **/
		vol3D_VolVisFrame *m_volVis;

    public:
		/** Creates or shows the vol3D_VolVisFrame*/
		void ShowVolVis(const wxIconBundle &bundle);
		/** Refresh visualization frame icon. */
		void RefreshVolVisIcon(const wxIconBundle &bundle);

    protected:
		/** Updates the vol3D_Visualization texture data*/
        virtual void UpdateVolVis() {}
		/** Updates the vol3D_Visualization isosurface data*/
        virtual void UpdateIsosurface() {}

        /////////////////////////////// Info frame ///////////////////////////////
    protected:
        /** Info frame window. */
        wxFrame *m_info_frame;        

    public:
        /** Show/hide info frame. */
        virtual void InfoFrame_Show(bool show);        
        /** Get info frame. */
        wxFrame *InfoFrame_Get() { return m_info_frame; }
        /** Is info frame shown? */
        bool InfoFrame_IsShown() { return (m_info_frame != NULL && m_info_frame->IsShown()); }
        /** Move info frame to appropriate position during moving. */
        void InfoFrame_Move();

    protected:
		/** Create info frame. */
		virtual void InfoFrame_Create(); 
        /** Update pixel info - intensity, mapped colour, etc. */
        virtual void InfoFrame_UpdatePixelInfo() = 0;
        /** Update image info in info frame. */
        virtual void InfoFrame_UpdateImageInfo() = 0;
        /** Update cut position and visbox info. */
        virtual void InfoFrame_UpdateCutAndVisBoxInfo() = 0;
	};
}

#endif
