/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * v4d_drwin.cpp
 *
 * 2D drawing window
 *
 * Petr Matula (pem@fi.muni.cz) 2006
 */

#include "v4d_basic.h"
#include <wx/log.h>
#include <wx/image.h>
#include <wx/dcmemory.h>
#include <math.h>
#include "v4d_drwin.h"
#include "v4d_canvas3d.h"
#include "mycursors.h"
#include "i3d/resolution.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

namespace v4d 
{
    // v4d_Mouse3dEvent
    wxDEFINE_EVENT(v4d_EVT_MOUSE3D, v4d_Mouse3dEvent);

    const i3d::Vector3d<float> v4d_Mouse3dEvent::Get3dPosInMicrons() const
    {
        i3d::Vector3d<float> pos;

        wxPoint win_pos = m_mouse2d.GetPosition();
        pos.x = m_dr_win->W2MPosX(win_pos.x);
        pos.y = m_dr_win->W2MPosY(win_pos.y);
        pos.z = m_dr_win->CutPos();

        return m_dr_win->Reorder(pos);
    }

    // v4d_DrWin3d
    BEGIN_EVENT_TABLE(v4d_DrWin3d, wxWindow)
        EVT_PAINT(v4d_DrWin3d::OnPaint)
        EVT_ERASE_BACKGROUND(v4d_DrWin3d::OnErase)
        EVT_KEY_DOWN(v4d_DrWin3d::OnChar)
        EVT_MOUSE_EVENTS(v4d_DrWin3d::OnMouseEvent)
        EVT_SET_FOCUS(v4d_DrWin3d::OnSetFocus)
        EVT_KILL_FOCUS(v4d_DrWin3d::OnKillFocus)
    END_EVENT_TABLE();

    v4d_DrWin3d::v4d_DrWin3d(v4d_Viewer3dCanvas *parent, wxWindowID id,
        const float &bb_w_, const float &bb_dw_, const float &bb_h_, const float &bb_dh_,
        const float &bb_d_, const float &bb_dd_,
        float &vis_w_, float &vis_dw_, float &vis_h_, float &vis_dh_,
        const float &res_w_, const float &res_h_,
        float &cut_w_, float &cut_h_, float &cut_d_,
        const float &step_w_, const float &step_h_, const float &step_d_,
        const bool &reverse_w_, const bool &reverse_h_, const bool &reverse_d_,
		  i4d::ImageIterator& ii_)
        : wxWindow(parent, id, wxDefaultPosition, wxDefaultSize, wxSUNKEN_BORDER | wxWANTS_CHARS),
        m_bmp(10,10),
        m_canvas(parent),
		  m_ii(ii_),

        r_bb_w(bb_w_), r_bb_dw(bb_dw_), 
        r_bb_h(bb_h_), r_bb_dh(bb_dh_), 
        r_bb_d(bb_d_), r_bb_dd(bb_dd_),

        r_vis_w(vis_w_), r_vis_dw(vis_dw_), 
        r_vis_h(vis_h_), r_vis_dh(vis_dh_),

        r_res_w(res_w_), r_res_h(res_h_),
        r_reverse_w(reverse_w_), r_reverse_h(reverse_h_), r_reverse_d(reverse_d_),
        r_cut_w(cut_w_), r_cut_h(cut_h_), r_cut_d(cut_d_),
        r_step_w(step_w_), r_step_h(step_h_), r_step_d(step_d_)
    {}

    void v4d_DrWin3d::OnErase(wxEraseEvent &evt)
    {
    }

    void v4d_DrWin3d::DrawCutPointer(wxDC &dc)
    {
        // Draw the cut pointer
        wxCoord x = M2WPosX(r_cut_w);
        wxCoord y = M2WPosY(r_cut_h);
        wxCoord w = m_bmp.GetWidth();
        wxCoord h = m_bmp.GetHeight();

        switch (m_canvas->GetPointerStyle().type)
        {
        case cp_FULL:
            dc.SetPen (wxPen (m_canvas->GetPointerStyle().colour, 1, wxPENSTYLE_SOLID));
            dc.DrawLine(0,y,w,y);
            dc.DrawLine(x,0,x,h);
            break;
        case cp_DOTTED:
            dc.SetPen (wxPen (m_canvas->GetPointerStyle().colour, 1, wxPENSTYLE_DOT));
            dc.DrawLine(0,y,w,y);
            dc.DrawLine(x,0,x,h);
            break;
        case cp_DASHED:
            dc.SetPen (wxPen (m_canvas->GetPointerStyle().colour, 1, wxPENSTYLE_SHORT_DASH));
            dc.DrawLine(0,y,w,y);
            dc.DrawLine(x,0,x,h);
            break;
        case cp_CIRCLE:
            dc.SetPen (wxPen (m_canvas->GetPointerStyle().colour, 1, wxPENSTYLE_SOLID));
            dc.SetBrush (*wxTRANSPARENT_BRUSH);
            dc.DrawCircle(x+1, y+1, CP_SHORT_LEN);
            /* no break -> short lines are wanted */
        case cp_SHORT:
            dc.SetPen (wxPen (m_canvas->GetPointerStyle().colour, 1, wxPENSTYLE_SOLID));
            dc.DrawLine(0, y, CP_SHORT_LEN, y);
            dc.DrawLine(w - CP_SHORT_LEN, y, w, y);
            dc.DrawLine(x, 0, x, CP_SHORT_LEN);
            dc.DrawLine(x, h - CP_SHORT_LEN, x, h);
            break;
        case cp_NONE:
            break;
        }
    }

	//calculates optimal top-left corner to place the annotation to
	wxPoint GetPointForAnnot(
		const v4d_TextPosition& position, //the preferred corner
		const wxPoint& offset,            //preferred offset from the corner (in px)
		const wxPoint& sizeW,             //size of the window (in px)
		const wxPoint& sizeA)             //size of the annotation (in px)
		                                  //for which the coordinate is calculated
	{
		wxPoint point(0, 0);
		switch (position)
		{
		case tp_TOP_LEFT:
			point = offset;
			break;
		case tp_BOTTOM_LEFT:
			point = wxPoint(offset.x, sizeW.y-sizeA.y-offset.y);
			break;
		case tp_TOP_RIGHT:
			point = wxPoint(sizeW.x-sizeA.x-offset.x, offset.y);
			break;
		case tp_BOTTOM_RIGHT:
			point = wxPoint(sizeW.x-sizeA.x-offset.x, sizeW.y-sizeA.y-offset.y);
			break;
		default:
			//hmm, rather do not change the "default" position
			break;
		}

		return point;
	}

	void v4d_DrWin3d::DrawAnnotationText(wxDC &dc)
	{
		//cache some relevant input data...
		const v4d_ProjectionAnnotationStyle& annotation
			= m_canvas->GetProjectionAnnotationStyle();

		//choose what to draw and where, actually
		const wxString* text=NULL;
		const v4d_TextPosition* corner=NULL;
		const wxPoint* offset=NULL;
		switch (thisWin())
		{
		case pt_XY:
			text=  &annotation.xyText;
			corner=&annotation.xyTextPos;
			offset=&annotation.xyOffset;
			break;
		case pt_XZ:
			text=  &annotation.xzText;
			corner=&annotation.xzTextPos;
			offset=&annotation.xzOffset;
			break;
		case pt_YZ:
			text=  &annotation.yzText;
			corner=&annotation.yzTextPos;
			offset=&annotation.yzOffset;
			break;
		default:
			return; //nothing to draw...
			break;
		}

		//do nothing if no text is to be displayed
		if (*text == wxEmptyString) return;

		//setup the drawing...
		dc.SetTextForeground(annotation.colour);
		dc.SetFont(annotation.font);

		//get size of the window to draw into
		wxPoint sizeW;
		GetClientSize(&sizeW.x, &sizeW.y);

		//get size of the text when rendered
		const wxSize textSize = dc.GetTextExtent(*text);
		wxPoint sizeA(textSize.GetWidth(),textSize.GetHeight());

		//where to start drawing?
		//gets top-left corner
		wxPoint pos=GetPointForAnnot(*corner,*offset,sizeW,sizeA);

		//calc bottom-left corner
		pos.y+=sizeA.y;

		//draw text
		dc.DrawText(*text, wxPoint(pos.x,pos.y-textSize.GetHeight()));
	}

	void v4d_DrWin3d::DrawTimeBar(wxDC &dc)
	{
		//cache some relevant input data...
		const v4d_TimeBarStyle& style = m_canvas->GetTimeBarStyle();
		const wxString text = "Time: ";
		const int numOfImages = m_canvas->GetImageCount();

		//do nothing if width value is not positive or there's "no time-lapse"
		if (style.width <= 0 || numOfImages <= 1)
			return;

		//setup the drawing...
		dc.SetFont(m_canvas->GetProjectionAnnotationStyle().font); //uses the one from Annotations...
		dc.SetTextForeground(m_canvas->GetProjectionAnnotationStyle().colour);
		dc.SetPen(wxPen(style.colour, style.width));

		//get size of the window to draw into
		wxPoint sizeW;
		GetClientSize(&sizeW.x, &sizeW.y);

		//get size of the text when rendered
		const wxSize textSize = dc.GetTextExtent(text);

		//additional offset to accommodate for thick lines of the bar
		const int barOffset=style.width/2;

		//calc size of the whole timebar (ignoring width of the lines)
		wxPoint sizeA(textSize.GetWidth()+barOffset+style.length,
		              std::max(textSize.GetHeight(),20));

		//where to start drawing?
		//gets top-left corner
		wxPoint pos=GetPointForAnnot(style.corner,style.offset,sizeW,sizeA);

		//calc bottom-left corner
		pos.y+=sizeA.y;

		//draw text
		dc.DrawText(text, wxPoint(pos.x,pos.y-textSize.GetHeight()));

		//draw bar baseline
		dc.SetTextForeground(style.colour);
		pos.x+=textSize.GetWidth()+barOffset;
		dc.DrawLine(pos,pos+wxPoint(style.length,0));

		//draw bar "time indicator"
		pos.x+=roundf( (float)style.length * (float)m_canvas->GetTime() / (float)(numOfImages-1) );
		dc.DrawLine(pos,pos+wxPoint(0,-20));
	}

	void v4d_DrWin3d::DrawScaleBar(wxDC &dc)
	{
		//cache some relevant input data...
		const v4d_ScaleBarStyle& style = m_canvas->GetScaleBarStyle();

		//determine precision of the float value (so that trailing zeros are not printed)
		long lValue=long(style.length*1000000.f) % 1000000;
		int lPrecision=6;
		while ((lPrecision > 0) && (lValue % 10 == 0))
		{ --lPrecision; lValue /= 10; }

		const wxString text = wxString::Format("%.*f %cm: ", lPrecision,style.length,0x00b5);
		//the 0x00b5 is UTF code for the Greek micro symbol

		//do nothing if width is not positive
		if (style.width <= 0)
			return;

		const int zoom = m_canvas->GetZoom() + 1;
		i3d::Vector3d<float> micronVector(float(style.length*zoom)); //inits all three elements
		i3d::Vector3d<i3d::PIXELS::T_offset> pixelVector;
		pixelVector = i3d::MicronsToPixels(micronVector, m_canvas->GetResolution());

		//choose what to draw
		int lengthX=0, lengthY=0;
		switch (thisWin())
		{
		case pt_XY:
			lengthX=pixelVector.x;
			lengthY=pixelVector.y;
			break;
		case pt_XZ:
			lengthX=pixelVector.x;
			lengthY=pixelVector.z;
			break;
		case pt_YZ:
			lengthX=pixelVector.z;
			lengthY=pixelVector.y;
			break;
		default:
			return; //nothing to draw...
			break;
		}

		//setup the drawing...
		//dc.SetFont(wxFont()); //keeps the one from Annotations (as a consequence)
		dc.SetPen(wxPen(style.colour, style.width));
		dc.SetTextForeground(m_canvas->GetProjectionAnnotationStyle().colour);

		//get size of the window to draw into
		wxPoint sizeW;
		GetClientSize(&sizeW.x, &sizeW.y);

		//get size of the text when rendered
		const wxSize textSize = dc.GetTextExtent(text);

		//additional offset to accommodate for thick lines of the bar
		const int barOffset=style.width/2;

		//calc size of the whole timebar (ignoring width of the lines)
		wxPoint sizeA(textSize.GetWidth()+barOffset+lengthX,
		              std::max(textSize.GetHeight(),lengthY));

		//where to start drawing?
		//gets top-left corner
		wxPoint pos=GetPointForAnnot(style.corner,style.offset,sizeW,sizeA);

		//draw text and bar scale bars
		switch (style.corner)
		{
		case tp_TOP_LEFT:
			dc.DrawText(text,pos);
			dc.SetTextForeground(style.colour);
			pos.x+=textSize.GetWidth()+barOffset;
			dc.DrawLine(pos,pos+wxPoint(lengthX,0));
			dc.DrawLine(pos,pos+wxPoint(0,lengthY));
			break;
		case tp_BOTTOM_LEFT:
			pos.y+=sizeA.y;
			dc.DrawText(text, wxPoint(pos.x,pos.y-textSize.GetHeight()));
			dc.SetTextForeground(style.colour);
			pos.x+=textSize.GetWidth()+barOffset;
			dc.DrawLine(pos,pos+wxPoint(lengthX,0));
			dc.DrawLine(pos,pos+wxPoint(0,-lengthY));
			break;
		case tp_TOP_RIGHT:
			dc.DrawText(text,pos);
			dc.SetTextForeground(style.colour);
			pos.x+=textSize.GetWidth()+barOffset;
			pos.x+=lengthX;
			dc.DrawLine(pos,pos+wxPoint(-lengthX,0));
			dc.DrawLine(pos,pos+wxPoint(0,lengthY));
			break;
		case tp_BOTTOM_RIGHT:
			pos.y+=sizeA.y;
			dc.DrawText(text, wxPoint(pos.x,pos.y-textSize.GetHeight()));
			dc.SetTextForeground(style.colour);
			pos.x+=textSize.GetWidth()+barOffset;
			pos.x+=lengthX;
			dc.DrawLine(pos,pos+wxPoint(-lengthX,0));
			dc.DrawLine(pos,pos+wxPoint(0,-lengthY));
			break;
		default:
			//hmm, rather do not paint at all
			break;
		}
	}

    /** Paint something to the window. */
    void v4d_DrWin3d::OnPaint(wxPaintEvent& event)
    {
        wxPaintDC dc(this);
        
        // Get client size
        int w, h;
        GetClientSize (&w, &h);

        // resize bitmap if needed 
        if (m_bmp.GetWidth() != w || m_bmp.GetHeight() != h)
        {
            m_bmp.Create(w,h);
        }

        wxMemoryDC memDC;
        memDC.SelectObject(m_bmp);

        // draw to memory dc
        memDC.SetBackground(*wxBLACK_BRUSH);
        memDC.Clear();
        m_canvas->Draw(memDC, thisWin());

        // add some "info graphics":
        // cut pointers -- always visible (unless expicitly "cancelled" in menu)
        DrawCutPointer(memDC);

        // time bar -- always visible (unless expicitly "cancelled" in menu)
        if (thisWin() == pt_XY)
        {			
        	DrawTimeBar(memDC);
        }

        // annotations only if 'Alt+/' requests them
        if (m_canvas->GetInfoFlag())
        {
        	DrawAnnotationText(memDC);
        	DrawScaleBar(memDC);
        }

        // copy memory dc to screen
        dc.Blit (0, 0, w, h, &memDC, 0, 0, wxCOPY);
        memDC.SelectObject(wxNullBitmap);
    }

    void v4d_DrWin3d::OnMouseEvent(wxMouseEvent &evt)
    {
        v4d_Mouse3dEvent mevt(v4d_EVT_MOUSE3D, evt, this);
        m_canvas->ProcessWindowEvent(mevt);
        evt.Skip();
    }

    void v4d_DrWin3d::OnSetFocus(wxFocusEvent& e)
    {
        SetCursor(MyCursorsFunc(v4d_CURSOR_CROSS));
    }

    void v4d_DrWin3d::OnKillFocus(wxFocusEvent& e)
    {
        SetCursor(wxNullCursor);
    }

    static inline void CenterComp(float &off, float x, float dx,
        float min_off, float max_off)
    {
        off = std::min (std::max(min_off, x - dx/2), max_off - dx);
    }

    void v4d_DrWin3d::CenterView(float x, float y)
    {
        CenterComp(r_vis_w, x, r_vis_dw, r_bb_w, r_bb_w + r_bb_dw);
        CenterComp(r_vis_h, y, r_vis_dh, r_bb_h, r_bb_h + r_bb_dh);
    }

    void v4d_DrWin3d::MoveCut(float x, float y)
    {
        float min_x = r_bb_w;
        float min_y = r_bb_h;
        float max_x = min_x + r_bb_dw;
        float max_y = min_y + r_bb_dh;

        r_cut_w = x;
        r_cut_h = y;

        if (r_cut_w < min_x)
            r_cut_w = min_x;

        if (r_cut_h < min_y)
            r_cut_h = min_y;

        if (r_cut_w > max_x)
            r_cut_w = max_x;

        if (r_cut_h > max_y)
            r_cut_h = max_y;
    }

    void v4d_DrWin3d::OnChar(wxKeyEvent& event)
    {
        wxPoint screenMousePosition, clientMousePosition;
        wxWindow *shot = wxFindWindowAtPointer(screenMousePosition);

        if (shot == this) {
            clientMousePosition = ScreenToClient(screenMousePosition);      
        } else {
            clientMousePosition.x = -1;
        }

        switch (event.GetKeyCode())
        {
        case WXK_UP:
            Up();
            break;

        case WXK_DOWN:
            Down();
            break;

        case WXK_LEFT:
            Left();
            break;

        case WXK_RIGHT:
            Right();
            break;

        case WXK_PAGEUP:
            PgUp();
            break;

        case WXK_PAGEDOWN:
            PgDown();
            break;

		case 61: //equal sign '='; on English keyboard this key holds the plus sign '+'
		case WXK_NUMPAD_ADD:
			Forward();
			break;

		case 45: //minus sign '-'
		case WXK_NUMPAD_SUBTRACT:
			Backward();
			break;

		case WXK_HOME:
			m_ii.ResetForForwardSweep(true);
			if (m_ii.IndexWasUpdated()) m_canvas->OnTimeChange();
			break;
		case WXK_END:
			m_ii.ResetForBackwardSweep(true);
			if (m_ii.IndexWasUpdated()) m_canvas->OnTimeChange();
			break;

        case WXK_SPACE:
            {
                if (PosInWin(clientMousePosition))    
                {
                    MoveCut(W2MPosX(clientMousePosition.x),W2MPosY(clientMousePosition.y));
                    m_canvas->OnCutChange(otherWins());
                }	  
            }
            break;

        case 'C':
        case 'c':
            if (event.ControlDown())
                event.Skip();
            else {
                if (event.AltDown()) {
                    m_canvas->CenterView();
                    m_canvas->OnVis3dChange(pt_All);
                }
                else if (PosInWin(clientMousePosition))	{
                    CenterView(W2MPosX(clientMousePosition.x),W2MPosY(clientMousePosition.y));
                    m_canvas->OnVis3dChange(pt_All);
                }
            }
            break;

        default:
            event.Skip();
        }
    }

    bool v4d_DrWin3d::Closer(float &v_off, float r_off, float step)
    {
        if (v_off > r_off)
            v_off -= step;
        else
            return false;

        if (v_off < r_off)
            v_off = r_off;
        return true;
    }

    bool v4d_DrWin3d::Further(float &v_off, float r_off, float step)
    {
        if (v_off < r_off)
            v_off += step;
        else
            return false;

        if (v_off > r_off)
            v_off = r_off;
        return true;
    }

    void v4d_DrWin3d::Up()
    {
        if (Closer(r_vis_h, r_bb_h, r_step_h))
        {
            m_canvas->OnVis3dChange(verticalWins());
        }
    }

    void v4d_DrWin3d::Down()
    {
        if (Further(r_vis_h, r_bb_h + r_bb_dh - r_vis_dh, r_step_h))
        {
            m_canvas->OnVis3dChange(verticalWins());	  
        }
    }

    void v4d_DrWin3d::Left()
    {
        if (Closer(r_vis_w, r_bb_w, r_step_w))
        {
            m_canvas->OnVis3dChange(horizontalWins());
        }
    }

    void v4d_DrWin3d::Right()
    {
        if (Further(r_vis_w, r_bb_w + r_bb_dw - r_vis_dw, r_step_w))
        {
            m_canvas->OnVis3dChange(horizontalWins());
        }
    }

    void v4d_DrWin3d::PgDown()
    {
        if (Further(r_cut_d, r_bb_d + r_bb_dd - r_step_d/2, r_step_d))
        {
            m_canvas->OnCutChange(thisWin());
        }
    }

    void v4d_DrWin3d::PgUp()
    {
        if (Closer(r_cut_d, r_bb_d + r_step_d/2, r_step_d)) 
        {
            m_canvas->OnCutChange(thisWin());
        }
    }

	void v4d_DrWin3d::Forward()
	{
		m_ii.AdvanceIndexForward();
		if (m_ii.IndexWasUpdated())
			m_canvas->OnTimeChange();
	}

	void v4d_DrWin3d::Backward()
	{
		m_ii.AdvanceIndexBackward();
		if (m_ii.IndexWasUpdated())
			m_canvas->OnTimeChange();
	}

    wxImage  v4d_DrWin3d::GetDrWinContent()
    {
        return m_bmp.ConvertToImage(); 
    }
}

