/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * v4d_drwin.h
 *
 * 2D drawing window
 *
 * Petr Matula (pem@fi.muni.cz) 2006
 */
#ifndef V4D_V4D_DRWIN_H
#define V4D_V4D_DRWIN_H

#include <wx/window.h>
#include <wx/bitmap.h>
#include "v4d_impexp.h"
#include "types/viewervars.h"
#include "types/projections.h"
#include <wx/textfile.h>
#include "types/v4d_imageiterator.h"

namespace v4d 
{
    // Forward declarations
    class v4d_DrWin3d;
    class v4d_Viewer3dCanvas;

    /** An event sent by DrWin3d on arbitrary mouse event. */
    class v4d_Mouse3dEvent 
        : public wxEvent
    {
    private:
        /** Orthogonal projection in which the event originated. */
        v4d_DrWin3d *m_dr_win;
        /** 2D mouse event. */
        wxMouseEvent m_mouse2d;

    public:
        /** Constructor. */
        v4d_Mouse3dEvent(wxEventType eventType, wxMouseEvent &evt, v4d_DrWin3d *dr_win)
            : wxEvent(evt.GetId(), eventType), m_dr_win(dr_win),  m_mouse2d(evt)
        {}

        /** Copy constructor. */
        v4d_Mouse3dEvent(const v4d_Mouse3dEvent& evt)
            : wxEvent(evt), m_dr_win(evt.m_dr_win), m_mouse2d(evt.m_mouse2d)
        {}

        /** Get projection in which the event originated. */
        v4d_DrWin3d *GetDrWin() const
        {
            return m_dr_win;
        }

        /** Return 2D mouse event. */
        const wxMouseEvent& Get2dMouseEvent() const
        { 
            return m_mouse2d; 
        }

        /** Return 3D position in microns. */
        const i3d::Vector3d<float> Get3dPosInMicrons() const;

        /** Clone method. */
        virtual wxEvent *Clone() const 
        { 
            return new v4d_Mouse3dEvent(*this); 
        }
    };

    wxDECLARE_EXPORTED_EVENT(V4D_DLLEXPORT, v4d_EVT_MOUSE3D, v4d_Mouse3dEvent);    

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                             DrWin3d                         //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    /** Base class of all three drawing windows. This class contains
    common functionality of xy, xz and yz windows. In fact the
    behaviour of all these three windows is the same. They differ only
    in mapping of x, y, and z components to width, height and depth in
    corresponding 2d windows. This correspondence is realised by
    references to the shared viewer variables (see v4d_ViewerVars). The
    right references are given in the constructors of the derived
    classes. \sa v4d_DrWinXY, v4d_DrWinXZ, v4d_DrWinYZ.
    */
    class v4d_DrWin3d 
        : public wxWindow
    {
        /** Buffer for memory drawings */
        wxBitmap m_bmp;
        /** Pointer to canvas object. */
        v4d_Viewer3dCanvas *m_canvas;
		  /** A reference to the image sequence browser. */
		  i4d::ImageIterator& m_ii;

    protected:
        ////////////////////////////////////////////////
        // References to the bounding box
        ////////////////////////////////////////////////
        /** Reference to the offset of the bounding box
        (x direction on the screen - w-direction) */
        const float &r_bb_w;  
        /** Reference to the size of the bounding box
        (x direction on the screen - w-direction ) */
        const float &r_bb_dw; 
        /** Reference to the offset of the bounding box
        (y direction on the screen - h-direction) */
        const float &r_bb_h;  
        /** Reference to the size of the bounding box
        (y direction on the screen - h-direction) */
        const float &r_bb_dh; 
        /** Reference to the offset of the bounding box
        (direction perpendicular to the screen - d-direction) */
        const float &r_bb_d;  
        /** Reference to the size of the bounding box
        (direction perpendicular to the screen - d-direction) */
        const float &r_bb_dd; 

        ////////////////////////////////////////////////
        // References to the visible part of the image
        ////////////////////////////////////////////////
        /** Reference to the offset of the visible part
        (x direction on the screen - w-direction) */
        float &r_vis_w;       
        /** Reference to the size of the visible part
        (x direction on the screen - w-direction) */
        float &r_vis_dw;      
        /** Reference to the offset of the visible part
        (y direction on the screen - h-direction) */
        float &r_vis_h;       
        /** Reference to the size of the visible part
        (y direction on the screen - h-direction) */
        float &r_vis_dh;      

        ////////////////////////////////////////////////
        // References to the screen resolution
        ////////////////////////////////////////////////
        /** Reference to the screen resolution in pixels
        per microns (x direction on the screen) */
        const float &r_res_w; 
        /** Reference to the screen resolution in pixels
        per microns (y direction on the screen) */
        const float &r_res_h; 

        ////////////////////////////////////////////////
        // References to the axis reverse flag
        ////////////////////////////////////////////////
        const bool &r_reverse_w; ///< Is w direction reversed?
        const bool &r_reverse_h; ///< Is h direction reversed?
        const bool &r_reverse_d; ///< Is d direction reversed?

        ////////////////////////////////////////////////
        // References to the position of the cut
        ////////////////////////////////////////////////
        /** Reference to the position of the cut
        (x direction on the screen - w-direction) */
        float &r_cut_w;       
        /** Reference to the position of the cut
        (y direction on the screen - h-direction) */
        float &r_cut_h;       
        /** Reference to the position of the cut
        (direction perpendicular to the screen - d-direction) */
        float &r_cut_d;       

        ////////////////////////////////////////////////
        // References to the steps
        ////////////////////////////////////////////////
        /** Reference to mstep - the number of microns that is
        added/subtracted to/from vis_w on a key press */
        const float &r_step_w;
        /** Reference to mstep - the number of microns that is
        added/subtracted to/from vis_h on a key press */
        const float &r_step_h;
        /** Reference to cstep - the number of microns that is
        added/subtracted to/from cut_d on a key press */
        const float &r_step_d;

        ////////////////////////////////////////////////
        // Local functions
        ////////////////////////////////////////////////
        /** Move \a v_off closer to \a r_off by a step \a step and
        simultaneously retain the condition \a r_off <= \a v_off.
        \returns \p true if \a v_off has changed\n \p false
        otherwise. */
        bool Closer(float &v_off, float r_off, float step);

        /** Move \a v_off closer to \a r_off by a step \a step and
        simultaneously retain the condition \a r_off >= \a v_off.
        \returns \p true if \a v_off has changed\n \p false
        otherwise. */
        bool Further(float &v_off, float r_off, float step);

        /** Move visible part up. */
        void Up();
        /** Move visible part down. */
        void Down();
        /** Move visible part left. */
        void Left();
        /** Move visible part right. */
        void Right();
        /** Move cut in d-direction up . */
        void PgUp();
        /** Move cut in d-direction down. */
        void PgDown();
		/** Move forward in time. */
		void Forward();
		/** Move backward in time. */
		void Backward();
        /** Set cut in wh-direction to [x,y]. */
        void MoveCut(float x, float y);
        /** Draw cut pointer position in current window */
        void DrawCutPointer(wxDC &dc);
        /** Draw information about current window (it is good for debugging purposes) */
        void DrawAnnotationText(wxDC &dc);
		/** Draw time bar */
		void DrawTimeBar(wxDC &dc);
		/** Draw scale bar */
		void DrawScaleBar(wxDC &dc);

    public:
        /** Default constructor. It initializes all class members, most
        importantly references to the shared viewer variables. */
        v4d_DrWin3d(v4d_Viewer3dCanvas *parent, wxWindowID id,
            const float &bb_w_, const float &bb_dw_, const float &bb_h_, const float &bb_dh_,
            const float &bb_d_, const float &bb_dd_,
            float &vis_w_, float &vis_dw_, float &vis_h_, float &vis_dh_,
            const float &res_w_, const float &res_h_,
            float &cut_w_, float &cut_h_, float &cut_d_,
            const float &step_w_, const float &step_h_, const float &step_d_,
            const bool &reverse_w_, const bool &reverse_h_, const bool &reverse_d_,
				i4d::ImageIterator& ii_);

        /** Get optimal size of the window (the smallest size to see whole bbox)*/
        virtual wxSize GetOptimalSize() { return wxSize(static_cast<int>(r_bb_dw * r_res_w),static_cast<int>(r_bb_dh * r_res_h)); };
        /** Function is called when a key is pressed and the drawing window
        has focus. */
        void OnChar(wxKeyEvent& event);
        /** On erase event handler. */
        void OnErase(wxEraseEvent& evt);
        /** Paint something to the window. */
        void OnPaint(wxPaintEvent& event);
        /** Handler called on mouse motion. */
        //void OnMouseMotion(wxMouseEvent& event);
        /** Handler called on all mouse events. */
        void OnMouseEvent(wxMouseEvent& evt);
        /** Update m_last_drwin pointer in focus events and change window cursor. */
        void OnSetFocus(wxFocusEvent& event);
        /** Set mouse cursor to default pointer*/
        void OnKillFocus(wxFocusEvent& event);
        /** Get content of projection window. */
        wxImage GetDrWinContent();

    public: 
        /** Test whether the given point \a p lies inside the current
        drawing window. \returns \p true if the test is positive\n \p
        false otherwise. */
        bool PosInWin(const wxPoint &p) {
            int w,h;
            GetClientSize(&w,&h);
            return (p.x >= 0) && (p.y >= 0) && (p.x < w) && (p.y < h);
        }

        /** Conversion from micrometers in the coordinate system of the
        image (w-direction) to the client window pixels in the x-direction of
        the drawing window.  */
        int M2WPosX(float pos) {
            return (int) ((pos - r_vis_w) * r_res_w + 0.5f);
        }
        /** Conversion from micrometers in the coordinate system of the
        image (h-direction) to the client window pixels in the y-direction of
        the drawing window.  */
        int M2WPosY(float pos) {
            return (int) ((pos - r_vis_h) * r_res_h + 0.5f);
        }

        /** Conversion from client pixels in the x-direction of the drawing
        window to micrometers in the coordinate system of the image(its
        w-direction).  */
        float W2MPosX(int pos) {
            if (r_reverse_w) pos = int(GetSize().GetWidth()) - pos - 1;
            return float(pos) / r_res_w + r_vis_w; }
        /** Conversion from client pixels in the y-direction of the drawing
        window to micrometers in the coordinate system of the image (its
        h-direction).  */
        float W2MPosY(int pos) {
            if (r_reverse_h) pos = int(GetSize().GetHeight()) - pos - 1;
            return float(pos)/ r_res_h + r_vis_h; }

        /** \returns current cut position in d direction. */
        float CutPos() { return r_cut_d; }

        /** Function centers visible part of the image so that the point
        [x,y] is as close as possible to the center of the visible part. */
        void CenterView(float x, float y);

        virtual v4d_ProjectionType thisWin() = 0;
        virtual v4d_ProjectionType otherWins() = 0;
        virtual v4d_ProjectionType verticalWins() = 0;
        virtual v4d_ProjectionType horizontalWins() = 0;

        virtual i3d::Vector3d<float> Reorder(const i3d::Vector3d<float> &v) const = 0;

        DECLARE_EVENT_TABLE()
    };


    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                             DrWinXY                         //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    /** XY drawing window derived from DrWin3d. */
    class v4d_DrWinXY
        : public v4d_DrWin3d
    {
    public:
        /** Default constructor. It sets appropriate references to the
        shared variables. */
        v4d_DrWinXY(v4d_Viewer3dCanvas *parent, v4d_ViewerVars &v, i4d::ImageIterator& ii) 
            : v4d_DrWin3d(parent, wxID_ANY,
              v.bbox.offset.x, v.bbox.size.x,
              v.bbox.offset.y, v.bbox.size.y,
              v.bbox.offset.z, v.bbox.size.z,
              v.vis3d.offset.x, v.vis3d.size.x,
              v.vis3d.offset.y, v.vis3d.size.y,
              v.res.x, v.res.y,
              v.cut.x, v.cut.y, v.cut.z,
              v.mstep.x, v.mstep.y, v.cstep.z,
              v.rev_axis[0], v.rev_axis[1], v.rev_axis[2],
				  ii) {};

          virtual v4d_ProjectionType thisWin() { return pt_XY; };
          virtual v4d_ProjectionType otherWins() { return pt_ButXY; };
          virtual v4d_ProjectionType verticalWins() { return pt_ButXZ; };
          virtual v4d_ProjectionType horizontalWins() { return pt_ButYZ; };
          virtual i3d::Vector3d<float> Reorder(const i3d::Vector3d<float> &v) const
          { return i3d::Vector3d<float>(v.x, v.y, v.z); };
    };

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                             DrWinXZ                         //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    /** XZ drawing window derived from DrWin3d. */
    class v4d_DrWinXZ
        : public v4d_DrWin3d
    {
    public:
        /** Default constructor. It sets appropriate references to the
        shared variables. */
        v4d_DrWinXZ(v4d_Viewer3dCanvas *parent, v4d_ViewerVars &v, i4d::ImageIterator& ii) 
            : v4d_DrWin3d(parent, wxID_ANY,
              v.bbox.offset.x, v.bbox.size.x,
              v.bbox.offset.z, v.bbox.size.z,
              v.bbox.offset.y, v.bbox.size.y,
              v.vis3d.offset.x, v.vis3d.size.x,
              v.vis3d.offset.z, v.vis3d.size.z,
              v.res.x, v.res.z,
              v.cut.x, v.cut.z, v.cut.y,
              v.mstep.x, v.mstep.z, v.cstep.y,
              v.rev_axis[0], v.rev_axis[1], v.rev_axis[2],
				  ii) {};

          virtual v4d_ProjectionType thisWin() { return pt_XZ; };
          virtual v4d_ProjectionType otherWins() { return pt_ButXZ; };
          virtual v4d_ProjectionType verticalWins() { return pt_ButXY; };
          virtual v4d_ProjectionType horizontalWins() { return pt_ButYZ; };
          virtual i3d::Vector3d<float> Reorder(const i3d::Vector3d<float> &v) const
          { return i3d::Vector3d<float>(v.x, v.z, v.y); };
    };

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                             DrWinYZ                         //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    /** YZ drawing window derived from DrWin3d. */
    class v4d_DrWinYZ
        : public v4d_DrWin3d
    {
    public:
        /** Default constructor. It sets appropriate references to the
        shared variables. */
        v4d_DrWinYZ(v4d_Viewer3dCanvas *parent, v4d_ViewerVars &v, i4d::ImageIterator& ii) 
            : v4d_DrWin3d(parent, wxID_ANY,
              v.bbox.offset.z, v.bbox.size.z,
              v.bbox.offset.y, v.bbox.size.y,
              v.bbox.offset.x, v.bbox.size.x,
              v.vis3d.offset.z, v.vis3d.size.z,
              v.vis3d.offset.y, v.vis3d.size.y,
              v.res.z, v.res.y,
              v.cut.z, v.cut.y, v.cut.x,
              v.mstep.z, v.mstep.y, v.cstep.x,
              v.rev_axis[0], v.rev_axis[1], v.rev_axis[2],
				  ii) {};

          virtual v4d_ProjectionType thisWin() { return pt_YZ; };
          virtual v4d_ProjectionType otherWins() { return pt_ButYZ; };
          virtual v4d_ProjectionType verticalWins() { return pt_ButXZ; };
          virtual v4d_ProjectionType horizontalWins() { return pt_ButXY; };
          virtual i3d::Vector3d<float> Reorder(const i3d::Vector3d<float> &v) const
          { return i3d::Vector3d<float>(v.z, v.y, v.x); };
    };
}

#endif
