/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * v4d_frame.h
 *
 * 3d image viewer frame
 *
 * Petr Matula (pem@fi.muni.cz) 2006
 */
#ifndef V4D_V4D_FRAME_H
#define V4D_V4D_FRAME_H

#include <stack>
#include <wx/frame.h>
#include <i3d/image3d.h>
#include <i3d/regions.h>
#include "v4d_impexp.h"
#include "types/evt_mapping.h"
#include "types/evt_overlaying.h"
#include "overlaying.h"
#include "dlg_levels.h"
#include "v4d_image4d.h"
#include "v4d_labeledimage4d.h"

namespace v4d 
{
    ///////////////////////////////////////////////////////////
    //                                                       //
    //             Menu and controls identifiers             //
    //                                                       //
    ///////////////////////////////////////////////////////////
    enum
    {
        // menu items
        SI_ZOOM_IN = wxID_HIGHEST,
        SI_ZOOM_OUT,
        SI_NORMAL_ZOOM,

        SI_SM_PLAY_STOPNOW,
		  SI_SM_SETRANGE,
        SI_SM_FPS_INCREASE,
        SI_SM_FPS_DECREASE,
        SI_SM_STEP_INCREASE,
        SI_SM_STEP_DECREASE,

        SI_CUT,
        SI_AXIS,
        SI_WINDOW,
        SI_SAVE_3D_DATA,
        SI_SAVE_VIEW,
        SI_SAVE_GALLERY,
        SI_ADJUST_FRAME_SIZE,
        SI_AUTO_ADJUST_FRAME_SIZE,
        SI_VISUALIZE_VOLUME,

        // put identifiers of commands requiring menu updates between 
        // FIRST and LAST ids
        SI_FIRST_UPDATE_UI,

        SI_SM_PLAY_FORWARD,
        SI_SM_PLAY_BACKWARD,
        SI_SM_PLAY_LOOP,
        SI_SM_PLAY_OSCILLATION,

        SI_TOGGLE_XY,
        SI_TOGGLE_XZ,
        SI_TOGGLE_YZ,
        SI_AF,
		  SI_SHOW_WINDOWS_ANNOTATIONS,
        SI_SHOW_INFO_WIN,
        SI_REVERSE_X,
        SI_REVERSE_Y,
        SI_REVERSE_Z,
        SI_AUTOFOCUS_XY,
        SI_AUTOFOCUS_XZ,
        SI_AUTOFOCUS_YZ,
        SI_AUTOFOCUS_XYZ,
        SI_REFRESH_ICON,
        SI_LAST_UPDATE_UI,
        SI_CLOSE_WINDOW,
        SI_LAST_MENU_ID,
    };

    class v4d_Viewer3dCanvas;
    class v4d_ExtensionModule;

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                        Viewer3d Frame                       //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    /** Viewer3d frame. This class has a menu, status bar,
        pointer to the main canvas (see Viewer3dCanvas) and in the future
        it will have a toolbar. When this frame is destroyed, the
        application exits. */
    class V4D_DLLEXPORT v4d_Viewer3dFrame 
        : public wxFrame
    {
    private:
        /** Identifier that was last assigned to a menu item. This variable is
            incremented when new menu items are added by extension modules. */
        int m_last_menu_id;
        /** Info about which extensions should be deleted by the frame. */
        std::stack<bool> m_delete_extension;
        /** Pointer to the canvas object. */
        v4d_Viewer3dCanvas *m_canvas;

    public:
        /** Default constructor. It creates and initializes the main
            window. \param title Title of the main window. */
        v4d_Viewer3dFrame(wxWindow* parent, wxWindowID id, const wxString& title, 
            const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxDefaultSize, 
            long style = wxDEFAULT_FRAME_STYLE);
        /** Destructor  */
        ~v4d_Viewer3dFrame();
        /** Add an extension module to the frame. */
        void AddExtensionModule(v4d_ExtensionModule *module, const wxString &menuname, bool delete_module = true);
        /** Append menu to the frame menu bar. This function creates a copy of the given menu. */
        void AppendMenu(const wxMenu *menu, const wxString &menu_name);
        /** Assign a different canvas. This allows assignment of a canvas derived 
        from v4d_Viewer3dCanvas to the frame.  */
        void SetCanvas(v4d_Viewer3dCanvas *canvas);
        /** Call OnAdjustFrameSize if the menu item SI_AUTO_ADJUST_FRAME_SIZE is checked 
            in MenuBar. This is called after zoom events. It is good to call this function 
            also after setting a new canvas.*/
        void ResizeFrameIfRequired();
        /** Refresh frame icon. */
        void RefreshIcon();
        /** Get pointer to canvas assigned to the frame */
        v4d_Viewer3dCanvas *GetCanvas() 
        { 
            return m_canvas; 
        }

    private:
        /** Function called when a user wants to show or hide a projection. */
        void OnToggleProjection(wxCommandEvent& event);
        /** Function for changing screen options. */
        void OnPreferences(wxCommandEvent& event);
        /** Function called when user wants to reverse axis orientation */
        void OnReverseAxis (wxCommandEvent& event);
        /** Function called when a user wants to switch on or switch off
        the autofocus mode. */
        void OnAutoFocus(wxCommandEvent& event);
        /** Function called when a user wants to zoom in. */
        void OnZoomIn(wxCommandEvent& event);
        /** Function called when a user wants to zoom out. */
        void OnZoomOut(wxCommandEvent& event);
        /** Function called when a user wants to set the scale 1:1. */
        void OnNormalZoom(wxCommandEvent& event);
        /** Function called on show axes annotation menu command. */
        void OnShowWindowsAnnotation(wxCommandEvent& evt);
        /** Function called on show info menu command. */
        void OnShowInfo(wxCommandEvent& evt);
        /** Save original 3d image data. */  
        void OnSave3dData(wxCommandEvent &event);
        /** Save actual 3d view for publication purposes. */  
        void OnSaveView(wxCommandEvent &event);
        /** Save slices of 3d data in chosen increments */
        void OnSaveGalleryView(wxCommandEvent &event);
        /** Adjust frame size to see maximal part of the bounding box */
        void OnAdjustFrameSize(wxCommandEvent& evt);
        /** Show 3D volume visualization window */
        void OnVisualizeVolume(wxCommandEvent& evt);
        /** Update menu according to frame state */
        void OnUpdateUI(wxUpdateUIEvent &evt);
        /** Mapping change handler. */
        void OnMappingChanged(v4d_MappingEvent &evt);
        /** Overlaying change handler. */
        void OnOverlayingChanged(v4d_OverlayingEvent &evt);
        /** Move canvas info frame on move events. */
        void OnMove(wxMoveEvent &evt);

		  /** Catches all 'Sequence' menu related commands and dispatches them
			* to the m_canvas because this guy has the v4d::ImageIterator object
			* that is actually the receiver of the commands...
			*/
		  void OnSequenceMenuCmd(wxCommandEvent& evt);

        /** Reset canvas to correspond to the current data. This function must be called
            after size changes of displayed image*/
        void Reset();

	 public:
		  /// allows one to set status \e text that vanishes after \e displayTime seconds
		  void SetInfoStatusText(const wxString& text,const int displayTime=5)
		  		{
					wxFrame::SetStatusText(text,1);

					//if the timer is already active, it gets reseted with this new one
					m_status_cleaner.Start(displayTime*1000,wxTIMER_ONE_SHOT);
				}
		  /// allows one to clear the status text 
		  void ClearInfoStatusText(void)
		  		{ wxFrame::SetStatusText("",1); }

    private:
		  class MyStatusTimer : public wxTimer
		  {
		    public:
		      MyStatusTimer(v4d_Viewer3dFrame* fr) : frame(fr) {};
		      virtual void Notify() { frame->ClearInfoStatusText(); };
		    private:
		      v4d_Viewer3dFrame* frame;
		  };
		  MyStatusTimer m_status_cleaner;

        DECLARE_EVENT_TABLE()
    };

	template <class T>
    v4d_Viewer3dFrame *v4d_CreateImage3dFrame(
		i4d::Image4d<T> *images,
		bool delete_image = true, 
        wxWindow *parent = NULL, 
        wxWindowID id = -1, 
        const wxString& title = "4D viewer", 
        const wxPoint& pos = wxDefaultPosition, 
        const wxSize& size = wxDefaultSize,
        long style = wxDEFAULT_FRAME_STYLE
        );

    /** Create a frame with v4d_LabeledImageViewer3dCanvas class. If some problems occurs then 
        v4d_OverlayingModuleException is raised. */
    template <class IDATA, class LABEL, class LDATA> 
    v4d_Viewer3dFrame *v4d_CreateLabeledImage3dFrame(
        i4d::Image4d<IDATA> *img, 
        i4d::LabeledImage4d<LABEL, LDATA> *limg,
        bool delete_image = true, 
        bool delete_lab_image = true, 
        wxWindow *parent = NULL, 
        wxWindowID id = -1, 
        const wxString &title = "3D viewer", 
        const wxPoint &pos = wxDefaultPosition, 
        const wxSize &size = wxDefaultSize, 
        long style = wxDEFAULT_FRAME_STYLE
        );
}

#endif
