/*
 * i4dcore library - 3D+t image representation library 
 *
 * Copyright (C) 2005-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * v4d_image4d.cc
 *
 * Functions for creating and manipulating time-lapse images.
 *
 * Vladimir Ulman (ulman@fi.muni.cz) 2007 and 2015
 */

#ifdef __GNUG__
#pragma interface "v4d_image4d.h"
#endif

#include <math.h>

/* MS Visual Studio doesn't know roundf() from math.h */
#ifdef _MSC_VER
#define roundf(x)	floorf((x)+0.5f)
#endif


#include <i3d/image3d.h>
#include "v4d_image4d.h"
#include <iostream>
#include "v4d_impexp.h"
#include <wx/log.h>
#include <wx/msgdlg.h>

using namespace i3d;

namespace i4d
{

//
// ************ generic (scalar) i4d::Image4d methods ************
//
//copy constructor
template <class VOXEL>
Image4d<VOXEL>::Image4d(const Image4d &seq, const i3d::VOI<i3d::PIXELS> *voi) :
	offset(), resolution(), size(), frames(0), timeStep(0),
	data(), names(), extremes()
{
	//we're in constructor: no data is present in the *this Image4d

	if (seq.TestConsistency())
		throw InternalException("Image4d, copy constructor: Source image is not consistent!");

	offset=seq.GetOffset();
	resolution=seq.GetResolution();
	timeStep=seq.GetTimeStep();

	if (voi) {
		size=voi->size;
		offset+=PixelsToMicrons(voi->offset,resolution);
	} else {
		size=seq.GetSize();
	}

	//since each frame in the source seq has the same metadata as its "parent" Image4d,
	//we may safely use the constructor() function "per partes" and we can be sure that
	//the newly created images will also have the same metadata (owing to the constructor())
	typename std::vector< Image3d<VOXEL>* >::const_iterator iter=seq.GetVoxelData().begin();
	while (iter != seq.GetVoxelData().end()) {
		Image3d<VOXEL> *im=new Image3d<VOXEL>(**iter,voi);
		data.push_back(im);

		names.Add(seq.GetImageName(frames));
		++frames;

		//update the (cached) min and max values
		if (frames == 1) extremes.Reset(*im);
		else extremes.Update(*im);

		++iter;
	}
}


template <class VOXEL>
void Image4d<VOXEL>::CopyFromVOI(const Image4d<VOXEL> &seq, const i3d::VOI<i3d::PIXELS> &voi)
{
	if (seq.TestConsistency())
		throw InternalException("Image4d, copy from VOI: Source image is not consistent!");

	offset=seq.GetOffset() + PixelsToMicrons(voi.offset, resolution);
	resolution=seq.GetResolution();
	timeStep=seq.GetTimeStep();
	size=voi.size;

	//since each frame in the source seq has the same metadata as its "parent" Image4d,
	//we may safely use the CopyFromVOI() function "per partes" and we can be sure that
	//the newly created images will also have the same metadata (owing to the function)
	typename std::vector< Image3d<VOXEL>* >::iterator d_iter=data.begin();
	typename std::vector< Image3d<VOXEL>* >::const_iterator s_iter=seq.GetVoxelData().begin();
	frames=0;
	while ((d_iter != data.end()) && (s_iter != seq.GetVoxelData().end())) {
		(*d_iter)->CopyFromVOI(**s_iter,voi);
		SetImageName(frames,seq.GetImageName(frames));
		++frames;

		//update the (cached) min and max values
		if (frames == 1) extremes.Reset(**d_iter);
		else extremes.Update(**d_iter);

		++d_iter; ++s_iter;
	}
	if (s_iter != seq.GetVoxelData().end())
	{
		//d_iter must be equal to data.end(), hence
		//there is left something to CopyFromVOI in the input seq 
		Image3d<VOXEL> *im=new Image3d<VOXEL>(**s_iter,voi);
		data.push_back(im);
		names.Add(seq.GetImageName(frames));
		++frames;

		//update the (cached) min and max values
		if (frames == 1) extremes.Reset(*im);
		else extremes.Update(*im);

		++s_iter;
	}
	else
	{
		//s_iter must be equal to seq.data.end(), hence
		//there is left something to remove from this sequence
		while (d_iter != data.end()) {
			delete *d_iter;
			*d_iter = NULL;
			++d_iter;
		}
		data.resize(frames);
		names.resize(frames);
	}
}


template <class VOXEL>
void Image4d<VOXEL>::ReadSequence(const wxArrayString &Names, const bool sourceIsBinary)
{
	this->ReadSequence_First(Names,sourceIsBinary);
	while (this->frames < Names.size())
		this->ReadSequence_Next(Names,sourceIsBinary);
}

template <class VOXEL>
void Image4d<VOXEL>::ReadSequence_First(const wxArrayString &Names, const bool sourceIsBinary)
{
	if (Names.empty())
		throw InternalException("ReadSequence(): array Names cannot be empty!");

	//clear any content this sequence image may have had...
	this->DestroyData();
	//frames == 0 due to the DestroyData()

	//read the first image
	wxLogMessage(wxString("I_Reading: ")+Names[this->frames]);
	Image3d<VOXEL>* im=new Image3d<VOXEL>;
	if (sourceIsBinary)
	{
		//opening a binary image and converting to the target voxel type
		Image3d<bool>* bim=new Image3d<bool>(Names[this->frames].mb_str(wxConvFile));
		CreateBinaryScalarImage(*bim,*im);
		delete bim;
	}
	else
	{
		//standard opening of the image file
		im->ReadImage(Names[this->frames].mb_str(wxConvFile));
		ReplaceInvalidFloats(*im);
	}

	//update the metadata of this Image4d
	this->offset=im->GetOffset();
	this->resolution=im->GetResolution();
	this->size=im->GetSize();
	this->data.push_back(im);
	this->names.Add(Names[this->frames]);
	this->frames=1;
	this->extremes.Reset(*im);
}

template <class VOXEL>
void Image4d<VOXEL>::ReadSequence_Next(const wxArrayString &Names, const bool sourceIsBinary)
{
	if (Names.empty())
		throw InternalException("ReadSequence(): array Names cannot be empty!");

	wxLogMessage(wxString("I_Reading: ")+Names[this->frames]);
	Image3d<VOXEL>* im=new Image3d<VOXEL>;
	if (sourceIsBinary)
	{
		//opening a binary image and converting to the target voxel type
		Image3d<bool>* bim=new Image3d<bool>(Names[this->frames].mb_str(wxConvFile));
		CreateBinaryScalarImage(*bim,*im);
		delete bim;
	}
	else
	{
		//standard opening of the image file,
		//also checks consistency on voxel type
		im->ReadImage(Names[this->frames].mb_str(wxConvFile));
		ReplaceInvalidFloats(*im);
	}
	
	//check consistency on metadata
	if ( (im->GetOffset() != this->offset)
		|| (im->GetResolution().GetRes() != this->resolution.GetRes())
		|| (im->GetSize() != this->size) )
		throw InternalException("ReadSequence(): The frame metadata differs from the sequence metadata!");

	this->data.push_back(im);
	this->names.Add(Names[this->frames]);
	++(this->frames);

	//update the (cached) min and max values
	this->extremes.Update(*im);
}


template <class VOXEL>
void Image4d<VOXEL>::SaveSequence(const char *NamePrefix,const int Width,const char *NameSuffix,
				 const int StartIndex,const int StopIndex)
{
	if (StopIndex < StartIndex)
		throw InternalException("SaveSequence(): StopIndex cannot be smaller than StartIndex!");

	if (this->TestConsistency())
		throw InternalException("SaveSequence(): The image is not consistent!");

	for (int i=StartIndex; i <= StopIndex; ++i) {
		//construct the filename
		wxString fn(
			wxString::FromUTF8(NamePrefix) + 
			wxString::Format( wxString::Format("%%0%dd",Width) ,i ) +
			wxString::FromUTF8(NameSuffix)
		);

		//report it
		//std::cout << "Saving frame no. " << i << " into file " << fn.str() << std::endl;
		wxLogMessage(wxString("Saving: ")+fn);

		//save it
		(this->data[i])->SaveImage(fn.mb_str(wxConvFile));
	}
}


template <class VOXEL>
void Image4d<VOXEL>::InsertFrame(i3d::Image3d<VOXEL>* const img,
		const wxString &name, const int time)
{
	//sanity internal test
	if ( (frames != data.size()) || (frames != names.size()) )
		throw InternalException("InsertFrame(): Image4d metadata is not consistent.");

	if (img == NULL)
		throw InternalException("InsertFrame(): NULL image pointer given.");

	//test the position
	if ((time < -1) || (time >= (int)frames))
		throw InternalException("InsertFrame(): Incorrect value of the time parameter.");

	//test metadata of the inserted image
	if ( (img->GetSize() != size)
		|| (img->GetOffset() != offset)
		|| (img->GetResolution().GetRes() != resolution.GetRes()) )
		throw InternalException("InsertFrame(): Inserted image is not consistent with the sequence!");

	if (time == -1)
	{
		data.push_back(img);
		names.Add(name);
		++frames;
	}
	else
	{
		//find the frame to insert before
		typename std::vector< Image3d<VOXEL>* >::iterator d_iter=data.begin();
		int i=time;
		while (i > 0) { ++d_iter; --i; }

		//insert the frame into the sequence
		data.insert(d_iter,img);
		names.Insert(name, time);
		++frames;
	}

	//update the (cached) min and max values
	if (frames == 1) extremes.Reset(*img);
	else extremes.Update(*img);
}


template <class VOXEL>
void Image4d<VOXEL>::InsertFrame(const i3d::Image3d<VOXEL>& img,
		const wxString &name,const int time)
{
	//make a dynamically allocated copy of the input image
	Image3d<VOXEL>* i=new Image3d<VOXEL>(img);
	this->InsertFrame(i,name,time);
}


template <class VOXEL>
void Image4d<VOXEL>::InsertFrame(const wxString &name,const int time)
{
	if (name.IsEmpty())
		throw InternalException("InsertFrame(): Empty image name given!");

	Image3d<VOXEL> *img=new Image3d<VOXEL>(name.mb_str(wxConvFile));
	this->InsertFrame(img,name,time);
}


template <class VOXEL>
void Image4d<VOXEL>::RemoveFrame(const int time)
{
	//sanity internal test
	if ( (frames != data.size()) || (frames != names.size()) )
		throw InternalException("InsertFrame(): Image4d metadata is not consistent.");

	//test the position
	if ((time < 0) || (time >= (int)frames))
		throw InternalException("InsertFrame(): Incorrect value of the time parameter.");

	//find the frame to be removed
	typename std::vector< Image3d<VOXEL>* >::iterator d_iter=data.begin();
	int i=time;
	while (i > 0) { ++d_iter; --i; }

	//should we update the (cached) min and max values - decision is later
	/* SAFER version
	VOXEL min,max;
	FindMinMax(**d_iter,min,max);
	*/

	//first, remove the image from memory
	delete *d_iter;

	//second, remove the reference
	data.erase(d_iter);
	names.RemoveAt(time);
	--frames;

	if (frames == 0) extremes.Reset();
	/*
	else if ((min <= minValue) || (max >= maxValue))
		  	//bad, the removed image was realizing either min or max value,
			//we therefore need to search for new ones...
				this->UpdateMinMaxValues();
		//THIS IS DANGEROUS: it can also be the case that raw-image data
		//is not realizing global min or max, but it may realize it for
		//some derived values!
		//
		//also: min <= minValue should be written differently for RGB...
	*/
	else this->UpdateMinMaxValues(); //SAFER version
	//SAFER version requires not to rewrite this function for different voxel types
}


//
// ************ FindMinMax() ************
//
template <class VOXEL>
void FindMinMax(const Image3d<VOXEL>& im,VOXEL& min,VOXEL& max)
{
	//sweeping pointers
	const VOXEL* p=im.GetFirstVoxelAddr();
	const VOXEL* pL=p+im.GetImageSize();

	//look for extremes
	min=max=*p;
	while (p != pL)
	{
		if      (*p < min) min=*p;
		else if (*p > max) max=*p;
		++p;
	}
}


template <typename T>
void FindMinMaxRGB(const Image3d<i3d::RGB_generic<T> >& im,
		i3d::RGB_generic<T>& min,i3d::RGB_generic<T>& max)
{
	//sweeping pointers
	const i3d::RGB_generic<T>* p=im.GetFirstVoxelAddr();
	const i3d::RGB_generic<T>* pL=p+im.GetImageSize();

	//look for extremes
	min=max=*p;
	while (p != pL)
	{
		if      (p->red   < min.red  ) min.red  =p->red  ;
		else if (p->red   > max.red  ) max.red  =p->red  ;

		if      (p->green < min.green) min.green=p->green;
		else if (p->green > max.green) max.green=p->green;

		if      (p->blue  < min.blue ) min.blue =p->blue ;
		else if (p->blue  > max.blue ) max.blue =p->blue ;

		++p;
	}
}


template <typename T>
void FindMinMax(const Image3d<std::complex<T> >& im,
		std::complex<T>& min,std::complex<T>& max,
		T& minM, T& maxM, T& minP, T& maxP)
{
	//sweeping pointers
	const std::complex<T>* p=im.GetFirstVoxelAddr();
	const std::complex<T>* pL=p+im.GetImageSize();

	//init extremes
	min=max=*p;
	minM=maxM=sqrt(p->real()*p->real() + p->imag()*p->imag());
	minP=maxP=std::atan2(p->imag(),p->real());

	//time-savers...
	T minR=min.real(); T minI=min.imag();
	T maxR=max.real(); T maxI=max.imag();

	//look for extremes
	while (p != pL)
	{
		//raw data
		if      (p->real() < minR) minR=p->real();
		else if (p->real() > maxR) maxR=p->real();

		if      (p->imag() < minI) minI=p->imag();
		else if (p->imag() > maxI) maxI=p->imag();

		//derived data
		const T mag=sqrt(p->real()*p->real() + p->imag()*p->imag());
		const T ph=std::atan2(p->imag(),p->real());

		if      (mag < minM) minM=mag;
		else if (mag > maxM) maxM=mag;

		if      (ph < minP) minP=ph;
		else if (ph > maxP) maxP=ph;

		++p;
	}

	//commit the time-savers...
	min=std::complex<T>(minR,minI);
	max=std::complex<T>(maxR,maxI);
}


/*
 * Explicit Instantiations
 */
template class V4D_DLLEXPORT Image4d<GRAY8>;
template class V4D_DLLEXPORT Image4d<GRAY16>;
template class V4D_DLLEXPORT Image4d<int>;
template class V4D_DLLEXPORT Image4d<size_t>;
template class V4D_DLLEXPORT Image4d<float>;
template class V4D_DLLEXPORT Image4d<double>;
template class V4D_DLLEXPORT Image4d<RGB>;
template class V4D_DLLEXPORT Image4d<RGB16>;
template class V4D_DLLEXPORT Image4d<std::complex<float> >;
template class V4D_DLLEXPORT Image4d<std::complex<double> >;


//
// ************ "fiend" conversion functions ************
//
template <class VIN, class VOUT> void FloatToGray(const Image4d <VIN> &fimg, Image4d <VOUT> &gimg)
{
	//resize (and possibly destroy any current data)
	gimg.MakeRoom(fimg.GetSize(),fimg.GetNumFrames());
	gimg.SetOffset(fimg.GetOffset());
	gimg.SetResolution(fimg.GetResolution());
	gimg.SetTimeStep(fimg.GetTimeStep());

	//since each frame in the source fimg has the same metadata as its "parent" Image4d,
	//we may safely use the FloatToGray() function "per partes" and we can be sure that
	//the newly converted images will also have the same metadata (owing to FloatToGray())
	typename std::vector< Image3d<VIN>* >::const_iterator iter=fimg.GetVoxelData().begin(); //source
	typename std::vector< Image3d<VOUT>* >::iterator d_iter=gimg.GetVoxelData().begin(); //source
	size_t cnt=0;
	while (iter != fimg.GetVoxelData().end()) {
		FloatToGray(**iter,**d_iter);
		gimg.SetImageName(cnt,fimg.GetImageName(cnt));

		++d_iter; ++iter;
		++cnt;
	}
}


template <class VIN, class VOUT> void FloatToGrayNoWeight(const Image4d <VIN> &fimg, Image4d <VOUT> &gimg)
{
	//resize (and possibly destroy any current data)
	gimg.MakeRoom(fimg.GetSize(),fimg.GetNumFrames());
	gimg.SetResolution(fimg.GetResolution());
	gimg.SetOffset(fimg.GetOffset());
	gimg.SetTimeStep(fimg.GetTimeStep());

	typename std::vector< Image3d<VIN>* >::const_iterator iter=fimg.GetVoxelData().begin(); //source
	typename std::vector< Image3d<VOUT>* >::iterator d_iter=gimg.GetVoxelData().begin(); //source
	size_t cnt=0;
	while (iter != fimg.GetVoxelData().end()) {
		FloatToGrayNoWeight(**iter,**d_iter);
		gimg.SetImageName(cnt,fimg.GetImageName(cnt));

		++d_iter; ++iter;
		++cnt;
	}
}


template <class VIN, class VOUT> void GrayToFloat(const Image4d <VIN> &gimg, Image4d <VOUT> &fimg)
{
	//resize (and possibly destroy any current data)
	fimg.MakeRoom(gimg.GetSize(),gimg.GetNumFrames());
	fimg.SetResolution(gimg.GetResolution());
	fimg.SetOffset(gimg.GetOffset());
	fimg.SetTimeStep(gimg.GetTimeStep());

	typename std::vector< Image3d<VIN>* >::const_iterator iter=gimg.GetVoxelData().begin(); //source
	typename std::vector< Image3d<VOUT>* >::iterator d_iter=fimg.GetVoxelData().begin(); //source
	size_t cnt=0;
	while (iter != gimg.GetVoxelData().end()) {
		GrayToFloat(**iter,**d_iter);
		fimg.SetImageName(cnt,gimg.GetImageName(cnt));

		++d_iter; ++iter;
		++cnt;
	}
}


template <class VOUT>
void CreateBinaryScalarImage(const i3d::Image3d<bool>& bimg, i3d::Image3d<VOUT>& img)
{
	//prepare the output image
	img.CopyMetaData(bimg);

	//sweeping pointers...
	const bool* inP=bimg.GetFirstVoxelAddr();
	const bool* const inPstop=inP+bimg.GetImageSize();
	VOUT* outP=img.GetFirstVoxelAddr();

	//the conversion...
	while (inP != inPstop)
	{
		*outP = (*inP)? 255:0;
		++inP; ++outP;
	}
}


//the function is empty (= does nothing) for "any" type
template <class VIN>
void ReplaceInvalidFloats(i3d::Image3d<VIN>& img)
{ }

//it does something, however, for floating-point voxel types
template <>
void ReplaceInvalidFloats(i3d::Image3d<float>& img)
{
	//problem indicator
	bool has_nan = false;

	//sweeping pointers...
	float* inP=img.GetFirstVoxelAddr();
	float* const inPstop=inP+img.GetImageSize();

	//the testing...
	while (inP != inPstop)
	{
#ifdef _MSC_VER
		if (!_finite(*inP))
#else
		if (!std::isfinite(*inP))
#endif
		{
			*inP = 0;
			has_nan = true;
		}
		++inP;
	}

	if (has_nan)
	{
		wxLogWarning("Invalid float values replaced by zeros.");
		wxMessageBox("The image contained invalid values (+/- infinity, NaN). These values were replaced by zeroes.", "Warning");
	}
}

template <typename F>
void ReplaceInvalidFloats(i3d::Image3d<std::complex<F> >& img)
{
	//problem indicator
	bool has_nan = false;

	//sweeping pointers...
	std::complex<F>* inP=img.GetFirstVoxelAddr();
	std::complex<F>* const inPstop=inP+img.GetImageSize();

	//the testing...
	while (inP != inPstop)
	{
		//makes no sense to repair individual elements (real or image) to me
		//as the original number (as well as its phase and magnitude) is lost anyway
#ifdef _MSC_VER
		if ( (!_finite(inP->real())) || (!_finite(inP->imag())) )
#else
		if ( (!std::isfinite(inP->real())) || (!std::isfinite(inP->imag())) )
#endif
		{
			*inP = std::complex<F>(0,0);
			has_nan = true;
		}
		++inP;
	}

	if (has_nan)
	{
		wxLogWarning("Invalid float values replaced by zeros.");
		wxMessageBox("The image contained invalid values (+/- infinity, NaN). These values were replaced by zeroes.", "Warning");
	}
}


/*
 * Explicit Instantiations
 */
template void FloatToGray(const Image4d <float> &fimg, Image4d <GRAY8> &gimg);
template void FloatToGray(const Image4d <float> &fimg, Image4d <GRAY16> &gimg);
template void FloatToGray(const Image4d <double> &fimg, Image4d <GRAY8> &gimg);
template void FloatToGray(const Image4d <double> &fimg, Image4d <GRAY16> &gimg);
template void FloatToGrayNoWeight(const Image4d <float> &fimg, Image4d <GRAY8> &gimg);
template void FloatToGrayNoWeight(const Image4d <float> &fimg, Image4d <GRAY16> &gimg);
template void FloatToGrayNoWeight(const Image4d <double> &fimg, Image4d <GRAY8> &gimg);
template void FloatToGrayNoWeight(const Image4d <double> &fimg, Image4d <GRAY16> &gimg);
template void GrayToFloat(const Image4d <GRAY8> &gimg, Image4d <float> &fimg);
template void GrayToFloat(const Image4d <GRAY16> &gimg, Image4d <float> &fimg);
template void GrayToFloat(const Image4d <GRAY8> &gimg, Image4d <double> &fimg);
template void GrayToFloat(const Image4d <GRAY16> &gimg, Image4d <double> &fimg);

/*
template void CreateBinaryScalarImage(const i3d::Image3d<bool>& bimg, i3d::Image3d<i3d::GRAY8>& img);
template void ReplaceInvalidFloats(i3d::Image3d<i3d::GRAY8>& img);
template void ReplaceInvalidFloats(i3d::Image3d<i3d::GRAY16>& img);
template void ReplaceInvalidFloats(i3d::Image3d<i3d::RGB>& img);
template void ReplaceInvalidFloats(i3d::Image3d<i3d::RGB16>& img);
template void ReplaceInvalidFloats(i3d::Image3d<std::complex<float> >& img);
*/
}
