/*
 * i4dcore library - 3D+t image representation library 
 *
 * Copyright (C) 2005-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * v4d_image4d.h
 *
 * Functions for creating and manipulating time-lapse images.
 *
 * Vladimir Ulman (ulman@fi.muni.cz) 2007 and 2015
 */

#ifndef __TIMELAPSEIMAGE__
#define __TIMELAPSEIMAGE__

#ifdef __GNUG__
#pragma interface "v4d_image4d.h"
#endif


#include <i3d/basic.h>
#include <i3d/vector3d.h>
#include <i3d/voi.h>
#include <i3d/resolution.h>
#include <i3d/image3d.h>
#include <wx/arrstr.h>
#include <valarray>
#include <vector>


/** \page libi4d 2D+t and 3D+t image representation library
 *
 * \section i4dintro Overview
 * This library provides a representation of 4D images. Clearly, it also handles
 * 1D, 2D and 3D images. The need for 4D images comes from the need to store 
 * time-lapse sequences of 3D images. Hence, we refer to such 3D image often as
 * the \e frame and to the whole time-lapse sequence as the \e sequence \e image.
 *
 * The 4D image is governed by the class <tt>Image4d\<VOXEL\></tt>. It aims to provide
 * the same support via its methods as <tt>i3d::Image3d\<VOXEL\></tt> does for 3D images.
 * In fact, this class is more or less a wrapper to \c Image3d owing to the memory model
 * employed. It also provides means to load/store 3D image sequence.
 *
 * Technically, 3D sequence image (= 4D image) is stored in a \b std::vector of \b i3d::Image3d.
 * That implies that separate frames lay in the memory separately while voxels within each
 * frame lay in a single continuous block. This setup allows for swaping some frames out of the
 * main memory as well as for high utilization of caches when working with single frame.
 *
 * \section Image4D The 3D+t sequence image
 * The 3D "sequence" image is a vector of <tt>i3d::Image3d\<VOXEL\></tt>. It provides almost all
 * methods, apropriately modified, that one may find in the class <tt>i3d::Image3d</tt>. The 3D sequence
 * image is implemented in the class \c Image4d whose documentation is the best source for details.
 *
 * \section Fileformats
 * The 3D sequence image, which is a vector of 3D images, is stored as a set of files supported
 * by \b i3dcore. These files obey some specified naming convention. This is explained further
 * in the documentaion of the function Image4d::ReadSequence().
 *
 * \section i4ddal Download and license information
 * The library is called \b i4dcore. It is a property of the CBIA.
 *
 * It is licensed under \ref gnulicense. It can be downloaded from the CBIA web pages:
 * http://cbia.fi.muni.cz/projects/optical-flow-downloads.html .
 *
 * \section i4ddeps Software dependencies
 * This library is written in pure C++. It is directly dependent on the \b i3dcore
 * ( http://cbia.fi.muni.cz/user_dirs/i3dlib_doc/i3dcore/ ) and libraries i3dcore further depends on.
 * \htmlonly
 * See <a href="http://cbia.fi.muni.cz/user_dirs/xulman/SW_roadmap.png">the Software map</a>.
 * \endhtmlonly
 *
 * \section i4dauthor Authors
 * This library was developed and is maintained by Vladimír Ulman <ulman@fi.muni.cz>
 * originally in 2007 and updated in 2015.
 */


/** \defgroup tlimain Timelapse Image Representations
 *
 * This module deals with \ref tli3D "2D+time" or \ref tli4D "3D+time" images.
 * The voxel intensity type may be i3d::GRAY8, i3d::GRAY16, float or double,
 * i3d::RGB, i3d::RGB16, or std::complex<float>.
 * Most of the functions in this module are instantiated for these voxel types.
 *
 * Timelapse image modificating functions belong here as well.
 *
 * \note These images are often termed "sequence images" throughout the documentation.
 */


namespace i4d
{
	//
	// -------------- 3D sequence images -------------- 
	//
	/** \defgroup tli4D 3D+time image handlers
	 * \ingroup tlimain
	 *
	 * 3D+time time-lapse images are represented by Image4d.
	 */
	// @{

	/// general helper function for calculating extremes...
	template <class VOXEL>
	void FindMinMax(const i3d::Image3d<VOXEL>& im,VOXEL& min,VOXEL& max);

	/// specialized helper function for calculating RGB extremes...
	template <typename T>
	void FindMinMaxRGB(const i3d::Image3d<i3d::RGB_generic<T> >& im,
			i3d::RGB_generic<T>& min,i3d::RGB_generic<T>& max);

	/// specialized helper function for calculating complex extremes...
	template <typename T>
	void FindMinMax(const i3d::Image3d<std::complex<T> >& im,
			std::complex<T>& min,std::complex<T>& max,
			T& minM, T& maxM, T& minP, T& maxP);


	/// class for holding extremas of scalar images, incl. modifier methods
	template <class VOXEL>
	class ExtremaValues {
	public:
		ExtremaValues(void) { Reset(); }

		void Reset(void)
		{
			minValue = std::numeric_limits<VOXEL>::max();
			maxValue = std::numeric_limits<VOXEL>::min();
		}

		void Reset(const i3d::Image3d<VOXEL>& im)
			{ FindMinMax(im,minValue,maxValue); }

		void Update(const i3d::Image3d<VOXEL>& im)
		{
			VOXEL min,max;
			FindMinMax(im,min,max);
			if (min < minValue) minValue=min;
			if (max > maxValue) maxValue=max;
		}

		VOXEL minValue, maxValue;
	};

	/// class for holding extremas of RGB images, incl. modifier methods
	template <typename T>
	class ExtremaValues<i3d::RGB_generic<T> > {
	public:
		ExtremaValues(void) { Reset(); }

		void Reset(void)
		{
			minValue.red = minValue.green
					= minValue.blue = std::numeric_limits<T>::max();
			maxValue.red = maxValue.green
					= maxValue.blue = std::numeric_limits<T>::min();
		}

		void Reset(const i3d::Image3d<i3d::RGB_generic<T> >& im)
			{ FindMinMaxRGB(im,minValue,maxValue); }

		void Update(const i3d::Image3d<i3d::RGB_generic<T> >& im)
		{
			i3d::RGB_generic<T> min,max;
			FindMinMaxRGB(im,min,max);

			if (min.red   < minValue.red  ) minValue.red  =min.red  ;
			if (min.green < minValue.green) minValue.green=min.green;
			if (min.blue  < minValue.blue ) minValue.blue =min.blue ;

			if (max.red   > maxValue.red  ) maxValue.red  =max.red  ;
			if (max.green > maxValue.green) maxValue.green=max.green;
			if (max.blue  > maxValue.blue ) maxValue.blue =max.blue ;
		}

		i3d::RGB_generic<T>  minValue, maxValue;
	};

	/// class for holding extremas of complex images, incl. modifier methods
	template <typename T>
	class ExtremaValues<std::complex<T> > {
	public:
		ExtremaValues(void) { Reset(); }

		void Reset(void)
		{
			minValue=std::complex<T>(std::numeric_limits<T>::max(),std::numeric_limits<T>::max());
			minM = minP = minValue.real();

			maxValue=std::complex<T>(std::numeric_limits<T>::min(),std::numeric_limits<T>::min());
			maxM = maxP = maxValue.real();
		}

		void Reset(const i3d::Image3d<std::complex<T> >& im)
			{ FindMinMax(im,minValue,maxValue,minM,maxM,minP,maxP); }

		void Update(const i3d::Image3d<std::complex<T> >& im)
		{
			std::complex<T> min,max;
			T minm,maxm;
			T minp,maxp;
			FindMinMax(im,min,max,minm,maxm,minp,maxp);

			//raw data
			if (min.real() < minValue.real()) minValue=std::complex<T>(min.real(),minValue.imag());
			if (min.imag() < minValue.imag()) minValue=std::complex<T>(minValue.real(),min.imag());

			if (max.real() > maxValue.real()) maxValue=std::complex<T>(max.real(),maxValue.imag());
			if (max.imag() > maxValue.imag()) maxValue=std::complex<T>(maxValue.real(),max.imag());

			//derived data
			if (minm < minM) minM=minm;
			if (maxm > maxM) maxM=maxm;
			if (minp < minP) minP=minp;
			if (maxp > maxP) maxP=maxp;
		}

		/// extremes of raw data
		std::complex<T>  minValue, maxValue;

		/// extremes of derived data: magnitude
		T minM,maxM;

		/// extremes of derived data: phase
		T minP,maxP;
	};


	/** \brief A class for storing sequence of 3D images.
	 * Basically, it consists of a vector of \c Image3d images where the position of the image
	 * in the vector represents the temporal position. In fact, the index to the vector is exactly
	 * the temporal (time) coordinate. An important consequence is that the entire 4D image is not
	 * layed continuously in the memory. Only each 3D image is continuous.
	 *
	 * This class tries to mimic the class \c Image3d. This class is not inherited from class \c Image3d.
	 * Most of the methods is preserved except, where appropriate, for the addition of fourth time
	 * coordinate. Some were renamed not to be confused with the-same-named methods in \c Image3d.
	 * Currently, histogram handling is not supported.
	 *
	 * \author Vladimír Ulman <xulman@fi.muni.cz>
	 * \date 2007, 2015
	 */
	template <class VOXEL>
	class Image4d {
	public:
		// constructors and destructors
		// ----------------------------
		explicit Image4d(const i3d::Offset* off_ = 0,
			const i3d::Resolution* res_ = 0,
			const float timeStep_ = 0.0f) : offset(),
			resolution(), size(), frames(0), timeStep(timeStep_),
			data(), names(), extremes()
		{
			if (off_) offset = *off_;
			if (res_) resolution = *res_;
		}

		/** Copy constructor. Only a portions, of each frame in the image sequence,
		 * corresponding to VOI are copied when VOI is set. The number of frames
		 * retains, only the dimension of frames is changed. Otherwise, if VOI == 0,
		 * the whole input image sequence is copied.
		 *
		 * VOI is in pixels and is relative to origin of image. New offset is recomputed
		 * apropriately. Offsets are simply handled, i.e. it's  supposed 1 pixel = 1 micron.
		 */
		Image4d(const Image4d<VOXEL> &seq, const i3d::VOI<i3d::PIXELS> *voi = 0);

		/** It extracts a portion, given by \e voi, of each frame from given sequence \e seq.
		 * The content of this \c Image4d is destroyed. The length of the input sequence \e seq
		 * need not be equal to this sequence. This sequence legnth will be updated accordingly.
		 */
		void CopyFromVOI(const Image4d<VOXEL> &seq, const i3d::VOI<i3d::PIXELS> &voi);

		/** Destructor: */
		~Image4d() { DestroyData(); }


		// Get image properties
		// --------------------

		/// Is the image empty? True if empty.
		bool IsEmpty(void) const			{
			return   (frames == 0) || (size.x == 0)
				|| (size.y == 0) || (size.z == 0);
		}

		/// Get the offset of this image.
		i3d::Offset& GetOffset(void) { return (offset); }
		const i3d::Offset& GetOffset(void) const { return (offset); }

		/// Get the spatial resolution of this image.
		i3d::Resolution& GetResolution(void) { return (resolution); }
		const i3d::Resolution& GetResolution(void) const { return (resolution); }

		/// Get the temporal resolution of this image.
		float GetTimeStep() const	{ return (timeStep); }


		/// Get the number of voxels in one frame.
		size_t GetFrameSize(void) const		{ return (size.x * size.y * size.z); }

		/// Get the number of voxels in one slice.
		size_t GetSliceSize(void) const		{ return (size.x * size.y); }

		/// Get the dimensions of the image frame.
		i3d::Vector3d<size_t> GetSize(void) const { return (size); }


		/// Get the width (horizontal axis) of the image.
		size_t GetWidth(void) const		{ return (size.x); }

		/// Get the height (vertical axis) of the image.
		size_t GetHeight(void) const		{ return (size.y); }

		/// Get number of slices per image.
		size_t GetNumSlices(void) const		{ return (size.z); }

		/// Get the number of frames in the sequence.
		size_t GetNumFrames(void) const		{ return (frames); }

		/// Get the width (horizontal axis) of the image.
		size_t GetSizeX(void) const		{ return (size.x); }

		/// Get the height (vertical axis) of the image.
		size_t GetSizeY(void) const		{ return (size.y); }

		/// Get number of slices per image.
		size_t GetSizeZ(void) const		{ return (size.z); }

		/// Get the number of frames in the sequence.
		size_t GetSizeT(void) const		{ return (frames); }

		/// Returns stored minimum value over all frames.
		VOXEL GetCachedMinValue(void) const		{ return (extremes.minValue); }

		/// Returns stored maximum value over all frames.
		VOXEL GetCachedMaxValue(void) const		{ return (extremes.maxValue); }

		/**
		 * Calculates this->extremes.minValue minimum value over all frames and returns
		 * it. Note that the function updates extremes.maxValue as well. One may use
		 * GetCachedMaxValue() safely afterwards to obtain current maximum value
		 * in the sequence as well.
		 */
		VOXEL GetUpdatedMinValue(void)
		{
			UpdateMinMaxValues();
			return (extremes.minValue);
		}

		/**
		 * Calculates this->extremes.maxValue maximum value over all frames and returns
		 * it. Note that the function updates extremes.minValue as well. One may use
		 * GetCachedMinValue() safely afterwards to obtain current minimum value
		 * in the sequence as well.
		 */
		VOXEL GetUpdatedMaxValue(void)
		{
			UpdateMinMaxValues();
			return (extremes.maxValue);
		}

		///Updates minValue and maxValue
		void UpdateMinMaxValues(void)
		{ 
			if (this->frames == 0) extremes.Reset();
			else
			{
				extremes.Reset(*(data[0]));
				for (size_t i=1; i < frames; ++i)
					extremes.Update(*(data[i]));
			}
		}

		/// Return read-only reference for the extremes class
		const ExtremaValues<VOXEL>& GetExtremes(void) const { return extremes; }


		// Set image properties
		// --------------------

		/// Set the offset for this image.
		void SetOffset(const i3d::Offset& off_)
		{
			offset = off_;

			typename std::vector< i3d::Image3d<VOXEL>* >::iterator iter = data.begin();
			while (iter != data.end()) {
				(*iter)->SetOffset(off_);
				++iter;
			}
		}

		/// Set the spatial resolution for this image.
		void SetResolution(const i3d::Resolution& res_)
		{
			resolution = res_;

			typename std::vector< i3d::Image3d<VOXEL>* >::iterator iter = data.begin();
			while (iter != data.end()) {
				(*iter)->SetResolution(res_);
				++iter;
			}
		}

		/// Get the temporal resolution of this image.
		void SetTimeStep(const float timeStep_) { timeStep=timeStep_; }


		// Indexing, indepent of frame number since only within Image3d
		// ------------------------------------------------------------

		/// Get the x-coordinate of the i-th voxel in the image data block.
		size_t GetX(const size_t i) const     { return (i % size.x); }

		/// Get the y-coordinate of the i-th voxel in the image data block.
		size_t GetY(const size_t i) const     { return ((i / size.x) % size.y); }

		/// Get the z-coordinate of the i-th voxel in the image data block.
		size_t GetZ(const size_t i) const     { return ((i / size.x) / size.y); }

		/// Get the position of the voxel in the image data block (first position is O).
		size_t GetInFrameIndex(const size_t x, const size_t y, const size_t z) const
			{ return (i3d::GetIndex(x, y, z, size)); }
		size_t GetInFrameIndex(const i3d::Vector3d<size_t>& v) const
			{ return (i3d::GetIndex(v.x, v.y, v.z, size)); }

		/** get the coords from the position in the image vector
		coords is Vector3d<int> type - for relative positions */
		i3d::Coords GetInFrameCoords(size_t i) const	  { return           i3d::Coords(GetX(i), GetY(i), GetZ(i)); }
		i3d::Vector3d<size_t> GetFramePos(size_t i) const { return i3d::Vector3d<size_t>(GetX(i), GetY(i), GetZ(i)); }


		// Voxel & data access, depends on the frame number
		// ------------------------------------------------
		//
		// !! ATTENTION, NO MEMORY IN-BOUND CHECKS ARE PERFORMED !!

		/** Get a voxel value (voxels are numbered from 0): */
		VOXEL GetVoxel(const size_t x, const size_t y, const size_t z, const size_t frame) const
		{
			return (data[frame]->GetVoxel(x, y, z));
		}

		/** Get a voxel value (voxels are numbered from 0) */
		VOXEL GetVoxel(const i3d::Vector3d<size_t>& v, size_t frame) const
		{
			return (data[frame]->GetVoxel(v.x, v.y, v.z));
		}

		/** Get a voxel value (voxels are numbered from 0) */
		VOXEL GetVoxel(const size_t i, const size_t frame) const
		{
			return (data[frame]->GetVoxel(i));
		}

		/** Set a voxel to value (voxels are numbered from 0) */
		void SetVoxel(const size_t x, const size_t y, const size_t z,
							const size_t frame, const VOXEL voxel)
		{
			data[frame]->SetVoxel(x, y, z, voxel);
		}

		/** Set a voxel to value (voxels are numbered from 0) */
		void SetVoxel(const i3d::Vector3d<size_t>& v,
							const size_t frame, const VOXEL voxel)
		{
			data[frame]->SetVoxel(v.x, v.y, v.z, voxel);
		}

		/** Set a voxel to value (voxels are numbered from 0) */
		void SetVoxel(const size_t i, const size_t frame, const VOXEL voxel)
		{
			data[frame]->SetVoxel(i, voxel);
		}


		/** Get a voxel address (voxels are numbered from 0) */
		VOXEL* GetVoxelAddr(const size_t x, const size_t y, const size_t z, const size_t frame)
		{
			return (data[frame]->GetVoxelAddr(x, y, z));
		}

		/** Get a voxel address (voxels are numbered from 0) */
		const VOXEL* GetVoxelAddr(const size_t x, const size_t y, const size_t z, const size_t frame) const
		{
			const typename std::valarray<VOXEL>& d = data[frame]->GetVoxelData();
			const VOXEL& v = d[i3d::GetIndex(x, y, z, size)];

			return (&v);
		}

		/** Get a voxel address (voxels are numbered from 0) */
		VOXEL* GetVoxelAddr(const size_t i, const size_t frame)
		{
			return (data[frame]->GetVoxelAddr(i));
		}

		/** Get a voxel address (voxels are numbered from 0) */
		const VOXEL* GetVoxelAddr(const size_t i, const size_t frame) const
		{
			const typename std::valarray<VOXEL>& d = data[frame]->GetVoxelData();
			const VOXEL& v = d[i];

			return (&v);
		}

		/** Get the first voxel address (voxels are numbered from 0) */
		VOXEL* GetFirstVoxelAddr(const size_t frame)
		{
			return (data[frame]->GetFirstVoxelAddr());
		}

		/** Get the first voxel address (voxels are numbered from 0) */
		const VOXEL* GetFirstVoxelAddr(const size_t frame) const
		{
			const typename std::valarray<VOXEL>& d = data[frame]->GetVoxelData();
			const VOXEL& v = d[0];

			return (&v);
		}


		/** Returns a reference to the property \e data of this \c Image4d
		 * which is of the \c std::vector< i3d::Image3d<VOXEL>* > type.
		 */
		std::vector< i3d::Image3d<VOXEL>* >& GetVoxelData(void)		{ return(data); }
		const std::vector< i3d::Image3d<VOXEL>* >& GetVoxelData(void) const	{ return(data); }


		/** Returns a direct reference to the \e data of particular frame \e frame
		 * which is of the \c std::valarray<VOXEL> data type.
		 */
		std::valarray<VOXEL>& GetVoxelData(const size_t frame)
			{ return (data[frame]->GetVoxelData()); }
		const std::valarray<VOXEL>& GetVoxelData(const size_t frame) const
		{
			const typename std::valarray<VOXEL>& d = data[frame]->GetVoxelData();
			return (d);
		}

		/** Returns a pointer to the \c Image3d (a frame) at given position. No bounds
		 * checking, thus may crash if called inappropriatly.
		 */
		i3d::Image3d<VOXEL>* GetImage(const size_t frame)
			{ return(data[frame]); }
		const i3d::Image3d<VOXEL>* GetImage(const size_t frame) const
			{ return(data[frame]); }

		i3d::Image3d<VOXEL>* operator[](const size_t frame)
			{ return(data[frame]); }
		const i3d::Image3d<VOXEL>* operator[](const size_t frame) const
			{ return(data[frame]); }

		/// Returns reference to a string with a name of the particular image (=file).
		const wxString& GetImageName(size_t frame) const
			{ return names[frame]; }

		/// Sets a string with a name of the particular image (=file).
		void SetImageName(const size_t frame, const wxString &name)
			{ names[frame]=name; }


		// Bounding box testing functions
		//-------------------------------

		/// Is (x,y,z) on the border of the frame?
		bool OnFrameBorder(const size_t x, const size_t y, const size_t z) const
		{
			return (size.x != 1 && (x == 0 || x == size.x - 1))
				|| (size.y != 1 && (y == 0 || y == size.y - 1))
				|| (size.z != 1 && (z == 0 || z == size.z - 1));
		}

		/// Is (x,y,z) (encoded in Vector3d) on the border of the frame?
		bool OnFrameBorder(const i3d::Vector3d<size_t> v) const
		{
			return OnFrameBorder(v.x, v.y, v.z);
		}

		//Is the given frame number the first or the last frame number of the sequence?
		bool OnSequenceBorder(const size_t frm) const
		{
			return ((frm == 0) || (frm == frames - 1));
		}

		bool OnBorder(const size_t x, const size_t y, const size_t z, const size_t frm) const
		{
			return (size.x != 1 && (x == 0 || x == size.x - 1))
				|| (size.y != 1 && (y == 0 || y == size.y - 1))
				|| (size.z != 1 && (z == 0 || z == size.z - 1))
				|| (frames != 1 && (frm == 0 || frm == frames - 1));
		}

		bool OnBorder(const i3d::Vector3d<size_t> v, const size_t frm) const
		{
			return OnBorder(v.x, v.y, v.z, frm);
		}

		/// Is the voxel [x,y,z] in this image?
		bool Include(const int x, const int y, const int z) const
		{
			return (x >= 0) && (y >= 0) && (z >= 0) &&
				(x <= (int)size.x - 1) &&
				(y <= (int)size.y - 1) &&
				(z <= (int)size.z - 1);
		}

		/// Is the voxel 'v' in this image?
		bool Include(const i3d::Vector3d<int>& v) const
		{
			return Include(v.x, v.y, v.z);
		}

		/// Is the voxel [x,y,z,t] in this image?
		bool Include(const int x, const int y, const int z, const int t) const
		{
			return (x >= 0) && (y >= 0) && (z >= 0) &&
				(x <= (int)size.x - 1) &&
				(y <= (int)size.y - 1) &&
				(z <= (int)size.z - 1) &&
				(t >= 0) && (t < (int)frames - 1);
		}

		/// Is the voxel 'v'+t in this image?
		bool Include(const i3d::Vector3d<int>& v, const int t) const
		{
			return Include(v.x, v.y, v.z, t);
		}


		// Loading and Saving of Image4d
		//--------------------------------------------------------------------------
		/**
		 * Reads sequence of images and stores them in \c Image4d.
		 * Previous data will be destroyed
		 *
		 * \param[in] Names				vector of paths to images
		 * \param[in] sourceIsBinary	do implicit conversion from binary data on disk to VOXEL data in RAM
		 *
		 * All processed images should be of the same dimension and have the same metadata.
		 *
		 * \throws InternalException when some image has different offset, resolution or size
		 * \throws InternalException when the function ContainsInvalidFloats() complains
		 */
		void ReadSequence(const wxArrayString &names, const bool sourceIsBinary=false);

	//private:
		/** Very internal functions that together do the job of Image4d::ReadSequence(),
		 * this one reads the very first image of the sequence and initiates local metadata. */
		void ReadSequence_First(const wxArrayString &names, const bool sourceIsBinary=false);
		/** Very internal functions that together do the job of Image4d::ReadSequence(),
		 * this one reads every next image on the list and checks for consistency. */
		void ReadSequence_Next(const wxArrayString &names, const bool sourceIsBinary=false);

	public:
		/**
		 * Stores \c Image4d sequence into several files. The specification of filenames
		 * is according to this example:
		 *
		 * to save files: green_0001.ics to green_0012.ics
		 * is translated into: SaveSequence("green_",4,".ics",1,12);
		 *
		 * \param[in] NamePrefix	the part of filename before the index number
		 * \param[in] Width		the width of zero-padded index number in the filename
		 * \param[in] NameSuffix	the part of filename after the index number
		 * \param[in] StartIndex	the first index number
		 * \param[in] StopIndex		the last index number
		 *
		 * All saved images will be of the same dimension and will have the same metadata.
		 *
		 * Parameters \e StartIndex and \e StopIndex are relevant both to filenames and indices
		 * within the vector of frames in the \c Image4d. In other words, only a subsequence can
		 * be saved owing to the two parameters.
		 */
		void SaveSequence(const char *NamePrefix, const int Width, const char *NameSuffix,
			const int StartIndex, const int StopIndex);


		// Low level functions
		//--------------------------------------------------------------------------
		/** Free memory occupied by data and set size to (0,0,0,\e frames). It scans the entire
		 * Image4d vector \e data (the 4D image sequence) and iteratively calls \c Image3d::DisposeData().
		 * In particular, the \c Image3d objects are not removed from memory.
		 * The \c std::vector \e data and \e names retain their lengths and content.
		 *
		 * In other words, the frames data is released from memory while frames themselves are preserved.
		 */
		void DisposeData()
		{
			size.x = size.y = size.z = 0;

			typename std::vector< i3d::Image3d<VOXEL>* >::iterator iter = data.begin();
			while (iter != data.end()) {
				(*iter)->DisposeData();
				++iter;
			}
		}

		/** Free memory occupied by data and set size to (0,0,0,0). It scans the entire
		 * Image4d vector \e data (the 4D image sequence) and iteratively destroys \c Image3d
		 * frames. The \c std::vector \e data is shortened to zero.
		 *
		 * In other words, the frames data is released from memory as well as the frames themselves.
		 */
		void DestroyData()
		{
			size.x = size.y = size.z = 0;

			typename std::vector< i3d::Image3d<VOXEL>* >::iterator iter = data.begin();
			while (iter != data.end()) {
				delete *iter; //destroy the Image3d object
				*iter = NULL;
				++iter;
			}

			frames = 0; //no Image3ds assigned
			data.clear(); //clear the Image4d vector
			names.clear();
		}

		/** Allocate/reallocate voxel memory. Set the image size and sequence length.
		 * The input are three dimensions of all frames and the number of frames.
		 *
		 * If the new sequence length is shorter than the original length then the appropriate
		 * number of frames is removed from memory (their destructor functions are called).
		 *
		 * \note There is absolutely no guarantee on the content of new 4D image. Original image is lost.
		 */
		void MakeRoom(const size_t width, const size_t height, const size_t slices, const size_t frms)
		{
			int frm_no = 0, frm_total = (int)frms;
			names.clear();

			typename std::vector< i3d::Image3d<VOXEL>* >::iterator iter = data.begin();
			//resize as much as you can within the existing sequence
			while ((iter != data.end()) && (frm_no < frm_total)) {
				(*iter)->MakeRoom(width, height, slices);
				(*iter)->SetOffset(offset);
				(*iter)->SetResolution(resolution);
				names.push_back(std::string("no file specified"));

				++iter; ++frm_no;
			}

			if (frm_no == frm_total) {
				//there may be possibly something left in the original sequence
				while (iter != data.end()) {
					delete *iter;
					*iter = NULL;
					++iter;
				}
				data.resize(frms);
				names.resize(frms);
			}
			else
			{
				//it must hold: frm_no < frm_total what implies iter == data.end(),
				//thus we need to add some new frames
				while (frm_no < frm_total) {
					i3d::Image3d<VOXEL> *im = new i3d::Image3d < VOXEL >;
					im->MakeRoom(width, height, slices);
					im->SetOffset(offset);
					im->SetResolution(resolution);

					data.push_back(im);
					names.push_back(std::string("no file specified"));
					++frm_no;
				}
				//data.resize() adjusts automatically during "push_backing"
			}

			size.x = width; size.y = height; size.z = slices;
			frames = frms;
		}

		/** Allocate/reallocate voxel memory. Set the image size and sequence length.
		 * The input are three dimensions (in the vector form) of all frames and the number of frames.
		 *
		 * If the new sequence length is shorter than the original length then the appropriate
		 * number of frames is removed from memory (their destructor functions are called).
		 *
		 * \note There is absolutely no guarantee on the content of new 4D image. Original image is lost.
		 */
		void MakeRoom(const i3d::Vector3d<size_t> &sz, const size_t frms)
			{ MakeRoom(sz.x, sz.y, sz.z, frms); }

		/** Inserts or appends the input image/frame \e img into the sequence right before the frame
		 * at index \e time. If \e time is -1 then the new image/frame is appended at the end of this sequence.
		 * Also the filename should be specified.
		 *
		 * The length of this sequence is increased by one.
		 *
		 * \note The inserted image metadata is checked against this sequence metadata.
		 * This, however, means that this sequence has to be initialized already (size,
		 * resolution, and offset must be set already).
		 *
		 * Use ReadSequence(single_image) to add first image into the sequence and
		 * initialize the sequence in this way.
		 *
		 * \note The sequence absorbs exactly the \e *img, no new copy of the input image is made.
		 * One should be careful and provide pointers only to dynamically created images.
		 */
		void InsertFrame(i3d::Image3d<VOXEL>* const img,const wxString &name,const int time=-1);

		/**
		 * Works exactly the same as the pointer-to-existing-image InsertFrame() except
		 * for the fact that this one make a copy of the input image. It is this copy
		 * that becomes part of this sequence.
		 */
		void InsertFrame(const i3d::Image3d<VOXEL>& img,const wxString &name,const int time=-1);

		/**
		 * Opens the filename \e name and inserts the input image/frame into
		 * the sequence right before the frame at index \e time. If \e time
		 * is -1 then the new image/frame is appended at the end of this sequence.
		 *
		 * The length of this sequence is increased by one.
		 *
		 * \note The inserted image metadata is checked against this sequence metadata.
		 * This, however, means that this sequence has to be initialized already (size,
		 * resolution, and offset must be set already).
		 *
		 * Use ReadSequence(single_image) to add first image into the sequence and
		 * initialize the sequence in this way.
		 */
		void InsertFrame(const wxString &name, const int time=-1);

		/**
		 * Removes image/frame from the sequence at index \e time.
		 *
		 * The length of this sequence is decreased (if not zero already).
		 */
		void RemoveFrame(const int time);

		/** Test internal consistency of \c Image4d. It controls that the length of frames
		 * vector \e data has the same length as stored in property \e frames. Moreover, it
		 * controls that each frame in the sequence has the same dimension as stated in the
		 * property \e size.
		 *
		 * It is not testing whether the minValue or maxValue hold valid values.
		 *
		 * The function returns zero when everything is OK.
		 */
		int TestConsistency(void) const {
			//test the lengths
			if (frames != data.size()) return(1);
			if (frames != names.size()) return(1);

			//test that each frame has the same metadata as the entire class Image4d
			typename std::vector< i3d::Image3d<VOXEL>* >::const_iterator iter = data.begin();
			while (iter != data.end()) {
				if ((*iter)->GetSize() != size) return(2);
				if ((*iter)->GetOffset() != offset) return(3);
				if ((*iter)->GetResolution().GetRes() != resolution.GetRes()) return(4);

				++iter;
			}

			return(0);
		}

	protected:
		/// position of the image in absolute coordinates in micrometers
		i3d::Offset offset;

		/// image resolution in pixels per micrometer
		i3d::Resolution resolution;

		/// size of the image (width x height x slices) in pixels 
		i3d::Vector3d<size_t> size;

		// number of frames
		size_t frames;

		/// delay between frames (in seconds)
		float timeStep;

		/// image data buffer... a std::vector of image3d
		std::vector< i3d::Image3d<VOXEL>* > data;

		/// list of image names, same order as data
		wxArrayString names;

		/// cached minimal and maximal values of the all frames
		ExtremaValues<VOXEL> extremes;
	};


	//conversion functions

	/** Type conversion of \c Image4d sequences. Input image is copied and converted from \c float
	 * or \c double to \c GRAY8 or \c GRAY16. The conversion linearly maps the interval of minimum
	 * to maximum value spotted in the input image to the interval 0 to max value. The max value is
	 * the maximum possible value that can be represented by the target type.
	 */
	template <class VIN, class VOUT> void FloatToGray(const Image4d <VIN> &fimg, Image4d <VOUT> &gimg);

	/** Type conversion of \c Image4d sequences. Input image is copied and converted from \c float
	 * or \c double to \c GRAY8 or \c GRAY16. The conversion just rounds a floating point numbers
	 * into integers. Negative values are changed to zero. Values overflowing the target type are
	 * set to the maximum possible value that can be represented by the target type.
	 */
	template <class VIN, class VOUT> void FloatToGrayNoWeight(const Image4d <VIN> &fimg, Image4d <VOUT> &gimg);

	/** Type conversion of \c Image4d sequences. Input image is copied and converted from \c GRAY8
	 * or \c GRAY16 to \c float or \c double.
	 */
	template <class VIN, class VOUT> void GrayToFloat(const Image4d <VIN> &gimg, Image4d <VOUT> &fimg);

	/** Helper type conversion of \c Image3d image from binary type to any
	 * multi-value scalar type, most likely to i3d::GRAY8 image.
	 */
	template <class VOUT> void CreateBinaryScalarImage(const i3d::Image3d<bool>& bimg, i3d::Image3d<VOUT>& img);

	/** Helper function to zero any voxel with invalid floating-point number.
	 */
	template <class VIN> void ReplaceInvalidFloats(i3d::Image3d<VIN>& img);

	// @}
}

#endif
