/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * v4d_imagecanvas3d.cpp
 *
 * Canvas with for drawing 3d images
 *
 * Petr Matula (pem@fi.muni.cz) 2006
 */

#include "v4d_basic.h"
#include <wx/image.h>
#include <wx/bitmap.h>
#include <i3d/transform.h>
#include <wx/busyinfo.h>
#include "v4d_imagecanvas3d.h"
#include "mapping.h"
#include "v4d_frame.h"
#include "wdr/info_wdr.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

namespace v4d 
{

/***************************************************************************/
//stolen from i3dlibs and modified for std::complex<> ...
/***************************************************************************/

  /**
   * Calculates the maximum projection \e dest from the input image \e src.
   * Both images are, however, expected to have voxels of std::complex<T> type.
   * Maximality is evaluated with the \e ext parameter, which is any class with
   * defined operator(): const SOME_SCALAR operator() (const std::complex<T> &) const .
   * An output pixel P is any such that \e exp(P) is maximum.
   */
  template < class VOXEL, class EXTRACTOR >
  void Get_AF_XY (const i3d::Image3d <VOXEL> &src, i3d::Image3d <VOXEL> &dest,
                  const EXTRACTOR &ext)
  {
    dest.MakeRoom(src.GetSizeX (), src.GetSizeY (), 1);
    dest.SetOffset(i3d::Offset(src.GetOffset().x, src.GetOffset().y, 0));
    dest.SetResolution(src.GetResolution());

    const VOXEL* src_data = src.GetFirstVoxelAddr ();
    VOXEL* dest_data = dest.GetFirstVoxelAddr ();

    const size_t max_z = src.GetSizeZ ();
    const size_t slice_size = src.GetSliceSize ();

    size_t i;
    VOXEL max_intensity;

    for (size_t j = 0; j < slice_size; j++)
      {
        i = j;
        max_intensity = src_data[i];
        for (size_t z = 1; z < max_z; z++)
          {
            i += slice_size;
            if (ext(max_intensity) < ext(src_data[i]))
                max_intensity=src_data[i];
          }
        dest_data[j] = max_intensity;
      }
  }

  /**
   * Calculates the maximum projection \e dest from the input image \e src.
   * Both images are, however, expected to have voxels of std::complex<T> type.
   * Maximality is evaluated with the \e ext parameter, which is any class with
   * defined operator(): const SOME_SCALAR operator() (const std::complex<T> &) const .
   * An output pixel P is any such that \e exp(P) is maximum.
   */
  template < class VOXEL, class EXTRACTOR >
  void Get_AF_XZ (const i3d::Image3d <VOXEL> &src, i3d::Image3d <VOXEL> &dest,
                  const EXTRACTOR &ext)
  {
    dest.MakeRoom(src.GetSizeX(), 1, src.GetSizeZ());
    dest.SetOffset(i3d::Offset(src.GetOffset().x, 0, src.GetOffset().z));
    dest.SetResolution(src.GetResolution());

    const VOXEL* src_data = src.GetFirstVoxelAddr ();
    VOXEL* dest_data = dest.GetFirstVoxelAddr ();

    const size_t max_y = src.GetSizeY ();
    const size_t size_x = src.GetSizeX ();
    const size_t slice_size = src.GetSliceSize ();
    const size_t new_size = src.GetSizeX () * src.GetSizeZ ();

    size_t i;
    VOXEL max_intensity;

    for (size_t j = 0; j < new_size; j++)
      {
        i = (j / size_x) * slice_size + (j % size_x);
        max_intensity = src_data[i];
        for (size_t y = 1; y < max_y; y++)
          {
            i += size_x;
            if (ext(max_intensity) < ext(src_data[i]))
                max_intensity=src_data[i];
          }
        dest_data[j] = max_intensity;
      }
  }

  /**
   * Calculates the maximum projection \e dest from the input image \e src.
   * Both images are, however, expected to have voxels of std::complex<T> type.
   * Maximality is evaluated with the \e ext parameter, which is any class with
   * defined operator(): const SOME_SCALAR operator() (const std::complex<T> &) const .
   * An output pixel P is any such that \e exp(P) is maximum.
   */
  template < class VOXEL, class EXTRACTOR >
  void Get_AF_YZ (const i3d::Image3d <VOXEL> &src, i3d::Image3d <VOXEL> &dest,
                  const EXTRACTOR &ext)
  {
    dest.MakeRoom (1, src.GetSizeY (), src.GetSizeZ ());
    dest.SetOffset (i3d::Offset(0,src.GetOffset().y,src.GetOffset().z));
    dest.SetResolution (src.GetResolution ());

    const VOXEL* src_data = src.GetFirstVoxelAddr ();
    VOXEL* dest_data = dest.GetFirstVoxelAddr ();

    const size_t size_x = src.GetSizeX ();
    const size_t slice_size = src.GetSliceSize ();
    const size_t size_y = src.GetSizeY ();
    const size_t new_size = src.GetSizeY () * src.GetSizeZ ();

    size_t i;
    VOXEL max_intensity;

    for (size_t j = 0; j < new_size; j++)
      {
        i = (j % size_y) * size_x + (j / size_y) * slice_size;
        max_intensity = src_data[i];
        for (size_t x = 1; x < size_x; x++)
          {
            i++;
            if (ext(max_intensity) < ext(src_data[i]))
                max_intensity=src_data[i];
          }
        dest_data[j] = max_intensity;
      }
  }


/***************************************************************************/
//now the extractors: scalar from complex
/***************************************************************************/

    template <typename T>
    class RealExtractor
    {
    public:
        const T operator()(const std::complex<T> &val) const
        { return( val.real() ); }
    };

    template <typename T>
    class ImagExtractor
    {
    public:
        const T operator()(const std::complex<T> &val) const
        { return( val.imag() ); }
    };

    template <typename T>
    class PhaseExtractor
    {
    public:
        const T operator()(const std::complex<T> &val) const
        { return( std::atan2(val.imag(),val.real()) ); }
    };

    template <typename T>
    class AmpExtractor
    {
    public:
        const T operator()(const std::complex<T> &val) const
        { return( sqrt(val.real()*val.real() + val.imag()*val.imag()) ); }
    };




    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                      ImageViewer3dCanvas                    //
    //                                                             //
    /////////////////////////////////////////////////////////////////

    template <class VOXEL> 
    v4d_ImageViewer3dCanvas<VOXEL>::v4d_ImageViewer3dCanvas(
        wxWindow *parent, const v4d_Mapper<VOXEL> &mapper, 
        i4d::Image4d<VOXEL> *img,
        bool delete_data, wxWindowID id, 
        const wxPoint& pos, const wxSize& size)
        : v4d_Viewer3dCanvas(parent, id, pos, size), 
        m_xy_origbuf(), m_xz_origbuf(), m_yz_origbuf(),
        m_xy_rgbbuf(), m_xz_rgbbuf(), m_yz_rgbbuf(), 
        m_xy_dirty(true), m_xz_dirty(true), m_yz_dirty(true), 
        /* TODO m_show_red(true), m_show_green(true), m_show_blue(true), */
		m_mapper(mapper), m_img(img), m_delete_data(delete_data)
    {
		if (m_img)
        {
            m_vv.bbox.offset = m_img->GetOffset();
            m_vv.bbox.size   = i3d::PixelsToMicrons(m_img->GetSize(), m_img->GetResolution());
            m_vv.orig_res    = i3d::Vector3d<float>(m_img->GetResolution().GetX(), m_img->GetResolution().GetY(), m_img->GetResolution().GetZ());
            m_vv.cstep.x     = m_vv.bbox.size.x / float(m_img->GetSizeX());
            m_vv.cstep.y     = m_vv.bbox.size.y / float(m_img->GetSizeY());
            m_vv.cstep.z     = m_vv.bbox.size.z / float(m_img->GetSizeZ());
            m_vv.cut.x       = m_vv.bbox.offset.x + m_vv.cstep.x * (float(m_img->GetSizeX() / 2) + 0.5);
            m_vv.cut.y       = m_vv.bbox.offset.y + m_vv.cstep.y * (float(m_img->GetSizeY() / 2) + 0.5);
            m_vv.cut.z       = m_vv.bbox.offset.z + m_vv.cstep.z * (float(m_img->GetSizeZ() / 2) + 0.5);

			m_vv.time		 = 0;
			m_ii.SetFirstLastIndex(0, m_img->GetNumFrames() - 1);
			m_ii.SetLastRangeIndex(m_img->GetNumFrames() - 1);
			m_xy_af.resize(m_img->GetNumFrames());
			m_xz_af.resize(m_img->GetNumFrames());
			m_yz_af.resize(m_img->GetNumFrames());
			m_xy_af_dirty.resize(m_img->GetNumFrames(), true);
			m_xz_af_dirty.resize(m_img->GetNumFrames(), true);
			m_yz_af_dirty.resize(m_img->GetNumFrames(), true);

            RecomputeVisible();
            CenterViewIfOut();
        }
    }

    template <class VOXEL> 
    v4d_ImageViewer3dCanvas<VOXEL>::~v4d_ImageViewer3dCanvas()
    {
        if (m_delete_data && m_img)
            delete m_img;
    }

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::MakeDirtyAndRefresh(v4d_ProjectionType pt)
    {
        if (pt & pt_XY)
            m_xy_dirty = true;
        if (pt & pt_XZ)
            m_xz_dirty = true;
        if (pt & pt_YZ)
            m_yz_dirty = true;

        Refresh(false);
    }

    template <class T>
    void v4d_ImageViewer3dCanvas<T>::GetSettings(v4d_ImageCanvasSettings &params) const
    {
        /* TODO zmenit setsettings
        params.show_red   = m_show_red;
        params.show_green = m_show_red;
        params.show_blue  = m_show_red;
        */
        v4d_Viewer3dCanvas::GetSettings(params);
    }


    template <class T>
    void v4d_ImageViewer3dCanvas<T>::SetSettings(const v4d_ImageCanvasSettings &params)
    {
        /* TODO zmenit setsettings
        m_show_red   = params.show_red;
        m_show_green = params.show_green;
        m_show_blue  = params.show_blue;
        */

        v4d_Viewer3dCanvas::SetSettings(params);
    }

    int ComputeFrom(float soff, float doff, float dres)
    {
        int m = static_cast<int>(round((soff - doff) * dres));
        return std::max<int>(m, 0);
    }

    int ComputeTo(float soff, float sres, float ssz, int dsz, float doff, float dres)
    {
        int m = static_cast<int>(ceil((ssz / sres + soff - doff) * dres));
        return std::min<int>(m, dsz);
    }

/***************************************************************************/
	
    ///wrapper for usual computation of the AF XY
    template <class VOXEL>
    void MyGet_AF_XY(const i3d::Image3d<VOXEL> &img3d, i3d::Image3d<VOXEL> &imgAF,
                     const v4d_Mapper<VOXEL> &mapper)
    {
        //wxLogMessage("univerzalni AF XY - casem me odstran");
        i3d::Get_AF_XY(img3d, imgAF);
    }

    ///wrapper for special computation of the AF XY from complex images
    template <class T>
    void MyGet_AF_XY(const i3d::Image3d<std::complex<T> > &img3d, i3d::Image3d<std::complex<T> > &imgAF,
                     const v4d_Mapper<std::complex<T> > &mapper)
    {
        const typename v4d_ComplexMappingModule<T>::ComplexMode showMode=
            (dynamic_cast<const v4d_ComplexMappingModule<T> &>(mapper)).GetShowMode();
        //wxLogMessage("specializace AF XY - casem me odstran, mode=%u", showMode);

        switch (showMode)
        {
        case v4d_ComplexMappingModule<T>::REAL:
            v4d::Get_AF_XY(img3d, imgAF, RealExtractor<T>() );
            break;
        case v4d_ComplexMappingModule<T>::IMAGINARY:
            v4d::Get_AF_XY(img3d, imgAF, ImagExtractor<T>() );
            break;
        case v4d_ComplexMappingModule<T>::PHASE:
            v4d::Get_AF_XY(img3d, imgAF, PhaseExtractor<T>() );
            break;
        case v4d_ComplexMappingModule<T>::AMPLITUDE:
            v4d::Get_AF_XY(img3d, imgAF, AmpExtractor<T>() );
            break;
        default:
            wxLogError("Don't know how to 'GetAF' in this mode of complex numbers.");
        }
    }

    ///wrapper for usual computation of the AF XZ
    template <class VOXEL>
    void MyGet_AF_XZ(const i3d::Image3d<VOXEL> &img3d, i3d::Image3d<VOXEL> &imgAF,
                     const v4d_Mapper<VOXEL> &mapper)
    {
        //wxLogMessage("univerzalni AF XZ - casem me odstran");
        i3d::Get_AF_XZ(img3d, imgAF);
    }

    ///wrapper for special computation of the AF XZ from complex images
    template <class T>
    void MyGet_AF_XZ(const i3d::Image3d<std::complex<T> > &img3d, i3d::Image3d<std::complex<T> > &imgAF,
                     const v4d_Mapper<std::complex<T> > &mapper)
    {
        const typename v4d_ComplexMappingModule<T>::ComplexMode showMode=
            (dynamic_cast<const v4d_ComplexMappingModule<T> &>(mapper)).GetShowMode();
        //wxLogMessage("specializace AF XZ - casem me odstran, mode=%u", showMode);

        switch (showMode)
        {
        case v4d_ComplexMappingModule<T>::REAL:
            v4d::Get_AF_XZ(img3d, imgAF, RealExtractor<T>() );
            break;
        case v4d_ComplexMappingModule<T>::IMAGINARY:
            v4d::Get_AF_XZ(img3d, imgAF, ImagExtractor<T>() );
            break;
        case v4d_ComplexMappingModule<T>::PHASE:
            v4d::Get_AF_XZ(img3d, imgAF, PhaseExtractor<T>() );
            break;
        case v4d_ComplexMappingModule<T>::AMPLITUDE:
            v4d::Get_AF_XZ(img3d, imgAF, AmpExtractor<T>() );
            break;
        default:
            wxLogError("Don't know how to 'GetAF' in this mode of complex numbers.");
        }
    }

    ///wrapper for usual computation of the AF YZ
    template <class VOXEL>
    void MyGet_AF_YZ(const i3d::Image3d<VOXEL> &img3d, i3d::Image3d<VOXEL> &imgAF,
                     const v4d_Mapper<VOXEL> &mapper)
    {
        //wxLogMessage("univerzalni AF YZ - casem me odstran");
        i3d::Get_AF_YZ(img3d, imgAF);
    }

    ///wrapper for special computation of the AF YZ from complex images
    template <class T>
    void MyGet_AF_YZ(const i3d::Image3d<std::complex<T> > &img3d, i3d::Image3d<std::complex<T> > &imgAF,
                     const v4d_Mapper<std::complex<T> > &mapper)
    {
        const typename v4d_ComplexMappingModule<T>::ComplexMode showMode=
            (dynamic_cast<const v4d_ComplexMappingModule<T> &>(mapper)).GetShowMode();
        //wxLogMessage("specializace AF YZ - casem me odstran, mode=%u", showMode);

        switch (showMode)
        {
        case v4d_ComplexMappingModule<T>::REAL:
            v4d::Get_AF_YZ(img3d, imgAF, RealExtractor<T>() );
            break;
        case v4d_ComplexMappingModule<T>::IMAGINARY:
            v4d::Get_AF_YZ(img3d, imgAF, ImagExtractor<T>() );
            break;
        case v4d_ComplexMappingModule<T>::PHASE:
            v4d::Get_AF_YZ(img3d, imgAF, PhaseExtractor<T>() );
            break;
        case v4d_ComplexMappingModule<T>::AMPLITUDE:
            v4d::Get_AF_YZ(img3d, imgAF, AmpExtractor<T>() );
            break;
        default:
            wxLogError("Don't know how to 'GetAF' in this mode of complex numbers.");
        }
    }
/***************************************************************************/

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::Draw(wxDC &dc, v4d_ProjectionType pt)
    {
        if (m_img == NULL)
            return;

        dc.SetTextForeground(*wxGREEN);
        if (pt == pt_XY)
        {
            if (m_xy_dirty && m_xy->IsShown())
            {
                // clear dirty flag
                m_xy_dirty = false;
                m_xy_remap = true;

                // resize buffer to the size of drwin
                wxCoord w,h;
                m_xy->GetClientSize(&w, &h);

                i3d::Vector3d<size_t> sz;
                sz.x = w;
                sz.y = h;
                sz.z = 1;

                if (m_xy_rgbbuf.GetWidth() != w || m_xy_rgbbuf.GetHeight() != h)
                {
                    m_xy_origbuf.MakeRoom(sz);
                    m_xy_rgbbuf.resize(w, h);
                }

                // Set offset and resolution of visible part
                i3d::Vector3d<float> off;
                off.x = m_vv.vis3d.offset.x;
                off.y = m_vv.vis3d.offset.y;
                off.z = GetContent(pt) == pc_Ortho ? m_vv.cut.z : 0;

                m_xy_origbuf.SetOffset(off);
                m_xy_origbuf.SetResolution(m_vv.res);

                // copy data to orig buffer
                if (GetContent(pt) == pc_Ortho)
                {
                    i3d::CopyToImage(*(m_img->GetImage(m_vv.time)), m_xy_origbuf);
                }
                else
                {
                    if (m_xy_af_dirty[m_vv.time])
                    {
                        m_xy_af_dirty[m_vv.time] = false;
						MyGet_AF_XY(*(m_img->GetImage(m_vv.time)), m_xy_af[m_vv.time], m_mapper);
                    }
                    i3d::CopyToImage(m_xy_af[m_vv.time], m_xy_origbuf);
                }
            }

            if (m_xy_remap)
            {
                m_xy_rgbbuf.resize((int)m_xy_origbuf.GetWidth(), (int)m_xy_origbuf.GetHeight());
                m_xy_remap = false;
                // map data to rgb
                m_mapper.Map(m_xy_origbuf.GetFirstVoxelAddr(), m_xy_rgbbuf);

                // reverse orientation and convert it to bitmap
                wxImage wx_img(m_xy_rgbbuf.GetWidth(), m_xy_rgbbuf.GetHeight(), (unsigned char*) m_xy_rgbbuf.GetData(), true);

                if (m_vv.rev_axis[0]) 
                {
                    wx_img = wx_img.Mirror(true);
                }

                if (m_vv.rev_axis[1]) 
                {
                    wx_img = wx_img.Mirror(false);
                }

                m_xy_bmp = wxBitmap(wx_img);

                // send request for refreshing icon 
                wxUpdateUIEvent e(SI_REFRESH_ICON);
                AddPendingEvent(e);
            }

            // draw bitmap on screen
            dc.DrawBitmap(m_xy_bmp, 0, 0);
        }


        if (pt == pt_XZ)
        {
            if (m_xz_dirty && m_xz->IsShown())
            {
                // clear dirty flag
                m_xz_dirty = false;
                m_xz_remap = true;

                // resize buffer to the size of drwin
                wxCoord w,h;
                m_xz->GetClientSize(&w, &h);

                i3d::Vector3d<size_t> sz;
                sz.x = w;
                sz.y = 1;
                sz.z = h;

                if (m_xz_rgbbuf.GetWidth() != w || m_xz_rgbbuf.GetHeight() != h)
                {
                    m_xz_origbuf.MakeRoom(sz);
                    m_xz_rgbbuf.resize(w, h);
                }


                // Set offset and resolution of visible part
                i3d::Vector3d<float> off;
                if (GetContent(pt) == pc_Ortho)
                {
                    off.x = m_vv.vis3d.offset.x;
                    off.y = m_vv.cut.y;
                    off.z = m_vv.vis3d.offset.z;

                }
                else
                {
                    off.x = m_vv.vis3d.offset.x;
                    off.y = 0;
                    off.z = m_vv.vis3d.offset.z;
                }
                m_xz_origbuf.SetOffset(off);
                m_xz_origbuf.SetResolution(m_vv.res);

                // copy data to orig buffer
                if (GetContent(pt) == pc_Ortho)
                {
					i3d::CopyToImage(*(m_img->GetImage(m_vv.time)), m_xz_origbuf);
                }
                else
                {
                    if (m_xz_af_dirty[m_vv.time])
                    {
                        m_xz_af_dirty[m_vv.time] = false;
						MyGet_AF_XZ(*(m_img->GetImage(m_vv.time)), m_xz_af[m_vv.time], m_mapper);
                    }
                    i3d::CopyToImage(m_xz_af[m_vv.time], m_xz_origbuf);
                }
            }
            if (m_xz_remap)
            {
                m_xz_remap = false;

                // map data to rgb
                m_mapper.Map(m_xz_origbuf.GetFirstVoxelAddr(), m_xz_rgbbuf);

                // reverse orientation and convert it to bitmap
                wxImage wx_img(m_xz_rgbbuf.GetWidth(), m_xz_rgbbuf.GetHeight(), (unsigned char*) m_xz_rgbbuf.GetData(), true);

                if (m_vv.rev_axis[0]) 
                {
                    wx_img = wx_img.Mirror(true);
                }

                if (m_vv.rev_axis[2]) 
                {
                    wx_img = wx_img.Mirror(false);
                }

                m_xz_bmp = wxBitmap(wx_img);
            }
            // draw bitmap on screen
            dc.DrawBitmap(m_xz_bmp, 0, 0);
        }

        if (pt == pt_YZ)
        {
            if (m_yz_dirty && m_yz->IsShown())
            {
                // clear dirty flag
                m_yz_dirty = false;
                m_yz_remap = true;

                // resize buffer to the size of drwin
                wxCoord w,h;
                m_yz->GetClientSize(&w, &h);

                i3d::Vector3d<size_t> sz;
                sz.x = 1;
                sz.y = h;
                sz.z = w;

                if (m_yz_rgbbuf.GetWidth() != h || m_yz_rgbbuf.GetHeight() != w)
                {
                    m_yz_origbuf.MakeRoom(sz);
                    m_yz_rgbbuf.resize(h, w);
                }


                // Set offset and resolution of visible part
                i3d::Vector3d<float> off;
                if (GetContent(pt) == pc_Ortho)
                {
                    off.x = m_vv.cut.x;
                    off.y = m_vv.vis3d.offset.y;
                    off.z = m_vv.vis3d.offset.z;
                }
                else
                {
                    off.x = 0;
                    off.y = m_vv.vis3d.offset.y;
                    off.z = m_vv.vis3d.offset.z;
                }
                m_yz_origbuf.SetOffset(off);
                m_yz_origbuf.SetResolution(m_vv.res);

                // copy data to orig buffer
                if (GetContent(pt) == pc_Ortho)
                {
					i3d::CopyToImage(*(m_img->GetImage(m_vv.time)), m_yz_origbuf);
                }
                else
                {
                    if (m_yz_af_dirty[m_vv.time])
                    {
                        m_yz_af_dirty[m_vv.time] = false;
						MyGet_AF_YZ(*(m_img->GetImage(m_vv.time)), m_yz_af[m_vv.time], m_mapper);
                    }
                    i3d::CopyToImage(m_yz_af[m_vv.time], m_yz_origbuf);
                }
            }

            if (m_yz_remap)
            {
                m_yz_remap = false;

                // map data to rgb
                m_mapper.Map(m_yz_origbuf.GetFirstVoxelAddr(), m_yz_rgbbuf);

                // reverse orientation and convert it to bitmap
                wxImage wx_img = wxImage(m_yz_rgbbuf.GetWidth(), m_yz_rgbbuf.GetHeight(), (unsigned char*) m_yz_rgbbuf.GetData(), true).Rotate90();

                if (m_vv.rev_axis[1]) 
                {
                    wx_img = wx_img.Mirror(false);
                }

                if (!m_vv.rev_axis[2]) 
                {
                    wx_img = wx_img.Mirror(true);
                }

                m_yz_bmp = wxBitmap(wx_img);
            }
            // draw bitmap on screen
            dc.DrawBitmap(m_yz_bmp, 0, 0);
        }

        InfoFrame_UpdatePixelInfo();
    }

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::OnCutChange(v4d_ProjectionType pt)
    {
        MakeDirtyAndRefresh(pt);
        InfoFrame_UpdateCutAndVisBoxInfo();
    }

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::OnVis3dChange(v4d_ProjectionType pt)
    {
        MakeDirtyAndRefresh(pt);
        InfoFrame_UpdateCutAndVisBoxInfo();
    }

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::OnContentChange(v4d_ProjectionType pt)
    {
        MakeDirtyAndRefresh(pt);
    }

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::MappingChanged(v4d_MappingEvent& evt)
    {
        m_xy_remap = true;
        m_xz_remap = true;
        m_yz_remap = true;

        Refresh(false);

        if (evt._refreshAll)
        {
            UpdateVolVis();
        }
    }

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::OnDataChange()
    {
        m_vv.bbox.offset = m_img->GetOffset();
        m_vv.bbox.size   = i3d::PixelsToMicrons(m_img->GetSize(), m_img->GetResolution());
        m_vv.orig_res    = i3d::Vector3d<float>(m_img->GetResolution().GetX(), m_img->GetResolution().GetX(), m_img->GetResolution().GetX());
        m_vv.cstep.x     = m_vv.bbox.size.x / float(m_img->GetSizeX());
        m_vv.cstep.y     = m_vv.bbox.size.y / float(m_img->GetSizeY());
        m_vv.cstep.z     = m_vv.bbox.size.z / float(m_img->GetSizeZ());

        RecomputeVisible();
        CenterViewIfOut();

        MakeDirtyAndRefresh(pt_All);
        UpdateVolVis();

        InfoFrame_Move();
        InfoFrame_UpdateImageInfo();
        InfoFrame_UpdateCutAndVisBoxInfo();
    }

	template <class VOXEL>
	void v4d_ImageViewer3dCanvas<VOXEL>::OnTimeChange()
	{
		m_vv.time=m_ii.GetCurrentIndex();

        MakeDirtyAndRefresh(pt_All);
        UpdateVolVis();
        GetParent() -> SetLabel(CreateTitle());
			//std::cout << "OnTimeChange(), time=" << m_vv.time << "\n";
	}

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::GetIcons(wxIconBundle &bundle)
    {
        wxIcon small_icon, large_icon;
        CreateIcon(small_icon, 16, 16, m_xy_bmp, m_xy_rgbbuf);
        CreateIcon(large_icon, 32, 32, m_xy_bmp, m_xy_rgbbuf);

        bundle.AddIcon(small_icon);
        bundle.AddIcon(large_icon);
    }

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::ReverseAxis(int axis)
    {
        switch (axis)
        {
        case 0: // reverse orientation of x axis in xy and xz projections

            if (m_xy -> IsShown())
            {
                m_xy_bmp = wxBitmap(m_xy_bmp.ConvertToImage().Mirror(true));
            }

            if (m_xz -> IsShown())
            {
                m_xz_bmp = wxBitmap(m_xz_bmp.ConvertToImage().Mirror(true));
            }

            break;

        case 1: // reverse orientation of y axis in xy and yz projections

            if (m_xy -> IsShown())
            {
                m_xy_bmp = wxBitmap(m_xy_bmp.ConvertToImage().Mirror(false));
            }

            if (m_yz -> IsShown())
            {

                m_yz_bmp = wxBitmap(m_yz_bmp.ConvertToImage().Mirror(false));
            }

            break;

        case 2: // reverse orientation of z axis in xz and yz projections

            if (m_xz -> IsShown())
            {
                m_xz_bmp = wxBitmap(m_xz_bmp.ConvertToImage().Mirror(false));
            }

            if (m_yz -> IsShown())
            {
                m_yz_bmp = wxBitmap(m_yz_bmp.ConvertToImage().Mirror(true));
            }

            break;
        }
    }

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::UpdateVolVis()
    {
        Buffer2d<i3d::RGB> rgbbufTemp;
        unsigned char *data;

        if(m_img == NULL)
            return;

        if (m_volVis && m_volVis->IsShown())
        {
            //show busy info
				#if !defined(__UNIX__) || !defined(__WXGTK__)
					//this misbehaves on Linux/Unix with GTK-based wxWidgets implementation
					wxWindowDisabler disableAll;
				#else
					wxLogMessage(wxString("Detected wxGTK on Linux: not disabling application other windows."));
				#endif
            wxBusyInfo wait("Updating 3D data...", m_volVis);

            // resize buffer to the size of the image
            i3d::Vector3d<size_t> size = m_img->GetSize();

            try
            {
                rgbbufTemp.resize((int)(size.x*size.y*size.z), 1);

                // map data to rgb
                m_mapper.Map(m_img->GetImage(m_vv.time)->GetFirstVoxelAddr(), rgbbufTemp);

                // Allocate buffer
                data = new unsigned char[4 * (size.x + 2) * (size.y + 2) * (size.z + 2)];
            }
            catch (std::bad_alloc &)
            {
                wxMessageDialog *msg = new wxMessageDialog(this, "Not enough memory to create the 3D texture!", "Error", wxOK);
                msg->ShowModal();
                delete msg;
                return;
            }

            i3d::RGB *src = rgbbufTemp.GetData();
            unsigned char *dst = data;

            // First slice is black
            size_t slice_sz = 4 * (size.x + 2) * (size.y + 2);
            memset(dst, 0, slice_sz);
            dst += slice_sz;

            for (size_t z = 0; z < size.z; z++)
            {
                // Last line on the slice is black
                dst += slice_sz - 4 * (size.x + 2);
                memset(dst, 0, 4 * (size.x + 2));

                // The line one before the last one
                dst -= 4 * (size.x + 2);

                for (size_t y = 0; y < size.y; y++)
                {
                    // First pixel black
                    *dst++ = 0;
                    *dst++ = 0;
                    *dst++ = 0;
                    *dst++ = 0;

                    for (size_t x = 0; x < size.x; x++)
                    {
                        //copy RGB
                        *dst++ = src->red;
                        *dst++ = src->green;
                        *dst++ = src->blue;

                        //compute alpha
                        *dst++ = (unsigned char)(255*sin((src->red*0.2989 + src->green*0.5870 + src->blue*0.1141)/255 * M_PI*0.5f));

                        // Next voxel
                        src++;
                    }

                    // Last pixel black
                    *dst++ = 0;
                    *dst++ = 0;
                    *dst++ = 0;
                    *dst++ = 0;

                    // Go to previous line
                    dst -= 8 * (size.x + 2);
                }

                // First line is black
                memset(dst, 0, 4 * (size.x + 2));

                // Go to next slice
                dst += slice_sz;
            }

            // Last slice is black
            memset(dst, 0, slice_sz);

            // Compute scale according to resolution and padded size
            i3d::Vector3d<float> res = m_img->GetResolution().GetRes();
            i3d::Vector3d<float> scale;
            scale.x = res.x;
            scale.y = res.y * float(size.x + 2) / float(size.y + 2);
            scale.z = res.z * float(size.x + 2) / float(size.z + 2);

            float dmax = std::max ((float(size.x) / float(size.x + 2)) / scale.x, 
                (float(size.y) / float(size.y + 2)) / scale.y);
            dmax = std::max ((float(size.z) / float(size.z + 2)) / scale.z, dmax);

            scale *= dmax;
            scale.z *= 0.5f;

            size.x += 2;
            size.y += 2;
            size.z += 2;

            // update the data
            m_volVis -> GetVolVisCanvas() -> SetVolumeData(data, size, scale);

            delete[] data;
        }
    }

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::CreateIcon(wxIcon &icon, int width, int height, const wxBitmap &bmp, const Buffer2d<i3d::RGB> &buff)
    {
        Buffer2d<i3d::RGB> tmp_buff(buff, width, height, m_vv.rev_axis[0], m_vv.rev_axis[1]);
        wxBitmap tmp_bmp(wxImage(tmp_buff.GetWidth(), tmp_buff.GetHeight(), (unsigned char*) tmp_buff.GetData(), true));

        // create transparent mask
        Buffer2d<i3d::RGB> mask_buff(width, height);
        i3d::RGB *data = mask_buff.GetData();
        int size_x = width, size_y = height;
        bool is_width_greater = true;

        if (buff.GetWidth() > buff.GetHeight())
        {
            size_y =static_cast<int>( (float) height / ((float) buff.GetWidth() / (float) buff.GetHeight()));
        }
        else
        {
            size_x = static_cast<int>((float) width / ((float) buff.GetHeight() / (float) buff.GetWidth()));
            is_width_greater = false;
        }

        int start_x = (width - size_x + 1) / 2;
        int start_y = (height - size_y + 1) / 2;
        int finish_x = (width + size_x) / 2;
        int finish_y = (height + size_y) / 2;

        if (is_width_greater)
        {
            for (int y = 0; y < start_y; y++)
                for (int x = 0; x < width; x++)
                    data[y * width + x] = i3d::RGB(0, 0, 0);

            for (int y = start_y; y < finish_y; y++)
                for (int x = 0; x < width; x++)
                    data[y * width + x] = i3d::RGB(255, 255, 255);

            for (int y = finish_y; y < height; y++)
                for (int x = 0; x < width; x++)
                    data[y * width + x] = i3d::RGB(0, 0, 0);
        }
        else
        {   for (int y = 0; y < height; y++)
                for (int x = 0; x < start_x; x++)
                    data[y * width + x] = i3d::RGB(0, 0, 0);

            for (int y = 0; y < height; y++)
                for (int x = start_x; x < finish_x; x++)
                    data[y * width + x] = i3d::RGB(255, 255, 255);

            for (int y = 0; y < height; y++)
                for (int x = finish_x; x < width; x++)
                    data[y * width + x] = i3d::RGB(0, 0, 0);
        }

        wxBitmap mask_bmp(wxImage(mask_buff.GetWidth(), mask_buff.GetHeight(), (unsigned char*) mask_buff.GetData(), true));
        wxMask *mask = new wxMask(mask_bmp, wxColour(0, 0, 0));
        tmp_bmp.SetMask(mask);
        icon.CopyFromBitmap(tmp_bmp);
    }

    template <class T>
    void v4d_ImageViewer3dCanvas<T>::Save3dData(const wxString& fname)
    {
        try
        {
			wxLogMessage(wxString("Saving: ")+fname);
			m_img->GetImage(m_vv.time)->SaveImage(fname.mb_str(wxConvFile));
        }
        catch(i3d::IOException &e)
        {
            wxLogError(wxString(e.what.c_str(), wxConvUTF8));
            return;
        }
    }

    template <class T>
    void v4d_ImageViewer3dCanvas<T>::SaveActualView(const char *fname, int d, v4d_ProjectionType projection)
    {
        int view_width, view_height;        //size of the view
        i3d::RGB * view_buffer = NULL;      // tmp view buffer
        i3d::Image3d < i3d::RGB > *view_img;
        wxImage xy_img = m_xy->GetDrWinContent();
        wxImage xz_img = m_xz->GetDrWinContent();
        wxImage yz_img = m_yz->GetDrWinContent();
        bool draw_xy, draw_xz, draw_yz;

        int x = xy_img.GetWidth();
        int y = xy_img.GetHeight();
        int z = xz_img.GetHeight();
        i3d::RGB * xy_data = (i3d::RGB *) xy_img.GetData();
        i3d::RGB * xz_data = (i3d::RGB *) xz_img.GetData();
        i3d::RGB * yz_data = (i3d::RGB *) yz_img.GetData();

        int i;

        draw_xy = IsDrWinShown(pt_XY) && (projection == pt_XY || projection == pt_All);
        draw_xz = IsDrWinShown(pt_XZ) && (projection == pt_XZ || projection == pt_All);
        draw_yz = IsDrWinShown(pt_YZ) && (projection == pt_YZ || projection == pt_All);

        if (!draw_xy && !draw_xz && !draw_yz)
        {
            wxMessageBox ("Selected projection(s) can not be saved\nbecause they are not visible!", "Error", wxOK | wxICON_HAND);
            return;
        }

        view_width = 0;
        view_width += ((draw_xy || draw_xz) && draw_yz ? d : 0);
        view_width += (draw_xy || draw_xz ? x : 0);
        view_width += (draw_yz ? z : 0);

        view_height = 0;
        view_height += ((draw_xy || draw_yz) && draw_xz ? d : 0);
        view_height += (draw_xy || draw_yz ? y : 0);
        view_height += (draw_xz ? z : 0);

        view_img = new i3d::Image3d < i3d::RGB > ();
        view_img->MakeRoom(view_width, view_height, 1);

        view_img->GetVoxelData() = i3d::RGB((i3d::GRAY8) 255, (i3d::GRAY8) 255, (i3d::GRAY8) 255);
        view_buffer = view_img->GetVoxelAddr(0);

        try
        {
            int tmp1, tmp2;
            i = 0;

            tmp1 = (draw_xy ||  draw_xz ? x : 0);
            tmp2 = ((draw_xy || draw_xz) && draw_yz ? d : 0);

            while (i < y)
            {
                if (draw_xy)
                    memcpy(static_cast < void *>(view_buffer + i * view_width),
                    static_cast < void *>(xy_data + i * tmp1),
                    tmp1 * sizeof(i3d::RGB));
                if (draw_xy && draw_xz)
                    memset(static_cast < void *>(view_buffer + i * view_width + tmp1),
                    255, tmp2 * sizeof(i3d::RGB));
                if (draw_yz)
                    memcpy(static_cast < void *>(view_buffer + i * view_width + tmp1 + tmp2),
                    static_cast < void *>(yz_data + i * z),
                    z * sizeof(i3d::RGB));
                ++i;
            }

            if (draw_xz)
            {
                tmp1 = (draw_xy || draw_yz ? y : 0);
                tmp2 = ((draw_xy || draw_yz) && draw_xz ? d : 0);

                if (draw_xy && draw_xz)
                    memset(static_cast < void *>(view_buffer + view_width * y), 255,
                    tmp2 * view_width * sizeof(i3d::RGB));

                i = 0;

                while (i < z)
                {
                    memcpy(static_cast<void *>(view_buffer + (tmp1 + tmp2 + i) * view_width),
                        static_cast< void *>(xz_data + i * x),
                        x * sizeof(i3d::RGB));
                    if (draw_yz)
                        memset(static_cast<void *>(view_buffer + (tmp1 + tmp2 + i) * view_width + x),
                        255, (z + tmp2) * sizeof(i3d::RGB));
                    ++i;
                }
            }
            view_img->SaveImage(fname);            
            delete view_img;
        }
        catch(i3d::IOException & e)
        {
            wxLogError(wxString(e.what.c_str(), wxConvUTF8));
            return;
        }
    }

    template <class T>
    void v4d_ImageViewer3dCanvas<T>::SaveGalleryView(const char *fname, int start, int end,
        int step, bool fixedRows, int count, int spacing)
    {
        i3d::Vector3d<float> cut, oldCut;
        int gallery_width, gallery_height, numSlices, rows, cols;
        i3d::RGB *view_buffer = NULL;
        i3d::Image3d < i3d::RGB > *view_img;
        wxImage xy_img;
        int sx = int(m_img->GetWidth());
        int sy = int(m_img->GetHeight());
        int i, j;
        int r, t;
        i3d::RGB *xy_data;

        // Get current cut
        oldCut.x = m_yz->CutPos();
        oldCut.y = m_xz->CutPos();
        oldCut.z = m_xy->CutPos();

        cut = oldCut;

        numSlices = (end-start+1) / step;

        if(fixedRows)
        {
            rows = count;
            cols = static_cast<int>(ceil((float)numSlices / (float)count));

            gallery_width = cols * (sx + spacing) - spacing;
            gallery_height = rows * (sy + spacing) - spacing;
        }
        else
        {
            rows = static_cast<int>(ceil((float)numSlices / (float)count));
            cols = count;

            gallery_width = cols * (sx + spacing) - spacing;
            gallery_height = rows * (sy + spacing) - spacing;
        }

        view_img = new i3d::Image3d<i3d::RGB>;
        view_img->MakeRoom(gallery_width, gallery_height, 1);

        view_img->GetVoxelData() = i3d::RGB((i3d::GRAY8) 255, (i3d::GRAY8) 255, (i3d::GRAY8) 255);
        view_buffer = view_img->GetVoxelAddr(0);

        i = j = 0;

        try
        {
            // Take incremental cuts in the z-dirrection of the image
            for(int z = start; z <= end; z += step)
            {
                cut.z = z / m_img->GetResolution().GetZ();
                ChangeCut(cut);
                Update();

                xy_img = m_xy->GetDrWinContent();
                xy_data = (i3d::RGB*) xy_img.GetData();

                for(t = 0, r = i*(sy+spacing)*(cols*(sx+spacing) - spacing); t < sy; t++, r += cols*(sx+spacing) - spacing)
                {
                    memcpy(static_cast<void*>(view_buffer + r + j*(sx+spacing)),
                        static_cast<void*>(xy_data + t*sx), sx*sizeof(i3d::RGB));
                }

                j++;
                if(j == cols)
                {
                    i++;
                    j = 0;
                }
            }

            view_img->SaveImage(fname);
            delete view_img;
        }
        catch(i3d::IOException &e)
        {
            // Change the view back the way it was
            ChangeCut(oldCut);

            wxLogError(wxString(e.what.c_str(), wxConvUTF8));
            return;
        }

        // Change the view back the way it was
        ChangeCut(oldCut);
    }

    template <class VOXEL>
    wxString v4d_ImageViewer3dCanvas<VOXEL>::GetImageTypeName() const
    {
        if (typeid(VOXEL) == typeid(i3d::GRAY8))
            return wxString("GRAY8");
        else if (typeid(VOXEL) == typeid(i3d::GRAY16))
            return wxString("GRAY16");
        else if (typeid(VOXEL) == typeid(float))
            return wxString("float");
        else if (typeid(VOXEL) == typeid(double))
            return wxString("double");
        else if (typeid(VOXEL) == typeid(bool))
            return wxString("BINARY");
        else if (typeid(VOXEL) == typeid(i3d::RGB))
            return wxString("RGB");
        else if (typeid(VOXEL) == typeid(i3d::RGB16))
            return wxString("RGB16");
		else if (typeid(VOXEL) == typeid(std::complex<float>))
			return wxString("complex");
        else
            return wxString("Unknown");
    }

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::GetColourInfo(v4d_ProjectionType pt,
        int x, int y, wxString &orig_col, wxColour &mapped_col) const
    {
        if (pt == pt_XY)
        {
            if (m_vv.rev_axis[0])
            {
                x = m_xy_rgbbuf.GetWidth() - 1 - x;
            }
            if (m_vv.rev_axis[1])
            {
                y = m_xy_rgbbuf.GetHeight() - 1 - y;
            }
            
            orig_col = VoxelToStr<VOXEL>(m_xy_origbuf.GetVoxel(x, y, 0));
            i3d::RGB col = m_xy_rgbbuf.GetDataAt(x, y);
            mapped_col = wxColour(col.red, col.green, col.blue);
        }

        if (pt == pt_XZ)
        {
            if (m_vv.rev_axis[0])
            {
                x = m_xz_rgbbuf.GetWidth() - 1 - x;
            }
            if (m_vv.rev_axis[2])
            {
                y = m_xz_rgbbuf.GetHeight() - 1 - y;
            }

            orig_col = VoxelToStr<VOXEL>(m_xz_origbuf.GetVoxel(x, 0, y));
            i3d::RGB col = m_xz_rgbbuf.GetDataAt(x, y);
            mapped_col = wxColour(col.red, col.green, col.blue);
        }

        if (pt == pt_YZ)
        {
            if (m_vv.rev_axis[2])
            {
                x = m_yz_rgbbuf.GetHeight() - 1 - x;
            }
            if (m_vv.rev_axis[1])
            {
                y = m_yz_rgbbuf.GetWidth() - 1 - y;
            }

            orig_col = VoxelToStr<VOXEL>(m_yz_origbuf.GetVoxel(0, y, x));
            i3d::RGB col = m_yz_rgbbuf.GetDataAt(y, x);
            mapped_col = wxColour(col.red, col.green, col.blue);
        }
    }

    //template <class T> //TODO
	 template <>
    void v4d_ImageViewer3dCanvas<std::complex<float> >::GetColourInfo(v4d_ProjectionType pt,
        int x, int y, wxString &orig_col, wxColour &mapped_col) const
    {
        if (pt == pt_XY)
        {
            if (m_vv.rev_axis[0])
            {
                x = m_xy_rgbbuf.GetWidth() - 1 - x;
            }
            if (m_vv.rev_axis[1])
            {
                y = m_xy_rgbbuf.GetHeight() - 1 - y;
            }
            
            orig_col = VoxelToStr(m_xy_origbuf.GetVoxel(x, y, 0),
					(dynamic_cast<const v4d_ComplexMappingModule<float> &>(m_mapper)).GetShowMode());
            i3d::RGB col = m_xy_rgbbuf.GetDataAt(x, y);
            mapped_col = wxColour(col.red, col.green, col.blue);
        }

        if (pt == pt_XZ)
        {
            if (m_vv.rev_axis[0])
            {
                x = m_xz_rgbbuf.GetWidth() - 1 - x;
            }
            if (m_vv.rev_axis[2])
            {
                y = m_xz_rgbbuf.GetHeight() - 1 - y;
            }

            orig_col = VoxelToStr(m_xz_origbuf.GetVoxel(x, 0, y),
					(dynamic_cast<const v4d_ComplexMappingModule<float> &>(m_mapper)).GetShowMode());
            i3d::RGB col = m_xz_rgbbuf.GetDataAt(x, y);
            mapped_col = wxColour(col.red, col.green, col.blue);
        }

        if (pt == pt_YZ)
        {
            if (m_vv.rev_axis[2])
            {
                x = m_yz_rgbbuf.GetHeight() - 1 - x;
            }
            if (m_vv.rev_axis[1])
            {
                y = m_yz_rgbbuf.GetWidth() - 1 - y;
            }

            orig_col = VoxelToStr(m_yz_origbuf.GetVoxel(0, y, x),
					(dynamic_cast<const v4d_ComplexMappingModule<float> &>(m_mapper)).GetShowMode());
            i3d::RGB col = m_yz_rgbbuf.GetDataAt(y, x);
            mapped_col = wxColour(col.red, col.green, col.blue);
        }
    }

    template <class VOXEL>
    void v4d_ImageViewer3dCanvas<VOXEL>::GetNoColourInfo(wxString& orig_col) const
    {
	 	orig_col="---";
	 }

    template <>
    void v4d_ImageViewer3dCanvas<std::complex<float> >::GetNoColourInfo(wxString& orig_col) const
    {
	 	orig_col="---\n ";
	 }

	template <class VOXEL>
	v4d_DrWin3d *v4d_ImageViewer3dCanvas<VOXEL>::MouseInProjection(const wxPoint& mpos, v4d_DrWin3d *proj, v4d_DrWin3d *old)
    {
        if (proj->IsShown())
        {
            wxPoint cpos = proj->ScreenToClient(mpos);
            wxSize csz = proj->GetClientSize();

            if (cpos.x >= 0 && cpos.y >= 0 && cpos.x < csz.GetWidth() && cpos.y < csz.GetHeight())
            {
                return proj;
            }
        }

        return old;
    }

    template <class T>
    void v4d_ImageViewer3dCanvas<T>::InfoFrame_UpdatePixelInfo()
    {
        if (m_info_frame && m_info_frame->IsShown())
        {
            wxNotebook *ntb = (wxNotebook *) m_info_frame->FindWindow(ID_INFO_NOTEBOOK);
            
            if (!ntb->GetSelection())
            {
                wxStaticText *box;
                wxPoint scr_pos = wxGetMousePosition(), win_pos;
                i3d::Vector3d<float> pos;
                v4d_DrWin3d *drwin = NULL;

                // Find in which projection is the mouse cursor
                drwin = MouseInProjection(scr_pos, m_xy, drwin);
                drwin = MouseInProjection(scr_pos, m_xz, drwin);
                drwin = MouseInProjection(scr_pos, m_yz, drwin);

                // Get coordinates in microns
                if (drwin)
                {
                    win_pos = drwin->ScreenToClient(scr_pos);
                    pos.x = drwin->W2MPosX(win_pos.x);
                    pos.y = drwin->W2MPosY(win_pos.y);
                    pos.z = drwin->CutPos();
                    pos = drwin->Reorder(pos);
                }

                if (drwin != NULL && InCO(m_vv.bbox, pos))
                {
                    box = (wxStaticText *) ntb->FindWindow(ID_POSITION_MICRONS);
                    box->SetLabel(wxString::Format("[%.3f, %.3f, %.3f]", pos.x, pos.y, pos.z));

                    i3d::Vector3d<int> pix_pos = (pos - m_img->GetOffset()) * m_img->GetResolution().GetRes();
                    box = (wxStaticText *) ntb->FindWindow(ID_POSITION_PIXELS);
                    box->SetLabel(wxString::Format("[%d, %d, %d, %d]", pix_pos.x, pix_pos.y, pix_pos.z, (int)m_vv.time));

                    wxString col_orig;
                    wxColour col_mapped;
                    GetColourInfo(drwin->thisWin(), win_pos.x, win_pos.y, col_orig, col_mapped);

                    box = (wxStaticText *) ntb->FindWindow(ID_INTENSITY_VALUE);
                    box->SetLabel(col_orig);

                    box = (wxStaticText *) ntb->FindWindow(ID_MAPPED_COLOUR);
                    box->SetLabel(wxString::Format("%3d %3d %3d", col_mapped.Red(), 
                        col_mapped.Green(), col_mapped.Blue()));

                    wxWindow *bmp = ntb->FindWindow(ID_COLOUR_WIN);
                    bmp->SetBackgroundColour(col_mapped);
                    bmp->ClearBackground();
                }
                else
                {
                    box = (wxStaticText *) ntb->FindWindow(ID_POSITION_MICRONS);
                    box->SetLabel("---");

                    box = (wxStaticText *) ntb->FindWindow(ID_POSITION_PIXELS);
                    box->SetLabel("---");

                    wxString col_orig;
						  GetNoColourInfo(col_orig);
                    box = (wxStaticText *) ntb->FindWindow(ID_INTENSITY_VALUE);
                    box->SetLabel(col_orig);

                    box = (wxStaticText *) ntb->FindWindow(ID_MAPPED_COLOUR);
                    box->SetLabel("---");

                    wxWindow *bmp = ntb->FindWindow(ID_COLOUR_WIN);
                    //bmp->SetBackgroundColour(box->GetBackgroundColour());
					bmp->SetBackgroundColour(wxColour((unsigned char)255,(unsigned char)255,
						                              (unsigned char)255,wxALPHA_TRANSPARENT));
					bmp->ClearBackground();
                }

                m_info_frame->Fit();
            }
        }
    }

    template <class T>
    void v4d_ImageViewer3dCanvas<T>::InfoFrame_UpdateImageInfo()
    {
        if (m_info_frame && m_info_frame->IsShown())
        {
            wxNotebook *ntb = (wxNotebook *) m_info_frame->FindWindow(ID_INFO_NOTEBOOK);
            wxStaticText *box;

            // Image info
            box = (wxStaticText *) ntb->FindWindow(ID_IMAGE_TYPE);
            box->SetLabel(GetImageTypeName());

            i3d::Vector3d<size_t> sz = m_img->GetSize();
            box = (wxStaticText *) ntb->FindWindow(ID_IMAGE_SIZE_PIXELS);
            box->SetLabel(wxString::Format("%3u x %3u x %3u x %3u",
					(unsigned int)sz.x, (unsigned int)sz.y, (unsigned int)sz.z,
					(unsigned int)(m_img->GetNumFrames()) ));

            i3d::Resolution res = m_img->GetResolution();
            box = (wxStaticText *) ntb->FindWindow(ID_IMAGE_SIZE_MICRONS);
            box->SetLabel(res.IsDefined() ? wxString::Format("%.3f x %.3f x %.3f", 
                sz.x/res.GetX(), sz.y/res.GetY(), sz.z/res.GetZ()) : wxString("undef"));

            box = (wxStaticText *) ntb->FindWindow(ID_PIXEL_SIZE);
            box->SetLabel(res.IsDefined() ? wxString::Format("%3u x %3u x %3u", 
                (unsigned int)(1000.0f/res.GetX()), (unsigned int)(1000.0f/res.GetY()), 
                (unsigned int)(1000.0f/res.GetZ())) : wxString("undef"));

            // Bbox info
            box = (wxStaticText *) ntb->FindWindow(ID_BBOX_OFFSET);
            box->SetLabel(wxString::Format("[%.3f, %.3f, %.3f]", m_vv.bbox.offset.x,
                m_vv.bbox.offset.y, m_vv.bbox.offset.z));

            box = (wxStaticText *) ntb->FindWindow(ID_BBOX_SIZE);
            box->SetLabel(wxString::Format("%.3f x %.3f x %.3f", m_vv.bbox.size.x,
                m_vv.bbox.size.y, m_vv.bbox.size.z));

            m_info_frame->Fit();
        }
    }

    template <class T>
    void v4d_ImageViewer3dCanvas<T>::InfoFrame_UpdateCutAndVisBoxInfo()
    {
        if (m_info_frame && m_info_frame->IsShown())
        {
            wxNotebook *ntb = (wxNotebook *) m_info_frame->FindWindow(ID_INFO_NOTEBOOK);
            wxStaticText *box;

            box = (wxStaticText *) ntb->FindWindow(ID_VIS3D_OFFSET);
            box->SetLabel(wxString::Format("[%.3f, %.3f, %.3f]", m_vv.vis3d.offset.x,
                m_vv.vis3d.offset.y, m_vv.vis3d.offset.z));

            box = (wxStaticText *) ntb->FindWindow(ID_VIS3D_SIZE);
            box->SetLabel(wxString::Format("%.3f x %.3f x %.3f", m_vv.vis3d.size.x,
                m_vv.vis3d.size.y, m_vv.vis3d.size.z));

            box = (wxStaticText *) ntb->FindWindow(ID_CUT_POSITION);
            box->SetLabel(wxString::Format("[%.3f, %.3f, %.3f]", m_vv.cut.x,
                m_vv.cut.y, m_vv.cut.z));

            m_info_frame->Fit();
        }
    }

    // INSTANTIATIONS
    template class V4D_DLLEXPORT v4d_ImageViewer3dCanvas<i3d::GRAY8>;
    template class V4D_DLLEXPORT v4d_ImageViewer3dCanvas<i3d::GRAY16>;
    template class V4D_DLLEXPORT v4d_ImageViewer3dCanvas<float>;
    template class V4D_DLLEXPORT v4d_ImageViewer3dCanvas<int>;
    template class V4D_DLLEXPORT v4d_ImageViewer3dCanvas<i3d::RGB>;
    template class V4D_DLLEXPORT v4d_ImageViewer3dCanvas<i3d::RGB16>;
    template class V4D_DLLEXPORT v4d_ImageViewer3dCanvas<std::complex<float> >;
}
