/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * v4d_imagecanvas3d.h
 *
 * Canvas for displaying images
 *
 * Petr Matula (pem@fi.muni.cz) 2006
 */
#ifndef V4D_V4D_IMAGECANVAS3D_H
#define V4D_V4D_IMAGECANVAS3D_H

#include <i3d/image3d.h>
#include "v4d_impexp.h"
#include "mapping.h"
#include "v4d_canvas3d.h"
#include "types/buffer2d.h"

namespace v4d 
{
	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                    v4d Canvas Settings                      //
	//                                                             //
	/////////////////////////////////////////////////////////////////
	/** Settings of an Image canvas. They can be retreived from the canvas and
	applied to the same or even a different instance. */
	struct v4d_ImageCanvasSettings : public v4d_CanvasSettings 
    {
		bool show_red;
		bool show_green;
		bool show_blue;
	};

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                   ImageViewer3dCanvas                       //
	//                                                             //
	/////////////////////////////////////////////////////////////////
	/** Canvas for displaying RGB images. It contains three drawing
	windows (for each orthogonal projection one). */
	template <class VOXEL> class V4D_DLLEXPORT v4d_ImageViewer3dCanvas 
        : public v4d_Viewer3dCanvas  
    {
	protected:
		i3d::Image3d<VOXEL> m_xy_origbuf;
		i3d::Image3d<VOXEL> m_xz_origbuf;
		i3d::Image3d<VOXEL> m_yz_origbuf;

		std::vector<i3d::Image3d<VOXEL> > m_xy_af;
		std::vector<i3d::Image3d<VOXEL> > m_xz_af;
		std::vector<i3d::Image3d<VOXEL> > m_yz_af;

		Buffer2d<i3d::RGB> m_xy_rgbbuf;
		Buffer2d<i3d::RGB> m_xz_rgbbuf;
		Buffer2d<i3d::RGB> m_yz_rgbbuf;

		wxBitmap m_xy_bmp;
		wxBitmap m_xz_bmp;
		wxBitmap m_yz_bmp;

		bool m_xy_dirty;
		bool m_xz_dirty;
		bool m_yz_dirty;

		bool m_xy_remap;
		bool m_xz_remap;
		bool m_yz_remap;

		std::vector<bool> m_xy_af_dirty;
		std::vector<bool> m_xz_af_dirty;
		std::vector<bool> m_yz_af_dirty;

		const v4d_Mapper<VOXEL> &m_mapper;

	protected:
		i4d::Image4d<VOXEL> *m_img;
		bool m_delete_data;
		/** Make dirty caches related to given projections and send repaint event */
		virtual void MakeDirtyAndRefresh(v4d_ProjectionType pt);

	public:
		/** Constructor. It creates an instance of the 3d canvas and
		initialize all its attributes according to the given image or to default
		values if img = NULL.  delete_data flag specifies wheather the image data
		is deleted in destructor or not. */
		v4d_ImageViewer3dCanvas(wxWindow *parent, const v4d_Mapper<VOXEL> &mapper, 
			i4d::Image4d<VOXEL> *img = NULL, bool delete_data = true,
			wxWindowID id = -1, const wxPoint& pos = wxDefaultPosition, 
			const wxSize& size = wxDefaultSize);
		
		/** Destructor. It frees all allocated memory. It can delete also image data
		if requested in constructor or in SetImage function. */
		virtual ~v4d_ImageViewer3dCanvas();

		/** \retuns list of wxIcons with different sizes. Each icon contains an image shown in xy window. 
		    If this window is hidden, then wxIcons remains unchanging. */
		virtual void GetIcons(wxIconBundle &bundle);
        /** Get image size. */
        virtual i3d::Vector3d<size_t> GetImageSize() { return m_img->GetSize(); }
		/** Get settings of the canvas. It can be used to store settings 
		before destroying and applying it later to a new instance 
		(\sa SetSettings())
		*/
		void GetSettings(v4d_ImageCanvasSettings &params) const;
		/** Set several parameters of a canvas. */
		void SetSettings(const v4d_ImageCanvasSettings &params);

		virtual int GetTime() { return m_vv.time; }
		virtual int GetImageCount() { return m_img->GetNumFrames(); }
		virtual i3d::Resolution GetResolution(){ return m_img->GetResolution(); }

		/** Draw RGB image projection to specified window. */
		virtual void Draw(wxDC &dc, v4d_ProjectionType pt);

		virtual void OnCutChange(v4d_ProjectionType pt);
		virtual void OnVis3dChange(v4d_ProjectionType pt);
		virtual void OnContentChange(v4d_ProjectionType pt);
		virtual void MappingChanged(v4d_MappingEvent &evt);
		virtual void OnDataChange();
		virtual void OnTimeChange();
		virtual void Save3dData(const wxString &fname);
		virtual void SaveActualView(const char *fname, int d, v4d_ProjectionType projection);
		virtual void SaveGalleryView(const char *fname, int start, int end, int step, bool fixedRows,
			int count, int spacing);

	protected:
		/** Reverse axis orientation (0=X 1=Y 2=Z) in all projections. */
		virtual void ReverseAxis(int axis);
		/** Updates the vol3D_Visualization data*/
		virtual void UpdateVolVis();

        /////////////////////////////// Info frame ///////////////////////////////

    protected:
		/** \returns Original and mapped colour at given position. */
		void GetColourInfo(v4d_ProjectionType pt, int x, int y, 
            wxString &orig_col, wxColour &mapped_col) const;
		/** \returns String, different for different voxel type, representing no_colour. */
		void GetNoColourInfo(wxString& col) const;

		v4d_DrWin3d *MouseInProjection(const wxPoint& mpos, v4d_DrWin3d *proj, v4d_DrWin3d *old);
        /** Get image type name. */
        wxString GetImageTypeName() const;
        /** Update pixel info - intensity, mapped colour, etc. */
        virtual void InfoFrame_UpdatePixelInfo();
        /** Update image info in info frame. */
        virtual void InfoFrame_UpdateImageInfo();
        /** Update cut position and visbox info. */
        virtual void InfoFrame_UpdateCutAndVisBoxInfo();

	private:
		/** Create icon of given size that contains data from given bitmap. */
		void CreateIcon(wxIcon &icon, int width, int height, const wxBitmap &bmp, const Buffer2d<i3d::RGB> &buff);
		/** Create title for the currently displayed time point in the sequence.*/
		virtual wxString CreateTitle() { return wxFileName::FileName(m_img->GetImageName(m_vv.time)).GetFullName(); };
	};
}

#endif
