/*
 * i4dcore library - 3D+t image representation library 
 *
 * Copyright (C) 2005-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * v4d_image4d.cc
 *
 * Functions for creating and manipulating time-lapse labeled images.
 *
 * Vladimir Ulman (ulman@fi.muni.cz) 2007 and 2015
 */

#ifdef __GNUG__
#pragma interface "v4d_labeledimage4d.h"
#endif

#include <math.h>

/* MS Visual Studio doesn't know roundf() from math.h */
#ifdef _MSC_VER
#define roundf(x)	floorf((x)+0.5f)
#endif


#include <i3d/regions.h>
#include <i3d/image3d.h>
#include "v4d_labeledimage4d.h"
#include <iostream>
#include "v4d_impexp.h"
#include <wx/log.h>

using namespace i3d;

namespace i4d
{

template <class LABEL, class VOXEL>
template <typename T>
void LabeledImage4d<LABEL,VOXEL>::ReadSequence(const wxArrayString &Names,const i4d::Image4d<T>* const refImg)
{
	this->ReadSequence_First(Names,refImg);
	while (this->frames < Names.size())
		this->ReadSequence_Next(Names);
}

template <class LABEL, class VOXEL>
template <typename T>
void LabeledImage4d<LABEL,VOXEL>::ReadSequence_First(const wxArrayString &Names,const i4d::Image4d<T>* const refImg)
{
	if (Names.IsEmpty())
		throw InternalException("ReadSequence(): array Names cannot be empty!");

	//clear any content this sequence image may have had...
	this->DestroyData();
	//frames == 0 due to the DestroyData()

	//read the first image
	wxLogMessage(wxString("L_Reading: ")+Names[this->frames]);
	Image3d<VOXEL> im;
	im.ReadImage(Names[this->frames].mb_str(wxConvFile));

	//if there is a reference image available, check for consistency against this one
	if (refImg)
	{
		if ( (im.GetOffset() != refImg->GetOffset())
			|| (im.GetResolution().GetRes() != refImg->GetResolution().GetRes())
			|| (im.GetSize() != refImg->GetSize()) )
			throw InternalException("ReadSequence(): The frame metadata differs from the reference metadata!");
	}

	//update the metadata of this Image4d
	this->offset=im.GetOffset();
	this->resolution=im.GetResolution();
	this->size=im.GetSize();
	this->data.push_back(CreateRegions(im));
	this->names.push_back(Names[this->frames]);
	this->frames=1;
}

template <class LABEL, class VOXEL>
void LabeledImage4d<LABEL,VOXEL>::ReadSequence_Next(const wxArrayString &Names)
{
	if (Names.IsEmpty())
		throw InternalException("ReadSequence(): array Names cannot be empty!");

	wxLogMessage(wxString("L_Reading: ")+Names[this->frames]);
	Image3d<VOXEL> im;
	im.ReadImage(Names[this->frames].mb_str(wxConvFile)); //also checks consistency on voxel type

	//check consistency on metadata
	if ( (im.GetOffset() != this->offset)
		|| (im.GetResolution().GetRes() != this->resolution.GetRes())
		|| (im.GetSize() != this->size) )
		throw InternalException("ReadSequence(): The frame metadata differs from the sequence metadata!");

	this->data.push_back(CreateRegions(im));
	this->names.Add(Names[this->frames]);
	++(this->frames);
}


template <class LABEL, class VOXEL>
void LabeledImage4d<LABEL,VOXEL>::SaveSequence(const char *NamePrefix,const int Width,const char *NameSuffix,
				 const int StartIndex,const int StopIndex)
{
	if (StopIndex < StartIndex)
		throw InternalException("SaveSequence(): StopIndex cannot be smaller than StartIndex!");

	if (this->TestConsistency())
		throw InternalException("SaveSequence(): The image is not consistent!");

	for (int i=StartIndex; i <= StopIndex; ++i) {
		//construct the filename
		wxString fn(
			wxString::FromUTF8(NamePrefix) + 
			wxString::Format( wxString::Format("%%0%dd",Width) ,i ) +
			wxString::FromUTF8(NameSuffix)
		);

		//report it
		//std::cout << "Saving frame no. " << i << " into file " << fn.str() << std::endl;
		wxLogMessage(wxString("Saving: ")+fn);

		//save it
		(this->data[i])->SaveImage(fn.mb_str(wxConvFile));
	}
}


template <class LABEL, class VOXEL>
i3d::LabeledImage3d<LABEL,VOXEL>* LabeledImage4d<LABEL,VOXEL>::CreateRegions(
		const i3d::Image3d<VOXEL>& img,
		const Neighbourhood &nb)
{
	i3d::LabeledImage3d<LABEL,VOXEL>* limg = new i3d::LabeledImage3d<LABEL,VOXEL>();

	if (overrideLabels)
	{
		limg -> CreateForegroundRegionsFF(img, nb);
	}
	else
	{
		limg -> CopyMetaData(img);

		LABEL val;

		for (size_t i = 0; i < img.GetImageSize(); ++i)
		{
			val = static_cast<LABEL>(img.GetVoxel(i));

			limg -> SetVoxel(i, val);

			if (val > LABEL(0))
			{
				if (limg -> components[val].volume == 0)
				{
					// create a new component
					limg -> components[val] = i3d::ComponentInfo<LABEL, VOXEL>(i3d::VOI<i3d::UPIXELS>(img.GetPos(i), i3d::Vector3d<size_t>(1, 1, 1)), 1, img.GetVoxel(i), limg -> OnBorder(i));
				}
				else
				{
					// extend a component
					limg -> components[val].volume += 1;
					limg -> components[val].border = limg -> components[val].border || limg -> OnBorder(i);
					limg -> components[val].voi += i3d::VOI<i3d::UPIXELS>(img.GetPos(i), i3d::Vector3d<size_t>(1, 1, 1));
				}
			}
		}
	}

	return limg;
}


template <class LABEL, class VOXEL>
void LabeledImage4d<LABEL,VOXEL>::RemoveFrame(const int time)
{
	//sanity internal test
	if ( (frames != data.size()) || (frames != names.size()) )
		throw InternalException("InsertFrame(): Image4d metadata is not consistent.");

	//test the position
	if ((time < 0) || (time >= (int)frames))
		throw InternalException("InsertFrame(): Incorrect value of the time parameter.");

	//find the frame to be removed
	typename std::vector< LabeledImage3d<LABEL, VOXEL>* >::iterator d_iter=data.begin();
	int i=time;
	while (i > 0) { ++d_iter; --i; }

	//first, remove the image from memory
	delete *d_iter;

	//second, remove the reference
	data.erase(d_iter);
	names.RemoveAt(time);
	--frames;
}


/*
 * Explicit Instantiations
 */
template class V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16,bool>;
template class V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY8>;
template class V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16,i3d::GRAY16>;

template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16,       bool>::ReadSequence(const wxArrayString&,const i4d::Image4d<i3d::GRAY8>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY8>::ReadSequence(const wxArrayString&,const i4d::Image4d<i3d::GRAY8>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16,i3d::GRAY16>::ReadSequence(const wxArrayString&,const i4d::Image4d<i3d::GRAY8>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16,       bool>::ReadSequence(const wxArrayString&,const i4d::Image4d<i3d::GRAY16>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY8>::ReadSequence(const wxArrayString&,const i4d::Image4d<i3d::GRAY16>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16,i3d::GRAY16>::ReadSequence(const wxArrayString&,const i4d::Image4d<i3d::GRAY16>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16,       bool>::ReadSequence(const wxArrayString&,const i4d::Image4d<float>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY8>::ReadSequence(const wxArrayString&,const i4d::Image4d<float>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16,i3d::GRAY16>::ReadSequence(const wxArrayString&,const i4d::Image4d<float>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16,       bool>::ReadSequence(const wxArrayString&,const i4d::Image4d<std::complex<float> >* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY8>::ReadSequence(const wxArrayString&,const i4d::Image4d<std::complex<float> >* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16,i3d::GRAY16>::ReadSequence(const wxArrayString&,const i4d::Image4d<std::complex<float> >* const);

template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, bool>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<i3d::GRAY8>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY8>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<i3d::GRAY8>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY16>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<i3d::GRAY8>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, bool>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<i3d::GRAY16>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY8>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<i3d::GRAY16>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY16>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<i3d::GRAY16>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, bool>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<float>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY8>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<float>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY16>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<float>* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, bool>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<std::complex<float> >* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY8>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<std::complex<float> >* const);
template void V4D_DLLEXPORT LabeledImage4d<i3d::GRAY16, i3d::GRAY16>::ReadSequence_First(const wxArrayString&, const i4d::Image4d<std::complex<float> >* const);
}
