/*
 * i4dcore library - 3D+t image representation library 
 *
 * Copyright (C) 2005-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * v4d_image4d.h
 *
 * Functions for creating and manipulating time-lapse labeled images.
 *
 * Vladimir Ulman (ulman@fi.muni.cz) 2015
 */

#ifndef __TIMELAPSELABELEDIMAGE__
#define __TIMELAPSELABELEDIMAGE__

#ifdef __GNUG__
#pragma interface "v4d_labeledimage4d.h"
#endif


#include <i3d/basic.h>
#include <i3d/vector3d.h>
#include <i3d/voi.h>
#include <i3d/resolution.h>
#include <i3d/regions.h>
#include <wx/arrstr.h>
#include <valarray>
#include <vector>
#include "v4d_image4d.h"


namespace i4d
{
	//
	// -------------- 3D sequence images -------------- 
	//
	/** 
	 * \ingroup tli4D
	 */
	// @{

	/** \brief A class for storing sequence of 3D images.
	 * Basically, it consists of a vector of \c LabeledImage3d images where the position of the image
	 * in the vector represents the temporal position. In fact, the index to the vector is exactly
	 * the temporal (time) coordinate. An important consequence is that the entire 4D image is not
	 * layed continuously in the memory. Only each 3D image is continuous.
	 *
	 * This class tries to mimic the class \c LabeledImage3d. This class is not inherited from it.
	 * Some of the methods are preserved except, where appropriate, for the addition of fourth time
	 * coordinate. Some were renamed not to be confused with the-same-named methods in \c LabeledImage3d.
	 *
	 * This is a minimalistic implementation at the moment.
	 *
	 * \author Vladimír Ulman <xulman@fi.muni.cz>
	 * \date 2015
	 */
	template <class LABEL, class VOXEL>
	class LabeledImage4d {
	public:
		// constructors and destructors
		// ----------------------------
		explicit LabeledImage4d(const bool overrideLabels_ = false,
			const i3d::Offset* off_ = 0,
			const i3d::Resolution* res_ = 0,
			const float timeStep_ = 0.0f) : offset(),
			resolution(), size(), frames(0), timeStep(timeStep_),
			data(), names(), overrideLabels(overrideLabels_)
		{
			if (off_) offset = *off_;
			if (res_) resolution = *res_;
		}

		/** Destructor: */
		~LabeledImage4d() { DestroyData(); }


		// Get image properties
		// --------------------

		/// Is the image empty? True if empty.
		bool IsEmpty(void) const			{
			return   (frames == 0) || (size.x == 0)
				|| (size.y == 0) || (size.z == 0);
		}

		/// Get the offset of this image.
		i3d::Offset& GetOffset(void) { return (offset); }
		const i3d::Offset& GetOffset(void) const { return (offset); }

		/// Get the spatial resolution of this image.
		i3d::Resolution& GetResolution(void) { return (resolution); }
		const i3d::Resolution& GetResolution(void) const { return (resolution); }

		/// Get the temporal resolution of this image.
		float GetTimeStep() const	{ return (timeStep); }


		/// Get the number of voxels in one frame.
		size_t GetFrameSize(void) const		{ return (size.x * size.y * size.z); }

		/// Get the number of voxels in one slice.
		size_t GetSliceSize(void) const		{ return (size.x * size.y); }

		/// Get the dimensions of the image frame.
		i3d::Vector3d<size_t> GetSize(void) const { return (size); }


		/// Get the width (horizontal axis) of the image.
		size_t GetWidth(void) const		{ return (size.x); }

		/// Get the height (vertical axis) of the image.
		size_t GetHeight(void) const		{ return (size.y); }

		/// Get number of slices per image.
		size_t GetNumSlices(void) const		{ return (size.z); }

		/// Get the number of frames in the sequence.
		size_t GetNumFrames(void) const		{ return (frames); }

		/// Get the width (horizontal axis) of the image.
		size_t GetSizeX(void) const		{ return (size.x); }

		/// Get the height (vertical axis) of the image.
		size_t GetSizeY(void) const		{ return (size.y); }

		/// Get number of slices per image.
		size_t GetSizeZ(void) const		{ return (size.z); }

		/// Get the number of frames in the sequence.
		size_t GetSizeT(void) const		{ return (frames); }


		// Set image properties
		// --------------------

		/// Set the offset for this image.
		void SetOffset(const i3d::Offset& off_)
		{
			offset = off_;

			typename std::vector< i3d::LabeledImage3d<LABEL,VOXEL>* >::iterator iter = data.begin();
			while (iter != data.end()) {
				(*iter)->SetOffset(off_);
				++iter;
			}
		}

		/// Set the spatial resolution for this image.
		void SetResolution(const i3d::Resolution& res_)
		{
			resolution = res_;

			typename std::vector< i3d::LabeledImage3d<LABEL,VOXEL>* >::iterator iter = data.begin();
			while (iter != data.end()) {
				(*iter)->SetResolution(res_);
				++iter;
			}
		}

		/// Get the temporal resolution of this image.
		void SetTimeStep(const float timeStep_) { timeStep=timeStep_; }


		// Indexing, indepent of frame number since only within Image3d
		// ------------------------------------------------------------

		/// Get the x-coordinate of the i-th voxel in the image data block.
		size_t GetX(const size_t i) const     { return (i % size.x); }

		/// Get the y-coordinate of the i-th voxel in the image data block.
		size_t GetY(const size_t i) const     { return ((i / size.x) % size.y); }

		/// Get the z-coordinate of the i-th voxel in the image data block.
		size_t GetZ(const size_t i) const     { return ((i / size.x) / size.y); }

		/// Get the position of the voxel in the image data block (first position is O).
		size_t GetInFrameIndex(const size_t x, const size_t y, const size_t z) const
			{ return (i3d::GetIndex(x, y, z, size)); }
		size_t GetInFrameIndex(const i3d::Vector3d<size_t>& v) const
			{ return (i3d::GetIndex(v.x, v.y, v.z, size)); }

		/** get the coords from the position in the image vector
		coords is Vector3d<int> type - for relative positions */
		i3d::Coords GetInFrameCoords(size_t i) const	  { return           i3d::Coords(GetX(i), GetY(i), GetZ(i)); }
		i3d::Vector3d<size_t> GetFramePos(size_t i) const { return i3d::Vector3d<size_t>(GetX(i), GetY(i), GetZ(i)); }


		// Voxel & data access, depends on the frame number
		// ------------------------------------------------
		//
		// !! ATTENTION, NO MEMORY IN-BOUND CHECKS ARE PERFORMED !!

		/** Get a voxel value (voxels are numbered from 0): */
		LABEL GetVoxel(const size_t x, const size_t y, const size_t z, const size_t frame) const
		{
			return (data[frame]->GetVoxel(x, y, z));
		}

		/** Get a voxel value (voxels are numbered from 0) */
		LABEL GetVoxel(const i3d::Vector3d<size_t>& v, size_t frame) const
		{
			return (data[frame]->GetVoxel(v.x, v.y, v.z));
		}

		/** Get a voxel value (voxels are numbered from 0) */
		LABEL GetVoxel(const size_t i, const size_t frame) const
		{
			return (data[frame]->GetVoxel(i));
		}

		/** Set a voxel to value (voxels are numbered from 0) */
		void SetVoxel(const size_t x, const size_t y, const size_t z,
							const size_t frame, const LABEL voxel)
		{
			data[frame]->SetVoxel(x, y, z, voxel);
		}

		/** Set a voxel to value (voxels are numbered from 0) */
		void SetVoxel(const i3d::Vector3d<size_t>& v,
							const size_t frame, const LABEL voxel)
		{
			data[frame]->SetVoxel(v.x, v.y, v.z, voxel);
		}

		/** Set a voxel to value (voxels are numbered from 0) */
		void SetVoxel(const size_t i, const size_t frame, const LABEL voxel)
		{
			data[frame]->SetVoxel(i, voxel);
		}


		/** Get a voxel address (voxels are numbered from 0) */
		LABEL* GetVoxelAddr(const size_t x, const size_t y, const size_t z, const size_t frame)
		{
			return (data[frame]->GetVoxelAddr(x, y, z));
		}

		/** Get a voxel address (voxels are numbered from 0) */
		const LABEL* GetVoxelAddr(const size_t x, const size_t y, const size_t z, const size_t frame) const
		{
			const typename std::valarray<LABEL>& d = data[frame]->GetVoxelData();
			const LABEL& v = d[i3d::GetIndex(x, y, z, size)];

			return (&v);
		}

		/** Get a voxel address (voxels are numbered from 0) */
		LABEL* GetVoxelAddr(const size_t i, const size_t frame)
		{
			return (data[frame]->GetVoxelAddr(i));
		}

		/** Get a voxel address (voxels are numbered from 0) */
		const LABEL* GetVoxelAddr(const size_t i, const size_t frame) const
		{
			const typename std::valarray<LABEL>& d = data[frame]->GetVoxelData();
			const LABEL& v = d[i];

			return (&v);
		}

		/** Get the first voxel address (voxels are numbered from 0) */
		LABEL* GetFirstVoxelAddr(const size_t frame)
		{
			return (data[frame]->GetFirstVoxelAddr());
		}

		/** Get the first voxel address (voxels are numbered from 0) */
		const LABEL* GetFirstVoxelAddr(const size_t frame) const
		{
			const typename std::valarray<LABEL>& d = data[frame]->GetVoxelData();
			const LABEL& v = d[0];

			return (&v);
		}


		/** Returns a reference to the property \e data of this \c LabeledImage4d
		 * which is of the \c std::vector< i3d::LabeledImage3d<LABEL,VOXEL>* > type.
		 */
		std::vector< i3d::LabeledImage3d<LABEL,VOXEL>* >& GetVoxelData(void)		{ return(data); }
		const std::vector< i3d::LabeledImage3d<LABEL,VOXEL>* >& GetVoxelData(void) const	{ return(data); }


		/** Returns a direct reference to the \e data of particular frame \e frame
		 * which is of the \c std::valarray<LABEL> data type.
		 */
		std::valarray<LABEL>& GetVoxelData(const size_t frame)
			{ return (data[frame]->GetVoxelData()); }
		const std::valarray<LABEL>& GetVoxelData(const size_t frame) const
		{
			const typename std::valarray<LABEL>& d = data[frame]->GetVoxelData();
			return (d);
		}

		/** Returns a pointer to the \c LabeledImage3d (a frame) at given position. No bounds
		 * checking, thus may crash if called inappropriatly.
		 */
		i3d::LabeledImage3d<LABEL,VOXEL>* GetImage(const size_t frame)
			{ return(data[frame]); }
		const i3d::LabeledImage3d<LABEL,VOXEL>* GetImage(const size_t frame) const
			{ return(data[frame]); }

		i3d::LabeledImage3d<LABEL,VOXEL>* operator[](const size_t frame)
			{ return(data[frame]); }
		const i3d::LabeledImage3d<LABEL,VOXEL>* operator[](const size_t frame) const
			{ return(data[frame]); }

		/// Returns reference to a string with a name of the particular image (=file).
		const wxString &GetImageName(const size_t frame) const
			{ return names[frame]; }

		/// Sets a string with a name of the particular image (=file).
		void SetImageName(const size_t frame, const wxString &name)
			{ names[frame]=name; }


		// Bounding box testing functions
		//-------------------------------

		/// Is (x,y,z) on the border of the frame?
		bool OnFrameBorder(const size_t x, const size_t y, const size_t z) const
		{
			return (size.x != 1 && (x == 0 || x == size.x - 1))
				|| (size.y != 1 && (y == 0 || y == size.y - 1))
				|| (size.z != 1 && (z == 0 || z == size.z - 1));
		}

		/// Is (x,y,z) (encoded in Vector3d) on the border of the frame?
		bool OnFrameBorder(const i3d::Vector3d<size_t> v) const
		{
			return OnFrameBorder(v.x, v.y, v.z);
		}

		//Is the given frame number the first or the last frame number of the sequence?
		bool OnSequenceBorder(const size_t frm) const
		{
			return ((frm == 0) || (frm == frames - 1));
		}

		bool OnBorder(const size_t x, const size_t y, const size_t z, const size_t frm) const
		{
			return (size.x != 1 && (x == 0 || x == size.x - 1))
				|| (size.y != 1 && (y == 0 || y == size.y - 1))
				|| (size.z != 1 && (z == 0 || z == size.z - 1))
				|| (frames != 1 && (frm == 0 || frm == frames - 1));
		}

		bool OnBorder(const i3d::Vector3d<size_t> v, const size_t frm) const
		{
			return OnBorder(v.x, v.y, v.z, frm);
		}

		/// Is the voxel [x,y,z] in this image?
		bool Include(const int x, const int y, const int z) const
		{
			return (x >= 0) && (y >= 0) && (z >= 0) &&
				(x <= (int)size.x - 1) &&
				(y <= (int)size.y - 1) &&
				(z <= (int)size.z - 1);
		}

		/// Is the voxel 'v' in this image?
		bool Include(const i3d::Vector3d<int>& v) const
		{
			return Include(v.x, v.y, v.z);
		}

		/// Is the voxel [x,y,z,t] in this image?
		bool Include(const int x, const int y, const int z, const int t) const
		{
			return (x >= 0) && (y >= 0) && (z >= 0) &&
				(x <= (int)size.x - 1) &&
				(y <= (int)size.y - 1) &&
				(z <= (int)size.z - 1) &&
				(t >= 0) && (t < (int)frames - 1);
		}

		/// Is the voxel 'v'+t in this image?
		bool Include(const i3d::Vector3d<int>& v, const int t) const
		{
			return Include(v.x, v.y, v.z, t);
		}


		// Loading and Saving of Image4d
		//--------------------------------------------------------------------------
		/**
		 * Reads sequence of images and stores them in \c Image4d.
		 * Previous data will be destroyed
		 *
		 * \param[in] Names	vector of paths to images
		 *
		 * All processed images should be of the same dimension and have the same metadata.
		 *
		 * \throws InternalException when some image has different offset, resolution or syze
		 */
		template <typename T>
		void ReadSequence(const wxArrayString &names,const i4d::Image4d<T>* const refImg=NULL);

	//private:
		/** Very internal functions that together do the job of LabeledImage4d::ReadSequence(),
		 * this one reads the very first image of the sequence and initiates local metadata. */
		template <typename T>
		void ReadSequence_First(const wxArrayString &names,const i4d::Image4d<T>* const refImg=NULL);
		/** Very internal functions that together do the job of LabeledImage4d::ReadSequence(),
		 * this one reads every next image on the list and checks for consistency. */
		void ReadSequence_Next(const wxArrayString &names);

	public:
		/**
		 * Stores \c Image4d sequence into several files. The specification of filenames
		 * is according to this example:
		 *
		 * to save files: green_0001.ics to green_0012.ics
		 * is translated into: SaveSequence("green_",4,".ics",1,12);
		 *
		 * \param[in] NamePrefix	the part of filename before the index number
		 * \param[in] Width		the width of zero-padded index number in the filename
		 * \param[in] NameSuffix	the part of filename after the index number
		 * \param[in] StartIndex	the first index number
		 * \param[in] StopIndex		the last index number
		 *
		 * All saved images will be of the same dimension and will have the same metadata.
		 *
		 * Parameters \e StartIndex and \e StopIndex are relevant both to filenames and indices
		 * within the vector of frames in the \c Image4d. In other words, only a subsequence can
		 * be saved owing to the two parameters.
		 */
		void SaveSequence(const char *NamePrefix, const int Width, const char *NameSuffix,
			const int StartIndex, const int StopIndex);


		// Low level functions
		//--------------------------------------------------------------------------
		/** Free memory occupied by data and set size to (0,0,0,\e frames). It scans the entire
		 * Image4d vector \e data (the 4D image sequence) and iteratively calls \c Image3d::DisposeData().
		 * In particular, the \c Image3d objects are not removed from memory.
		 * The \c std::vector \e data and \e names retain their lengths and content.
		 *
		 * In other words, the frames data is released from memory while frames themselves are preserved.
		 */
		void DisposeData()
		{
			size.x = size.y = size.z = 0;

			typename std::vector< i3d::LabeledImage3d<LABEL,VOXEL>* >::iterator iter = data.begin();
			while (iter != data.end()) {
				(*iter)->DisposeData();
				++iter;
			}
		}

		/** Free memory occupied by data and set size to (0,0,0,0). It scans the entire
		 * Image4d vector \e data (the 4D image sequence) and iteratively destroys \c Image3d
		 * frames. The \c std::vector \e data is shortened to zero.
		 *
		 * In other words, the frames data is released from memory as well as the frames themselves.
		 */
		void DestroyData()
		{
			size.x = size.y = size.z = 0;

			typename std::vector< i3d::LabeledImage3d<LABEL,VOXEL>* >::iterator iter = data.begin();
			while (iter != data.end()) {
				delete *iter; //destroy the Image3d object
				*iter = NULL;
				++iter;
			}

			frames = 0; //no Image3ds assigned
			data.clear(); //clear the Image4d vector
			names.clear();
		}

		/** Allocate/reallocate voxel memory. Set the image size and sequence length.
		 * The input are three dimensions of all frames and the number of frames.
		 *
		 * If the new sequence length is shorter than the original length then the appropriate
		 * number of frames is removed from memory (their destructor functions are called).
		 *
		 * \note There is absolutely no guarantee on the content of new 4D image. Original image is lost.
		 */
		void MakeRoom(const size_t width, const size_t height, const size_t slices, const size_t frms)
		{
			int frm_no = 0, frm_total = (int)frms;
			names.clear();

			typename std::vector< i3d::LabeledImage3d<LABEL,VOXEL>* >::iterator iter = data.begin();
			//resize as much as you can within the existing sequence
			while ((iter != data.end()) && (frm_no < frm_total)) {
				(*iter)->MakeRoom(width, height, slices);
				(*iter)->SetOffset(offset);
				(*iter)->SetResolution(resolution);
				names.push_back(std::string("no file specified"));

				++iter; ++frm_no;
			}

			if (frm_no == frm_total) {
				//there may be possibly something left in the original sequence
				while (iter != data.end()) {
					delete *iter;
					*iter = NULL;
					++iter;
				}
				data.resize(frms);
				names.resize(frms);
			}
			else
			{
				//it must hold: frm_no < frm_total what implies iter == data.end(),
				//thus we need to add some new frames
				while (frm_no < frm_total) {
					i3d::LabeledImage3d<LABEL,VOXEL> *im = new i3d::LabeledImage3d<LABEL,VOXEL>;
					im->MakeRoom(width, height, slices);
					im->SetOffset(offset);
					im->SetResolution(resolution);

					data.push_back(im);
					names.push_back(std::string("no file specified"));
					++frm_no;
				}
				//data.resize() adjusts automatically during "push_backing"
			}

			size.x = width; size.y = height; size.z = slices;
			frames = frms;
		}

		/** Allocate/reallocate voxel memory. Set the image size and sequence length.
		 * The input are three dimensions (in the vector form) of all frames and the number of frames.
		 *
		 * If the new sequence length is shorter than the original length then the appropriate
		 * number of frames is removed from memory (their destructor functions are called).
		 *
		 * \note There is absolutely no guarantee on the content of new 4D image. Original image is lost.
		 */
		void MakeRoom(const i3d::Vector3d<size_t> &sz, const size_t frms)
			{ MakeRoom(sz.x, sz.y, sz.z, frms); }

	protected:
		/**
		 * Creates a labeled image from the given image \e img according to the
		 * state given in LabeledImage4d::overrideLabels parameter. If this parameter
		 * is true then the input parameter \e nb is considered and both input
		 * parameters are supplied to the LabeledImage3d::CreateForegroundRegionsFF().
		 * Otherwise, this function just sweeps the input image \e img and creates
		 * i3d::ComponentInfo accordingly on labels it possibly finds in the \e img.
		 */
		i3d::LabeledImage3d<LABEL,VOXEL>* CreateRegions(const i3d::Image3d<VOXEL>& img,
		                                           const i3d::Neighbourhood &nb = i3d::nb3D_6);

	public:
		/**
		 * Removes image/frame from the sequence at index \e time.
		 *
		 * The length of this sequence is decreased (if not zero already).
		 */
		void RemoveFrame(const int time);

		/** Test internal consistency of \c Image4d. It controls that the length of frames
		 * vector \e data has the same length as stored in property \e frames. Moreover, it
		 * controls that each frame in the sequence has the same dimension as stated in the
		 * property \e size.
		 *
		 * It is not testing whether the minValue or maxValue hold valid values.
		 *
		 * The function returns zero when everything is OK.
		 */
		int TestConsistency(void) const {
			//test the lengths
			if (frames != data.size()) return(1);
			if (frames != names.size()) return(1);

			//test that each frame has the same metadata as the entire class Image4d
			typename std::vector< i3d::LabeledImage3d<LABEL,VOXEL>* >::const_iterator iter = data.begin();
			while (iter != data.end()) {
				if ((*iter)->GetSize() != size) return(2);
				if ((*iter)->GetOffset() != offset) return(3);
				if ((*iter)->GetResolution().GetRes() != resolution.GetRes()) return(4);

				++iter;
			}

			return(0);
		}

	protected:
		/// position of the image in absolute coordinates in micrometers
		i3d::Offset offset;

		/// image resolution in pixels per micrometer
		i3d::Resolution resolution;

		/// size of the image (width x height x slices) in pixels 
		i3d::Vector3d<size_t> size;

		// number of frames
		size_t frames;

		/// delay between frames (in seconds)
		float timeStep;

		/// image data buffer... a std::vector of labeled image3d
		std::vector< i3d::LabeledImage3d<LABEL,VOXEL>* > data;

		/// list of image names, same order as data
		wxArrayString names;

		/**
		 * Should keep labels in the newly opened masks or re-label?
		 *
		 * All methods that incorporate new images into this sequence
		 * should consider the value of this parameter. To make all
		 * images in the sequence consistent, this parameter can be
		 * set only in the constructor, i.e., only once at the very
		 * beginning; and cannot be altered anytime later.
		 */
		bool overrideLabels;
	};

	// @}
}

#endif
