/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * v4d_labeledimagecanvas3d.cpp
 *
 * Canvas for drawing 3d images together with labeled images
 *
 * Martin Maka (xmaska@fi.muni.cz) 2007
 */

#include "v4d_basic.h"
#include <i3d/transform.h>
#include <wx/busyinfo.h>
#include <wx/minifram.h>
#include <map>
#include "v4d_frame.h"
#include "v4d_labeledimagecanvas3d.h"
#include "palette.h"
#include "wdr/info_wdr.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

namespace v4d 
{

	/***************************************************************************/
	//stolen from i3dlibs and modified for std::complex<> ...
	/***************************************************************************/

	/**
	* Calculates the maximum projection \e dest from the input image \e src.
	* Both images are, however, expected to have voxels of std::complex<T> type.
	* Maximality is evaluated with the \e ext parameter, which is any class with
	* defined operator(): const SOME_SCALAR operator() (const std::complex<T> &) const .
	* An output pixel P is any such that \e exp(P) is maximum.
	*/
	template < class VOXEL, class EXTRACTOR >
	void Get_AF_XY(const i3d::Image3d <VOXEL> &src, i3d::Image3d <VOXEL> &dest,
		const EXTRACTOR &ext)
	{
		dest.MakeRoom(src.GetSizeX(), src.GetSizeY(), 1);
		dest.SetOffset(i3d::Offset(src.GetOffset().x, src.GetOffset().y, 0));
		dest.SetResolution(src.GetResolution());

		const VOXEL* src_data = src.GetFirstVoxelAddr();
		VOXEL* dest_data = dest.GetFirstVoxelAddr();

		const size_t max_z = src.GetSizeZ();
		const size_t slice_size = src.GetSliceSize();

		size_t i;
		VOXEL max_intensity;

		for (size_t j = 0; j < slice_size; j++)
		{
			i = j;
			max_intensity = src_data[i];
			for (size_t z = 1; z < max_z; z++)
			{
				i += slice_size;
				if (ext(max_intensity) < ext(src_data[i]))
					max_intensity = src_data[i];
			}
			dest_data[j] = max_intensity;
		}
	}

	/**
	* Calculates the maximum projection \e dest from the input image \e src.
	* Both images are, however, expected to have voxels of std::complex<T> type.
	* Maximality is evaluated with the \e ext parameter, which is any class with
	* defined operator(): const SOME_SCALAR operator() (const std::complex<T> &) const .
	* An output pixel P is any such that \e exp(P) is maximum.
	*/
	template < class VOXEL, class EXTRACTOR >
	void Get_AF_XZ(const i3d::Image3d <VOXEL> &src, i3d::Image3d <VOXEL> &dest,
		const EXTRACTOR &ext)
	{
		dest.MakeRoom(src.GetSizeX(), 1, src.GetSizeZ());
		dest.SetOffset(i3d::Offset(src.GetOffset().x, 0, src.GetOffset().z));
		dest.SetResolution(src.GetResolution());

		const VOXEL* src_data = src.GetFirstVoxelAddr();
		VOXEL* dest_data = dest.GetFirstVoxelAddr();

		const size_t max_y = src.GetSizeY();
		const size_t size_x = src.GetSizeX();
		const size_t slice_size = src.GetSliceSize();
		const size_t new_size = src.GetSizeX() * src.GetSizeZ();

		size_t i;
		VOXEL max_intensity;

		for (size_t j = 0; j < new_size; j++)
		{
			i = (j / size_x) * slice_size + (j % size_x);
			max_intensity = src_data[i];
			for (size_t y = 1; y < max_y; y++)
			{
				i += size_x;
				if (ext(max_intensity) < ext(src_data[i]))
					max_intensity = src_data[i];
			}
			dest_data[j] = max_intensity;
		}
	}

	/**
	* Calculates the maximum projection \e dest from the input image \e src.
	* Both images are, however, expected to have voxels of std::complex<T> type.
	* Maximality is evaluated with the \e ext parameter, which is any class with
	* defined operator(): const SOME_SCALAR operator() (const std::complex<T> &) const .
	* An output pixel P is any such that \e exp(P) is maximum.
	*/
	template < class VOXEL, class EXTRACTOR >
	void Get_AF_YZ(const i3d::Image3d <VOXEL> &src, i3d::Image3d <VOXEL> &dest,
		const EXTRACTOR &ext)
	{
		dest.MakeRoom(1, src.GetSizeY(), src.GetSizeZ());
		dest.SetOffset(i3d::Offset(0, src.GetOffset().y, src.GetOffset().z));
		dest.SetResolution(src.GetResolution());

		const VOXEL* src_data = src.GetFirstVoxelAddr();
		VOXEL* dest_data = dest.GetFirstVoxelAddr();

		const size_t size_x = src.GetSizeX();
		const size_t slice_size = src.GetSliceSize();
		const size_t size_y = src.GetSizeY();
		const size_t new_size = src.GetSizeY() * src.GetSizeZ();

		size_t i;
		VOXEL max_intensity;

		for (size_t j = 0; j < new_size; j++)
		{
			i = (j % size_y) * size_x + (j / size_y) * slice_size;
			max_intensity = src_data[i];
			for (size_t x = 1; x < size_x; x++)
			{
				i++;
				if (ext(max_intensity) < ext(src_data[i]))
					max_intensity = src_data[i];
			}
			dest_data[j] = max_intensity;
		}
	}


	/***************************************************************************/
	//now the extractors: scalar from complex
	/***************************************************************************/

	template <typename T>
	class RealExtractor
	{
	public:
		const T operator()(const std::complex<T> &val) const
		{
			return(val.real());
		}
	};

	template <typename T>
	class ImagExtractor
	{
	public:
		const T operator()(const std::complex<T> &val) const
		{
			return(val.imag());
		}
	};

	template <typename T>
	class PhaseExtractor
	{
	public:
		const T operator()(const std::complex<T> &val) const
		{
			return(std::atan2(val.imag(), val.real()));
		}
	};

	template <typename T>
	class AmpExtractor
	{
	public:
		const T operator()(const std::complex<T> &val) const
		{
			return(sqrt(val.real()*val.real() + val.imag()*val.imag()));
		}
	};

	/***************************************************************************/

	///wrapper for usual computation of the AF XY
	template <class VOXEL>
	void MyGet_AF_XY(const i3d::Image3d<VOXEL> &img3d, i3d::Image3d<VOXEL> &imgAF,
		const v4d_Mapper<VOXEL> &mapper)
	{
		//wxLogMessage("univerzalni AF XY - casem me odstran");
		i3d::Get_AF_XY(img3d, imgAF);
	}

	///wrapper for special computation of the AF XY from complex images
	template <class T>
	void MyGet_AF_XY(const i3d::Image3d<std::complex<T> > &img3d, i3d::Image3d<std::complex<T> > &imgAF,
		const v4d_Mapper<std::complex<T> > &mapper)
	{
		const typename v4d_ComplexMappingModule<T>::ComplexMode showMode =
			(dynamic_cast<const v4d_ComplexMappingModule<T> &>(mapper)).GetShowMode();
		//wxLogMessage("specializace AF XY - casem me odstran, mode=%u", showMode);

		switch (showMode)
		{
		case v4d_ComplexMappingModule<T>::REAL:
			v4d::Get_AF_XY(img3d, imgAF, RealExtractor<T>());
			break;
		case v4d_ComplexMappingModule<T>::IMAGINARY:
			v4d::Get_AF_XY(img3d, imgAF, ImagExtractor<T>());
			break;
		case v4d_ComplexMappingModule<T>::PHASE:
			v4d::Get_AF_XY(img3d, imgAF, PhaseExtractor<T>());
			break;
		case v4d_ComplexMappingModule<T>::AMPLITUDE:
			v4d::Get_AF_XY(img3d, imgAF, AmpExtractor<T>());
			break;
		default:
			wxLogError("Don't know how to 'GetAF' in this mode of complex numbers.");
		}
	}

	///wrapper for usual computation of the AF XZ
	template <class VOXEL>
	void MyGet_AF_XZ(const i3d::Image3d<VOXEL> &img3d, i3d::Image3d<VOXEL> &imgAF,
		const v4d_Mapper<VOXEL> &mapper)
	{
		//wxLogMessage("univerzalni AF XZ - casem me odstran");
		i3d::Get_AF_XZ(img3d, imgAF);
	}

	///wrapper for special computation of the AF XZ from complex images
	template <class T>
	void MyGet_AF_XZ(const i3d::Image3d<std::complex<T> > &img3d, i3d::Image3d<std::complex<T> > &imgAF,
		const v4d_Mapper<std::complex<T> > &mapper)
	{
		const typename v4d_ComplexMappingModule<T>::ComplexMode showMode =
			(dynamic_cast<const v4d_ComplexMappingModule<T> &>(mapper)).GetShowMode();
		//wxLogMessage("specializace AF XZ - casem me odstran, mode=%u", showMode);

		switch (showMode)
		{
		case v4d_ComplexMappingModule<T>::REAL:
			v4d::Get_AF_XZ(img3d, imgAF, RealExtractor<T>());
			break;
		case v4d_ComplexMappingModule<T>::IMAGINARY:
			v4d::Get_AF_XZ(img3d, imgAF, ImagExtractor<T>());
			break;
		case v4d_ComplexMappingModule<T>::PHASE:
			v4d::Get_AF_XZ(img3d, imgAF, PhaseExtractor<T>());
			break;
		case v4d_ComplexMappingModule<T>::AMPLITUDE:
			v4d::Get_AF_XZ(img3d, imgAF, AmpExtractor<T>());
			break;
		default:
			wxLogError("Don't know how to 'GetAF' in this mode of complex numbers.");
		}
	}

	///wrapper for usual computation of the AF YZ
	template <class VOXEL>
	void MyGet_AF_YZ(const i3d::Image3d<VOXEL> &img3d, i3d::Image3d<VOXEL> &imgAF,
		const v4d_Mapper<VOXEL> &mapper)
	{
		//wxLogMessage("univerzalni AF YZ - casem me odstran");
		i3d::Get_AF_YZ(img3d, imgAF);
	}

	///wrapper for special computation of the AF YZ from complex images
	template <class T>
	void MyGet_AF_YZ(const i3d::Image3d<std::complex<T> > &img3d, i3d::Image3d<std::complex<T> > &imgAF,
		const v4d_Mapper<std::complex<T> > &mapper)
	{
		const typename v4d_ComplexMappingModule<T>::ComplexMode showMode =
			(dynamic_cast<const v4d_ComplexMappingModule<T> &>(mapper)).GetShowMode();
		//wxLogMessage("specializace AF YZ - casem me odstran, mode=%u", showMode);

		switch (showMode)
		{
		case v4d_ComplexMappingModule<T>::REAL:
			v4d::Get_AF_YZ(img3d, imgAF, RealExtractor<T>());
			break;
		case v4d_ComplexMappingModule<T>::IMAGINARY:
			v4d::Get_AF_YZ(img3d, imgAF, ImagExtractor<T>());
			break;
		case v4d_ComplexMappingModule<T>::PHASE:
			v4d::Get_AF_YZ(img3d, imgAF, PhaseExtractor<T>());
			break;
		case v4d_ComplexMappingModule<T>::AMPLITUDE:
			v4d::Get_AF_YZ(img3d, imgAF, AmpExtractor<T>());
			break;
		default:
			wxLogError("Don't know how to 'GetAF' in this mode of complex numbers.");
		}
	}


    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                      ImageViewer3dCanvas                    //
    //                                                             //
    /////////////////////////////////////////////////////////////////

    template <class IDATA, class LABEL, class LDATA> 
    v4d_LabeledImageViewer3dCanvas<IDATA, LABEL, LDATA>::v4d_LabeledImageViewer3dCanvas 
        (wxWindow *parent, const v4d_Mapper<IDATA> &mapper, v4d_Overlayer<LABEL, LDATA> &overlayer, 
        i4d::Image4d<IDATA> *imgs, bool delete_data, wxWindowID id, const wxPoint &pos, const wxSize &size)
        : v4d_ImageViewer3dCanvas<IDATA>(parent, mapper, imgs, delete_data, id, pos, size),
        m_xy_overlay(true), m_xz_overlay(true), m_yz_overlay(true), m_overlayer(overlayer)
	{
		m_lab_xy_origbuf = m_overlayer.Buffer_Create(v4d_OVERLAY_WIN_XY);
		m_lab_xz_origbuf = m_overlayer.Buffer_Create(v4d_OVERLAY_WIN_XZ);
		m_lab_yz_origbuf = m_overlayer.Buffer_Create(v4d_OVERLAY_WIN_YZ);
		m_lab_xy_af = m_overlayer.Buffer_Create(v4d_OVERLAY_WIN_XY_AF);
		m_lab_xz_af = m_overlayer.Buffer_Create(v4d_OVERLAY_WIN_XZ_AF);
		m_lab_yz_af = m_overlayer.Buffer_Create(v4d_OVERLAY_WIN_YZ_AF); 
	}

    template <class IDATA, class LABEL, class LDATA> 
    v4d_LabeledImageViewer3dCanvas<IDATA, LABEL, LDATA>::~v4d_LabeledImageViewer3dCanvas()
    {
    }

    template <class IDATA, class LABEL, class LDATA>
    void v4d_LabeledImageViewer3dCanvas<IDATA, LABEL, LDATA>::Draw(wxDC &dc, v4d_ProjectionType pt)
    {
        if (m_img == NULL)
            return;

        // If empty image is supplied don't do any overlaying
        if (m_overlayer.IsEmpty(false, m_vv.time))
        {
            v4d_ImageViewer3dCanvas<IDATA>::Draw (dc, pt);
            return;
        }

        dc.SetTextForeground(*wxGREEN);
        if (pt == pt_XY)
        {
            if (m_xy_dirty && m_xy->IsShown())
            {
                // clear dirty flag
                m_xy_dirty = false;
                m_xy_remap = true;
                m_xy_overlay = true;

                // resize buffers to the size of drwin
                wxCoord w,h;
                m_xy->GetClientSize(&w, &h);

                // Set size of visible part
                i3d::Vector3d<size_t> sz;
                sz.x = w;
                sz.y = h;
                sz.z = 1;

                if (m_xy_rgbbuf.GetWidth() != w || m_xy_rgbbuf.GetHeight() != h)
                {
                    m_xy_origbuf.MakeRoom(sz);
                    m_xy_rgbbuf.resize(w, h);
                }

                // Set offset and resolution of visible part
                i3d::Vector3d<float> off;
                off.x = m_vv.vis3d.offset.x;
                off.y = m_vv.vis3d.offset.y;
                off.z = GetContent(pt) == pc_Ortho ? m_vv.cut.z : 0;

                m_xy_origbuf.SetOffset (off);
                m_xy_origbuf.SetResolution (m_vv.res);
                m_overlayer.Buffer_SetDim (m_lab_xy_origbuf, sz, off, m_vv.res);
                m_overlayer.Buffer_SetDim (m_lab_xy_af, sz, off, m_vv.res);

                // copy data to orig buffers
                if (GetContent(pt) == pc_Ortho)
                {
					i3d::CopyToImage(*(m_img->GetImage(m_vv.time)), m_xy_origbuf);
                    m_overlayer.Buffer_Render (m_lab_xy_origbuf, m_vv.time);
                }
                else
                {
                    if (m_xy_af_dirty[m_vv.time])
                    {
                        m_xy_af_dirty[m_vv.time] = false;
						MyGet_AF_XY(*(m_img->GetImage(m_vv.time)), m_xy_af[m_vv.time], m_mapper);
                    }

                    i3d::CopyToImage(m_xy_af[m_vv.time], m_xy_origbuf);                  
                }
            }

            if (m_xy_remap || m_xy_overlay)
            {              
                if (GetContent(pt) != pc_Ortho)
					m_overlayer.Buffer_Render(m_lab_xy_af, m_vv.time);

                // map data to rgb
                m_mapper.Map(m_xy_origbuf.GetFirstVoxelAddr(), m_xy_rgbbuf);
                m_xy_remap = false;

                // overlay data
                if (GetContent(pt) != pc_Ortho)
                    m_overlayer.Overlay(m_lab_xy_af, m_xy_rgbbuf);
                else
                    m_overlayer.Overlay(m_lab_xy_origbuf, m_xy_rgbbuf);
                m_xy_overlay = false;

                // reverse orientation and convert it to bitmap
                wxImage wx_img(m_xy_rgbbuf.GetWidth(), m_xy_rgbbuf.GetHeight(), (unsigned char*) m_xy_rgbbuf.GetData(), true);

                if (m_vv.rev_axis[0]) 
                {
                    wx_img = wx_img.Mirror(true);
                }

                if (m_vv.rev_axis[1]) 
                {
                    wx_img = wx_img.Mirror(false);
                }

                m_xy_bmp = wxBitmap(wx_img);

                // send request for refreshing icon 
                wxUpdateUIEvent e(SI_REFRESH_ICON);
                wxEvtHandler::AddPendingEvent(e);
            }

            // draw bitmap on screen
            dc.DrawBitmap(m_xy_bmp, 0, 0);
        }

        if (pt == pt_XZ)
        {
            if (m_xz_dirty && m_xz->IsShown())
            {
                // clear dirty flag
                m_xz_dirty = false;
                m_xz_remap = true;
                m_xz_overlay = true;

                // resize buffer to the size of drwin
                wxCoord w,h;
                m_xz->GetClientSize(&w, &h);

                // Set size of visible part
                i3d::Vector3d<size_t> sz;
                sz.x = w;
                sz.y = 1;
                sz.z = h;

                if (m_xz_rgbbuf.GetWidth() != w || m_xz_rgbbuf.GetHeight() != h)
                {
                    m_xz_origbuf.MakeRoom(sz);
                    m_xz_rgbbuf.resize(w, h);
                }

                // Set offset and resolution of visible part
                i3d::Vector3d<float> off;
                off.x = m_vv.vis3d.offset.x;
                off.y = GetContent(pt) == pc_Ortho ? m_vv.cut.y : 0;
                off.z = m_vv.vis3d.offset.z;

                m_xz_origbuf.SetOffset (off);
                m_xz_origbuf.SetResolution (m_vv.res);
                m_overlayer.Buffer_SetDim (m_lab_xz_origbuf, sz, off, m_vv.res);
                m_overlayer.Buffer_SetDim (m_lab_xz_af, sz, off, m_vv.res);

                // copy data to orig buffer
                if (GetContent(pt) == pc_Ortho)
                {
					i3d::CopyToImage(*(m_img->GetImage(m_vv.time)), m_xz_origbuf);
                    m_overlayer.Buffer_Render (m_lab_xz_origbuf, m_vv.time);
                }
                else
                {
                    if (m_xz_af_dirty[m_vv.time])
                    {
                        m_xz_af_dirty[m_vv.time] = false;
						MyGet_AF_XZ(*(m_img->GetImage(m_vv.time)), m_xz_af[m_vv.time], m_mapper);
                    }

                    i3d::CopyToImage(m_xz_af[m_vv.time], m_xz_origbuf);                  
                }
            }

            if (m_xz_remap || m_xz_overlay)
            {
                if (GetContent(pt) != pc_Ortho)
                    m_overlayer.Buffer_Render (m_lab_xz_af, m_vv.time);

                // map data to rgb
                m_mapper.Map(m_xz_origbuf.GetFirstVoxelAddr(), m_xz_rgbbuf);
                m_xz_remap = false;

                // overlay data
                if (GetContent(pt) == pc_Ortho)
                    m_overlayer.Overlay(m_lab_xz_origbuf, m_xz_rgbbuf);
                else
                    m_overlayer.Overlay(m_lab_xz_af, m_xz_rgbbuf);
                m_xz_overlay = false;

                // reverse orientation and convert it to bitmap
                wxImage wx_img(m_xz_rgbbuf.GetWidth(), m_xz_rgbbuf.GetHeight(), (unsigned char*) m_xz_rgbbuf.GetData(), true);

                if (m_vv.rev_axis[0]) 
                {
                    wx_img = wx_img.Mirror(true);
                }

                if (m_vv.rev_axis[2]) 
                {
                    wx_img = wx_img.Mirror(false);
                }

                m_xz_bmp = wxBitmap(wx_img);
            }

            // draw bitmap on screen
            dc.DrawBitmap(m_xz_bmp, 0, 0);
        }

        if (pt == pt_YZ)
        {
            if (m_yz_dirty && m_yz->IsShown())
            {
                // clear dirty flag
                m_yz_dirty = false;
                m_yz_remap = true;
                m_yz_overlay = true;

                // resize buffer to the size of drwin
                wxCoord w,h;
                m_yz->GetClientSize(&w, &h);

                i3d::Vector3d<size_t> sz;
                sz.x = 1;
                sz.y = h;
                sz.z = w;

                if (m_yz_rgbbuf.GetWidth() != h || m_yz_rgbbuf.GetHeight() != w)
                {
                    m_yz_origbuf.MakeRoom(sz);
                    m_yz_rgbbuf.resize(h, w);
                }

                // Set offset and resolution of visible part
                i3d::Vector3d<float> off;
                off.x = GetContent(pt) == pc_Ortho ? m_vv.cut.x : 0;
                off.y = m_vv.vis3d.offset.y;
                off.z = m_vv.vis3d.offset.z;

                m_yz_origbuf.SetOffset (off);
                m_yz_origbuf.SetResolution (m_vv.res);
                m_overlayer.Buffer_SetDim (m_lab_yz_origbuf, sz, off, m_vv.res);
                m_overlayer.Buffer_SetDim (m_lab_yz_af, sz, off, m_vv.res);

                // copy data to orig buffer
                if (GetContent(pt) == pc_Ortho)
                {
					i3d::CopyToImage(*(m_img->GetImage(m_vv.time)), m_yz_origbuf);
                    m_overlayer.Buffer_Render (m_lab_yz_origbuf, m_vv.time);
                }
                else
                {
                    if (m_yz_af_dirty[m_vv.time])
                    {
                        m_yz_af_dirty[m_vv.time] = false;
						MyGet_AF_YZ(*(m_img->GetImage(m_vv.time)), m_yz_af[m_vv.time], m_mapper);
                    }

                    i3d::CopyToImage(m_yz_af[m_vv.time], m_yz_origbuf);
                }
            }

            if (m_yz_remap || m_yz_overlay)
            {
                if (GetContent(pt) != pc_Ortho)
                    m_overlayer.Buffer_Render (m_lab_yz_af, m_vv.time);

                // map data to rgb
                m_mapper.Map(m_yz_origbuf.GetFirstVoxelAddr(), m_yz_rgbbuf);
                m_yz_remap = false;

                // overlay data
                if (GetContent(pt) == pc_Ortho)
                    m_overlayer.Overlay(m_lab_yz_origbuf, m_yz_rgbbuf);
                else
                    m_overlayer.Overlay(m_lab_yz_af, m_yz_rgbbuf);
                m_yz_overlay = false;

                // reverse orientation and convert it to bitmap
                wxImage wx_img = wxImage(m_yz_rgbbuf.GetWidth(), m_yz_rgbbuf.GetHeight(), (unsigned char*) m_yz_rgbbuf.GetData(), true).Rotate90();

                if (m_vv.rev_axis[1]) 
                {
                    wx_img = wx_img.Mirror(false);
                }

                if (!m_vv.rev_axis[2]) 
                {
                    wx_img = wx_img.Mirror(true);
                }

                m_yz_bmp = wxBitmap(wx_img);
            }

            // draw bitmap on screen
            dc.DrawBitmap(m_yz_bmp, 0, 0);
        }

        this->InfoFrame_UpdatePixelInfo();
    }

    template <class IDATA, class LABEL, class LDATA>
    void v4d_LabeledImageViewer3dCanvas<IDATA, LABEL, LDATA>::OverlayingChanged(v4d_OverlayingEvent &evt)
    {
        m_xy_overlay = true;
        m_xz_overlay = true;
        m_yz_overlay = true;

        if (evt.ShouldReload3d())
        {
            m_overlayer.LoadOverlay (false, m_img->GetSize(), m_img->GetOffset(), m_img->GetResolution());
            this->MakeDirtyAndRefresh (pt_All);
            if (this->m_volVis && this->m_volVis->IsShown()) 
            {
                this->UpdateIsosurface();
            }
        }
        else
        {
            this->Refresh(false);
        }

        //overlaying is enabled
        if(this->m_volVis)
        {
            v4d_Palette pal;

            if (m_overlayer.GetPalette() == NULL)
            {
                v4d_GetPalette(v4dPAL_OVERLAY, pal);
            }
            else
            {
                i3d::RGB *overPal = m_overlayer.GetPalette();
                for(size_t i = 0; i < m_overlayer.GetPaletteSize(); i++)
                    pal.push_back(overPal[i]);
            }

            if (m_overlayer.GetOverlaying())
            {
                this->m_volVis->GetVolVisCanvas()->ShowIsosurface(true, pal);
            }
            else
            {
                this->m_volVis->GetVolVisCanvas()->ShowIsosurface(false, pal);
            }

            this->m_volVis->GetVolVisCanvas()->Refresh(false);
        }
    }

    template <class IDATA, class LABEL, class LDATA>
    void v4d_LabeledImageViewer3dCanvas<IDATA, LABEL, LDATA>::OnDataChange()
    {
        m_xy_overlay = true;
        m_xz_overlay = true;
        m_yz_overlay = true;

        v4d_ImageViewer3dCanvas<IDATA>::OnDataChange();
    }

	template <class IDATA, class LABEL, class LDATA>
    void v4d_LabeledImageViewer3dCanvas<IDATA, LABEL, LDATA>::OnTimeChange()
    {
        m_xy_overlay = true;
        m_xz_overlay = true;
        m_yz_overlay = true;

        v4d_ImageViewer3dCanvas<IDATA>::OnTimeChange();
		this->UpdateIsosurface();
    }

    template <class IDATA, class LABEL, class LDATA>
    void v4d_LabeledImageViewer3dCanvas<IDATA, LABEL, LDATA>::UpdateIsosurface()
    {
        vol3D_IsoSurface<LABEL> marchingCube;
        unsigned char sqX, sqY, sqZ;
        i3d::Vector3d<float> res, scale, cell, trans;
        i3d::Vector3d<size_t> size;

        const i3d::LabeledImage3d<LABEL, LDATA> *limg = m_overlayer.GetLabeledImage(false, false, m_vv.time);

		if(m_img == NULL || limg == NULL || !limg->GetImageSize())
            return;

        if(this->m_volVis)
        {
            //show busy info
            wxWindowDisabler disableAll;
            wxBusyInfo wait("Updating isosurface data...", this->m_volVis);

            size = m_img->GetSize();
            sqX = sqY = sqZ = 0;

            for(unsigned short padX = 1; padX < size.x + 2; padX <<= 1, sqX++);  
            for(unsigned short padY = 1; padY < size.y + 2; padY <<= 1, sqY++);
            for(unsigned short padZ = 1; padZ < size.z + 2; padZ <<= 1, sqZ++);

            res = limg->GetResolution().GetRes();

            //Compute scale according to resolution and padded size
            scale.x = res.x;
            scale.y = res.y * float(1<<sqX) / float(1<<sqY);
            scale.z = res.z * float(1<<sqX) / float(1<<sqZ);

            float dmax = std::max((float(size.x) / float(1<<sqX)) / scale.x, 
                (float(size.y) / float(1<<sqY)) / scale.y);
            dmax = std::max((float(size.z) / float(1<<sqZ)) / scale.z, dmax);

            scale *= dmax;

            cell.x = 2.0f/(scale.x*(1<<sqX));
            cell.y = 2.0f/(scale.y*(1<<sqY));
            cell.z = 2.0f/(scale.z*(1<<sqZ));

            //create a map converting labels into sequence information
            typename vol3D_IsoSurface<LABEL>::vol3D_ComponentTable componentInfo;
            size_t j = 0;

            for (typename i3d::ComponentInfo<LABEL, LDATA>::Container::const_iterator i = limg->components.begin();
                i != limg->components.end(); i++, j++)
            {
                componentInfo.insert(std::pair<LABEL,size_t>((LABEL)i->first,j));
            }

            marchingCube.GenerateSegmented(limg->GetFirstVoxelAddr(), componentInfo,
                size.x, size.y, size.z, cell.x, cell.y, cell.z);

            if(!marchingCube.IsValid())
            {
                wxMessageDialog *msg = new wxMessageDialog(this,
                    "Error occured during creation of an isosurface!", "Error", wxOK);
                msg->ShowModal();
                delete msg;
                return;
            }

            v4d_Palette pal;

            if (m_overlayer.GetPalette() == NULL)
            {
                v4d_GetPalette(v4dPAL_OVERLAY, pal);
            }
            else
            {
                i3d::RGB *overPal = m_overlayer.GetPalette();
                for(size_t i = 0; i < m_overlayer.GetPaletteSize(); i++)
                    pal.push_back(overPal[i]);
            }

            if(m_overlayer.GetOverlaying())
                this->m_volVis->GetVolVisCanvas()->ShowIsosurface(true, pal);
            else
                this->m_volVis->GetVolVisCanvas()->ShowIsosurface(false, pal);

            trans.x = -cell.x*size.x/2.0f;
            trans.y = -cell.y*size.y/2.0f;
            trans.z = -cell.z*size.z/2.0f;

            this->m_volVis->GetVolVisCanvas()->SetIsoSurfaceData(marchingCube.GetComponentVector(), trans);
        }
    }

	template <class IDATA, class LABEL, class LDATA>
	void v4d_LabeledImageViewer3dCanvas<IDATA, LABEL, LDATA>::GetLabelInfo(v4d_ProjectionType pt, int x, int y, wxString &col_label) const
	{
		if (pt == pt_XY)
		{
			if (m_vv.rev_axis[0])
			{
				x = m_xy_rgbbuf.GetWidth() - 1 - x;
			}
			if (m_vv.rev_axis[1])
			{
				y = m_xy_rgbbuf.GetHeight() - 1 - y;
			}

			col_label = VoxelToStr<LABEL>(m_overlayer.Buffer_GetValue(GetContent(pt) == pc_Ortho ? m_lab_xy_origbuf : m_lab_xy_af, i3d::Vector3d<size_t>(x, y, 0)));
		}

		if (pt == pt_XZ)
		{
			if (m_vv.rev_axis[0])
			{
				x = m_xy_rgbbuf.GetWidth() - 1 - x;
			}
			if (m_vv.rev_axis[2])
			{
				y = m_xy_rgbbuf.GetHeight() - 1 - y;
			}

			col_label = VoxelToStr<LABEL>(m_overlayer.Buffer_GetValue(GetContent(pt) == pc_Ortho ? m_lab_xz_origbuf : m_lab_xz_af, i3d::Vector3d<size_t>(x, 0, y)));
		}

		if (pt == pt_YZ)
		{
			if (m_vv.rev_axis[2])
			{
				x = m_xy_rgbbuf.GetHeight() - 1 - x;
			}
			if (m_vv.rev_axis[1])
			{				
				y = m_xy_rgbbuf.GetWidth() - 1 - y;
			}

			col_label = VoxelToStr<LABEL>(m_overlayer.Buffer_GetValue(GetContent(pt) == pc_Ortho ? m_lab_yz_origbuf : m_lab_yz_af, i3d::Vector3d<size_t>(0, y, x)));
		}
	}

	template <class IDATA, class LABEL, class LDATA>
	void v4d_LabeledImageViewer3dCanvas<IDATA, LABEL, LDATA>::InfoFrame_Create()
	{
		m_info_frame = new wxMiniFrame(this, wxID_ANY, "Info frame", wxDefaultPosition, wxDefaultSize, 0); 
		m_info_frame->SetMaxSize(wxSize(380,200));
		OverlayInfoWindowFunc(m_info_frame); 
	}

	template <class IDATA, class LABEL, class LDATA>
	void v4d_LabeledImageViewer3dCanvas<IDATA, LABEL, LDATA>::InfoFrame_UpdatePixelInfo()
	{
		if (m_info_frame && m_info_frame->IsShown())
		{
			wxNotebook *ntb = (wxNotebook *)m_info_frame->FindWindow(ID_INFO_NOTEBOOK);

			if (!ntb->GetSelection())
			{
				wxStaticText *box;
				wxPoint scr_pos = wxGetMousePosition(), win_pos;
				i3d::Vector3d<float> pos;
				v4d_DrWin3d *drwin = NULL;

				// Find in which projection is the mouse cursor
				drwin = this->MouseInProjection(scr_pos, m_xy, drwin);
				drwin = this->MouseInProjection(scr_pos, m_xz, drwin);
				drwin = this->MouseInProjection(scr_pos, m_yz, drwin);

				// Get coordinates in microns
				if (drwin)
				{
					win_pos = drwin->ScreenToClient(scr_pos);
					pos.x = drwin->W2MPosX(win_pos.x);
					pos.y = drwin->W2MPosY(win_pos.y);
					pos.z = drwin->CutPos();
					pos = drwin->Reorder(pos);
				}

				if (drwin != NULL && InCO(m_vv.bbox, pos))
				{
					box = (wxStaticText *)ntb->FindWindow(ID_POSITION_MICRONS);
					box->SetLabel(wxString::Format("[%.3f, %.3f, %.3f]", pos.x, pos.y, pos.z));

					i3d::Vector3d<int> pix_pos = (pos - m_img->GetOffset()) * m_img->GetResolution().GetRes();
					box = (wxStaticText *)ntb->FindWindow(ID_POSITION_PIXELS);
					box->SetLabel(wxString::Format("[%d, %d, %d, %d]", pix_pos.x, pix_pos.y, pix_pos.z, (int)m_vv.time));

					wxString col_orig;
					wxColour col_mapped;
					this->GetColourInfo(drwin->thisWin(), win_pos.x, win_pos.y, col_orig, col_mapped);

					box = (wxStaticText *)ntb->FindWindow(ID_INTENSITY_VALUE);
					box->SetLabel(col_orig);

					wxString col_label;
					GetLabelInfo(drwin->thisWin(), win_pos.x, win_pos.y, col_label);
					box = (wxStaticText *)ntb->FindWindow(ID_OVERLAY_VALUE);
					box->SetLabel(col_label);

					box = (wxStaticText *)ntb->FindWindow(ID_MAPPED_COLOUR);
					box->SetLabel(wxString::Format("%3d %3d %3d", col_mapped.Red(),
						col_mapped.Green(), col_mapped.Blue()));

					wxWindow *bmp = ntb->FindWindow(ID_COLOUR_WIN);
					bmp->SetBackgroundColour(col_mapped);
					bmp->ClearBackground();
				}
				else
				{
					box = (wxStaticText *)ntb->FindWindow(ID_POSITION_MICRONS);
					box->SetLabel("---");

					box = (wxStaticText *)ntb->FindWindow(ID_POSITION_PIXELS);
					box->SetLabel("----");

					wxString col_orig;
					this->GetNoColourInfo(col_orig);
					box = (wxStaticText *)ntb->FindWindow(ID_INTENSITY_VALUE);
					box->SetLabel(col_orig);

					box = (wxStaticText *)ntb->FindWindow(ID_MAPPED_COLOUR);
					box->SetLabel("---");

					box = (wxStaticText *)ntb->FindWindow(ID_OVERLAY_VALUE);
					box->SetLabel("---");

					wxWindow *bmp = ntb->FindWindow(ID_COLOUR_WIN);
					//bmp->SetBackgroundColour(box->GetBackgroundColour());
					bmp->SetBackgroundColour(wxColour((unsigned char)255, (unsigned char)255,
						                              (unsigned char)255, wxALPHA_TRANSPARENT));
					bmp->ClearBackground();
				}

				m_info_frame->Fit();
			}
		}
	}

    // EXPLICIT INSTANTIATIONS
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::GRAY8, i3d::GRAY16, bool>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::GRAY8, i3d::GRAY16, i3d::GRAY8>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::GRAY8, i3d::GRAY16, i3d::GRAY16>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<float, i3d::GRAY16, bool>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<float, i3d::GRAY16, i3d::GRAY8>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<float, i3d::GRAY16, i3d::GRAY16>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<int, i3d::GRAY16, bool>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<int, i3d::GRAY16, i3d::GRAY8>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<int, i3d::GRAY16, i3d::GRAY16>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::GRAY16, i3d::GRAY16, bool>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::GRAY16, i3d::GRAY16, i3d::GRAY8>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::GRAY16, i3d::GRAY16, i3d::GRAY16>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::RGB, i3d::GRAY16, bool>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::RGB, i3d::GRAY16, i3d::GRAY8>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::RGB, i3d::GRAY16, i3d::GRAY16>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::RGB16, i3d::GRAY16, bool>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::RGB16, i3d::GRAY16, i3d::GRAY8>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<i3d::RGB16, i3d::GRAY16, i3d::GRAY16>;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<std::complex<float>, i3d::GRAY16, bool >;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<std::complex<float>, i3d::GRAY16, i3d::GRAY8 >;
	template class V4D_DLLEXPORT v4d_LabeledImageViewer3dCanvas<std::complex<float>, i3d::GRAY16, i3d::GRAY16 >;
}
