/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * viewer3d.cpp
 *
 * Simple application for displaying images in v4d Windows
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006
 * 
 */

#include "v4d_basic.h"

#ifndef MSC_VER
	#include "icons/viewer4d.xpm"
#endif

#include <typeinfo>
#include <locale.h>
#include <float.h>
#include <math.h>
#include <cmath>
#include <wx/filedlg.h>
#include <wx/ipc.h>
#include <wx/snglinst.h>
#include <wx/filename.h>
#include <wx/filefn.h>
#include <i3d/image3d.h>
#include "types/prior_log.h"
#include "v4d_frame.h"
#include "v4d_imagecanvas3d.h"
#include "overlaying.h"
#include "wdr/viewer_about_wdr.h"
#include "dlg_openfile.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

namespace viewer3d
{
    //########################################################//
	//
	//                     Declarations
	//
	//       can be moved to a header file in future
	//
	//########################################################//

    /////////////////////////////////////////////////////////////////
	//                                                             //
	//                            MyServer                         //
	//                                                             //
	/////////////////////////////////////////////////////////////////
	class MyServer: public wxServer
	{
	  public:
		wxConnectionBase *OnAcceptConnection(const wxString &topic);
	};

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                          MyConnection                       //
	//                                                             //
	/////////////////////////////////////////////////////////////////
	class MyConnection : public wxConnection
	{
	  public:
        /** Constructor. */
		MyConnection() {};
		/** Destructor. */
        ~MyConnection() {};
		virtual bool OnExec(const wxString &topic, const wxString &data);
	};

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                            MyClient                         //
	//                                                             //
	/////////////////////////////////////////////////////////////////
	class MyClient: public wxClient
	{
	  public:
        /** Constructor. */ 
		MyClient() {};
        wxConnectionBase *OnMakeConnection() { return new MyConnection(); };
	};

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                              MyApp                          //
	//                                                             //
	/////////////////////////////////////////////////////////////////
	/** Main application class. This class is created when the program
	  * starts. */
	class MyApp : public wxApp
	{
	  private:
        wxSingleInstanceChecker *m_single_instance_checker;
		MyServer *m_server;

	  public:
        /** Constructor. */
		MyApp(): wxApp(), m_single_instance_checker(NULL), m_server(NULL) {};
     	/** Funcition called by wxWindows during initialization of the
		application. It creates the main application window. \sa MainFrame*/
		virtual bool OnInit();
		virtual int OnExit();
	};

    /////////////////////////////////////////////////////////////////
	//                                                             //
	//                            MainFrame                        //
	//                                                             //
	/////////////////////////////////////////////////////////////////
	/** Main application window. This class has a menu, status bar
	  * and in the future it will have a toolbar. When this frame 
	  * is destroyed, the application exits. */
    class MainFrame : public wxFrame
    {
	  private:
		wxLogTextCtrl *m_log;
		wxTextCtrl *log_ctrl;
		wxLogChain *m_log_chain;
		int m_num_children; // number of child windows
		bool m_show_frame; // Is the main application window shown?
		wxArrayString m_image_names; // Names of the images

	  public:
		/** Default constructor. It creates and initializes the main
	      * window. \param title Title of the main window. */
		MainFrame(const wxString& title, int id = -1);
        
        /** Destructor. */
		~MainFrame()
        {}

		/** Function called when a user wants open new window of given size for labeled images showing on given 
            screen position.  The function returns a pointer to new window. The parameter suppress_showing 
            allows show or not show the created window on the screen. */
		v4d::v4d_Viewer3dFrame *OpenWin(const wxArrayString &fnames, const wxArrayString &lnames,
            const wxPoint &position = wxDefaultPosition, const wxSize &size = wxDefaultSize, 
            bool suppress_showing = false);

		/** Function called when a user wants open new window of given size for images showing on given
		screen position.  The function returns a pointer to new window. The parameter suppress_showing
		allows show or not show the created window on the screen. */
		v4d::v4d_Viewer3dFrame *OpenWin(const wxArrayString &names,
			const wxPoint &position = wxDefaultPosition, const wxSize &size = wxDefaultSize,
			bool suppress_showing = false);

		/**
		 * Opens the OpenFileDialog, process images as requested in the dialog,
		 * and fires OpenWin() appropriately.
		 */
		void HandleOpenFileDialog(wxArrayString& names);
		
	  private:
        /** Create the File menu. This menu can be created in different way according to the value of 
            the parameter main_frame. If main_frame is true, then the menu is created for the application 
            main window. */
        wxMenu *CreateFileMenu(bool main_frame = true);
						
        /** Template function for opening new window of given size for labeled images showing on given 
            screen position. The function returns a pointer to new window. 
        
				The last parameter indicates whether the data on the disk is binary or not.
				If it is binary, it is implicitly converted to VOXEL type -- a type used for viewing.
        */
		template <class T, class LABEL, class VOXEL>
		v4d::v4d_Viewer3dFrame *CreateWin(const wxArrayString &fnames, const wxArrayString &lnames,
            const wxPoint &position, const wxSize &size,
				const bool sourceIsBinary=false);

		/** Template function for opening new window of given size for images showing on given
		screen position. The function returns a pointer to new window.

		The last parameter indicates whether the data on the disk is binary or not.
		If it is binary, it is implicitly converted to VOXEL type -- a type used for viewing.
		*/
		template <class VDATA>
		v4d::v4d_Viewer3dFrame *CreateWin(const wxArrayString &names,
			const wxPoint &position, const wxSize &size,
			const bool sourceIsBinary=false);
		
        /** Show or hide a given frame. */
		void ShowWin(v4d::v4d_Viewer3dFrame *frame, bool show = true);
        
        /** Reload the image data in a given frame. */
		void ReloadFrame(v4d::v4d_Viewer3dFrame *frame);
        
        /** Function called when a user wants open the new image. */
		void OnOpen(wxCommandEvent &evt);
		/** Function called when a user wants to display some info about the application. */
		void OnAbout(wxCommandEvent& evt);
		/** Function called when a user wants reload image data. */
		void OnReload(wxCommandEvent &evt);
        /** Function called when a user wants quit the application. */
		void OnQuit(wxCommandEvent& evt);
		/** Function called when a user wants close a child window. */
		void OnCloseChild(wxUpdateUIEvent &evt);
		/** Function called when a user wants show or hide main application window. */
		void OnShowMainFrame(wxCommandEvent &evt);
		/** Update menu according to frame state. */
		void OnUpdateUI(wxUpdateUIEvent &evt);

		///last folder used for opening files
		wxString openDialogFolder=wxEmptyString;

		DECLARE_EVENT_TABLE()
	};

    enum { // identifiers
		SI_QUIT = 3000,
		SI_OPEN,
		SI_ABOUT,
		SI_RELOAD,
		SI_SHOW_MAIN_FRAME,
	};

	//########################################################//
	//
	//                End --- Declarations --- End
	//
	//########################################################//

} // namespace viewer3d

namespace viewer3d 
{
	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                          Event tables                       //
	//                                                             //
	/////////////////////////////////////////////////////////////////
	// the event tables connect the wxWindows events with the functions (event
	// handlers) which process them. It can be also done at run-time, but for the
	// simple menu events like this the static method is much simpler.
	
	BEGIN_EVENT_TABLE(MainFrame, wxFrame)
		EVT_MENU(SI_OPEN,  MainFrame::OnOpen)
		EVT_MENU(SI_RELOAD, MainFrame::OnReload)
		EVT_MENU(SI_SHOW_MAIN_FRAME, MainFrame::OnShowMainFrame)
		EVT_MENU(SI_ABOUT,  MainFrame::OnAbout)
		EVT_MENU(SI_QUIT,  MainFrame::OnQuit)
		EVT_UPDATE_UI(v4d::SI_CLOSE_WINDOW, MainFrame::OnCloseChild)
		EVT_UPDATE_UI(SI_SHOW_MAIN_FRAME, MainFrame::OnUpdateUI)
		EVT_UPDATE_UI(SI_RELOAD, MainFrame::OnUpdateUI)
	END_EVENT_TABLE()

} // namespace viewer3d

	// ----------------------------------------------------------- //
    //                        Implementation                       //
    // ----------------------------------------------------------- //

	/////////////////////////////////////////////////////////////////
    //                                                             //
    //                           Main loop                         //
    //                                                             //
    /////////////////////////////////////////////////////////////////

    IMPLEMENT_APP(viewer3d::MyApp)
		
    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                      Local definitions                      //
    //                                                             //
    /////////////////////////////////////////////////////////////////

	#define LABELED_IMAGE_SEPARATOR " + "

	/////////////////////////////////////////////////////////////////
    //                                                             //
    //                       Local functions                       //
    //                                                             //
    /////////////////////////////////////////////////////////////////
	
	/** Create a frame title from an image file name and a labeled image file name. */
	wxString MergeFileNames(const wxString &fname, const wxString &lname)
	{
		if (fname == wxEmptyString)
		{
			wxLogError("The file name must be defined!");
			return wxString("Unknown");
		}

		if (lname == wxEmptyString)
		{
			return fname;
		}
		else
		{
			return wxString::Format("%s%s%s", fname.mb_str(wxConvFile),
			            LABELED_IMAGE_SEPARATOR, lname.mb_str(wxConvFile));
		}
	}

namespace viewer3d 
{
	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                           MyServer                          //
	//                                                             //
	/////////////////////////////////////////////////////////////////

	wxConnectionBase *MyServer::OnAcceptConnection(const wxString &topic)
	{
		if (topic == "Viewer4d")
		{
			// Check that there are no modal dialogs active
			wxWindowList::compatibility_iterator node = wxTopLevelWindows.GetFirst();
			while (node)
			{
				wxDialog *dialog = wxDynamicCast(node -> GetData(), wxDialog);

				if (dialog && dialog -> IsModal())
				{
					return NULL;
				}	

				node = node->GetNext();
			}
			
			return new MyConnection();
		}
		else
		{
			return NULL;
		}
	}

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                         MyConnection                        //
	//                                                             //
	/////////////////////////////////////////////////////////////////

	bool MyConnection::OnExec(const wxString &topic, const wxString &data)
	{
		MainFrame *frame = wxDynamicCast(wxGetApp().GetTopWindow(), MainFrame);
	
		if (data.IsEmpty())
		{
			// Just raise the main application window
			if (frame)
			{
				frame -> Show(TRUE);
				frame -> Raise();
			}
		}
		else
		{
			//unpack filenames from wxString::data to wxArrayString::names
			wxString dd(data); //writeable copy
			wxArrayString names; //output format

			int sepPosition;
			while ((sepPosition=dd.Find(wxString("AWDfhjRFTuioeHQ"))) != wxNOT_FOUND)
			{
				//read the portion before the magic separator string
				names.Add(dd.Mid(0,sepPosition));

				//now clear the read part including the separator
				dd=dd.Mid(sepPosition+15);
			}

			//since the magic separator string was not (already) found,
			//there is one last filename remaining
			names.Add(dd);
			frame->HandleOpenFileDialog(names);
			//the HandleOpenFileDialog eventually ends up calling frame->OpenWin()
		}
		
		return true;
	}

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                              MyApp                          //
	//                                                             //
	/////////////////////////////////////////////////////////////////
	// `Main program' equivalent: the program execution "starts" here
	bool MyApp::OnInit()
	{	
		// the C locales are the only allowed!
		setlocale(LC_NUMERIC, "C");

		m_single_instance_checker = new wxSingleInstanceChecker;

		if (m_single_instance_checker -> IsAnotherRunning())
		{
			// OK, there is another one running, so try to connect to it
			// and send it any filename before exiting.
			MyClient *client = new MyClient();
			// ignored under DDE, host name in TCP/IP based classes
			wxString hostName = "localhost";
			// Create the connection
			wxConnectionBase *connection = client -> MakeConnection(hostName, "4242", "Viewer4d");
			if (connection)
			{
				if (argc == 1)
				{
					// Ask the other instance to raise itself
					connection -> Execute(wxString(""));
				}
				else
				{
					// Ask the other instance to open files from the command line
					//
					// since this instance may be called from different folder then
					// the server  one, we rather supply files with their full paths
					wxString names;
					for (int i = 1; i < argc; i++) {
						if (i > 1) names.Append(wxString("AWDfhjRFTuioeHQ"));
						wxFileName FileName(argv[i]);
						if (FileName.IsAbsolute())
							names.Append(wxString(argv[i]));
						else {
							FileName.MakeAbsolute(wxGetCwd());
							names.Append(FileName.GetFullPath());
						}
					}
					connection -> Execute(names);
				}

				connection -> Disconnect();
				delete connection;
			}
			else
			{
				wxMessageBox("Sorry, the existing instance may be too busy too respond.\nPlease close any open dialogs and retry.", 
					         "Viewer4d", wxICON_INFORMATION | wxOK);
			}

			delete client;
			return false;
		}
		else
		{
			// Create a new server
			m_server = new MyServer();
			
			if (!m_server -> Create("4242"))
			{
				wxLogWarning("Failed to create an IPC service.");
			}
		}

		// Create the main application window
		MainFrame *frame = new MainFrame("Viewer4d", -1);
        
		// show it and tell the application that it's our main window
		// @@@ what does it do exactly, in fact? is it necessary here?
		frame -> SetSize( wxSize(900,250) );
		frame -> Show(TRUE);
		SetTopWindow(frame);

		// Process command line arguments and open corresponding v4d windows
		// (also expand all filenames to their full paths)
		wxArrayString names;
		for (int i = 1; i < argc; i++)
		{
			wxFileName FileName(argv[i]);
			if (FileName.IsAbsolute())
				names.Add(wxString(argv[i]));
			else {
				FileName.MakeAbsolute(wxGetCwd());
				names.Add(FileName.GetFullPath());
			}
		}
		if (names.GetCount() > 0)
			frame->HandleOpenFileDialog(names);
			//the HandleOpenFileDialog eventually ends up calling frame->OpenWin()

		// success: wxApp::OnRun() will be called which will enter the main message
		// loop and the application will run. If we returned FALSE here, the
		// application would exit immediately.
		return TRUE;
	}

	int MyApp::OnExit() 
	{ 
		if (m_single_instance_checker) 
			delete m_single_instance_checker; 

		if (m_server)
			delete m_server;
		
		return 0; 
	}

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                           Main Frame                        //
	//                                                             //
	/////////////////////////////////////////////////////////////////

	// frame constructor
	MainFrame::MainFrame(const wxString& title, int id): 
		wxFrame(NULL, id, title, wxDefaultPosition,wxDefaultSize, wxDEFAULT_FRAME_STYLE), 
		m_num_children(0), m_show_frame(true)
	{
		SetIcon(wxICON(icon_viewer3d));

		v4d::PriorityLog *prior_log = new v4d::PriorityLog(wxLOG_Error);
		delete ::wxLog::SetActiveTarget(prior_log);

		log_ctrl = new wxTextCtrl(this, -1, wxEmptyString, wxDefaultPosition, 
            wxDefaultSize, wxTE_MULTILINE | wxTE_READONLY);
		m_log = new wxLogTextCtrl(log_ctrl);
		m_log_chain = new wxLogChain(m_log);

		////////////////////////////////
		// Menu File
		////////////////////////////////
		wxMenu *menuFile = CreateFileMenu();

	    ////////////////////////////////
		// MenuBar
		////////////////////////////////
		wxMenuBar *menuBar = new wxMenuBar;
		menuBar -> Append(menuFile, "&File");
		// ... and attach this menu bar to the frame
		SetMenuBar(menuBar);

	    // create status bar with 3 panes
		int st_widths[2] = { -1, 60 };
		CreateStatusBar(2);
		SetStatusWidths (2, st_widths);
	}


	bool CheckImagesTypeLogErrors(i3d::ImgVoxelType vt, const wxArrayString& names)
	{
		for (size_t i = 0; i < names.GetCount(); i++)
		{
			if (vt != i3d::ReadImageType(names.Item(i).mb_str(wxConvFile)))
			{
				wxLogError("%s: Formats of images are not same", names.Item(i).mb_str(wxConvFile));
				return false;
			}
		}

		return true;
	}

	v4d::v4d_Viewer3dFrame *MainFrame::OpenWin(const wxArrayString &names,
		const wxPoint &position, const wxSize &size, bool suppress_showing)
	{

		i3d::ImgVoxelType vt;
		v4d::v4d_Viewer3dFrame *frame = NULL;

		try
		{
			vt = i3d::ReadImageType(names.Item(0).mb_str(wxConvFile));
			if (!CheckImagesTypeLogErrors(vt, names))
				return NULL;

			switch (vt)
			{
			case i3d::Gray8Voxel:
				frame = CreateWin<i3d::GRAY8>(names, position, size);
				break;

			case i3d::Gray16Voxel:
				frame = CreateWin<i3d::GRAY16>(names, position, size);
				break;

			case i3d::RGBVoxel:
				frame = CreateWin<i3d::RGB>(names, position, size);
				break;

			case i3d::RGB16Voxel:
				frame = CreateWin<i3d::RGB16>(names, position, size);
				break;

			case i3d::FloatVoxel:
				frame = CreateWin<float>(names, position, size);
				break;

			case i3d::BinaryVoxel:
				wxLogWarning("Binary images will be represented as GRAY8, which may consume more memory."); 
				frame = CreateWin<i3d::GRAY8>(names, position, size, true);
				break;
			case i3d::Complex32Voxel:
				frame = CreateWin<std::complex<float> >(names, position, size);
				break;
			default:
				wxLogError("%s: Unknown image type", names.begin()->mb_str(wxConvFile));
				return NULL;
			}

			ShowWin(frame, !suppress_showing);
			return frame;
		}
		catch (i3d::IOException &e)
		{
			wxLogError("%s", e.what.c_str());
		}

		return NULL;
	}
	
	v4d::v4d_Viewer3dFrame *MainFrame::OpenWin(const wxArrayString &fnames, const wxArrayString &lnames,
		const wxPoint &position, const wxSize &size, bool suppress_showing)
	{
		i3d::ImgVoxelType vt;
		i3d::ImgVoxelType vt_l;
		v4d::v4d_Viewer3dFrame *frame = NULL;

		try
		{
			vt = i3d::ReadImageType(fnames.Item(0).mb_str(wxConvFile));
			if (!CheckImagesTypeLogErrors(vt, fnames))
				return NULL;

			vt_l = i3d::ReadImageType(lnames.Item(0).mb_str(wxConvFile));
			if (!CheckImagesTypeLogErrors(vt_l, lnames))
				return NULL;

			if (vt == i3d::Gray8Voxel && vt_l == i3d::BinaryVoxel)
			{
				frame = CreateWin<i3d::GRAY8, i3d::GRAY16, bool>(fnames, lnames, position, size);
			}
			else if (vt == i3d::Gray8Voxel && vt_l == i3d::Gray8Voxel)
			{
				frame = CreateWin<i3d::GRAY8, i3d::GRAY16, i3d::GRAY8>(fnames, lnames, position, size);
			}
			else if (vt == i3d::Gray8Voxel && vt_l == i3d::Gray16Voxel)
			{
				frame = CreateWin<i3d::GRAY8, i3d::GRAY16, i3d::GRAY16>(fnames, lnames, position, size);
			}

			else if (vt == i3d::Gray16Voxel && vt_l == i3d::BinaryVoxel)
			{
				frame = CreateWin<i3d::GRAY16, i3d::GRAY16, bool>(fnames, lnames, position, size);
			}
			else if (vt == i3d::Gray16Voxel && vt_l == i3d::Gray8Voxel)
			{
				frame = CreateWin<i3d::GRAY16, i3d::GRAY16, i3d::GRAY8>(fnames, lnames, position, size);
			}
			else if (vt == i3d::Gray16Voxel && vt_l == i3d::Gray16Voxel)
			{
				frame = CreateWin<i3d::GRAY16, i3d::GRAY16, i3d::GRAY16>(fnames, lnames, position, size);
			}

			else if (vt == i3d::FloatVoxel && vt_l == i3d::BinaryVoxel)
			{
				frame = CreateWin<float, i3d::GRAY16, bool>(fnames, lnames, position, size);
			}
			else if (vt == i3d::FloatVoxel && vt_l == i3d::Gray8Voxel)
			{
				frame = CreateWin<float, i3d::GRAY16, i3d::GRAY8>(fnames, lnames, position, size);
			}
			else if (vt == i3d::FloatVoxel && vt_l == i3d::Gray16Voxel)
			{
				frame = CreateWin<float, i3d::GRAY16, i3d::GRAY16>(fnames, lnames, position, size);
			}

			else if (vt == i3d::Complex32Voxel && vt_l == i3d::BinaryVoxel)
			{
				frame = CreateWin<std::complex<float>, i3d::GRAY16, bool>(fnames, lnames, position, size);
			}
			else if (vt == i3d::Complex32Voxel && vt_l == i3d::Gray8Voxel)
			{
				frame = CreateWin<std::complex<float>, i3d::GRAY16, i3d::GRAY8>(fnames, lnames, position, size);
			}
			else if (vt == i3d::Complex32Voxel && vt_l == i3d::Gray16Voxel)
			{
				frame = CreateWin<std::complex<float>, i3d::GRAY16, i3d::GRAY16>(fnames, lnames, position, size);
			}

			else if (vt == i3d::BinaryVoxel && vt_l == i3d::BinaryVoxel)
			{
				wxLogWarning("Binary data images will be represented as GRAY8, which may consume more memory."); 
				frame = CreateWin<i3d::GRAY8, i3d::GRAY16, bool>(fnames, lnames, position, size, true);
			}
			else if (vt == i3d::BinaryVoxel && vt_l == i3d::Gray8Voxel)
			{
				wxLogWarning("Binary data images will be represented as GRAY8, which may consume more memory."); 
				frame = CreateWin<i3d::GRAY8, i3d::GRAY16, i3d::GRAY8>(fnames, lnames, position, size, true);
			}
			else if (vt == i3d::BinaryVoxel && vt_l == i3d::Gray16Voxel)
			{
				wxLogWarning("Binary data images will be represented as GRAY8, which may consume more memory."); 
				frame = CreateWin<i3d::GRAY8, i3d::GRAY16, i3d::GRAY16>(fnames, lnames, position, size, true);
			}
			else
			{
				wxLogError("%s: Unknown image type", lnames.Item(0).mb_str(wxConvFile));
				return NULL;
			}

			ShowWin(frame, !suppress_showing);
			return frame;
		}
		catch (i3d::IOException &e)
		{
			wxLogError("%s", e.what.c_str());
		}

		return NULL;
	}

	wxMenu *MainFrame::CreateFileMenu(bool main_window)
	{
		wxMenu *menuFile = new wxMenu;
		menuFile -> Append(SI_OPEN, "&Open\tCtrl-O", "Open new image");
		menuFile -> AppendSeparator();

		if (main_window)
		{
			menuFile -> Append(SI_RELOAD, "&Reload All\tCtrl-R", "Reload image data in all opened windows");
		}
		else
		{
			menuFile -> Append(SI_RELOAD, "&Reload\tCtrl-R", "Reload image data");
		}

		menuFile -> AppendSeparator();
		menuFile -> AppendCheckItem(SI_SHOW_MAIN_FRAME, "&Show log window\tCtrl-L", "Show/hide log window");
		menuFile -> AppendSeparator();
		menuFile -> Append(SI_ABOUT, "&About", "About application");
		menuFile -> AppendSeparator();
		menuFile -> Append(SI_QUIT, "&Quit\tCtrl-Q", "Quit application");

		return menuFile;
	}


	template <class VDATA>
	v4d::v4d_Viewer3dFrame *MainFrame::CreateWin(const wxArrayString &names,
		const wxPoint &position, const wxSize &size,
		const bool sourceIsBinary)
	{
		i4d::Image4d<VDATA>* images=NULL;
		v4d::v4d_Viewer3dFrame *frame = NULL;

		try
		{
			images = new i4d::Image4d<VDATA>;
			images->ReadSequence(names,sourceIsBinary);
			frame = v4d::v4d_CreateImage3dFrame(images, true, this, -1, "4D viewer", position, size);
			
			m_num_children++;
			frame->GetMenuBar()->Insert(0, CreateFileMenu(false), "File");
			frame->SetTitle(wxFileName::FileName(names[0]).GetFullName());

			// 2D data => show only XY projection window 
			if (images->GetSizeZ() == 1)
			{
				frame->GetCanvas()->ShowDrWin(v4d::pt_ButXY, false);
			}

			m_image_names = names;
			return frame;
			
		}
		catch (i3d::LibException &e)
		{
			wxLogError("%s: i3d exception: %s", __FUNCTION__, e.what.c_str());
		}
		catch (std::exception &e)
		{
			wxLogError("%s: Exception: %s", __FUNCTION__, e.what());
		}
		catch (...)
		{
			wxLogError("%s: Unexpected exception.", __FUNCTION__);
		}

		return NULL;
	}

	template <class T, class LABEL, class VOXEL>
	v4d::v4d_Viewer3dFrame *MainFrame::CreateWin(const wxArrayString &fnames, const wxArrayString &lnames,
        const wxPoint &position, const wxSize &size,
		  const bool sourceIsBinary)
	{		
		i4d::Image4d<T> *images = NULL;
		i4d::LabeledImage4d<LABEL, VOXEL> *limages = NULL;
        v4d::v4d_Viewer3dFrame *frame = NULL;
        
        try
        {
			images = new i4d::Image4d<T>;
			limages = new i4d::LabeledImage4d<LABEL, VOXEL>;

			//images->ReadSequence(fnames,sourceIsBinary);
			//limages->ReadSequence(lnames,images);            

			//read the both (raw data + overlay mask) first frames
			images->ReadSequence_First(fnames,sourceIsBinary);
			limages->ReadSequence_First(lnames,images);            

			size_t framesRead=1;
			//note: it is checked earlier that fnames.size() == lnames.size()
			while (framesRead < fnames.size())
			{
				images->ReadSequence_Next(fnames,sourceIsBinary);
				limages->ReadSequence_Next(lnames);            
				++framesRead;
			}
            
			frame = v4d::v4d_CreateLabeledImage3dFrame(images, limages, true, true, this, -1, "4D viewer", position, size);

			m_num_children++;
			frame->GetMenuBar()->Insert(0, CreateFileMenu(false), "File");
			frame->SetTitle(wxFileName::FileName(fnames[0]).GetFullName());

			// 2D data => show only XY projection window 
			if (images->GetSizeZ() == 1)
			{
				frame->GetCanvas()->ShowDrWin(v4d::pt_ButXY, false);
			}
        }
        catch (i3d::LibException &e)
		{
			wxLogError("%s: i3d exception: %s", __FUNCTION__, e.what.c_str());
		}
        catch (std::exception &e)
		{
            wxLogError("%s: Exception: %s", __FUNCTION__, e.what());
		}        
        catch (...)
		{
            wxLogError("%s: Unexpected exception.", __FUNCTION__);
		}
		
		return frame;
	}

	void MainFrame::ShowWin(v4d::v4d_Viewer3dFrame *frame, bool show)
	{
		if (frame)
		{
			frame -> Show(show);
			
			if (show)
			{
				frame -> Update();
			}
		}
	}

	void MainFrame::ReloadFrame(v4d::v4d_Viewer3dFrame *frame)
	{
		if (!frame)
		{
			wxLogError("Invalid pointer to a frame window!");
			return;
		}

		v4d::v4d_Viewer3dFrame *new_frame = OpenWin(m_image_names, frame -> GetPosition(), frame -> GetSize(), true);

		if (new_frame)
		{
			v4d::v4d_CanvasSettings settings;
			frame -> GetCanvas() -> GetSettings(settings);
			frame -> Close();

			new_frame -> GetCanvas() -> SetSettings(settings);
			new_frame -> ResizeFrameIfRequired();
			ShowWin(new_frame);
		}
	}

	void MainFrame::OnOpen(wxCommandEvent &evt)
	{
		wxArrayString emptyNames;
		this->HandleOpenFileDialog(emptyNames);
	}

	void MainFrame::HandleOpenFileDialog(wxArrayString& names)
	{
		//a flag to indicate detection of an incomplete command line,
		//in which case we will open nothing
		bool directOpening=false;

		//TODO: setup a reading mode object here READINGMODE

		//are there any cmd line paramters?
		//(most parameters are, however, filenames...)
		if (names.GetCount() > 0)
		{
			//try to recognize a command in the first parameter
			if (names[0].Find(wxString("-help")) != wxNOT_FOUND)
			{
				std::cout << "some help here TODO...\n";
				std::cout << "some help here...\n";
				std::cout << "some help here...\n";
				return;
			}
			else
			if (names[0].Find(wxString("-seq")) != wxNOT_FOUND)
			{
				//okay, the first parameter is actually a command
				if (names.GetCount() > 1)
				{
					//remove the command from the parameter list
					names.RemoveAt(0);

					//setup the READINGMODE object accordingly

					directOpening=true;
				}
				else return; //no files to operate the command on, quit.
			}
			else
			if (names[0].Find(wxString("-overlaypair")) != wxNOT_FOUND)
			{
				if (names.GetCount() > 1)
				{
					names.RemoveAt(0);

					//setup the READINGMODE object

					directOpening=true;
				}
				else return;
			}
			else
			if (names[0].Find(wxString("-complexpair")) != wxNOT_FOUND)
			{
				if (names.GetCount() > 1)
				{
					names.RemoveAt(0);

					//setup the READINGMODE object

					directOpening=true;
				}
				else return;
			}
		}

		if (directOpening)
		{
			//open directly the list
			//(supply also the READINGMODE object)
			OpenWin(names);
		}
		else
		{
			//prepare for the dialog
			v4d::OpenFileDialog* dlg=NULL;
			int retVal=wxID_CANCEL;

			//init and process the dialog
			try {
				const wxArrayString* pNames=(names.GetCount() == 0)? NULL : &names;
				dlg=new v4d::OpenFileDialog(this,openDialogFolder,pNames);
				retVal=dlg->ShowModal();
			}
			//"excuse" the dialog internal/programming errors
			catch (i3d::InternalException& e)
			{
				wxMessageBox("An internal error occured while processing the Open dialog:\n\n"
				              +e.what,"Viewer4d internal error", wxICON_EXCLAMATION, this);
				retVal=wxID_CANCEL;
			}
			catch (...)
			{
				wxMessageBox("An internal error occured while processing the Open dialog.",
				             "Viewer4D internal error", wxICON_EXCLAMATION, this);
				retVal=wxID_CANCEL;
			}

			//was the dialog closed correctly with the OK button?
			if (retVal == wxID_OK)
			{
				//remember the last used folder
				dlg->GetCurrentFolder(openDialogFolder);

				//retrieve the files to be opened...
				wxArrayString names;
				dlg->GetFileNames(names);

				//retrieve the request on how to open the files
				// READINGMODE

				size_t processedNames=0;
				while (processedNames < names.GetCount())
				{
					//setup the single job arrays
					wxArrayString jobNames,jobLabels;

					do {
						//add to the image list
						jobNames.Add(names[processedNames++]);
						if (dlg->GetCurrentGroupMode() > 1)
						            jobNames.Add(names[processedNames++]);
						if (dlg->GetCurrentGroupMode() == 5)
						            jobNames.Add(names[processedNames++]);

						//possibly add to the overlay list
						if ((dlg->GetCurrentGroupMode() == 1)
						 || (dlg->GetCurrentGroupMode() == 3))
							jobLabels.Add(names[processedNames++]);

					} while ((!dlg->GetCurrentSepWinMode()) && (processedNames < names.GetCount()));

					//open it... (supply also the READINGMODE object)
					if ((dlg->GetCurrentGroupMode() == 1)
					 || (dlg->GetCurrentGroupMode() == 3))
							//overlay mode
					      OpenWin(jobNames,jobLabels);
					else
							//normal mode
					      OpenWin(jobNames);
				}
			}

			if (dlg) delete dlg;
		}
	}

	void MainFrame::OnReload(wxCommandEvent &evt)
	{
		// The function GetEventObject returns a pointer to a menu, for which the event was generated 
		wxMenu *menu = wxDynamicCast(evt.GetEventObject(), wxMenu);
		wxWindow *evt_window = wxDynamicCast(menu -> GetMenuBar() -> GetParent(), wxWindow);

		if (evt_window == this)
		{
			wxWindowList children = this -> GetChildren();

			wxWindowList::compatibility_iterator node = children.GetFirst();
			while (node)
			{
				if (typeid(*(node -> GetData())) == typeid(v4d::v4d_Viewer3dFrame))
				{
					ReloadFrame(wxDynamicCast(node -> GetData(), v4d::v4d_Viewer3dFrame));
				}
					
				node = node -> GetNext();
			}

			SetFocus();
		}
		else
		{
			ReloadFrame(wxDynamicCast(evt_window, v4d::v4d_Viewer3dFrame));
		}
	}

	void MainFrame::OnQuit(wxCommandEvent &evt)
	{
		Close(TRUE);
	}

	void MainFrame::OnAbout(wxCommandEvent &evt)
	{
        wxDialog dlg(this, wxID_ANY, "About Viewer4D", wxDefaultPosition, wxDefaultSize);
        
        AboutViewerDialogFunc(&dlg);
        wxStaticText *built = (wxStaticText *)dlg.FindWindow(ID_ABOUT_BUILD_DATE);
        built->SetLabel("Built: " __DATE__);

        dlg.ShowModal();
	}

	void MainFrame::OnCloseChild(wxUpdateUIEvent &evt)
	{
		if (m_num_children == 0)
        {
			wxLogError("The main application has not any child window!");
        }

		m_num_children--;

		if (m_num_children == 0 && !m_show_frame)
        {
			Close(TRUE);
        }
	}

	void MainFrame::OnShowMainFrame(wxCommandEvent &evt)
	{
		m_show_frame = !m_show_frame;

		if (m_num_children == 0 && !m_show_frame)
        {
			Close(TRUE);
        }
		
		Show(m_show_frame);
	}

	void MainFrame::OnUpdateUI(wxUpdateUIEvent &evt)
	{
		switch (evt.GetId())
		{
			case SI_SHOW_MAIN_FRAME:
				evt.Enable(m_num_children > 0);
				evt.Check(m_show_frame);
				break;
        
			case SI_RELOAD:
				evt.Enable(m_num_children > 0);
				break;
		}
	}

} // namespace viewer3d
