/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * vol3D_canvas.cpp
 *
 * implementation of vol3D_VolVisCanvas
 *
 * Marek Vinkler (172521@mail.muni.cz) 2008
 * Ondrej Danek (xdanek2@fi.muni.cz) 2010
 */

#include "GL/glew.h" // must be included first
#include "v4d_basic.h"
#include <wx/dcclient.h> 
#include "vol3D_canvas.h"

#ifdef _DEBUG
	#define new DEBUG_NEW
#endif

namespace v4d
{
	BEGIN_EVENT_TABLE(vol3D_VolVisCanvas, wxGLCanvas)
		EVT_SIZE(vol3D_VolVisCanvas::OnSize)
		EVT_PAINT(vol3D_VolVisCanvas::OnPaint)
		EVT_CHAR(vol3D_VolVisCanvas::OnChar)
		EVT_LEFT_DOWN(vol3D_VolVisCanvas::OnMouseEvent)
		EVT_LEFT_UP(vol3D_VolVisCanvas::OnMouseEvent)
		EVT_RIGHT_DOWN(vol3D_VolVisCanvas::OnMouseEvent)
		EVT_RIGHT_UP(vol3D_VolVisCanvas::OnMouseEvent)
		EVT_MOTION(vol3D_VolVisCanvas::OnMouseMoveEvent)
		EVT_LEAVE_WINDOW(vol3D_VolVisCanvas::OnMouseMoveEvent)
		EVT_ERASE_BACKGROUND(vol3D_VolVisCanvas::OnEraseBackground)
	END_EVENT_TABLE()

	int vol3D_VolVisCanvas::AttributeList[] = { WX_GL_RGBA, WX_GL_DOUBLEBUFFER }; 

	vol3D_VolVisCanvas::vol3D_VolVisCanvas(wxStatusBar *statusBar, wxWindow *parent, wxWindowID id,
        const wxPoint& pos, const wxSize& size, long style, const wxString& name, int *gl_attrib)
        : wxGLCanvas(parent, id, gl_attrib, pos, size, style | wxFULL_REPAINT_ON_RESIZE, name), 
        _statusBar(statusBar), _context(NULL)
	{
		//initialize visualization parameters
		Reset();

		//initialize the data
		int ii, gg;
		float temp;

		//create the points used to draw the slices
		for (ii=0;ii<8;ii++)
		{
			for (gg=0;gg<8;gg++)
			{
				_points[ii*8+gg][0]=(float)gg/7.0f-0.5f;
				_points[ii*8+gg][1]=(float)ii/7.0f-0.5f;
				_points[ii*8+gg][2]=-0.5f;
				temp=_points[ii*8+gg][0]*_points[ii*8+gg][0]+
						 _points[ii*8+gg][1]*_points[ii*8+gg][1]+
						 _points[ii*8+gg][2]*_points[ii*8+gg][2];
				temp=1.0f/(float)sqrt(temp);
				_points[ii*8+gg][0]*=temp;
				_points[ii*8+gg][1]*=temp;
				_points[ii*8+gg][2]*=temp;
				_points[ii*8+gg][0]*=2.0f;
				_points[ii*8+gg][1]*=2.0f;
				_points[ii*8+gg][2]+=1.0f;
			}
		}

		//_pal.erase(_pal.begin());

		_has_data = false;
		_drawIsosurface = false;
	}

	vol3D_VolVisCanvas::~vol3D_VolVisCanvas()
	{
		glDeleteTextures(1, &_volumeTexture);

		//delete previous buffers
		for(std::vector<ComponentDrawInfo>::iterator j = _cdiVec.begin(); j != _cdiVec.end(); j++)
		{
			glDeleteBuffers(1, &j->_vertexVBO);
			glDeleteBuffers(1, &j->_indexVBO);
			glDeleteBuffers(1, &j->_normalVBO);
		}

        delete _context;
	}

	void vol3D_VolVisCanvas::Reset()
	{
		//intialization of class variables
		_slices = 100;
		_intensity = 0.5f;
		_density = 0.5f;
		_trackingMotion = GL_FALSE;
		_trackingTranslation = GL_FALSE;
		_angle = 0.0f;
		_axis[0] = 1.0f; _axis[1] = 0.0f; _axis[2] = 0.0f;
		_pAngle = 0.0f;
		_pAxis[0] = 1.0f; _pAxis[1] = 0.0f; _pAxis[2] = 0.0f;
		_mode = MoveNone;
		_viewDistance = -5.0f;
		_user[0] = 1.0; _user[1] = 0.0; _user[2] = 0.0; _user[3] = 1.0;
		_gbPlane = false;

        _showBox = true;
        _showClipPlane = true;

        _bcgColour[0] = 0.0f;
        _bcgColour[1] = 0.0f;
        _bcgColour[2] = 0.0f;

		memset(_objectXform, 0, 16*sizeof(float));
		_objectXform[0] = 1.0f;
		_objectXform[5] = 1.0f;
		_objectXform[10] = 1.0f;
		_objectXform[15] = 1.0f;

		memset(_planeXform, 0, 16*sizeof(float));
		_planeXform[0] = 1.0f;
		_planeXform[5] = 1.0f;
		_planeXform[10] = 1.0f;
		_planeXform[15] = 1.0f;
	}

	bool vol3D_VolVisCanvas::Init()
	{
		GetParent()->Show(true);

        // Create OpenGL context
		_context = new wxGLContext(this);
        _context->SetCurrent(*this);

        // Init GLEW
        if (glewInit() != GLEW_OK)
        {
            wxLogError("Unable to init GLEW.");
            return false;
        }

		// Print some OpenGL info
		wxLogMessage("Starting OpenGL window");
		wxLogMessage("Vendor: %s", (const char *)glGetString(GL_VENDOR));
		wxLogMessage("Renderer: %s", (const char *)glGetString(GL_RENDERER));
		wxLogMessage("Version: %s", (const char *)glGetString(GL_VERSION));

        // Check OpenGL version
        if (!GLEW_VERSION_2_0)
        {
            wxLogError("OpenGL 2.0 or newer driver is required.");
            return false;
        }

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		//setup the 3D textures
		//give _volumeTexture a unique ID, with glGenTextures same IDs are used in different instances of vol3D_VolVisCanvas
		_volumeTexture = wxNewId();
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		glDepthFunc(GL_LESS);
		glEnable(GL_DEPTH_TEST);

		//lights setup
		GLfloat light_diffuseColor[] = {1.0f, 1.0f, 1.0f, 1.0f};
		GLfloat light_ambientColor[] = {0.2f, 0.2f, 0.2f, 1.0f};
		GLfloat light_specularColor[] = {1.0f, 1.0f, 1.0f, 1.0f};

		glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuseColor);
		glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambientColor);
		glLightfv(GL_LIGHT0, GL_SPECULAR, light_specularColor);

		glShadeModel(GL_SMOOTH);

		//global material setup
		GLfloat ambientColor[] = {0.25f, 0.25f, 0.25f, 1.0f};
		GLfloat specularColor[] = {0.6f, 0.6f, 0.6f, 1.0f};
		GLfloat shininess = 70.0f;

		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambientColor);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specularColor);
		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

		//Sets the correct size and projection
		ResizeViewPort();
		
		return true;
	}

	void vol3D_VolVisCanvas::OnPaint(wxPaintEvent& WXUNUSED(evt))
	{
		// This is a dummy, to avoid an endless succession of paint messages.
		// OnPaint handlers must always create a wxPaintDC.
		wxPaintDC dc(this);

        if (_context != NULL)
        {
            _context->SetCurrent(*this);

		    Draw();
		    SwapBuffers();
        }
	}

	void vol3D_VolVisCanvas::OnSize(wxSizeEvent& evt)
	{
        if (_context != NULL)
        {
            ResizeViewPort();
        }
	}

    void vol3D_VolVisCanvas::ResizeViewPort()
    {
        _context->SetCurrent(*this);

        // Get client area size
        int width, height;
        GetClientSize(&width, &height);

        // Setup viewport
        glViewport(0, 0, width, height);

        // Setup projection matrix: gluPerspective(40.0, width / height, 0.1, 50.0);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();

        GLdouble f = 1.0f / tan(20.0 * M_PI / 180.0), zNear = 0.1, zFar = 50.0;
        GLdouble aspect = (GLdouble) width / (GLdouble) height;

        GLdouble pm[16];
        memset(pm, 0, 16 * sizeof(GLdouble));
        pm[0] = f / aspect;
        pm[5] = f;
        pm[10] = (zNear + zFar) / (zNear - zFar); 
        pm[11] = -1;
        pm[14] = (2 * zNear * zFar) / (zNear - zFar);

        glMultMatrixd(pm);
        glMatrixMode(GL_MODELVIEW);

        _screenWidth=width;
        _screenHeight=height;
    }

	void vol3D_VolVisCanvas::OnEraseBackground(wxEraseEvent& WXUNUSED(evt))
	{
		// Do nothing, to avoid flashing.
	}

	void vol3D_VolVisCanvas::OnChar(wxKeyEvent& evt)
	{
		switch(evt.GetKeyCode())
		{
		case WXK_ESCAPE: //escape
			GetParent()->Close();
			break;

		case '+': //increase the number of slices
			_slices+=10;
			break;

		case '-': //decrease the number of slices
			_slices-=10;
			if (_slices<0)
				_slices=0;
			break;

		case WXK_SPACE: //switch between controlling the view and controlling the plane
			_gbPlane=!_gbPlane;
			break;

		case WXK_UP:
			_intensity+=1.0f/32.0f;
			if (_intensity>1.0f)
				_intensity=1.0f;
			break;

		case WXK_DOWN:
			_intensity-=1.0f/32.0f;
			if (_intensity<0.0f)
				_intensity=0.0f;
			break;

		case WXK_RIGHT:
			_density+=1.0f/32.0f;
			if (_density>1.0f)
				_density=1.0f;
			break;

		case WXK_LEFT:
			_density-=1.0f/32.0f;
			if (_density<0.0f)
				_density=0.0f;
			break;

		default: //unknown command, let it go up the handler hierarchy
			evt.Skip();
			return;
		}

		Refresh(false);
	}

	void vol3D_VolVisCanvas::OnMouseEvent(wxMouseEvent& evt)
	{
		long x, y, elapsed;
		
		x = evt.GetX();
		y = evt.GetY();

		elapsed = _stopWatch.Time();
		_stopWatch.Start();

		if (evt.LeftDown())
		{
			StartMotion(x, y, elapsed);
			// skip this event otherwise the window cannot obtain the focus 
			evt.Skip();
		}
		else if (evt.LeftUp())
		{
			StopMotion();
		}
		
		if (evt.RightDown())
		{
			if (!_trackingTranslation)
			{
				_trackingTranslation = GL_TRUE;
				_transPosY = y;
			}
		}
		else if(evt.RightUp())
		{
			_trackingTranslation = GL_FALSE;
		}

		Refresh(false);
	}

	void vol3D_VolVisCanvas::OnMouseMoveEvent(wxMouseEvent& evt)
	{
		float curPos[3], dx, dy, dz;
		long x, y, elapsed;
		
		x = evt.GetX();
		y = evt.GetY();

		if (_trackingMotion)
		{
			Ptov(x, y, _screenWidth, _screenHeight, curPos);

			dx = curPos[0] - _lastPos[0];
			dy = curPos[1] - _lastPos[1];
			dz = curPos[2] - _lastPos[2];

			if (!_gbPlane)
			{
				_angle = 90.0 * sqrt(dx*dx + dy*dy + dz*dz);

				_axis[0] = _lastPos[1]*curPos[2] - _lastPos[2]*curPos[1];
				_axis[1] = _lastPos[2]*curPos[0] - _lastPos[0]*curPos[2];
				_axis[2] = _lastPos[0]*curPos[1] - _lastPos[1]*curPos[0];
			}
			else
			{
				_pAngle = 90.0 * sqrt(dx*dx + dy*dy + dz*dz);

				_pAxis[0] = _lastPos[1]*curPos[2] - _lastPos[2]*curPos[1];
				_pAxis[1] = _lastPos[2]*curPos[0] - _lastPos[0]*curPos[2];
				_pAxis[2] = _lastPos[0]*curPos[1] - _lastPos[1]*curPos[0];
			}

			elapsed = _stopWatch.Time();
			_stopWatch.Start();

			_lastTime = elapsed;
			_lastPos[0] = curPos[0];
			_lastPos[1] = curPos[1];
			_lastPos[2] = curPos[2];

			Refresh(false);
		}

		if (_trackingTranslation)
		{
			if (!_gbPlane)
			{
				_viewDistance += 0.01f * (y - _transPosY);
				_transPosY = y;
			}
			else
			{
				_user[3]-= 0.01f * (y - _transPosY);
				_transPosY = y;
			}

			Refresh(false);
		}

		if (evt.Leaving())
		{
			_trackingMotion = GL_FALSE;
			_trackingTranslation = GL_FALSE;
			_mode = MoveNone;
			_angle = 0.0f;
			_pAngle = 0.0f;
		}
	}

	void vol3D_VolVisCanvas::SetView(float distance, float ax, float ay, float az, float ux, float uy, float uz)
	{
        //Compute: gluLookAt(-ax, -ay, -az, 0.0f, 0.0f, 0.0f, ux, uy, uz); glTranslatef(-ax, -ay, -az);
        i3d::Vector3d<float> f(ax, ay, az);
        i3d::Vector3d<float> up(ux, uy, uz);
        i3d::Vector3d<float> s, u;

        s = i3d::Cross(f, up);
        u = i3d::Cross(s, f);

        _objectXform[0] = s.x; _objectXform[4] = s.y; _objectXform[8] = s.z; _objectXform[12] = 0.0f;
        _objectXform[1] = u.x; _objectXform[5] = u.y; _objectXform[9] = u.z; _objectXform[13] = 0.0f;
        _objectXform[2] = -f.x; _objectXform[6] = -f.y; _objectXform[10] = -f.z; _objectXform[14] = 0.0f;
        _objectXform[3] = 0.0f; _objectXform[7] = 0.0f; _objectXform[11] = 0.0f; _objectXform[15] = 1.0f;

		_viewDistance = distance;

		Refresh(false);
	}

    void vol3D_VolVisCanvas::SaveViewMatrix()
    {
        wxFileDialog dlg(GetParent(), "Save view matrix as", wxEmptyString, wxEmptyString, 
            "Viewer3D view matrix (*.v4dview)|*.v4dview",
            wxFD_SAVE | wxFD_OVERWRITE_PROMPT);

        if (dlg.ShowModal() == wxID_OK)
        {
            FILE *f = fopen(dlg.GetPath().mb_str(wxConvFile), "wt");

            if (f == NULL)
            {
                wxMessageBox("Unable to create file.");
            }
            else
            {
                fprintf(f, "%.11f ", _viewDistance);
                
                for (size_t i = 0; i < 16; ++i)
                {
                    fprintf(f, "%.11f ", _objectXform[i]);
                }

                fprintf(f, "%.11f ", _user[3]);

                for (size_t i = 0; i < 16; ++i)
                {
                    fprintf(f, "%.11f ", _planeXform[i]);
                }

                fclose(f);
            }            
        }
    }

    void vol3D_VolVisCanvas::LoadViewMatrix()
    {
        wxFileDialog dlg(GetParent(), "Choose a view to load", wxEmptyString, wxEmptyString, 
            "Viewer3D view matrix (*.v4dview)|*.v4dview|All files (*.*) | *.*", 
            wxFD_OPEN | wxFD_FILE_MUST_EXIST);

        if (dlg.ShowModal() == wxID_OK)
        {
            FILE *f = fopen(dlg.GetPath().mb_str(wxConvFile), "rt");

            if (f == NULL)
            {
                wxMessageBox("Unable to open file.");
            }
            else
            {
                int ignoredRetVal;
					 ignoredRetVal=fscanf(f, "%f", &_viewDistance);
                
                for (size_t i = 0; i < 16; ++i)
                {
						  ignoredRetVal=fscanf(f, "%f", &_objectXform[i]);
                }

					 ignoredRetVal+=fscanf(f, "%lf", &_user[3]);

                for (size_t i = 0; i < 16; ++i)
                {
						  ignoredRetVal=fscanf(f, "%f", &_planeXform[i]);
                }

                fclose(f);
            }                        
        }
    }

    void vol3D_VolVisCanvas::SetVolumeData
		(const unsigned char *rgbabuf, const i3d::Vector3d<size_t> &size, const i3d::Vector3d<float> &scale)
	{
		_has_data = false;

		if (size.x > GL_MAX_3D_TEXTURE_SIZE || 
			size.y > GL_MAX_3D_TEXTURE_SIZE ||
			size.z > GL_MAX_3D_TEXTURE_SIZE)
		{
			wxLogError("Sorry. The data is too big and cannot be displayed");
			return;
		}

		_scale = scale;

		//load the volume textures
		if (glTexImage3DEXT != NULL)
		{
			//load full-size RGBA texture
			float color[]={1.0f,1.0f,1.0f,1.0f};

			glBindTexture(GL_TEXTURE_3D,_volumeTexture);
			glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_WRAP_R,GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexParameterfv(GL_TEXTURE_3D,GL_TEXTURE_BORDER_COLOR,color);
			glTexImage3D(GL_TEXTURE_3D, 0, GL_RGBA8, (GLsizei)size.x, (GLsizei)size.y,
                (GLsizei)size.z, 0, GL_RGBA, GL_UNSIGNED_BYTE, rgbabuf);

			if (glGetError() != GL_NO_ERROR)
			{
				wxLogError("Sorry. Creating the 3D texture failed (the image is probably too big), no data will be displayed.");
				return;
			}

			_has_data = true;

			Refresh(false);
		}
	}

	void vol3D_VolVisCanvas::SetIsoSurfaceData(const vol3D_ComponentVector &comps, const i3d::Vector3d<float> &trans)
	{
		_trans = trans;

		//delete previous buffers
		for(std::vector<ComponentDrawInfo>::iterator j = _cdiVec.begin(); j != _cdiVec.end(); j++)
		{
			glDeleteBuffers(1, &j->_vertexVBO);
			glDeleteBuffers(1, &j->_indexVBO);
			glDeleteBuffers(1, &j->_normalVBO);
		}

		_cdiVec.resize(comps.size());

		//copy the new data
		vol3D_ComponentVector::const_iterator i = comps.begin();

		for(std::vector<ComponentDrawInfo>::iterator j = _cdiVec.begin(); j != _cdiVec.end(); i++, j++)
		{
			//create buffers
			//give VBOs a unique ID, with glGenBuffers same IDs are used in different instances of vol3D_VolVisCanvas
			j->_vertexVBO = wxNewId();
			j->_indexVBO = wxNewId();
			j->_normalVBO = wxNewId();


			//vertex data
			glBindBuffer(GL_ARRAY_BUFFER, j->_vertexVBO);
			glBufferData(GL_ARRAY_BUFFER, i->GetVertexCount()*3*sizeof(GLfloat),
				i->GetVertexList(), GL_STATIC_DRAW);

			//triangle data
			j->_triangleCount = (GLuint)i->GetTriangleCount();
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, j->_indexVBO);
			glBufferData(GL_ELEMENT_ARRAY_BUFFER, i->GetTriangleCount()*3*sizeof(GLuint),
				i->GetTriangleList(), GL_STATIC_DRAW);

			//use normals
			j->_useLighting = false;
			if(i->GetNormalList() != NULL)
			{
				glBindBuffer(GL_ARRAY_BUFFER, j->_normalVBO);
				glBufferData(GL_ARRAY_BUFFER, i->GetVertexCount()*3*sizeof(GLfloat),
					i->GetNormalList(), GL_STATIC_DRAW);
				j->_useLighting = true;
			}

			//color information
			j->_color = i->GetColor();
		}
	}

	void vol3D_VolVisCanvas::ShowIsosurface(bool show, v4d_Palette pal)
	{
		_drawIsosurface = show;
		_pal = pal;
	}

	void vol3D_VolVisCanvas::StartMotion(int x, int y, int time)
	{
		if (_gbPlane)
			_mode = MovePlane;
		else
			_mode = MoveObject;

		_trackingMotion = GL_TRUE;

		_lastTime = time;

		Ptov(x, y, _screenWidth, _screenHeight, _lastPos);
	}

	void vol3D_VolVisCanvas::StopMotion()
	{
		_trackingMotion = GL_FALSE;
		_mode = MoveNone;
		_angle = 0.0f;
		_pAngle = 0.0f;
	}

	void vol3D_VolVisCanvas::Ptov(int x, int y, int width, int height, float v[3]) const
	{
		float d, a;

		// project x,y onto a hemi-sphere centered within width, height
		v[0] = (2.0*x - width) / width;
		v[1] = (height - 2.0*y) / height;
		d = sqrt(v[0]*v[0] + v[1]*v[1]);
		v[2] = cos((M_PI/2.0) * ((d < 1.0) ? d : 1.0));

		a = 1.0 / sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);

		v[0] *= a;
		v[1] *= a;
		v[2] *= a;
	}

	void vol3D_VolVisCanvas::InvertMat(float mat[16]) const
	{
		float temp;

		temp=mat[1];
		mat[1]=mat[4];
		mat[4]=temp;

		temp=mat[2];
		mat[2]=mat[8];
		mat[8]=temp;

		temp=mat[6];
		mat[6]=mat[9];
		mat[9]=temp;
	  
		mat[12]=-mat[12];
		mat[13]=-mat[13];
		mat[14]=-mat[14];
	}

	void vol3D_VolVisCanvas::Draw()
	{
		//texgen planes
		float xPlane[] = {1.0f, 0.0f, 0.0f, 0.0f};
		float yPlane[] = {0.0f, 1.0f, 0.0f, 0.0f};
		float zPlane[] = {0.0f, 0.0f, 1.0f, 0.0f};

		float mat[16];
		int ii,gg,hh;
		float temp[3];
		char str[256];
		float plane[12];
		GLfloat light_position[4] = {0.0f, 2.0f, 10.0f, 0.0f};

		glEnable(GL_DEPTH_TEST);

		glClearColor(_bcgColour[0], _bcgColour[1], _bcgColour[2], 0.0f);
		glClearDepth(1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		if (!_has_data || !glTexImage3DEXT)
		{
			return;
		}

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(0, 0, _viewDistance);

		glLightfv(GL_LIGHT0, GL_POSITION, light_position);

		if (_mode == MoveObject)
		{
			// Have OpenGL compute the new transformation (simple but slow)
			glPushMatrix();
			glLoadIdentity();
			glRotatef(_angle, _axis[0], _axis[1], _axis[2]);
			glMultMatrixf((GLfloat *) _objectXform);
			glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat *) _objectXform);
			glPopMatrix();
		}

		glMultMatrixf((GLfloat *) _objectXform);

		glGetFloatv(GL_MODELVIEW_MATRIX,mat);
		InvertMat(mat);

		//want rotation only, zero the translation
		mat[12]=0.0f;
		mat[13]=0.0f;
		mat[14]=0.0f;

		if (_mode == MovePlane)
		{
			//handle the plane rotations
			temp[0]=mat[0]*_pAxis[0]+mat[4]*_pAxis[1]+mat[8]*_pAxis[2];
			temp[1]=mat[1]*_pAxis[0]+mat[5]*_pAxis[1]+mat[9]*_pAxis[2];
			temp[2]=mat[2]*_pAxis[0]+mat[6]*_pAxis[1]+mat[10]*_pAxis[2];

			glPushMatrix();
			glLoadIdentity();
			glRotatef(_pAngle, temp[0], temp[1], temp[2]);

			glMultMatrixf(_planeXform);
			glGetFloatv(GL_MODELVIEW_MATRIX,_planeXform);
			glPopMatrix();
		}

		//get the eqn for the user plane
		_user[0]=-_planeXform[0];
		_user[1]=-_planeXform[1];
		_user[2]=-_planeXform[2];

		//setup the texture coord generation
		glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);
		glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);
		glTexGeni(GL_R,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);  

		//create the points for the corners of the clip plane
		for (ii=0;ii<4;ii++)
		{
			plane[ii*3+0]=_planeXform[0]*_user[3];
			plane[ii*3+1]=_planeXform[1]*_user[3];
			plane[ii*3+2]=_planeXform[2]*_user[3];
			plane[ii*3+0]+=_planeXform[4]*((ii<2)?-0.8f:0.8f);
			plane[ii*3+1]+=_planeXform[5]*((ii<2)?-0.8f:0.8f);
			plane[ii*3+2]+=_planeXform[6]*((ii<2)?-0.8f:0.8f);
			plane[ii*3+0]+=_planeXform[8]*((ii&0x1)?0.8f:-0.8f);
			plane[ii*3+1]+=_planeXform[9]*((ii&0x1)?0.8f:-0.8f);
			plane[ii*3+2]+=_planeXform[10]*((ii&0x1)?0.8f:-0.8f);
		}

		//replace the plane opposite the viewer with the user controlled one
		glClipPlane(GL_CLIP_PLANE0, _user);		
		glEnable(GL_CLIP_PLANE0);

		//draw the isosurface
		if(_drawIsosurface)
		{
			//isosurface material
			GLfloat diffuseColor[4];

			for(std::vector<ComponentDrawInfo>::iterator i = _cdiVec.begin(); i != _cdiVec.end(); i++)
			{
				glEnableClientState(GL_VERTEX_ARRAY);
				if(i->_useLighting)
					glEnableClientState(GL_NORMAL_ARRAY);
				glEnableClientState(GL_INDEX_ARRAY);

				glBindBuffer(GL_ARRAY_BUFFER, i->_vertexVBO);
				glVertexPointer(3, GL_FLOAT, 0, (char *) NULL);

				//set the propper color
				if(_pal.size() != 0)
				{
					diffuseColor[0] = _pal[i->_color % _pal.size()].red / 255.0f;
					diffuseColor[1] = _pal[i->_color % _pal.size()].green / 255.0f;
					diffuseColor[2] = _pal[i->_color % _pal.size()].blue / 255.0f;
					diffuseColor[3] = 1.0f;
				}

				if(i->_useLighting)
				{
					glEnable(GL_LIGHTING);
					glEnable(GL_LIGHT0);

					glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuseColor);

					glBindBuffer(GL_ARRAY_BUFFER, i->_normalVBO);
					glNormalPointer(GL_FLOAT, 0, (char *) NULL);
				}
				else
				{
					glColor3fv(diffuseColor);
				}

				glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, i->_indexVBO);
				glIndexPointer(GL_UNSIGNED_BYTE, 0, (char *) NULL);

				//draw the data
				glPushMatrix();
				glTranslatef(_trans.x, _trans.y, _trans.z);
				glDrawElements(GL_TRIANGLES, i->_triangleCount*3, GL_UNSIGNED_INT, (char *) NULL);
				glPopMatrix();

				glDisableClientState(GL_VERTEX_ARRAY);
				glDisableClientState(GL_INDEX_ARRAY);

				if(i->_useLighting)
				{
					glDisable(GL_LIGHTING);
					glDisable(GL_LIGHT0);
					glDisableClientState(GL_NORMAL_ARRAY);
				}
			}
		}
		
		//set the color for the slices
		glColor4f(1.0f,1.0f,1.0f,_intensity);

		//enable the alpha blending/testing
		glEnable(GL_ALPHA_TEST);
		glAlphaFunc(GL_GREATER,_density*_intensity);

		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);

		//set up the texture matrix
		glMatrixMode(GL_TEXTURE);
		glLoadIdentity();
		glEnable(GL_TEXTURE_3D);

		glTranslatef(0.5f,0.5f,0.5f);
		glScalef(0.5f,0.5f,1.0f);
		glScalef(_scale.x, _scale.y, _scale.z);

		glMultMatrixf(mat);
		glTranslatef(0.0f,0.0f,-_viewDistance);

		//set modelview to identity
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		//setup the tex gen
		glTexGenfv(GL_S,GL_EYE_PLANE,xPlane);
		glTexGenfv(GL_T,GL_EYE_PLANE,yPlane);
		glTexGenfv(GL_R,GL_EYE_PLANE,zPlane);

		glEnable(GL_TEXTURE_GEN_S);
		glEnable(GL_TEXTURE_GEN_T);
		glEnable(GL_TEXTURE_GEN_R);

		//bind the correct texture
		glBindTexture(GL_TEXTURE_3D,_volumeTexture);
		glTranslatef(0.0f,0.0f,_viewDistance);

		//draw the slices
		for (ii=0;ii<_slices; ii++)
		{
			glPushMatrix();
			glTranslatef(0.0f,0.0f,-1.0f +(float)ii*2.0f/(float)(_slices-1));
			glBegin(GL_QUADS);

			for(hh=0;hh<7;hh++)
			{
				for(gg=0;gg<7;gg++)
				{
					glVertex3fv(_points[hh*8+gg]);
					glVertex3fv(_points[hh*8+gg+1]);
					glVertex3fv(_points[(hh+1)*8+gg+1]);
					glVertex3fv(_points[(hh+1)*8+gg]);
				}
			}
			glEnd();
			glPopMatrix();
		}

		glPopMatrix();

		glDisable(GL_CLIP_PLANE0);
        glDisable(GL_ALPHA_TEST);

        if (_showClipPlane)
        {
            //draw the slice plane across to get a better image
            glDepthMask(GL_FALSE);
            glBegin(GL_QUADS);
            glColor4f(1.0f,1.0f,1.0f,1.0f); // With slice update
            glVertex3fv(&plane[0]);
            glVertex3fv(&plane[3]);
            glVertex3fv(&plane[9]);
            glVertex3fv(&plane[6]);
            glEnd();
            glDepthMask(GL_TRUE);
        }

		glDisable(GL_TEXTURE_3D);
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		glDisable(GL_TEXTURE_GEN_R);

		//draw the box framing everything
        if (_showBox)
        {
            glBegin(GL_LINES);
            glColor3f(0.0f,0.0f,0.7f);
            glVertex3f(1.0f,1.0f,1.0f);
            glVertex3f(1.0f,1.0f,-1.0f);
            glVertex3f(1.0f,-1.0f,1.0f);
            glVertex3f(1.0f,-1.0f,-1.0f);
            glVertex3f(-1.0f,1.0f,1.0f);
            glVertex3f(-1.0f,1.0f,-1.0f);
            glVertex3f(-1.0f,-1.0f,1.0f);
            glVertex3f(-1.0f,-1.0f,-1.0f);
            glVertex3f(1.0f,1.0f,1.0f);
            glVertex3f(1.0f,-1.0f,1.0f);
            glVertex3f(1.0f,1.0f,-1.0f);
            glVertex3f(1.0f,-1.0f,-1.0f);
            glVertex3f(-1.0f,1.0f,1.0f);
            glVertex3f(-1.0f,-1.0f,1.0f);
            glVertex3f(-1.0f,1.0f,-1.0f);
            glVertex3f(-1.0f,-1.0f,-1.0f);
            glVertex3f(1.0f,1.0f,1.0f);
            glVertex3f(-1.0f,1.0f,1.0f);
            glVertex3f(1.0f,1.0f,-1.0f);
            glVertex3f(-1.0f,1.0f,-1.0f);
            glVertex3f(1.0f,-1.0f,1.0f);
            glVertex3f(-1.0f,-1.0f,1.0f);
            glVertex3f(1.0f,-1.0f,-1.0f);
            glVertex3f(-1.0f,-1.0f,-1.0f);
            glEnd();
        }

		//draw clip plane
        if (_showClipPlane)
        {
            glBegin(GL_LINES);
            glColor3f(1.0f,0.0f,0.0f);
            glVertex3fv(&plane[0]);
            glVertex3fv(&plane[3]);
            glVertex3fv(&plane[3]);
            glVertex3fv(&plane[9]);
            glVertex3fv(&plane[9]);
            glVertex3fv(&plane[6]);
            glVertex3fv(&plane[6]);
            glVertex3fv(&plane[0]);
            glEnd();
        }

		if (_gbPlane)
			_statusBar->SetStatusText("Mode: Moving Plane", 0);
		else
			_statusBar->SetStatusText("Mode: Moving Object", 0);

		sprintf(str, "# slices: %d, intensity: %f, threshold: %f",
			_slices, _intensity, _density);
		_statusBar->SetStatusText(wxString(str,wxConvUTF8), 1);
	}

} //namespace v4d
