/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * vol3D_isosurface.cpp
 *
 * implementation of vol3D_IsoSurface class providing the Marching Cube algorithm
 * the input arrays should be in Image3D data format
 *
 * Marek Vinkler (172521@mail.muni.cz) 2008
 * 
 */

#include "v4d_basic.h"
#include <cmath>
#include <ctime>
#include <string.h>
#include <algorithm>
#include "vol3D_isosurface.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

namespace v4d
{
	//Lookup tables used in the construction of the isosurface
	const size_t edgeTable[256] = {
		0x0  , 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c,
		0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
		0x190, 0x99 , 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c,
		0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
		0x230, 0x339, 0x33 , 0x13a, 0x636, 0x73f, 0x435, 0x53c,
		0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
		0x3a0, 0x2a9, 0x1a3, 0xaa , 0x7a6, 0x6af, 0x5a5, 0x4ac,
		0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
		0x460, 0x569, 0x663, 0x76a, 0x66 , 0x16f, 0x265, 0x36c,
		0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
		0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0xff , 0x3f5, 0x2fc,
		0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
		0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55 , 0x15c,
		0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
		0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc ,
		0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
		0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc,
		0xcc , 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
		0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c,
		0x15c, 0x55 , 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
		0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc,
		0x2fc, 0x3f5, 0xff , 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
		0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c,
		0x36c, 0x265, 0x16f, 0x66 , 0x76a, 0x663, 0x569, 0x460,
		0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac,
		0x4ac, 0x5a5, 0x6af, 0x7a6, 0xaa , 0x1a3, 0x2a9, 0x3a0,
		0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c,
		0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33 , 0x339, 0x230,
		0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c,
		0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99 , 0x190,
		0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c,
		0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x0
	};

	const int triTable[256][16] = {
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
		{3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
		{3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
		{3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
		{9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
		{9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
		{2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
		{8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
		{9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
		{4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
		{3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
		{1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
		{4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
		{4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
		{9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
		{5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
		{2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
		{9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
		{0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
		{2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
		{10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
		{4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
		{5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
		{5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
		{9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
		{0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
		{1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
		{10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
		{8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
		{2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
		{7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
		{9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
		{2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
		{11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
		{9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
		{5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
		{11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
		{11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
		{1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
		{9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
		{5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
		{2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
		{0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
		{5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
		{6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
		{3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
		{6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
		{5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
		{1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
		{10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
		{6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
		{8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
		{7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
		{3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
		{5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
		{0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
		{9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
		{8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
		{5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
		{0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
		{6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
		{10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
		{10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
		{8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
		{1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
		{3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
		{0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
		{10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
		{3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
		{6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
		{9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
		{8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
		{3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
		{6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
		{0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
		{10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
		{10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
		{2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
		{7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
		{7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
		{2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
		{1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
		{11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
		{8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
		{0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
		{7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
		{10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
		{2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
		{6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
		{7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
		{2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
		{1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
		{10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
		{10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
		{0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
		{7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
		{6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
		{8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
		{9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
		{6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
		{4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
		{10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
		{8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
		{0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
		{1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
		{8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
		{10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
		{4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
		{10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
		{5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
		{11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
		{9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
		{6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
		{7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
		{3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
		{7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
		{9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
		{3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
		{6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
		{9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
		{1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
		{4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
		{7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
		{6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
		{3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
		{0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
		{6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
		{0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
		{11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
		{6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
		{5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
		{9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
		{1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
		{1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
		{10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
		{0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
		{5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
		{10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
		{11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
		{9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
		{7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
		{2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
		{8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
		{9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
		{9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
		{1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
		{9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
		{9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
		{5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
		{0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
		{10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
		{2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
		{0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
		{0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
		{9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
		{5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
		{3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
		{5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
		{8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
		{0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
		{9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
		{0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
		{1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
		{3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
		{4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
		{9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
		{11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
		{11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
		{2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
		{9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
		{3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
		{1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
		{4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
		{4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
		{0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
		{3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
		{3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
		{0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
		{9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
		{1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
		{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
	};

	vol3D_Component::vol3D_Component()
	{
		_tableIndex = 0;
		_numVertices = 0;
		_numTriangles = 0;
		_vertexList = NULL;
		_triangleList = NULL;
		_normalList = NULL;
		_color = 0;
	}

	vol3D_Component::~vol3D_Component()
	{		
		if(_vertexList != NULL)
			delete[] _vertexList;
		if(_triangleList != NULL)
			delete[] _triangleList;
		if(_normalList != NULL)
			delete[] _normalList;
	}

	size_t vol3D_Component::GetVertexCount() const
	{
		return _numVertices;
	}

	float* vol3D_Component::GetVertexList() const
	{
		return _vertexList;
	}

	size_t vol3D_Component::GetTriangleCount() const
	{
		return _numTriangles;
	}

	size_t* vol3D_Component::GetTriangleList() const
	{
		return _triangleList;
	}

	float* vol3D_Component::GetNormalList() const
	{
		return _normalList;
	}

	size_t vol3D_Component::GetColor() const
	{
		return _color;
	}

	template <class T>
	vol3D_IsoSurface<T>::vol3D_IsoSurface()
	{
		_cellLengthX = 0;
		_cellLengthY = 0;
		_cellLengthZ = 0;
		_cellsX = 0;
		_cellsY = 0;
		_cellsZ = 0;
		_scalarField = NULL;
		_isoLevel = 0;
		_valid = false;
	}

	template <class T>
	vol3D_IsoSurface<T>::~vol3D_IsoSurface()
	{
		Delete();
	}

	template <class T>
	void vol3D_IsoSurface<T>::GenerateSegmented(const T* scalarField, 
        const vol3D_ComponentTable &componentInfo, size_t cellsX, size_t cellsY,
		size_t cellsZ, float cellLengthX, float cellLengthY, float cellLengthZ)
	{
		TriangleID triangle;
		size_t id, j, x, y, z;
		Point3DID pt;
		vol3D_ComponentVector::iterator i;
		typename vol3D_ComponentTable::const_iterator info;
		const T *memPtr;
		std::set<size_t> hit;
		typename std::set<size_t>::const_iterator iHit;

		if(_valid)
			Delete();

		_cellsX = cellsX;
		_cellsY = cellsY;
		_cellsZ = cellsZ;
		_cellLengthX = cellLengthX;
		_cellLengthY = cellLengthY;
		_cellLengthZ = cellLengthZ;
		_scalarField = scalarField;
		_sliceSize = _cellsX*_cellsY;
		_rowSize = _cellsX;

		_compVec.resize(componentInfo.size());

		for(info = componentInfo.begin(); info != componentInfo.end(); info++)
		{
			_compVec[info->second]._tableIndex = 0xff;
			_compVec[info->second]._color = (size_t)info->first;
		}

		//beginning of the last line of the first slice of _scalarField
		memPtr = _scalarField + _sliceSize - _rowSize;

		//Generate isosurface
		for(z = 0; z < _cellsZ-1; z++)
		{
			for(y = 0; y < _cellsY-1; y++)
			{
				for(x = 0; x < _cellsX-1; x++)
				{
					//Calculate table lookup index from those
					//vertices which are below the isolevel
					if(*memPtr)
					{
						info = componentInfo.find(*memPtr);
						_compVec[info->second]._tableIndex ^= 1;
						hit.insert(info->second);
					}
					if(*(memPtr-_rowSize))
					{
						info = componentInfo.find(*(memPtr-_rowSize));
						_compVec[info->second]._tableIndex ^= 2;
						hit.insert(info->second);
					}
					if(*(memPtr-_rowSize+1))
					{
						info = componentInfo.find(*(memPtr-_rowSize+1));
						_compVec[info->second]._tableIndex ^= 4;
						hit.insert(info->second);
					}
					if(*(memPtr+1))
					{
						info = componentInfo.find(*(memPtr+1));
						_compVec[info->second]._tableIndex ^= 8;
						hit.insert(info->second);
					}
					if(*(memPtr+_sliceSize))
					{
						info = componentInfo.find(*(memPtr+_sliceSize));
						_compVec[info->second]._tableIndex ^= 16;
						hit.insert(info->second);
					}
					if(*(memPtr+_sliceSize-_rowSize))
					{
						info = componentInfo.find(*(memPtr+_sliceSize-_rowSize));
						_compVec[info->second]._tableIndex ^= 32;
						hit.insert(info->second);
					}
					if(*(memPtr+_sliceSize-_rowSize+1))
					{
						info = componentInfo.find(*(memPtr+_sliceSize-_rowSize+1));
						_compVec[info->second]._tableIndex ^= 64;
						hit.insert(info->second);
					}
					if(*(memPtr+_sliceSize+1))
					{
						info = componentInfo.find(*(memPtr+_sliceSize+1));
						_compVec[info->second]._tableIndex ^= 128;
						hit.insert(info->second);
					}

					//Now create a triangulation of the isosurface in this cell
					for(iHit = hit.begin(); iHit != hit.end(); iHit++)
					{
						if(edgeTable[_compVec[*iHit]._tableIndex] != 0)
						{
							if(edgeTable[_compVec[*iHit]._tableIndex] & 8)
							{
								pt = CalculateIntersection(x, y, z, 3, false);
								id = 3*((memPtr-_scalarField));
								_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
							}
							if(edgeTable[_compVec[*iHit]._tableIndex] & 1)
							{
								pt = CalculateIntersection(x, y, z, 0, false);
								id = 3*((memPtr-_scalarField)) + 1;
								_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
							}
							if(edgeTable[_compVec[*iHit]._tableIndex] & 256)
							{
								pt = CalculateIntersection(x, y, z, 8, false);
								id = 3*((memPtr-_scalarField)) + 2;
								_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
							}
							
							if(x == _cellsX - 2)
							{
								if(edgeTable[_compVec[*iHit]._tableIndex] & 4)
								{
									pt = CalculateIntersection(x, y, z, 2, false);
									id = 3*((memPtr-_scalarField) + 1) + 1;
									_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
								}
								if(edgeTable[_compVec[*iHit]._tableIndex] & 2048)
								{
									pt = CalculateIntersection(x, y, z, 11, false);
									id = 3*((memPtr-_scalarField) + 1) + 2;
									_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
								}
							}
							if(y == _cellsY - 2)
							{
								if(edgeTable[_compVec[*iHit]._tableIndex] & 2)
								{
									pt = CalculateIntersection(x, y, z, 1, false);
									id = 3*((memPtr-_scalarField) - _rowSize);
									_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
								}
								if(edgeTable[_compVec[*iHit]._tableIndex] & 512)
								{
									pt = CalculateIntersection(x, y, z, 9, false);
									id = 3*((memPtr-_scalarField) - _rowSize) + 2;
									_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
								}
							}
							if(z == _cellsZ - 2)
							{
								if(edgeTable[_compVec[*iHit]._tableIndex] & 16)
								{
									pt = CalculateIntersection(x, y, z, 4, false);
									id = 3*((memPtr-_scalarField) + _sliceSize) + 1;
									_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
								}
								if(edgeTable[_compVec[*iHit]._tableIndex] & 128)
								{
									pt = CalculateIntersection(x, y, z, 7, false);
									id = 3*((memPtr-_scalarField) + _sliceSize);
									_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
								}
							}
							if((x == _cellsX - 2) && (y == _cellsY - 2))
							{
								if(edgeTable[_compVec[*iHit]._tableIndex] & 1024)
								{
									pt = CalculateIntersection(x, y, z, 10, false);
									id = 3*((memPtr-_scalarField) - _rowSize + 1) + 2;
									_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
								}
							}
							if((x == _cellsX - 2) && (z == _cellsZ - 2))
							{
								if(edgeTable[_compVec[*iHit]._tableIndex] & 64)
								{
									pt = CalculateIntersection(x, y, z, 6, false);
									id = 3*((memPtr-_scalarField) + _sliceSize + 1) + 1;
									_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
								}
							}
							if((y == _cellsY - 2) && (z == _cellsZ - 2))
							{
								if(edgeTable[_compVec[*iHit]._tableIndex] & 32)
								{
									pt = CalculateIntersection(x, y, z, 5, false);
									id = 3*((memPtr-_scalarField) + _sliceSize-_rowSize);
									_compVec[*iHit]._pointMap.insert(ID2Point3DID::value_type(id, pt));
								}
							}

							//Create the triangles
							for(j = 0; triTable[_compVec[*iHit]._tableIndex][j] != -1; j += 3)
							{
								triangle.pointID[0] = GetEdgeID(memPtr, triTable[_compVec[*iHit]._tableIndex][j]);
								triangle.pointID[1] = GetEdgeID(memPtr, triTable[_compVec[*iHit]._tableIndex][j+1]);
								triangle.pointID[2] = GetEdgeID(memPtr, triTable[_compVec[*iHit]._tableIndex][j+2]);

								_compVec[*iHit]._triVec.push_back(triangle);
							}
						}
						_compVec[*iHit]._tableIndex = 0xff;
					}
					hit.clear();
					memPtr++;
				}
				memPtr -= ((_rowSize<<1)-1);
			}
			memPtr += (_sliceSize<<1) - _rowSize;
		}

		RenameVerticesAndTriangles();
		CalculateNormals();
 		_valid = true;
	}

	template <class T>
	bool vol3D_IsoSurface<T>::IsValid() const
	{
		return _valid;
	}

	template <class T>
	void vol3D_IsoSurface<T>::Delete()
	{
		_cellLengthX = 0;
		_cellLengthY = 0;
		_cellLengthZ = 0;
		_cellsX = 0;
		_cellsY = 0;
		_cellsZ = 0;
		
		_compVec.clear();

		_scalarField = NULL;
		_isoLevel = 0;
		_valid = false;
	}

	template <class T>
	int vol3D_IsoSurface<T>::GetVolumeLengths(float& volLengthX, float& volLengthY, float& volLengthZ) const
	{
		if(IsValid())
		{
			volLengthX = _cellLengthX*_cellsX;
			volLengthY = _cellLengthY*_cellsY;
			volLengthZ = _cellLengthZ*_cellsZ;
			return 1;
		}
		else
			return -1;
	}

	template <class T>
	const vol3D_ComponentVector& vol3D_IsoSurface<T>::GetComponentVector() const
	{
		return _compVec;
	}

	template <class T>
	size_t vol3D_IsoSurface<T>::GetEdgeID(const T *memPtr, size_t edgeNo) const
	{
		switch(edgeNo)
		{
		case 0:
			return 3*((memPtr-_scalarField)) + 1;
		case 1:
			return 3*((memPtr-_scalarField) - _rowSize);
		case 2:
			return 3*((memPtr-_scalarField) + 1) + 1;
		case 3:
			return 3*((memPtr-_scalarField));
		case 4:
			return 3*((memPtr-_scalarField) + _sliceSize) + 1;
		case 5:
			return 3*((memPtr-_scalarField) + _sliceSize-_rowSize);
		case 6:
			return 3*((memPtr-_scalarField) + _sliceSize + 1) + 1;
		case 7:
			return 3*((memPtr-_scalarField) + _sliceSize);
		case 8:
			return 3*((memPtr-_scalarField)) + 2;
		case 9:
			return 3*((memPtr-_scalarField) - _rowSize) + 2;
		case 10:
			return 3*((memPtr-_scalarField) - _rowSize + 1) + 2;
		case 11:
			return 3*((memPtr-_scalarField) + 1) + 2;
		default:
			// Invalid edge no.
			return -1;
		}
	}

	template <class T>
	Point3DID vol3D_IsoSurface<T>::CalculateIntersection(size_t x, size_t y, size_t z,
		size_t edgeNo, bool interpolate) const
	{
		float x1, y1, z1, x2, y2, z2;
		size_t v1x = x, v1y = y, v1z = z;
		size_t v2x = x, v2y = y, v2z = z;
		Point3DID intersection;
		
		switch(edgeNo)
		{
		case 0:
			v2y += 1;
			break;
		case 1:
			v1y += 1;
			v2x += 1;
			v2y += 1;
			break;
		case 2:
			v1x += 1;
			v1y += 1;
			v2x += 1;
			break;
		case 3:
			v1x += 1;
			break;
		case 4:
			v1z += 1;
			v2y += 1;
			v2z += 1;
			break;
		case 5:
			v1y += 1;
			v1z += 1;
			v2x += 1;
			v2y += 1;
			v2z += 1;
			break;
		case 6:
			v1x += 1;
			v1y += 1;
			v1z += 1;
			v2x += 1;
			v2z += 1;
			break;
		case 7:
			v1x += 1;
			v1z += 1;
			v2z += 1;
			break;
		case 8:
			v2z += 1;
			break;
		case 9:
			v1y += 1;
			v2y += 1;
			v2z += 1;
			break;
		case 10:
			v1x += 1;
			v1y += 1;
			v2x += 1;
			v2y += 1;
			v2z += 1;
			break;
		case 11:
			v1x += 1;
			v2x += 1;
			v2z += 1;
			break;
		}

		x1 = v1x*_cellLengthX;
		y1 = v1y*_cellLengthY;
		z1 = v1z*_cellLengthZ;
		x2 = v2x*_cellLengthX;
		y2 = v2y*_cellLengthY;
		z2 = v2z*_cellLengthZ;

		if(interpolate)
		{
			T val1 = _scalarField[MemAddr(v1x, v1y, v1z)];
			T val2 = _scalarField[MemAddr(v2x, v2y, v2z)];
			intersection = Interpolate(x1, y1, z1, x2, y2, z2, val1, val2);
		}
		else
		{
			intersection.x = (x1+x2)/2.0f; intersection.y = (y1+y2)/2.0f; intersection.z = (z1+z2)/2.0f;
		}
		
		return intersection;
	}

	template <class T>
	Point3DID vol3D_IsoSurface<T>::Interpolate(float x1, float y1, float z1,
		float x2, float y2, float z2, T val1, T val2) const
	{
		Point3DID interpolation;
		float mu;

		if(fabs((double)_isoLevel-val1) < 0.00001)
		{
			interpolation.x = x1; interpolation.y = y1; interpolation.z = z1;
			return interpolation;
		}
		if(fabs((double)_isoLevel-val2) < 0.00001)
		{
			interpolation.x = x2; interpolation.y = y2; interpolation.z = z2;
			return interpolation;
		}
		if(fabs((double)val1-val2) < 0.00001)
		{
			interpolation.x = x1; interpolation.y = y1; interpolation.z = z1;
			return interpolation;
		}


		mu = (float)(_isoLevel - val1)/(float)(val2 - val1);
		interpolation.x = x1 + mu*(x2 - x1);
		interpolation.y = y1 + mu*(y2 - y1);
		interpolation.z = z1 + mu*(z2 - z1);

		return interpolation;
	}

	template <class T>
	void vol3D_IsoSurface<T>::RenameVerticesAndTriangles()
	{
		size_t nextID, i;
		float *pVec;
		size_t *pTri;
		ID2Point3DID::iterator mapIterator;
		TriangleVector::iterator vecIterator;

		for(vol3D_ComponentVector::iterator comp = _compVec.begin(); comp != _compVec.end(); comp++)
		{
			nextID = 0;
			mapIterator = comp->_pointMap.begin();
			vecIterator = comp->_triVec.begin();

			comp->_numVertices = comp->_pointMap.size();
			comp->_vertexList = new float[comp->_numVertices*3];
			pVec = comp->_vertexList;
			
			//Rename and copy vertices so that they can be efficiently accessed
			while(mapIterator != comp->_pointMap.end())
			{
				(*mapIterator).second.newID = nextID;
				nextID++;

				*pVec = (*mapIterator).second.x;
				*(pVec+1) = (*mapIterator).second.y;
				*(pVec+2) = (*mapIterator).second.z;
				pVec += 3;

				mapIterator++;
			}

			comp->_numTriangles = comp->_triVec.size();
			comp->_triangleList = new size_t[comp->_numTriangles*3];
			pTri = comp->_triangleList;

			//Now rename triangles
			while(vecIterator != comp->_triVec.end())
			{
				for(i = 0; i < 3; i++, pTri++)
				{
					*pTri = comp->_pointMap[(*vecIterator).pointID[i]].newID;
				}
				vecIterator++;
			}

			comp->_pointMap.clear();
			comp->_triVec.clear();
		}
	}

	template <class T>
	void vol3D_IsoSurface<T>::CalculateNormals()
	{
		i3d::Vector3d<float> vec1, vec2, normal;
		size_t id0, id1, id2;

		for(vol3D_ComponentVector::iterator comp = _compVec.begin(); comp != _compVec.end(); comp++)
		{
			//number of normals = number of vertexes * dimension
			comp->_normalList = new float[comp->_numVertices*3];
			
			//Set all normals to 0
			memset(comp->_normalList, 0, comp->_numVertices*3*sizeof(float));

			//Calculate normals
			for(size_t i = 0; i < comp->_numTriangles; i++)
			{
				id0 = comp->_triangleList[3*i];
				id1 = comp->_triangleList[3*i+1];
				id2 = comp->_triangleList[3*i+2];

				vec1.x = comp->_vertexList[3*id1+0] - comp->_vertexList[3*id0+0];
				vec1.y = comp->_vertexList[3*id1+1] - comp->_vertexList[3*id0+1];
				vec1.z = comp->_vertexList[3*id1+2] - comp->_vertexList[3*id0+2];
				vec2.x = comp->_vertexList[3*id2+0] - comp->_vertexList[3*id0+0];
				vec2.y = comp->_vertexList[3*id2+1] - comp->_vertexList[3*id0+1];
				vec2.z = comp->_vertexList[3*id2+2] - comp->_vertexList[3*id0+2];

				normal.x = vec1.z*vec2.y - vec1.y*vec2.z;
				normal.y = vec1.x*vec2.z - vec1.z*vec2.x;
				normal.z = vec1.y*vec2.x - vec1.x*vec2.y;

				comp->_normalList[3*id0+0] += normal.x;
				comp->_normalList[3*id0+1] += normal.y;
				comp->_normalList[3*id0+2] += normal.z;
				comp->_normalList[3*id1+0] += normal.x;
				comp->_normalList[3*id1+1] += normal.y;
				comp->_normalList[3*id1+2] += normal.z;
				comp->_normalList[3*id2+0] += normal.x;
				comp->_normalList[3*id2+1] += normal.y;
				comp->_normalList[3*id2+2] += normal.z;
			}

			//Normalize normals
			for(size_t i = 0; i < comp->_numVertices; i++)
			{
				float length = sqrt(comp->_normalList[3*i+0]*comp->_normalList[3*i+0]
					+ comp->_normalList[3*i+1]*comp->_normalList[3*i+1]
					+ comp->_normalList[3*i+2]*comp->_normalList[3*i+2]);
				comp->_normalList[3*i+0] /= length;
				comp->_normalList[3*i+1] /= length;
				comp->_normalList[3*i+2] /= length;
			}
		}
	}

	//EXPLICIT INSTANTIATIONS
	template class vol3D_IsoSurface<bool>;
	template class vol3D_IsoSurface<i3d::GRAY8>;
    template class vol3D_IsoSurface<i3d::GRAY16>;
    template class vol3D_IsoSurface<float>;
	template class vol3D_IsoSurface<double>;
    template class vol3D_IsoSurface<int>;
	template class vol3D_IsoSurface<size_t>;

    template class vol3D_IsoSurface<i3d::RGB>;
    template class vol3D_IsoSurface<i3d::RGB16>;
}
