/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* vol3D_isosurface.h
 *
 * Marching cube algorithm for I3D library
 *
 * Marek Vinkler (172521@mail.muni.cz) 2008
 */

#ifndef V4D_VOL3D_ISOSURFACE_H
#define V4D_VOL3D_ISOSURFACE_H

#include <map>
#include <set>
#include <vector>
#include <i3d/vector3d.h>

namespace v4d
{
	struct Point3DID
	{
		size_t newID;
		float x, y, z;
	};

	struct TriangleID
	{
		size_t pointID[3];
	};

	typedef std::map<size_t, Point3DID> ID2Point3DID;
	typedef std::vector<TriangleID> TriangleVector;

	//class which holds information about single component in the scalar field
	class vol3D_Component
	{
		//All vol3D_IsoSurface are friends - can directly operate on private members
		template<class T>	friend class vol3D_IsoSurface;
	public:
		//Constructor and destructor
		vol3D_Component();
		~vol3D_Component();
		//Returns the number of vertices
		size_t GetVertexCount() const;
		//Returns the list of vertices
		float* GetVertexList() const;
		//Returns the number of triangle primitives created
		size_t GetTriangleCount() const;
		//Returns the triangle list
		size_t* GetTriangleList() const;
		//Returns the list of vertex normals
		float* GetNormalList() const;
		//Returns the color
		size_t GetColor() const;

	private:
		//Table index of this component
		size_t _tableIndex;
		//List of POINT3Ds which form the isosurface
		ID2Point3DID _pointMap;
		//List of Triangles which form the triangulation of the isosurface
		TriangleVector _triVec;
		//The number of vertices which make up the isosurface
		size_t _numVertices;
		//The number of triangles which make up the isosurface
		size_t _numTriangles;
		//The vertices which make up the isosurface
		float *_vertexList;
		//The triangle indices which make up the isosurface
		size_t *_triangleList;
		//The normals for each vertex
		float *_normalList;
		//color index of the component
		size_t _color;
	};

	typedef std::vector<vol3D_Component> vol3D_ComponentVector;

	template <class T>
	class vol3D_IsoSurface
	{
	public:
		typedef std::map<T, size_t> vol3D_ComponentTable;

		//Constructor and destructor
		vol3D_IsoSurface();
		~vol3D_IsoSurface();
		
		//Generates the isosurface from the scalar field contained in the buffer scalarField[]
		//grayscale data version
		/*void Generate(const T* scalarField, double isoLevel, size_t cellsX, size_t cellsY, size_t cellsZ,
			float cellLengthX, float cellLengthY, float cellLengthZ, bool interpolate = true);*/
		//segmentation data version
		void GenerateSegmented(const T* scalarField, const vol3D_ComponentTable &componentInfo, size_t cellsX,
			size_t cellsY, size_t cellsZ, float cellLengthX, float cellLengthY, float cellLengthZ);

		//Returns true if a valid surface has been generated
		bool IsValid() const;

		//Deletes the isosurface
		void Delete();

		//Returns the length, width, and height of the volume in which the
		//isosurface in enclosed in.  Returns -1 if the surface is not valid
		int GetVolumeLengths(float& volLengthX, float& volLengthY, float& volLengthZ) const;

		//Returns a vector of created isosurface components
		const vol3D_ComponentVector& GetComponentVector() const;

	private:
		//Returns the edge ID
		size_t GetEdgeID(const T *memPtr, size_t edgeNo) const;

		//Calculates the intersection point of the isosurface with an edge
		Point3DID CalculateIntersection(size_t x, size_t y, size_t z,
			size_t edgeNo, bool interpolate) const;

		//Interpolates between two grid points to produce the point at which
		//the isosurface intersects an edge
		Point3DID Interpolate(float x1, float y1, float z1, float x2, float y2, float z2, T val1, T val2) const;

		//Renames vertices and triangles so that they can be accessed more efficiently
		void RenameVerticesAndTriangles();

		//Calculates the normals
		void CalculateNormals();

		//Calculates the memory adress of a cell
		size_t MemAddr(size_t x, size_t y, size_t z) const
    	{
	    	return z*_sliceSize + ((_cellsY-1) - y)*_rowSize + x;
	    }

		//Vector of found components
		vol3D_ComponentVector _compVec;

		//No. of cells in x, y, and z directions
		size_t _cellsX, _cellsY, _cellsZ;

		//Cell length in x, y, and z directions
		float _cellLengthX, _cellLengthY, _cellLengthZ;

		//The buffer holding the scalar field
		const T *_scalarField;

		//The size of a slice
		size_t _sliceSize;

		//The size of a row
		size_t _rowSize;

		//The isosurface value
		double _isoLevel;

		//Indicates whether a valid surface is present
		bool _valid;
	};
}

#endif //__VOL3D_ISOSURFACE_H__

