/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * wxtools.cpp
 *
 * Some wxWidgets improvements
 *
 * by Petr Matula (September 2004)
 */

#include "v4d_basic.h"
#include "wxtools.h"

#ifdef _DEBUG
    #define new DEBUG_NEW
#endif

namespace v4d
{
    ///////////////////////////////////////////////////////////
    //                                                       //
    //             Some useful global functions              //
    //                                                       //
    ///////////////////////////////////////////////////////////
    void SavewxString (wxString &s, std::ostream &o, bool space)
    {
        if (s.Find (' ') == -1)
            o << s.c_str ();
        else
            o << '\"' << s.c_str() << '\"';

        if (space)
            o << ' ';
    }


    void LoadwxString (wxString &s, std::istream &i)
    {
        char lookfor = ' ', znak = ' ';

        s.clear ();

        // v batchcropu neni \n potreba. 
        // OVERIT - zda to nedela neplechu pri nacitani algoritmu
        while (!i.eof() && isspace (znak)) // && znak != '\n') 
            znak = i.get ();

        if (znak == '\"')
            lookfor = znak;
        else
            i.putback (znak);

        while (!i.eof() && znak != '\n')
        {
            znak = i.get();

            if (!i.eof())
            {
                if (znak == '\n')
                    i.putback (znak);
                else 
                    if (znak == lookfor || (lookfor == ' ' && znak == '\t'))
                        break;
                    else
                        s += znak;
            }
        }
    }

	template <class T> wxString VoxelToStr(const T &val)
    {
        return wxString::Format("%d", int(val));
    }

    template <> V4D_DLLEXPORT wxString VoxelToStr(const float &val)
    {
        return wxString::Format("%g",val);
    }

    template <> V4D_DLLEXPORT wxString VoxelToStr(const i3d::RGB &val)
    {
        return wxString::Format("%3d %3d %3d", int(val.red), int(val.green), int(val.blue));
    }

    template <> V4D_DLLEXPORT wxString VoxelToStr(const i3d::RGB16 &val)
    {
        return wxString::Format("%3d %3d %3d", int(val.red), int(val.green), int(val.blue));
    }

    template <> V4D_DLLEXPORT wxString VoxelToStr(const std::complex<float> &val,
        const int showMode)
    {
		  switch (showMode)
		  {
		  case 2:
			  //showing also phase
			  return wxString::Format("%3f + %3fi\nPhase: %3f rad",
			  				val.real(), val.imag(),
							std::atan2(val.imag(), val.real()) );
		  case 3:
			  //showing also magnitudes
			  return wxString::Format("%3f + %3fi\nAmplitude: %3f",
			  				val.real(), val.imag(),
							sqrtf(val.real()*val.real() + val.imag()*val.imag()) );
		  default:
			  //showing just the complex number
			  return wxString::Format("%3f + %3fi\n ", val.real(), val.imag() );
		  }
    }

    template <> V4D_DLLEXPORT wxString VoxelToStr(const std::complex<double> &val,
        const int showMode)
    {
		  switch (showMode)
		  {
		  case 2:
			  //showing also phase
			  return wxString::Format("%3g + %3gi\nPhase: %3g rad",
			  				val.real(), val.imag(),
							std::atan2(val.imag(), val.real()) );
		  case 3:
			  //showing also magnitudes
			  return wxString::Format("%3g + %3gi\nAmplitude: %3g",
			  				val.real(), val.imag(),
							sqrt(val.real()*val.real() + val.imag()*val.imag()) );
		  default:
			  //showing just the complex number
			  return wxString::Format("%3g + %3gi\n ", val.real(), val.imag() );
		  }
    }

	template V4D_DLLEXPORT wxString VoxelToStr(const i3d::GRAY8 &value);
	template V4D_DLLEXPORT wxString VoxelToStr(const i3d::GRAY16 &value);
	template V4D_DLLEXPORT wxString VoxelToStr(const int &value);
	
    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                         Validate                            //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    template <class T> bool RealValidator<T>::Validate(wxWindow *parent)
    {
        if (!wxTextValidator::Validate(parent))
            return false;

        bool ok(true);
        wxString errmsg;

        if (check_min || check_max)
        {

            wxTextCtrl *control = (wxTextCtrl *) m_validatorWindow ;

            T dbl;
            if (GetRealValue(control, &dbl))
            {
                if (check_min && check_max && (dbl < min_val || dbl > max_val))
                {
                    wxString fmt;
                    fmt = "The value " + format_str + " is out of range ["
                        + format_str + "," + format_str + "]";
                    errmsg.Printf(fmt.c_str(), dbl, min_val, max_val);
                    ok = false;
                }

                if (check_min && !check_max && (dbl < min_val))
                {
                    wxString fmt;
                    fmt = "The value cannot be less than " + format_str;
                    errmsg.Printf(fmt.c_str(), min_val);
                    ok = false;
                }

                if (!check_min && check_max && (dbl > max_val))
                {
                    wxString fmt;
                    fmt = "The value cannot be greater than " + format_str;
                    errmsg.Printf(fmt.c_str(), max_val);
                    ok = false;
                }
            }
            else
            {
                errmsg.Printf("The value '%s' cannot be converted to a double", control->GetValue().c_str());
                ok = false;
            }
        }
        if (!ok)
        {
            wxMessageBox(errmsg, "Validation conflict", wxOK | wxICON_EXCLAMATION, parent);
            m_validatorWindow->SetFocus();
        }

        return ok;
    }

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                     Transfer To Window                      //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    template <class T> bool RealValidator<T>::TransferToWindow(void)
    {
        if (GetTextEntry() == NULL)
            return false;

        wxTextCtrl *control = (wxTextCtrl *) m_validatorWindow ;
        control->SetValue(wxString::Format(format_str, my_dbl));

        return true;
    }
    template <class T> void RealValidator<T>::SetMin(T m)
    {
        min_val = m;
    }
    template <class T> void RealValidator<T>::SetMax(T m)
    {

        max_val = m;
    }

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                    Transfer From Window                     //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    template <class T> bool RealValidator<T>::TransferFromWindow(void)
    {
        if (GetTextEntry() == NULL)
            return false;

        wxTextCtrl *control = (wxTextCtrl *) m_validatorWindow ;

        return GetRealValue(control, &my_dbl);
    }

    template <class T> bool RealValidator<T>::GetRealValue(wxTextCtrl *textctrl, T *val)
    {
        wxMessageBox(wxString::Format("RealValidator::GetRealValue: Only float and double types are supported."));
        return false;
    }

    template <> bool RealValidator<double>::GetRealValue(wxTextCtrl *textctrl, double *val)
    {
        return textctrl->GetValue().ToDouble(val);
    }

    template <> bool RealValidator<float>::GetRealValue(wxTextCtrl *textctrl, float *val)
    {
        double dbl;
        bool ok = textctrl->GetValue().ToDouble(&dbl);
        *val = dbl;
        return ok;
    }

    template class V4D_DLLEXPORT RealValidator<double>;
    template class V4D_DLLEXPORT RealValidator<float>;

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                         Validate                            //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    template <class T> bool IntValidator<T>::Validate(wxWindow *parent)
    {
        if (!wxTextValidator::Validate(parent))
            return false;

        bool ok(true);
        wxString errmsg;

        if (check_min || check_max)
        {
            wxTextCtrl *control = (wxTextCtrl *) m_validatorWindow ;

            long i;
            if (control->GetValue().ToLong(&i))
            {
                if (check_min && check_max && (i < static_cast<long>(min_val) || i > static_cast<long>(max_val)))
                {
                    wxString fmt;
                    fmt = "The value " + format_str + " is out of range [" + format_str + "," + format_str + "]";
                    errmsg.Printf(fmt.c_str(), i, min_val, max_val);
                    ok = false;
                }

                if (check_min && !check_max && (i < static_cast<long>(min_val)))
                {
                    wxString fmt;
                    fmt = "The value cannot be less than " + format_str;
                    errmsg.Printf(fmt.c_str(), min_val);
                    ok = false;
                }

                if (!check_min && check_max && (i > static_cast<long>(max_val)))
                {
                    wxString fmt;
                    fmt = "The value cannot be greater than " + format_str;
                    errmsg.Printf(fmt.c_str(), max_val);
                    ok = false;
                }
            }
            else
            {
                errmsg.Printf("The value '%s' cannot be converted to a long", control->GetValue().c_str());
                ok = false;
            }
        }
        if (!ok)
        {
            wxMessageBox(errmsg, "Validation conflict", wxOK | wxICON_EXCLAMATION, parent);
            m_validatorWindow->SetFocus();
        }

        return ok;
    }

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                     Transfer To Window                      //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    template <class T> bool IntValidator<T>::TransferToWindow(void)
    {
        if (GetTextEntry() == NULL)
            return false;

        wxTextCtrl *control = (wxTextCtrl *) m_validatorWindow ;
        control->SetValue(wxString::Format(format_str, my_long));

        return true;
    }

    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                    Transfer From Window                     //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    template <class T> bool IntValidator<T>::TransferFromWindow(void)
    {
        if (GetTextEntry() == NULL)
            return false;

        wxTextCtrl *control = (wxTextCtrl *) m_validatorWindow ;

        return GetIntValue(control, &my_long);
    }

    template <class T> bool IntValidator<T>::GetIntValue(wxTextCtrl *textctrl, T *val)
    {
        wxMessageBox(wxString::Format("IntValidator::GetIntValue: Only long, int, unsigned int, Gray8 and Gray16 types are supported."));
        return false;
    }

    template <> bool IntValidator<long>::GetIntValue(wxTextCtrl *textctrl, long *val)
    {
        return textctrl->GetValue().ToLong(val);
    }

    template <> bool IntValidator<int>::GetIntValue(wxTextCtrl *textctrl, int *val)
    {
        long lval;
        bool ok = textctrl->GetValue().ToLong(&lval);
        *val = static_cast<int>(lval);
        return ok;
    }

    template <> bool IntValidator<unsigned int>::GetIntValue(wxTextCtrl *textctrl, unsigned int *val)
    {
        unsigned long lval;
        bool ok = textctrl->GetValue().ToULong(&lval);
        *val = static_cast<unsigned int>(lval);
        return ok;
    }

    template <> bool IntValidator<i3d::GRAY8>::GetIntValue(wxTextCtrl *textctrl, i3d::GRAY8 *val)
    {
        long lval;
        bool ok = textctrl->GetValue().ToLong(&lval);
        *val = static_cast<i3d::GRAY8>(lval);
        return ok;
    }

    template <> bool IntValidator<i3d::GRAY16>::GetIntValue(wxTextCtrl *textctrl, i3d::GRAY16 *val)
    {
        long lval;
        bool ok = textctrl->GetValue().ToLong(&lval);
        *val = static_cast<i3d::GRAY16>(lval);
        return ok;
    }

    template class V4D_DLLEXPORT IntValidator<long>;
    template class V4D_DLLEXPORT IntValidator<int>;
    template class V4D_DLLEXPORT IntValidator<unsigned int>;
    template class V4D_DLLEXPORT IntValidator<unsigned long>;
    template class V4D_DLLEXPORT IntValidator<i3d::GRAY8>;
    template class V4D_DLLEXPORT IntValidator<i3d::GRAY16>;

#if defined(_M_X64) || (defined(__GNUC__) && (__SIZEOF_INT__ != __SIZEOF_SIZE_T__) && (__SIZEOF_LONG__ != __SIZEOF_SIZE_T__))
    // On 64-bit platforms size_t is usually not the same as unsigned int
    template class V4D_DLLEXPORT IntValidator<size_t>;
#endif
}
