/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-*-c++-*-
 * wxtools.h
 *
 * Some wxWidgets improvements
 *
 * by Petr Matula (September 2004)
 */

#ifndef V4D_WXTOOLS_H
#define V4D_WXTOOLS_H

#include <typeinfo>
#include <iostream>
#include <wx/valtext.h>
#include <wx/textctrl.h> 
#include <i3d/basic.h>
#include "v4d_impexp.h"

namespace v4d
{
    ///////////////////////////////////////////////////////////
    //                                                       //
    //             Some useful global functions              //
    //                                                       //
    ///////////////////////////////////////////////////////////
    /** Output a wxString into a stream. If the string contains spaces then 
    quotation marks are added around the string. */
    V4D_DLLEXPORT void SavewxString (wxString &s, std::ostream &o, bool add_space = false);

    /** Load a string previously saved with call to SaveString function 
    from a stream. Skips all white space in front of the string except
    new line characters. */
    V4D_DLLEXPORT void LoadwxString (wxString &s, std::istream &o);

	/** Convert given scalar voxel value to a wxString. */
	template <class T> V4D_DLLEXPORT wxString VoxelToStr(const T &val);

	/** Convert given complex voxel value to a wxString. */
	template <class T> V4D_DLLEXPORT wxString VoxelToStr(const std::complex<T> &val,
        const int showMode);

	
    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                       RealValidator                         //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    /** Validator class for validating floating point numbers in wxTextCtrl. */
    template <class T> class V4D_DLLEXPORT RealValidator: public wxTextValidator
    {
    protected:

        /** Should be the min limit checked?  */
        bool check_min;
        /** Should be the max limit checked?  */
        bool check_max;

        /** Minimal allowed value. It is checked only if \p check_min == \p true  */
        T min_val;
        /** Maximal allowed value. It is checked only if \p check_max == \p true  */
        T max_val;
        /** Reference to a double variable */
        T &my_dbl;
        /** Format string used for displaying double value */
        wxString format_str;
        /** String buffer */
        wxString strbuf;
        bool GetRealValue(wxTextCtrl *textctrl, T *val);

    public:
        /** Constructor. No check on limits is done. */
        RealValidator(T &dbl, const wxString &format = wxEmptyString)
            : wxTextValidator(wxFILTER_NUMERIC, &strbuf),
              check_min(false), check_max(false), my_dbl(dbl), format_str(format) 
          { 
              if (format_str.IsEmpty())
              {
                  if (typeid(T) == typeid(double))
                      format_str = "%g";
                  else
                      format_str = "%f";
              }
          }

          /** Constructor. The values must be in interval [minval, maxval]. */
          RealValidator(T &dbl, T minval, T maxval,
              bool chmin = true, bool chmax = true, const wxString &format = wxEmptyString)
              : wxTextValidator(wxFILTER_NUMERIC, &strbuf),
              check_min(chmin), check_max(chmax), min_val(minval), max_val(maxval),
              my_dbl(dbl), format_str(format) 
          {
              if (format_str.IsEmpty())
              {
                  if (typeid(T) == typeid(double))
                      format_str = "%g";
                  else
                      format_str = "%f";
              }
          }

          /** Destructor */
          ~RealValidator() 
          {}

          /** \returns copy of itself */
          virtual wxObject *Clone(void) const { return new RealValidator(*this); }

          /** Validate the control */
          virtual bool Validate(wxWindow *parent);

          /** Transfer the value from my_dbl variable to the text control  */
          virtual bool TransferToWindow(void);

          /** Transfer the value from the text control to my_dbl variable */
          virtual bool TransferFromWindow(void);
          T GetMin() {return min_val;}
          T GetMax() {return max_val;}
          void SetMin(T m);
          void SetMax(T m);
    };

    typedef RealValidator<double> DoubleValidator;
    typedef RealValidator<float> FloatValidator;



    /////////////////////////////////////////////////////////////////
    //                                                             //
    //                        IntValidator                         //
    //                                                             //
    /////////////////////////////////////////////////////////////////
    /** Validator class for validating integer numbers in wxTextCtrl. */
    template <class T> class V4D_DLLEXPORT IntValidator: public wxTextValidator
    {
    protected:

        /** Should be the min limit checked?  */
        bool check_min;
        /** Should be the max limit checked?  */
        bool check_max;

        /** Minimal allowed value. It is checked only if \p check_min == \p true  */
        T min_val;
        /** Maximal allowed value. It is checked only if \p check_max == \p true  */
        T max_val;
        /** Reference to a long variable */
        T &my_long;
        /** Format string used for displaying long value */
        wxString format_str;
        /** String buffer */
        wxString strbuf;
        bool GetIntValue(wxTextCtrl *textctrl, T *val);
    public:
        /** Constructor. No check on limits is done. */
        IntValidator(T &i, const wxString &format = "%d"):
          wxTextValidator(wxFILTER_NUMERIC, &strbuf),
              check_min(false), check_max(false), my_long(i), format_str(format) 
          {}

          /** Constructor. The values must be in longerval [minval, maxval]. */
          IntValidator(T &i, T minval, T maxval, bool only_positive = false,
              bool chmin = true, bool chmax = true, const wxString &format = "%d")
              : wxTextValidator(wxFILTER_INCLUDE_CHAR_LIST, &strbuf),
              check_min(chmin), check_max(chmax), min_val(minval), max_val(maxval),
              my_long(i), format_str(format) 
          {
              wxArrayString sl;

              sl.Add(wxString("0"));
              sl.Add(wxString("1"));
              sl.Add(wxString("2"));
              sl.Add(wxString("3"));
              sl.Add(wxString("4"));
              sl.Add(wxString("5"));
              sl.Add(wxString("6"));
              sl.Add(wxString("7"));
              sl.Add(wxString("8"));
              sl.Add(wxString("9"));
              if (only_positive == false)
                  sl.Add(wxString("-"));
              SetIncludes(sl);
          };


          /** Destructor */
          ~IntValidator() {}

          /** \returns copy of itself */
          virtual wxObject *Clone(void) const { return new IntValidator(*this); }

          /** Validate the control */
          virtual bool Validate(wxWindow *parent);

          /** Transfer the value from my_long variable to the text control  */
          virtual bool TransferToWindow(void);

          /** Transfer the value from the text control to my_long variable */
          virtual bool TransferFromWindow(void);
    };

    typedef IntValidator<long> LongValidator;
    typedef IntValidator<i3d::GRAY8> Gray8Validator;
    typedef IntValidator<i3d::GRAY16> Gray16Validator;
}

#endif
