/*
 *  v4dwins
 *
 *  Copyright (C) 2006 - 2009   Centre for Biomedical Image Analysis (CBIA)
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* vol3D_visualization.h
 *
 * Volume visualization canvas exploiting OpenGL
 *
 * Radeon Volume Visualization
 * '+'     - Increase number of slices
 * '-'     - Decrease number of slices
 * <SPACE> - Switch between moving clip plane and object
 * <UP>    - Increase the opacity of slices
 * <DOWN>  - Decrease the opacity of slices
 * <LEFT>  - Decrease minimum density displayed
 * <RIGHT> - Increase minimum density displayed
 * <ESC>   - Quit
 *
 * Marek Vinkler (172521@mail.muni.cz) 2008
 */

#ifndef V4D_VOL3D_CANVAS_H
#define V4D_VOL3D_CANVAS_H

#include <wx/glcanvas.h>
#include <wx/stopwatch.h> 
#include <wx/statusbr.h> 
#include <i3d/vector3d.h>
#include "palette.h"
#include "vol3D_isosurface.h"

#if !wxUSE_GLCANVAS
    #error "OpenGL required: set wxUSE_GLCANVAS to 1 and rebuild the wxWidgets library"
#endif

namespace v4d
{
	/** Volume visualization canvas exploiting OpenGL. */
	class vol3D_VolVisCanvas : public wxGLCanvas
	{
    protected:
		//useful data type declaration
		enum MoveModes 
        { 
            MoveNone, 
            MoveObject, 
            MovePlane 
        };

		struct ComponentDrawInfo
		{
			GLuint _vertexVBO, _indexVBO, _normalVBO; //VBOs handles
			GLuint _triangleCount; //elements count
			bool _useLighting; //use lightning
			size_t _color; //color index
		};

		// list of attributes
		static int AttributeList[];

    protected:
		//scale information
		i3d::Vector3d<float> _scale;
		bool _has_data;

		//wxWidgets classes
		wxStopWatch _stopWatch;
		wxStatusBar *_statusBar; 

		//draw Isosurface
		bool _drawIsosurface;

		//vector of components draw information
		std::vector<ComponentDrawInfo> _cdiVec;

		//overlaying palette
		v4d_Palette _pal;

		//vector of isosurface translation
		i3d::Vector3d<float> _trans;

		//default window 
		GLint _screenHeight;
		GLint _screenWidth;

		//number of slices to draw
		int _slices;

		//intensity and density functions
		float _intensity;
		float _density; 

		//texture used to draw the 3D volume
		GLuint _volumeTexture;

		//points of the slices
		float _points[64][3];

		// OpenGL rendering context
		wxGLContext *_context;

        // Rendering options
        bool _showBox;
        bool _showClipPlane;
        GLfloat _bcgColour[3];

		/****** Mouse Code ******/

		// simple trackball-like motion control
		GLboolean _trackingMotion;
		GLboolean _trackingTranslation;
		float _lastPos[3], _transPosY;
		int _lastTime;
		float _angle, _axis[3];
		float _pAngle, _pAxis[3];
		MoveModes _mode;
		float _viewDistance;
		GLdouble _user[4];
		bool _gbPlane;
		GLfloat _objectXform[16];
		GLfloat _planeXform[16];

	public:
		/** Constructor. */
		vol3D_VolVisCanvas(wxStatusBar *statusBar, wxWindow *parent, wxWindowID id = wxID_ANY,
						   const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxDefaultSize, 
						   long style = 0, const wxString& name = "VolGLCanvas", int *gl_attrib = AttributeList);
		
        /** Destructor. */
		~vol3D_VolVisCanvas();

		bool Init();

		/** Event handlers. */
		void OnPaint(wxPaintEvent& evt);
		void OnSize(wxSizeEvent& evt);
		void OnEraseBackground(wxEraseEvent& evt);
		void OnChar(wxKeyEvent& evt);
		void OnMouseEvent(wxMouseEvent& evt);
		void OnMouseMoveEvent(wxMouseEvent& evt);

		void SetView(float distance, float ax, float ay, float az, float ux, float uy, float uz);
		void SetVolumeData(const unsigned char *rgbabuf, const i3d::Vector3d<size_t> &size, const i3d::Vector3d<float> &scale);
		void SetIsoSurfaceData(const vol3D_ComponentVector &comps, const i3d::Vector3d<float> &trans);
		void ShowIsosurface(bool show, v4d_Palette pal);

        void ShowBox(bool show)
        {
            _showBox = show;
            Refresh();
        }

        void ShowClipPlane(bool show)
        {
            _showClipPlane = show;
            Refresh();
        }

        void SetBcgColour(GLfloat r, GLfloat g, GLfloat b)
        {
            _bcgColour[0] = r;
            _bcgColour[1] = g;
            _bcgColour[2] = b;
            Refresh();
        }

        void SaveViewMatrix();
        void LoadViewMatrix();

		void Reset();

	private:
        void ResizeViewPort();
		void StartMotion(int x, int y, int time);
		void StopMotion();
		void Ptov(int x, int y, int width, int height, float v[3]) const;
		void InvertMat(float mat[16]) const;
		void Draw();

    private:
		DECLARE_EVENT_TABLE()
	};

} // namespace v4d

#endif //__VOL3D_CANVAS_H__
